Package: prova
Title: Nonparametric Probabilistic-Statistical Variate Analysis with
        Automated Markov-Chain Monte Carlo
Version: 1.0.0
Authors@R: c(
  person("PierGianLuca", "Porta Mana", ,
  "pgl@portamana.org", role = c("aut", "cre", "cph"),
  comment = c(ORCID = "0000-0002-6070-0784")),
  person("Aurora", "Grefsrud", ,
  "agre@hvl.no", role = c("ctb"),
  comment = c(ORCID = "0009-0005-9063-4131")),
  person("Håkon", "Mydland", ,
  "haakon.mydland@gmail.com", role = c("ctb"),
  comment = c(ORCID = "0009-0004-9558-8894")),
  person("Maksim", "Ohvrill", ,
  "maksimohvrill@hotmail.com", role = c("ctb"),
  comment = c(ORCID = "")),
  person("Simen", "Hesthamar Hauge", ,
  "simen@hhnet.no", role = c("ctb"),
  comment = c(ORCID = "0009-0005-6158-2846"))
  )
Description: Calculate posterior joint and conditional probabilities, probability distributions of population frequencies, and information-theoretic measures, by means of Bayesian nonparametric methods. Data imputation is automatic and done in a principled way. Markov-chain Monte Carlo calculations are automatically handled and do not require user supervision. Applications range from statistical estimation and probabilistic hypothesis testing to evidence-based inference and decision making, in a wide range of disciplines from astrophysics to medicine. For more details and examples see for instance Porta Mana et al. (2026) <doi:10.31219/osf.io/8nr56>, Dunson & Bhattacharya (2011) <doi:10.1093/acprof:oso/9780199694587.003.0005>, Lindley & Novick (1981) <doi:10.1214/aos/1176345331>, Bernardo & Smith (2000) <doi:10.1002/9780470316870>, Müller et al. (2015) <doi:10.1007/978-3-319-18968-0>. Requires the packages 'Nimble', 'parallel', 'extraDistr'.
License: AGPL (>= 3)
URL: https://pglpm.github.io/prova/, https://github.com/pglpm/prova/
Encoding: UTF-8
Depends: R (>= 4.5.0)
Imports: extraDistr, parallel
Suggests: nimble (>= 1.4.2), knitr, rmarkdown
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
LazyData: true
NeedsCompilation: no
Packaged: 2026-07-07 12:26:48 UTC; pglpm
Author: PierGianLuca Porta Mana [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-6070-0784>),
  Aurora Grefsrud [ctb] (ORCID: <https://orcid.org/0009-0005-9063-4131>),
  Håkon Mydland [ctb] (ORCID: <https://orcid.org/0009-0004-9558-8894>),
  Maksim Ohvrill [ctb],
  Simen Hesthamar Hauge [ctb] (ORCID:
    <https://orcid.org/0009-0005-6158-2846>)
Maintainer: PierGianLuca Porta Mana <pgl@portamana.org>
Repository: CRAN
Date/Publication: 2026-07-16 12:50:08 UTC
Built: R 4.5.2; ; 2026-07-16 13:01:44 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(hist,probability)
S3method(plot,probability)
S3method(print,probability)
export(Pr)
export(flexiplot)
export(learn)
export(metadatatemplate)
export(mutualinfo)
export(plotquantiles)
export(pread.csv)
export(pwrite.csv)
export(qPr)
export(rPr)
export(vrtgrid)
import(grDevices)
import(graphics)
import(parallel)
import(stats)
import(utils)
importFrom(extraDistr,rbern)
importFrom(extraDistr,rcat)

<!-- badges: start -->
  [![Zenodo](https://zenodo.org/badge/DOI/10.5281/zenodo.19136582.svg)](https://doi.org/10.5281/zenodo.17226082)
<!-- badges: end -->

*[Looking for the R package "inferno"? you came to the right place! It's been renamed "Prova"]*

<img src="https://github.com/pglpm/prova/raw/main/man/figures/prova_symbol.jpg" alt="Ensemble of densities" width="100%"/>

"prova" /'prɔva/ (Italian)

- [(noun)](https://dictionary.cambridge.org/dictionary/italian-english/prova): test, trial, assessment, proof, evidence, sign, indication, try, attempt.
- [(verb)](https://dictionary.cambridge.org/dictionary/italian-english/provare): test!, try out!, assess!, attempt!, prove!, demonstrate!, show!


# ***Prova***: probabilistic-statistical variate analysis, nonparametric and with automated Markov-chain Monte Carlo

This repository provides an R package and some theoretical background to perform probabilistic and statistical data analysis and inference. These are its main features:

- Any combination of **binary**, **nominal**, **ordinal**, **continuous** variates. Continuous variates can be bounded or unbounded, and also rounded or discretized.
- **No modelling assumptions** such as gaussianity, linearity, or any other kind of model. The analysis and inferences are fully [non-parametric](https://dictionary.apa.org/nonparametric).
- **No assumptions about functional dependence** between variates. The analysis and inferences are therefore more general than those by neural networks, random forests, or similar machine-learning algorithms.
- **Automatic [imputation](https://dictionary.apa.org/imputation)** of missing data: all sample data are used, even those that lacks some variate values. The imputation is done with a principled method (the marginalization rule of probability theory), rather than ad-hoc procedures.
- Easy and straightforward **subgroup analyses** and **stratified analyses**, for any division of variates, with full statistical details.
- **Quantification of [generalizability](https://dictionary.apa.org/generalizability)** beyond the finite sample size. In other words, quantification of uncertainty of results regarding the whole, unsampled, population.
- Straightforward use within **decision theory**, such as **clinical decision-making**. Users can immediately combine the probabilistic results with any measures of utilities, such as [quality-adjusted life years](https://dictionary.apa.org/quality-adjusted-life-years).
- **Quantification of associations** between any kinds of variates, without modelling assumptions (gaussianity, linearity, etc.), thanks to the use of [*mutual information*](https://electropedia.org/iev/iev.nsf/display?openform&ievref=171-07-26).
- **[Base-rate](https://dictionary.apa.org/base-rate-fallacy) correction** for inferences about out-of-population data, by means of Bayes's theorem.
- **Automated Markov-chain Monte Carlo** computation. Users unfamiliar with Monte Carlo methods don't have to worry, because the computations are handled automatically.

The package at bottom does Bayesian nonparametric inference (also called "density inference" or "inference under exchangeability"), which makes all features above possible.

The [introductory vignette](https://pglpm.github.io/prova/articles/intro.html) explains, with a guided example, most of the features above, as well as the main ideas and functions. It can be particularly useful for researchers who are more familiar with traditional "frequentist" statistics but would like to try the Bayesian approach. See the [post](https://www.apadivisions.org/division-7/publications/newsletters/developmental/2018/07/bayesian-statistics) by Barbara W. Sarnecka, frequentist statistician turned Bayesian, for a brilliant overview of the Bayesian advantages. The [vignette about mutual information](https://pglpm.github.io/prova/articles/mutualinfo.html) explains the use of this powerful measure of association.

The package is under continuous development, but the core functionalities work and have been tested in concrete research questions; see [example applications](#example-applications) below.

The package internally does the computations necessary for Bayesian inference by means of Monte Carlo methods thanks to the R package [**Nimble**](https://r-nimble.org/). As already mentioned, this computation is automated. Users familiar with Monte Carlo methods can still access computational details and can even change some of the computation hyperparameters.

## Installation

**You need to have installed the package [**Nimble**](https://r-nimble.org/), *at least version 1.4.2*.** Please follow its [installation instructions](https://r-nimble.org/manual/cha-installing-nimble.html) for your operating system.

Newer versions of **Prova** can be installed with
```
remotes::install_github('pglpm/prova')
```


## Documentation

The vignette [*An introduction to probabilistic-statistical variate analysis*](https://pglpm.github.io/prova/articles/intro.html) is a step-by-step introduction to **Prova** and also to Bayesian nonparametrics. It guides you through a concrete example with various kinds of inferences. You may also try to follow it using a dataset of your own.

Other tutorials are available at [pglpm.github.io/prova](https://pglpm.github.io/prova/), or can be accessed in an R session with `browseVignettes('prova')`.

A summary of the theoretical foundations, including further references, is available in [this draft](https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf). The main idea for the internal mathematical representation comes from [Dunson & Bhattacharya](https://doi.org/10.1093/acprof:oso/9780199694587.003.0005) and [Ishwaran & Zarepour](https://doi.org/10.2307/3315951).

For a low-level course on Bayesian nonparametric inference and Decision Theory see [Data Science and AI Prototyping](https://pglpm.github.io/ADA511/).


## Example applications

- [*Personalized prognosis & treatment using an optimal predictor machine: An example study on conversion from Mild Cognitive Impairment to Alzheimer's Disease*](https://doi.org/10.31219/osf.io/8nr56).

- [*Don't guess what's true: choose what's optimal. A probability transducer for machine-learning classifiers*](https://doi.org/10.31219/osf.io/vct9y)

- [*Does the evaluation stand up to evaluation? A first-principle approach to the evaluation of classifiers*](https://doi.org/10.31219/osf.io/7rz8t)

- [*Calibrated and uncertain? Evaluating uncertainty estimates in
binary classification models*](https://doi.org/10.1088/2632-2153/ae45ed)


## Projects using **Prova**:

- [InfernoCalibNet](https://m4siko.github.io/InfernoCalibNet/): uncertainty-aware predictions for medical AI using CNN and Bayesian nonparametrics framework
- [parkinsonbayes](https://github.com/pglpm/parkinsonbayes/): Examples of Bayesian nonparametric inference for studies of Parkinson's Disease
- [Inferno-App](https://github.com/Myddis/Inferno-App/): PySide6 application that integrates Python and R functionality using the **Inferno** (old version of **Prova**) R package.


## Contact

Please report bugs and request features or specific documentation on [GitHub Issues](https://github.com/pglpm/prova/issues).
If you have other questions about application, theory, technical implementation, feel free to contact Luca <pglXYZ@portamanaXYZ.org> (remove 'XYZ' for anti-spam purposes).


## Disclaimer

No large language models were used in the production of this software and of its documents.

#' Calculate quantiles
#'
#' @description This function calculates the quantiles of posterior probabilities and posterior conditional probabilities. It also outputs the variability of such quantiles if more training data were available.
#'
#' @details This function calculates the quantiles of \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data})} or of \eqn{\mathrm{Pr}(Y = y \vert X \le x, \text{data})} or combinations thereof, at specified cumulative-probability levels. In other words, it calculates the values of \eqn{Y} having specified cumulative probabilities or conditional probabilities. It also calculates the variability of those quantiles if more learning data were provided. It is somewhat analogous to the `q`-variants of R distribution functions, such as [stats::qnorm()]. The variability can be expressed in the form of quantiles, samples, or both, as in the [Pr()] function. If several joint values are given for the probability levels and for `X`, the function creates a 2D grid of results for all possible combinations of the given probability levels and `X` values. Each variate in the argument `X` can be specified either as a point-value \eqn{X = x} or as a left-open interval \eqn{X \le x} or as a right-open interval \eqn{X \ge x}, through the argument `tails`.
#'
#' @param p Numeric vector of probability levels. Default: `c(0.25, 0.5, 0.75)`.
#' @param Yname Character vector: name of variate whose quantiles will be computed.
#' @param X Matrix or data.table or `NULL` (default): set of values of variates on which we want to condition. If `NULL`, no conditioning is made (except for conditioning on the learning dataset and prior assumptions). One variate per column, one set of values per row.
#' @param learnt Either a character with the name of a directory or full path for a 'learnt.rds' object, produced by the [learn()] function, or such an object itself.
#' @param tails Named vector or list, or `NULL` (default). The names must match some or all of the variates in arguments `X`. For variates in this list, the probability conditional is understood in a semi-open interval sense: \eqn{X \le x} or \eqn{X \ge x}, an so on. See analogous argument in [Pr()].
#' @param priorY Reserved for use in future versions of the package.
## #' @param priorY Numeric vector with the same length as the rows of `Y`, or `TRUE`, or `NULL` (default): prior probabilities or base rates for the `Y` values. If `TRUE`, the prior probabilities are assumed to be all equal. For the moment only the value `NULL` is accepted.
#' @param nsamples Integer or `NULL` or `'all'` (default): desired number of samples of the variability of the quantile for `Y`. If `NULL`, no samples are reported. If `'all'` (or `Inf`), all samples obtained by the [learn()] function are used.
#' @param quantiles Numeric vector, between 0 and 1, or `NULL`: desired quantiles of the variability of the quantile for `Y`. Default `c(0.055, 0.25, 0.75, 0.945)`, that is, the 5.5%, 25%, 75%, 94.5% quantiles (these are typical quantile values in the Bayesian literature: they give 50% and 89% credibility intervals, which correspond to 1 shannons and 0.5 shannons of uncertainty). If `NULL`, no quantiles are calculated.
#' @param parallel Logical or positive integer or cluster object. `TRUE` (default): use roughly half of available cores; `FALSE`: use serial computation; integer: use this many cores. It can also be a cluster object previously created with [parallel::makeCluster()]; in this case the parallel computation will use this object.
#' @param sep character, default `','`: character to separate variate names and values
#' @param solidus character, default `'|'`: character prepended to names of the variates in the conditional (typically the `X` variates).
#' @param verbose Logical, default `FALSE`: give messages about parallel processing?
#' @param keepYX Logical, default `TRUE`: keep a copy of the `Y` and `X` arguments in the output? This is used for the plot method.
#' @param tol numeric positive: tolerance in the calculation of quantiles. Default: `.Machine$double.eps * 10` (typically `2.22045e-15`).

#'
#' @return A list of the following elements:
#' - `values`: a matrix with the requested \eqn{Y}-quantiles `p` conditional on the requested \eqn{X}-values in `X`, for all combinations of `p` (rows) and `X` (columns).
#' - `quantiles` (possibly `NULL`): an array with the variability quantiles (3rd dimension of the array) for the quantiles of the `value` element.
#' - `samples` (possibly `NULL`): an array with the variability samples (3rd dimension of the array) for such quantiles.
#' - `Y`, `X`: copies of the `Y` and `X` arguments.
#'
#' @references
#' - Porta Mana (2025): *What's special about 89% credibility intervals?* <doi:10.5281/zenodo.17072199>.
#'
#' @seealso
#' [learn()], which generates the `learnt` objects required by `qPr()`.
#'
#' [Pr()] to calculate joint and conditional probabilities.
#'
#' [rPr()] to generate datapoints.
#'
#' @examples
#' ### WARNING: the following examples, if run, might even take a minute or more.
#'
#' \donttest{
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## ## Example 1:
#' ## Calculate the 5.5%-, 50%-, and 94.5%-quantiles for the variate "bill lengt",
#' ## that is, the values of "bill length" having such cumulative probabilities
#'
#' quants <- qPr(
#'   Yname = 'bill_len',
#'   learnt = learnt, parallel = 1
#' )
#'
#' ## display the quantile values
#' quants$values
#'
#' ## verify these values using Pr():
#' probs <- Pr(
#'   Y = data.frame(bill_len = c(quants$values)),
#'   tails = list(bill_len = -1),
#'   learnt = learnt, parallel = 1
#' )
#'
#' ## the cumulative probabilities are indeed 0.055, 0.5, 0.945 within numerical error:
#' probs$values
#'
#' ## display the variability about the quantiles
#' quants$quantiles
#'
#'
#' ## ## Example 2:
#' ## Calculate the 5.5%-, 50%-, and 94.5%-quantiles for the variate "bill lengt",
#' ## for the subpopulation of species 'Adelie'
#'
#' quants <- qPr(
#'   Yname = 'bill_len',
#'   X = data.frame(species = 'Adelie'),
#'   learnt = learnt, parallel = 1
#' )
#'
#' ## display the quantile values
#' quants$values
#'
#' ## verify these values using Pr():
#' probs <- Pr(
#'   Y = data.frame(bill_len = c(quants$values)),
#'   X = data.frame(species = 'Adelie'),
#'   tails = list(bill_len = -1),
#'   learnt = learnt, parallel = 1)
#'
#' ## the cumulative probabilities are indeed 0.055, 0.5, 0.945 within numerical error:
#' probs$values
#' }
#'
#' @import parallel
#' @import utils
#' @import stats
#'
#' @export
qPr <- function(
    p = c(0.25, 0.5, 0.75),
    Yname,
    X = NULL,
    learnt,
    tails = NULL,
    priorY = NULL,
    nsamples = 'all',
    quantiles = c(0.055, 0.5, 0.945),
    parallel = TRUE,
    sep = ',',
    solidus = '|',
    verbose = FALSE,
    keepYX = TRUE,
    tol = .Machine$double.eps * 10

) {
    ## #' @param usememory Logical, default `TRUE`: save partial results to disc, to avoid excessive RAM use. (For the moment only possible value is `TRUE`.)
    usememory <- TRUE

    Qerror <- pnorm(c(-1, 1))

#### Requested parallel processing
    ## NB: doesn't make sense to have more cores than chains
    closeexit <- FALSE
    if ('cluster' %in% class(parallel)){
        ## user provides a cluster object
        cl <- parallel
    } else if (isTRUE(parallel)) {
        ## user wants us to register a parallel backend
        ## and to choose number of cores
        ncores <- max(1,
            floor(parallel::detectCores() / 2))
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
        if(verbose){message('Registered ', capture.output(print(cl)), '.')}
    } else if (isFALSE(parallel)) {
        ## user wants us not to use parallel cores
        ncores <- 1
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
    } else if (is.numeric(parallel) &&
                   is.finite(parallel) && parallel >= 1) {
        ## user wants us to register 'parallel' # of cores
        ncores <- parallel
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
        if(verbose){message('Registered ', capture.output(print(cl)), '.')}
    } else {
        stop("Unknown value of argument 'parallel'.")
    }

    ## Close parallel connections if any were opened
    if(closeexit) {
        closecoresonexit <- function(){
            if(verbose){message('Closing connections to cores.')}
            parallel::stopCluster(cl)
            ## parallel::setDefaultCluster(NULL)
        }
        on.exit(closecoresonexit())
    }

    ## Extract Monte Carlo output & auxmetadata
    ## If learnt is a string, check if it's a folder name or file name
    if (is.character(learnt)) {
        ## Check if 'learnt' is a folder containing learnt.rds
        if (file_test('-d', learnt) &&
                file.exists(file.path(learnt, 'learnt.rds'))) {
            learnt <- readRDS(file.path(learnt, 'learnt.rds'))
        } else {
            ## Assume 'learnt' the full path of learnt.rds
            ## possibly without the file extension '.rds'
            learnt <- paste0(sub('.rds$', '', learnt), '.rds')
            if (file.exists(learnt)) {
                learnt <- readRDS(learnt)
            } else {
                stop('The argument "learnt" must be a folder containing learnt.rds, or the path to an rds-file containing the output from "learn()".')
            }
        }
    }
    ## Add check to see that learnt is correct type of object?
    auxmetadata <- learnt$auxmetadata
    learnt$auxmetadata <- NULL
    learnt$auxinfo <- NULL
    ncomponents <- nrow(learnt$W)
    nmcsamples <- ncol(learnt$W)

    if(is.numeric(nsamples)){
        if(is.na(nsamples) || nsamples < 1) {
            nsamples <- NULL
        } else if(!is.finite(nsamples)) {
            nsamples <- nmcsamples
        }
    } else if (is.character(nsamples) && nsamples == 'all'){
        nsamples <- nmcsamples
    }


    if(length(Yname) > 1){stop('Specify only one variate in "Yname".')}

    if(all(is.na(X))){X <- NULL}
    if(!is.null(X)){X <- as.data.frame(X)}
    Xv <- colnames(X)

    if(!is.null(tails)){
        tails <- as.list(tails)
        if(is.null(names(tails))) {
            stop('Missing variate names in "tails"')
        }
    }

    tailscentre <- list('==', 0, '0', NULL)
    tailsleft <- list('<=', -1, '-1', 'left')
    tailsright <- list('>=', 1, '+1', 'right')
    tailsvalues <- c(tailscentre, tailsleft, tailsright)

    ## Consistency checks

    if (!all(Yname %in% auxmetadata$name)) {
        stop('unknown Y variate ',
            paste0(Yname[!(Yname %in% auxmetadata$name)], collapse = ' '),
            '\n')
    }
    if (auxmetadata[auxmetadata$name == Yname, 'mcmctype'] %in% c('B', 'N')){
        stop('quantiles are undefined for binary and nominal variates.')
    }
    if (length(unique(Yname)) != length(Yname)) {
        stop('duplicate Y variates\n')
    }

    if (!all(Xv %in% auxmetadata$name)) {
        stop('unknown X variate ',
            paste0(Xv[!(Xv %in% auxmetadata$name)], collapse = ' '),
            '\n')
    }
    if (length(unique(Xv)) != length(Xv)) {
        stop('duplicate X variates\n')
    }

    if (any(Yname %in% Xv)) {
        stop('overlap in Y and X variates\n')
    }

    tailsv <- names(tails)
    if (!all(tailsv %in% Xv)) {
        warning('"tails" variate ',
            paste0(tailsv[!(tailsv %in% Xv)], collapse = ' '),
            ' not among X; ignored\n')
        tails <- tails[(tailsv %in% Xv)]
        tailsv <- names(tails)
    }
    if (length(unique(tailsv)) != length(tailsv)) {
        stop('duplicate "tails" variates\n')
    }
    if(!all(tails %in% tailsvalues)) {
        stop('"tails" values must be ',
            paste0(tailsvalues, collapse = ' '), '\n')
    }

    ## transform 'tails' to -1, +1
    ## +1: '<=',    -1: '>='
    ## this is opposite of the argument convention because
    ## interval probabilities are calculated with `lower.tail = TRUE`
    ## eg:
    ## pnorm(x, mean, sd, lower.tail = FALSE) ==
    ##     pnorm(-x, -mean, sd, lower.tail = TRUE)
    tails[tails %in% tailscentre] <- NULL
    cleft <- tails %in% tailsleft
    cright <- tails %in% tailsright
    tails[cleft] <- +1
    tails[cright] <- -1
    tails <- unlist(tails)

    ## Check if a prior for Y is given, in that case Y and X will be swapped
    if (isFALSE(priorY) || is.null(priorY)) {
        priorY <- NULL
        doquantiles <- !is.null(quantiles)
    } else {
        if(is.null(X)){ stop('X must be non-null if priorY is given') }

        if(!isTRUE(priorY) && length(priorY) != nrow(Y)) {
            stop('priorY must have as many elements a the rows of Y')
        }

        doquantiles <- FALSE
        if(!is.null(quantiles)) {
            ## message('For the moment, ',
            ##     'computation of quantiles is affected by a larger error',
            ##     'if "priorY" is specified.')
            nsamples0 <- nsamples
            nsamples <- max(nsamples0, 1200L)
        }

        ## if priorY is TRUE, the user wants a uniform prior distribution
        if(isTRUE(priorY)){
            priorY <- 1 + numeric(nrow(Y))
        } else {
            ## Check for invalid probabilities
            if (!is.numeric(priorY) || any(priorY < 0)) {
                stop('priorY contains invalid probabilities')
            }
        }

        ## Swap X and Y, to use Bayes's theorem
        . <- Y
        Y <- X
        X <- .
        rm(.)
        . <- Yname
        Yname <- Xv
        Xv <- .
        rm(.)
    }

    nY <- length(p)
    nX <- max(nrow(X), 1L)
    auxY <- auxmetadata[auxmetadata$name == Yname, ]

    dosamples <- !is.null(nsamples)


#### tmp dir where to save X and Y objects
    temporarydir <- tempdir()


#### Calculate and save arrays for X values:
    if (is.null(X)) {
        lprobX <- log(learnt$W)
        saveRDS(lprobX,
            file.path(temporarydir,
                paste0('__X', 1, '__.rds'))
        )
    } else {
        ## Construction of the arguments for util_lprobs, X argument
        lpargs <- util_lprobsargsyx(
            x = X,
            auxmetadata = auxmetadata,
            learnt = learnt,
            tails = tails
        )

        invisible(parallel::parLapply(cl = cl,
            X = lpargs$xVs,
            fun = util_lprobsbase,
            params = lpargs$params,
            logW = c(log(learnt$W)),
            temporarydir = temporarydir,
            lab = '__X'
        ))
    }

#### Determine the type of Y variate, set params accordingly
    if(auxY$mcmctype == 'O'){
        params1 = log(apply(
            learnt$Oprob[auxY$indexpos + seq_len(auxY$Nvalues), ,],
            c(2, 3), cumsum
        ))
        params2 <- NULL
        util_qYX <- util_qYXdiscr
    } else if(auxY$mcmctype == 'R'){
        params1 <- learnt$Rmean[auxY$id, ,]
        params2 <- learnt$Rsd[auxY$id, ,]
        util_qYX <- util_qYXcont
    } else if(auxY$mcmctype == 'D'){
        params1 <- learnt$Dmean[auxY$id, ,]
        params2 <- learnt$Dsd[auxY$id, ,]
        util_qYX <- util_qYXcont
    } else if(auxY$mcmctype == 'C'){
        params1 <- learnt$Cmean[auxY$id, ,]
        params2 <- learnt$Csd[auxY$id, ,]
        util_qYX <- util_qYXcont
    } else {
        stop('type of Yname not found')
    }

#### Calculation with all pY and X combinations
    ## keys <- c('values', 'quantiles', 'samples', 'values.MCaccuracy', 'quantiles.MCaccuracy')
    keys <- c('values', 'quantiles', 'samples')
    ##
    combfnr <- function(...){setNames(do.call(mapply,
        c(FUN = `rbind`, lapply(X = ..., FUN = `[`, keys, drop = FALSE))),
        keys)}
    ## combfnc <- function(...){setNames(do.call(mapply, c(FUN=cbind, lapply(list(...), `[`, keys))), keys)}

    out <- combfnr(parallel::parApply(cl = cl,
            X = expand.grid(pY = p, jx = seq_len(nX)),
            MARGIN = 1,
            FUN = util_qYX,
            params1 = params1, params2 = params2,
            auxmetadata = auxY,
            temporarydir = temporarydir, usememory = usememory,
            doquantiles = doquantiles, quantiles = quantiles,
            dosamples = dosamples, nsamples = nsamples,
            Qerror = Qerror,
            tol = tol
    ))

    ## clean temp files
    if(usememory) {
        unlink(x = sapply(seq_len(nX), function(jx){
            file.path(temporarydir, paste0('__X', jx, '__.rds'))
            }))
    }


    ## if(is.null(priorY)){
    ##     y <- Y
    ##     y[, colnames(Y) %in% tailsv] <- NA
    ##     jacobians <- exp(rowSums(
    ##         as.matrix(vtransform(y,
    ##             auxmetadata = auxmetadata,
    ##             logjacobianOr = TRUE)),
    ##         na.rm = TRUE
    ##     ))
    ##     rm(y)
    ## }

    ## transform to grid
    ## in the output-list elements the Y & X values are the rows
    ## dim(out$values.MCaccuracy) <-
    dim(out$values) <- c(nY, nX)

    ## dimension & value names for variates
    if(!is.null(X)){
        Xnames <- setNames(object = list(
            apply(X = X, MARGIN = 1, FUN = paste0, collapse = sep,
                simplify = TRUE)),
            nm = paste0(if(is.null(priorY)){solidus},
                paste0(colnames(X), collapse = sep)) )
    } else {
        Xnames <- list(NULL)
    }
    if(!is.null(Yname)){
        Ynames <- setNames(object = list(p), nm = Yname)
        ## Ynames <- setNames(object = list(
        ##     apply(X = Y, MARGIN = 1, FUN = paste0, collapse = sep,
        ##         simplify = TRUE)),
        ##     nm = paste0(colnames(Y), collapse = sep)
        ## )
    } else {
        Ynames <- list(NULL)
    }

    if(is.null(priorY)){

        dimnames(out$values) <- c(Ynames, Xnames)
        ## dimnames(out$values.MCaccuracy) <- c(Ynames, Xnames)

    } else {
        ## Bayes's theorem
        out$values <- t(priorY * t(out$values))
        out$values.MCaccuracy <- NULL
        normf <- rowSums(out$values, na.rm = TRUE)
        out$values <- t(out$values/normf)

        dimnames(out$values) <- c(Xnames, Ynames)
    }

    if(dosamples){
        temp <- list(sample = round(seq(1, nmcsamples, length.out = nsamples)))

        ## transform to grid
        dim(out$samples) <- c(nY, nX, nsamples)

        if(is.null(priorY)){
            ## multiply by jacobian factors
            ## out$samples <- out$samples * jacobians
            dimnames(out$samples) <- c(Ynames, Xnames, temp)
        } else {
            ## Bayes's theorem
            out$samples <- priorY * aperm(a = out$samples, perm = c(2, 1, 3),
                resize = TRUE)
            normf <- c(t(colSums(out$samples, na.rm = TRUE)))
            out$samples <- aperm(a = aperm(a = out$samples, perm = NULL,
                resize = TRUE) / normf, perm = NULL,
                resize = TRUE)

            dimnames(out$samples) <- c(Xnames, Ynames, temp)
        }

    }

    if(doquantiles){
        temp <- list(Q = names(quantile(x = 1, probs = quantiles, names = TRUE)))
        if(is.null(priorY)){
            ## transform to grid
            dim(out$quantiles) <- c(nY, nX, length(quantiles))
            ## dim(out$quantiles.MCaccuracy) <- c(nY, nX, length(quantiles))
            ## multiply by jacobian factors
            ## out$quantiles <- out$quantiles * jacobians
            ## out$quantiles.MCaccuracy <- signif(x = out$quantiles.MCaccuracy * jacobians,
            ##     digits = 2)

            dimnames(out$quantiles) <- c(Ynames, Xnames, temp)
            ## dimnames(out$quantiles.MCaccuracy) <- c(Ynames, Xnames, temp)
        } else {
            ## calculate quantiles from samples
            out$quantiles <- aperm(
                a = apply(X = out$samples, MARGIN = c(1, 2), FUN = quantile,
                    probs = quantiles, type = 6,
                    na.rm = TRUE, names = FALSE,
                    simplify = TRUE),
                perm = c(2, 3, 1), resize = TRUE)

            ## adjust number of samples as originally requested
            if(is.null(nsamples0)) {
                out$samples <- NULL
            } else if(nsamples0 < nsamples) {
                out$samples <-out$samples[ , ,
                    round(seq(1, nsamples, length.out = nsamples0))]
            }

            dimnames(out$quantiles) <- c(Xnames, Ynames, temp)
            ## dimnames(out$quantiles.MCaccuracy) <- c(Xnames, Ynames, temp)
        }
    }

    if(isTRUE(keepYX)){
        ## save Y and X values in the output; useful for plotting methods
        if(is.null(priorY)){
            out$pY <- Ynames
            out$X <- X
        } else {
            out$Y <- X
            out$X <- Y
            }
    }

    class(out) <- 'probability'
    out
}

#' @name prova.data
#' @rdname prova.data
#'
#' @title Write and read CSV files in **Prova**
#'
#' @description Utility functions to read and write CSV files in the format required by **Prova**
#'
#' @param x The object to be written. Preferably a matrix or data frame; if not, it is attempted to coerce `x` to a [data frame][base::data.frame()]. See [utils::write.csv()].
#'
#' @param file Either a character naming a file or a connection open for writing or reading. See [utils::write.csv()] and [utils::read.csv()].
#' @param ... Other arguments to be passed to [utils::write.csv()] or [utils::read.csv()]. Arguments 'row.names', 'quote', 'na', 'na.strings', 'tryLogical', 'sep', 'dec' are not allowed.
#'
#' @return `pread.csv` returns a [data frame][base::data.frame()] containing a representation of the data in the file; see [utils::read.csv()]. `pwrite.csv' returns `NULL` [invisibly][base::invisible()].
#'
#' @details
#' The functions [learn()] and [metadatatemplate()] accept CSV files formatted as follows:
#'
#' - Decimal values should be separated by a *dot*; no comma should be used to separate thousands etc. Example: `86342.75`.
#' - Character and names should be quoted in single or double quotes. Example: `"female"`.
#' - Values should be separated by *commas*, not by tabs or semicolons.
#' - Missing values should be simply *empty*, not denoted by "NA", "missing", "-", or similar.
#' - Preferably there should not be factors (see [base::factor]); use character names instead.
#'
#' The utility functions `pwrite.csv()` and `pread.csv()` are wrappers to [utils::write.csv()] and [utils::read.csv()] that set appropriate default parameters according to the formatting rules above.
#'
#' @seealso
#' [metadatatemplate()] to help writing metadata files.
#'
#' [learn()], which needs a metadata data-frame or CSV file.
#'
#' @examples
#' ## Save the 'penguins' dataset in a (temporary) file
#' filename <- tempfile(fileext = '.csv')
#'
#' pwrite.csv(penguins, file = filename)
#'
#' ## check first few lines of the raw file
#' writeLines(readLines(filename, n = 10))
#'
NULL

#' @rdname prova.data
#'
#' @import utils
#'
#' @export
pwrite.csv <- function(x, file, ...){
    write.csv(x = x, file = file,
        row.names = FALSE,
        quote = TRUE,
        # sep = ", ",
        # dec = ".",
        na = '',
        ...)
}

#' @rdname prova.data
#'
#' @import utils
#'
#' @export
pread.csv <- function(file, ...){
    read.csv(file = file,
        na.strings = '',
        stringsAsFactors = FALSE,
        tryLogical = FALSE,
        sep = ",",
        dec = ".",
        ...)
}

#' Cleanup a learn()-output directory
#'
#' @details
#' For deeper monitoring of the MCMC, the user can require the 'learn()' function not to clean intermediate MCMC-related files generated during the computation.
#'
#' The present function can be used to remove these intermediate files from the output directory created by 'learn()'.
#'
#' @return No return value; called for side effects.
#'
#' @keywords internal
util_cleanup <- function(path){
    if(all(file.exists(file.path(path, c(
        'learnt.rds',
        'MCtraces.rds',
        'MCtraces.pdf',
        'log-1.log',
        'plotquantiles_learnt.pdf',
        'plotsamples_learnt.pdf'
    ))))){
        invisible(file.remove(dir(path,
            pattern = paste0('^___.*\\..*$'),
            full.names = TRUE
        )))
    } else {
        stop("This doesn't look like a 'learn()' output directory")
    }
    invisible()
}


#' Join '____tempPtraces-' files
#'
#' @details
#' For deeper monitoring of the MCMC, the user can require the 'learn()' function not to clean intermediate MCMC-related files generated during the computation. The files with prefix '____tempPtraces-' contain chunks of MCMC traces.
#'
#' The present function can be used to join them into a single trace.
#'
#' @return A [data frame][base::data.frame()] of MCMC traces.
#'
#' @keywords internal
util_joinPtraces <- function(path){
    Plist <- list.files(path = path, pattern = '^____tempPtraces-.*\\.rds$')
    chainlist <- unique(sub(
        pattern = '^____tempPtraces-([0-9]+)-.*',
        replacement = '\\1',
        x = Plist
    ))

    for(achain in chainlist){
        chunklist <- grepv(
            pattern = paste0('^____tempPtraces-', achain, '-.*\\.rds$'),
            x = Plist
        )
        chunks <- sort(as.numeric(unique(sub(
            pattern = paste0('^____tempPtraces-', achain, '-([0-9]+)\\.rds$'),
        replacement = '\\1',
        x = chunklist
        ))))

        Ptraces <- NULL
        for(achunk in chunks){
            Ptraces <- rbind(Ptraces,
                readRDS(file.path(path,
                    paste0('____tempPtraces-', achain, '-', achunk, '.rds')
                ))
            )
        }
        saveRDS(Ptraces, file.path(path,
            paste0('_allPtraces-', achain, '.rds')
        ))
    }
    message('\nDone.\n')
}

#' Plot numeric or character values
#'
#' @description
#' Plot function that modifies and expands the **graphics** package's [graphics::matplot()] function in several ways.
#'
#' @details
#' This function is essentially a wrapper around [graphics::matplot()], augmenting the latter with some additional features useful for plotting data and results handled by **Prova**. Some of the additional features provided by `flexiplot` are the following:
#'
#' - Either or both `x` and `y` arguments can be of class [`base::character`]. In this case, axes labels corresponding to the unique values are used (see arguments `xdomain` and `ydomain`). This makes it easier to plot nominal and ordinal variates.
#' - A jitter can also be added to the generated points, via the `xjitter` and `yjitter` switches. This makes it easier to generate scatter plots of nominal and ordinal variates.
#' - It is possible to specify only a lower or upper limit in the `xlim` and `ylim` arguments, letting the other limit to be found automatically. This can be useful in plotting probabilities, in cases where we want to specify the lower, `0` limit, but want the upper limit to simply be the the maximum probability.
#' - Transparency of lines or markers can be specified through argument `alpha.f`.
#' - The plotting style is different, and default argument `type = 'l'` (line plot) rather than `type = 'p'` (point plot).
#'
#' See the package's vignettes for more examples.
#'
#' @param x Numeric or character: vector of x-coordinates. If missing, a numeric vector `1:...` is created having as many values as the rows of `y`.
#' @param y Numeric or character: vector of y coordinates. If missing, a numeric vector `1:...` is created having as many values as the rows of `x`.
#' @param xdomain,ydomain Character or numeric or `NULL` (default): vector of possible values of the variables represented in the `x`- and `y`-axes, in case the `x` or `y` argument is a character vector. The ordering of the values is respected. If `NULL`, then `unique(x)` or `unique(y)` is used.
#' @param xlim,ylim `NULL` (default) or a vector of two values. In the latter case, if any of the two values is not finite (including `NA` or `NULL`), then the `min` or `max` `x`- or `y`-coordinates of the plotted points are used.
#' @param grid Logical: whether to plot a light grid. Default `TRUE`.
#' @param alpha.f Numeric, default 1: opacity of the colours, `0` being completely invisible and `1` completely opaque.
#' @param xjitter,yjitter Logical or `NULL` (default): add [base::jitter()] to `x`- or `y`-values? Useful when plotting discrete variates. If `NULL`, jitter is added if the values are of character (or factor) class.
#' @param type,lty,lwd,pch,col,xlab,ylab,add,cex.main see analogous arguments in [graphics::matplot()].
#' @param ... Other parameters to be passed to [graphics::matplot()].
#'
#' @return `NULL`, [invisibly][base::invisible()]; produces a plot, see [graphics::matplot()].
#'
#' @seealso
#' [Pr()] to calculate posterior probabilities and quantiles.
#'
#' [plot.probability()] to directly plot posterior probabilities and quantiles contained in a probability object.
#'
#' [plotquantiles()] to plot quantile ranges.
#'
#' @examples
#' ## Scatter plot of the 'island' vs 'species' nominal variates of the penguins dataset;
#' ## note how jitter is automatically added:
#' flexiplot(x = penguins[, 'species'], y = penguins[, 'island'])
#'
#'
#' ## Scatter plot of the 'bill_len' vs 'species' variates of the penguins dataset:
#' flexiplot(x = penguins[, 'species'], y = penguins[, 'bill_len'])
#'
#' ## We can add jitter to separate the nominal values:
#' flexiplot(x = penguins[, 'species'], y = penguins[, 'bill_len'],
#'   xjitter = TRUE)
#'
#'
#' ## Scatter plot of the 'bill_len' vs 'body_mass' variates;
#' ## in this case we must specify the scatter-plot option `type = 'p'`:
#' flexiplot(x = penguins[, 'body_mass'], y = penguins[, 'bill_len'],
#'   type = 'p')
#'
#' ## Calculate the values of a normal distribution in a restricted range
#' x <- seq(from = -2, to = 2, length.out = 127)
#' y <- dnorm(x, mean = 0, sd = 1)
#'
#' ## plot the distribution, with 0 as the lower plot range:
#' flexiplot(x = x, y = y, ylim = c(0, NA))
#'
#' @import grDevices
#' @import graphics
#'
#' @export
flexiplot <- function(
    x, y,
    type = NULL,
    lty = c(1, 2, 4, 3, 6, 5),
    lwd = 2,
    pch = c(1, 2, 0, 5, 6, 3), #, 4,
    col = palette(),
    xlab = NULL, ylab = NULL,
    xlim = NULL, ylim = NULL,
    add = FALSE,
    xdomain = NULL, ydomain = NULL,
    alpha.f = 1,
    xjitter = NULL,
    yjitter = NULL,
    ## c( ## Tol's colour-blind-safe scheme
    ##     '#4477AA',
    ##     '#EE6677',
    ##     '#228833',
    ##     '#CCBB44',
    ##     '#66CCEE',
    ##     '#AA3377' #, '#BBBBBB'
    ## ),
    grid = TRUE,
    cex.main = 1,
    ...
){
    xat <- yat <- xaxp <- yaxp <- NULL

    if(!missing(x) && is.factor(x)){
        if(is.null(xlab)){ xlab <- deparse1(substitute(x)) }
        x <- as.character(x)
    }
    if(!missing(y) && is.factor(y)){
        if(is.null(ylab)){ ylab <- deparse1(substitute(y)) }
        y <- as.character(y)
    }

    if(missing('x') && !missing('y')){
        x <- y
        x[] <- rep(seq_len(NROW(y)), each = NCOL(y))
        if(is.null(ylab)){ ylab <- deparse1(substitute(y)) }
        if(is.null(yjitter)){ yjitter <- FALSE }
        if(is.null(xdomain) && is.null(xlim)){
            xat <- seq_len(NCOL(y))
            xdomain <- NA
            ## if(!is.null(xjitter)){
            ##     xlim <- range(x) + c(-0.04, 0.04)
            ## }
            if(is.null(xlab)){ xlab <- NA }
            if(is.null(type)){ type <- 'p' }
        }
    } else if(!missing('x') && missing('y')){
        y <- x
        y[] <- rep(seq_len(NROW(x)), each = NCOL(x))
        if(is.null(xlab)){ xlab <- deparse1(substitute(x)) }
        if(is.null(xjitter)){ xjitter <- FALSE }
        if(is.null(ydomain) && is.null(ylim)){
            yat <- seq_len(NCOL(x))
            ydomain <- NA
            ## if(!is.null(yjitter)){
            ##     ylim <- range(y) + c(-0.04, 0.04)
            ## }
            if(is.null(ylab)){ ylab <- NA }
            if(is.null(type)){ type <- 'p' }
        }
    } else if(!missing('x') && !missing('y')){
        if(is.null(xlab)){ xlab <- deparse1(substitute(x)) }
        if(is.null(ylab)){ ylab <- deparse1(substitute(y)) }
    } else {
        stop('Arguments "x" and "y" cannot both be missing')
    }

    if(NROW(y) == 1 && NCOL(y) == NCOL(x)){
        y <- rep(y, each = NROW(x))
        dim(y) <- dim(x)
        if(is.null(type)){ type <- 'p' }
    }
    if(NROW(x) == 1 && NCOL(x) == NCOL(y)){
        x <- rep(x, each = NROW(y))
        dim(x) <- dim(y)
        if(is.null(type)){ type <- 'p' }
    }

    if(is.character(x) && is.character(y)) {
        if(is.null(xjitter)){xjitter <- TRUE}
        if(is.null(yjitter)){yjitter <- TRUE}
    }

    ## if x is character, convert to numeric
    if(is.character(x)){
        if(is.null(xdomain)){ xdomain <- unique(x) }
        ## we assume the user has sorted the values in a meaningful order
        ## because the lexical order may not be correct
        ## (think of values like 'low', 'medium', 'high')
        . <- dim(x)
        x <- as.numeric(factor(x, levels = xdomain))
        dim(x) <- .
        xat <- seq_along(xdomain)
        xaxp <- c(range(xat), length(xat) - 1)
        if(is.null(type)){ type <- 'p' }
    }
    if(isTRUE(xjitter)){
        xaxp <- c(range(xat) + c(-0.5, 0.5), length(xat))
        ## xaxp <- c(range(xat), length(xat) - 1)
        x <- jitter(x, factor = 5/3)
    }

    ## if y is character, convert to numeric
    if(is.character(y)){
        if(is.null(ydomain)){ ydomain <- unique(y) }
        ## we assume the user has sorted the values in a meaningful order
        ## because the lexical order may not be correct
        ## (think of values like 'low', 'medium', 'high')
        . <- dim(y)
        y <- as.numeric(factor(y, levels = ydomain))
        dim(y) <- .
        yat <- seq_along(ydomain)
        yaxp <- c(range(yat), length(yat) - 1)
        if(is.null(type)){ type <- 'p' }
    }
    if(isTRUE(yjitter)){
        yaxp <- c(range(yat) + c(-0.5, 0.5), length(yat))
        ## yaxp <- c(range(yat), length(yat) - 1)
        y <- jitter(y, factor = 5/3)
    }

    ## Syntax of xlim and ylim that allows
    ## for the specification of only upper- or lower-bound
    if(length(xlim) == 2){
        if(is.null(xlim[1]) || !is.finite(xlim[1])){ xlim[1] <- min(x[is.finite(x)]) }
        if(is.null(xlim[2]) || !is.finite(xlim[2])){ xlim[2] <- max(x[is.finite(x)]) }
    }
    if(length(ylim) == 2){
        if(is.null(ylim[1]) || !is.finite(ylim[1])){ ylim[1] <- min(y[is.finite(y)]) }
        if(is.null(ylim[2]) || !is.finite(ylim[2])){ ylim[2] <- max(y[is.finite(y)]) }
    }

    if(is.null(type)){ type <- 'l' }

    if(is.na(alpha.f)){alpha.f <- 1}
    col <- adjustcolor(col, alpha.f = alpha.f)

    graphics::matplot(x, y, xlim = xlim, ylim = ylim, type = type, axes = FALSE,
        col = col, lty = lty, lwd = lwd, pch = pch, cex.main = cex.main,
        add = add, xlab = xlab, ylab = ylab, ...)
    if(!add){
        graphics::axis(1, at = xat, labels = xdomain, tick = !grid,
            col = 'black', lwd = 1, lty = 1, ...)
        graphics::axis(2, at = yat, labels = ydomain, tick = !grid,
            col = 'black', lwd = 1, lty = 1, ...)
        if(grid){
            ## Save and restore user's par()
            oldpar <- par(no.readonly = TRUE)
            on.exit(par(oldpar))
            if(!is.null('xaxp')){ par(xaxp = xaxp) }
            if(!is.null('yaxp')){ par(yaxp = yaxp) }
            graphics::grid(nx = NULL, ny = NULL, lty = 1, col = '#BBBBBB80')
        }
    }
    invisible()
}


#' Plot pairs of quantiles
#'
#' @description
#' Utility function to plot pairs of quantiles obtained with [Pr()].
#'
#' @param x Numeric or character: vector of x-coordinates. See [flexiplot()].
#' @param y Numeric: a matrix having as many rows as `x` and an even number of columns, with one column per quantile. Typically these quantiles have been obtained with [Pr()], as their `$quantiles` value. This value is a three-dimensional array, and one of its columns (corresponding to the possible values of the `X` argument of [Pr()]) or one of its rows (corresponding to the possible values of the `Y` argument of [Pr()]) should be selected before being used as `y` input.
#' @param xdomain Character or numeric or `NULL` (default): vector of possible values of the variable represented in the x-axis, if the `x` argument is a character vector. The ordering of the values is respected. If `NULL`, then `unique(x)` is used.
#' @param alpha.f Numeric, default 0.25: opacity of the quantile bands, `0` being completely invisible and `1` completely opaque.
#' @param col Fill colour of the quantile bands. Can be specified in any of the usual ways, see for instance [grDevices::col2rgb()]. Default `#4477AA`.
#' @param border Fill colour of the quantile bands. Can be specified in any of the usual ways, see for instance [grDevices::col2rgb()]. If `NA` (default), no border is drawn.
#' @param type see analogous argument in [flexiplot()].
#' @param ... Other parameters to be passed to [flexiplot()].
#'
#' @return `NULL`, [invisibly][base::invisible()]; produces a plot, see [graphics::matplot()].
#'
#' @seealso
#' [Pr()] to calculate posterior probabilities and quantiles.
#'
#' [plot.probability()] to directly plot posterior probabilities and quantiles contained in a probability object.
#'
#' [flexiplot()] for more general plots.
#'
#' @examples
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## create a grid of values for variate "bill length",
#' ## based on the information in the dataset and metadata:
#' values <- vrtgrid(vrt = 'bill_len', learnt = learnt)
#'
#' ## calculate the probabilities and quantiles
#' probs <- Pr(Y = data.frame(bill_len = values), learnt = learnt, parallel = 1)
#'
#' ## plot the quantiles, setting lower plot range to zero
#' plotquantiles(x = values, y = probs$quantiles[, 1, ], ylim = c(0, NA),
#'   xlab = 'bill length', ylab = 'probability')
#'
#' ## add a plot of the probabilities in thick dashed red
#' flexiplot(x = values, y = probs$values, lwd = 5, lty = 2, col = 2, add = TRUE)
#'
#' @import grDevices
#'
#' @export
plotquantiles <- function(
    x, y,
    xdomain = NULL,
    alpha.f = 0.25,
    col = palette(),
    ##     c( ## Tol's colour-blind-safe scheme
    ##     '#4477AA',
    ##     '#EE6677',
    ##     '#228833',
    ##     '#CCBB44',
    ##     '#66CCEE',
    ##     '#AA3377' #, '#BBBBBB'
    ## ),
    border = NA,
    type = 'n',
    ...
){
    ## ## TODO: modify so that a vertical plot is also possible
    if(!is.matrix(y) || ncol(y) %% 2 != 0) {
        stop('"y" must be a matrix with an even number of columns.')
    }
    nquant <- ncol(y)

    isfin <- ( (is.numeric(x) & is.finite(x)) | !is.na(x)) &
        apply(y, 1, function(xx){all(is.finite(xx))})
    x <- unname(x[isfin])
    y <- unname(y[isfin, , drop = FALSE])

    ##
    ## col[!grepl('^#',col)] <- palette()[as.numeric(col[!grepl('^#',col)])]
    ## if(is.na(alpha.f)){alpha.f <- 1}
    ## col <- adjustcolor(col, alpha.f = alpha.f)
    ## if(is.na(alpha)){alpha <- ''}
    ## else if(!is.character(alpha)){alpha <- alpha2hex(alpha)}
    ## if(!(is.na(col) | nchar(col)>7)){col <- paste0(col, alpha)}
    ##
    flexiplot(x = x, y = y, xdomain = xdomain, type = 'n',
        xjitter = FALSE, yjitter = FALSE, ...)

    ## if x is character, convert to numeric
    if(is.character(x)){
        if(is.null(xdomain)){ xdomain <- unique(x) }
        ## we assume the user has sorted the values in a meaningful order
        ## because the lexical order may not be correct
        ## (think of values like 'low', 'medium', 'high')
        x <- as.numeric(factor(x, levels = xdomain))
    }
    if(is.na(alpha.f)){alpha.f <- 1}
    col <- adjustcolor(col, alpha.f = alpha.f)
    for(ii in seq_len(nquant/2)) {
        graphics::polygon(x=c(x, rev(x)), y=c(y[,ii], rev(y[, nquant + 1 - ii])),
            col = col, border = border)
    }
    invisible()
}


#' Plot an object of class "probability"
#'
#' @description
#' This [base::plot()] method is a utility to plot probabilities obtained with [Pr()], as well as their variabilities. The probabilities are plotted either against `Y`, with one curve for each value of `X`, or vice versa.
#'
#' @param x Object of class "probability", obtained with [Pr()].
#' @param variability One of the values `'quantiles'`, `'samples'`, `'none'` (equivalent to `NA` or `FALSE`), or `NULL` (default), in which case the variability available in `p` is used. This argument chooses how to represent the variability of the probability; see [Pr()]. If the requested variability is not available in the object `p`, then a warning is issued and no variability is plotted.
#' @param subset Named list or named vector: which variate values to display. For the variates corresponding to the names in this list, only the vector of values corresponding to that variate is displayed.
#'
#' @param PvsY Logical or `NULL`: should probabilities be plotted against their `Y` argument? If `NULL`, the argument between `Y` and `X` having larger number of values is chosen. As many probability curves will be plotted as the number of values of the other argument.
#' @param legend One of the values `'bottomright'`, `'bottom'`, `'bottomleft'`, `'left'`, `'topleft'`, `'top'`, `'topright'`, `'right'`, `'center'` (see [graphics::legend()]): plot a legend at that position. A value `FALSE` or any other does not plot any legend. Default `'top'`.
#' @param alpha.f Numeric, default 0.25: opacity of the colours, `0` being completely invisible and `1` completely opaque.
#' @param var.alpha.f Numeric: opacity of the quantile bands or of the samples, `0` being completely invisible and `1` completely opaque.
#' @param lty,lwd,col,type,xlab,ylab,main,ylim,grid,add see analogous arguments in [graphics::matplot()]
#' @param ... Other parameters to be passed to [flexiplot()].
#'
#' @return `NULL`, [invisibly][base::invisible()]; produces a plot, see [graphics::matplot()].
#'
#' @seealso
#' [Pr()] to calculate posterior probabilities and quantiles.
#'
#' [hist.probability()] to plot the variability of the probabilities as a distribution.
#'
#' [flexiplot()] (on which `plot.probability()` is based) for more general plots.
#'
#' [plotquantiles()] to plot quantile ranges.
#'
#' @examples
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## create a grid of values for variate "bill length",
#' ## based on the information in the dataset and metadata:
#' values <- vrtgrid(vrt = 'bill_len', learnt = learnt)
#'
#' ## calculate the probabilities and quantiles
#' probs <- Pr(Y = data.frame(bill_len = values), learnt = learnt, parallel = 1)
#'
#' ## plot the probabilities and quantiles
#' plot(probs)
#'
#' @import grDevices
#'
#' @export
plot.probability <- function(
    x,
    variability = NULL,
    subset = NULL,
    PvsY = NULL,
    legend = 'top',
    lty = c(1, 2, 4, 3, 6, 5),
    lwd = 2,
    col = palette(),
    type = NULL,
    ##     c( ## Tol's colour-blind-safe scheme, or palette()
    ##     '#4477AA',
    ##     '#EE6677',
    ##     '#228833',
    ##     '#CCBB44',
    ##     '#66CCEE',
    ##     '#AA3377' #, '#BBBBBB'
    ## ),
    alpha.f = 1,
    var.alpha.f = NULL,
    xlab = NULL,
    ylab = NULL,
    main = NULL,
    ylim = c(0, NA),
    grid = TRUE,
    add = FALSE,
    ...
){
    ## Replace object x keeping only values given in 'subset'
    if(!is.null(subset)){
        x <- prsubset(x, subset = subset)
    }

    ## If there's only one probability it doesn't make sense to plot anything:
    ## print() the result instead
    if(length(x$values) == 1){
        return(print(x))
    }



    ## Check how we should represent the variability
    ## The user can choose among three options
    ## provided that option is available in argument 'x'
    if(is.null(variability)) { # User is not choosing
        ## We choose 'quantiles' or what's available
        if(!is.null(x$quantiles)) {
            variability <- 'quantiles'
        } else if(!is.null(x$samples)){
            variability <- 'samples'
        } else {
            variability <- 'none'
        }
    } else { # User is choosing
        if(is.na(variability) || isFALSE(variability)){ variability <- 'none'}

        ## handle shortenings
        variability <- match.arg(variability, c('quantiles', 'samples', 'none'))

        ## handle impossible requests
        if(
        (variability == 'quantiles' && is.null(x$quantiles)) ||
            (variability == 'samples' && is.null(x$samples))
        ) {
            message('Requested variability not available. Omitting its plot.')
            variability <- 'none'
        }
    }

    Ylen <- nrow(x$values)
    Xlen <- ncol(x$values)

    ## Rename the variability object so as to avoid if-else below
    if(variability == 'quantiles'){
        mainpercentiles <- c(5.5, 94.5) # By default we choose an 89% band
        pvar <- x$quantiles
        ## if we are plotting more than one curve, keep only the 89% band
        if(Xlen > 1 && Ylen > 1){
            qnames <- as.numeric(sub('%', '', dimnames(pvar)[[3]]))
            choosepercentiles <- sapply(mainpercentiles,
                function(xx){which.min(abs(qnames - xx))})
            pvar <- pvar[, , choosepercentiles, drop = FALSE]
            qnames <- as.numeric(sub('%', '', dimnames(pvar)[[3]]))
        }
        qnames <- as.numeric(sub('%', '', dimnames(pvar)[[3]]))
        if(is.null(var.alpha.f)){var.alpha.f <- 0.25}
    } else if(variability == 'samples'){
        pvar <- x$samples
        if(is.null(var.alpha.f)){var.alpha.f <- 1/ceiling(sqrt(dim(pvar)[3]))}
    } else {
        pvar <- NULL
    }

    ## Handle the case of missing Y and X items in 'x'
    if(is.null(x$Y)){
        x$Y <- data.frame(Y = paste0('Y', seq_len(Ylen)))
        if(Xlen > 1){
            x$X <- data.frame(X = paste0('X', seq_len(Xlen)))
        }
    }

    ## If 'PvsY' is NULL, then we guess that the longest between Y and X
    ## is meant to be abscissa
    if(is.null(PvsY)){ PvsY <- (Ylen >= Xlen) }

    if(isTRUE(PvsY)){
        xxx <- x$Y
        leg <- x$X
        tempxlab <- 'Y'
    } else {
        xxx <- x$X
        leg <- x$Y
        tempxlab <- 'X'
        x$values <- t(x$values)
        if(!is.null(pvar)){ pvar <- aperm(pvar, c(2, 1, 3)) }
    }

    ## If the abscissa has more than one variate,
    ## then it becomes tricky to understand which of these we must plot against
    ## Heuristic: if there's one variate with as many unique elements as xxx,
    ## then use that one. Otherwise use a generic 'Y...'
    if(ncol(xxx) == 1){
        tempxlab <- colnames(xxx)
        xxx <- unlist(xxx)
    } else {
        uniquevrts <- apply(xxx, 2, function(xx){length(unique(xx))})
        toselect <- which(uniquevrts == nrow(xxx))[1]
        if(is.na(toselect)){
            xxx <- seq_len(nrow(xxx))
        } else {
            tempxlab <- colnames(xxx)[toselect]
            xxx <- xxx[, toselect]
        }
    }

    if(is.null(xlab)){xlab <- tempxlab}
    if(missing(main)){
        main <- paste0('prob. or freq.  ',
            paste0(names(x$Y), collapse = ', '),
            if(!is.null(x$X)){paste0(' | ', paste0(names(x$X), collapse = ', '))}
        )
        if(variability == 'quantiles'){
            main <- paste0(main, '  [',
                paste0(round(qnames, 1), '%', collapse = ', '),
                ']')
        }
    }
    if(is.null(ylab)){
        ylab <- 'probability or rel.frequency'
    }

    if(is.null(type)){
        if(is.character(xxx)){type <- 'b'} else {type <- 'l'}
    }

    ## Plot the variability first
    ## find maximum and minimum y-value first, if needed
    if(is.na(ylim[2])){
        ylim[2] <- max(pvar, x$values)
    }
    if(is.na(ylim[1])){
        ylim[1] <- min(pvar, x$values)
    }
    if(variability == 'quantiles'){
        for(i in seq_len(dim(pvar)[2])){
            plotquantiles(x = unlist(xxx), y = pvar[, i, ],
                col = col[(i - 1) %% length(col) + 1],
                alpha.f = var.alpha.f,
                lty =  lty[(i - 1) %% length(lty) + 1],
                xlab = xlab,
                ylab = ylab,
                ylim = ylim,
                main = main,
                grid = grid,
                add = (add || i > 1),
                ...)
            add <- TRUE
        }

    } else if(variability == 'samples'){
        ## the samples are plotted alternating between the different subgroups,
        ## rather than one group at a time, in order to avoid that
        ## the samples of the last subgroup cover the previous ones
        nx <- dim(pvar)[2]
        dim(pvar) <- c(dim(pvar)[1], prod(dim(pvar)[-1]))
        flexiplot(x = xxx, y = pvar,
            type = type,
            col = col[(seq_len(nx) - 1) %% length(col) + 1],
            alpha.f = var.alpha.f,
            lty =  1, #lty[(seq_len(nx) - 1) %% length(lty) + 1],
            lwd = 0.5, #lwd[(seq_len(nx) - 1) %% length(lwd) + 1] / 4,
            xlab = xlab,
            ylab = ylab,
            ylim = ylim,
            main = main,
            grid = grid,
            xjitter = FALSE,
            yjitter = FALSE,
            add = add,
            ...)
        add <- TRUE
    }

    ## Plot the probabilities
    flexiplot(x = xxx, y = x$values,
        type = type,
        col = col,
        alpha.f = alpha.f,
        lty = lty,
        lwd = lwd,
        xlab = xlab,
        ylab = ylab,
        ylim = ylim,
        main = main,
        grid = grid,
        xjitter = FALSE,
        yjitter = FALSE,
        add = add,
        ...)

    ## Plot legends
    if(!is.null(leg)){
        ##  && is.character(legend) &&
        ## (legend %in%
        ##      c("bottomright", "bottom", "bottomleft", "left", "topleft",
        ##          "top", "topright", "right", "center"))
        graphics::legend(x = legend,
            legend = apply(leg, 1, function(xxx){
                paste0(paste0(names(xxx), ' = ', xxx), collapse = ', ')
            }),
            bty = 'n',
            col = col,
            lty = lty,
            lwd = lwd,
            ...)
    }
    invisible()
}


#' Plot the variability of an object of class "probability" as a histogram
#'
#' @description
#' The posterior probabilities calculated with the [Pr()] function, and outputted as a `probability` object, have an associated variability that comes from the finite size of the data sample. This variability can be interpreted in two ways:
#'
#' - How the probabilities would change, if we could collect a very large (infinite) amount of additional data, and how likely would such change be;
#' - The relative frequency of a particular variate value in the full (sampled and unsampled) population is unknown; we can quantify our uncertainty about this relative frequency with a probability distribution.
#'
#' The `hist()` method for a `probability` object is a utility to visualize this kind of variability, in the form of a distribution.
#'
#' @param x Object of class "probability", obtained with [Pr()].
#' @param subset Named list or named vector: which variate values to display. For the variates corresponding to the names in this list, only the vector of values corresponding to that variate is displayed.
#' @param breaks `NULL` or as in function [graphics::hist()]. If `NULL` (default), an optimal number of breaks for each probability distribution is computed.
#' @param fill.alpha.f Numeric, default 0.125: opacity of the histogram filling. `0` means no filling.
#' @param legend One of the values `"bottomright"`, `"bottom"`, `"bottomleft"`, `"left"`, `"topleft"`, `"top"`, `"topright"`, `"right"`, `"center"` (see [graphics::legend()]): plot a legend at that position. A value `FALSE` or any other does not plot any legend. Default `"top"`.
#' @param showmean Logical, default `TRUE`: show the means of the probability distributions? The means correspond to the probabilities about the next observed unit.
#' @param lty,lwd,col,alpha.f,xlab,ylab,xlim,ylim,main,grid,add see analogous arguments in [graphics::matplot()]
#' @param ... Other parameters to be passed to [flexiplot()].
#'
#' @return [Invisibly][base::invisible()], an object of class ["histogram"][graphics::hist()].
#'
#' @seealso
#' [Pr()] to calculate posterior probabilities and quantiles.
#'
#' [plot.probability()] to plot the posterior probabilities.
#'
#' [flexiplot()] (on which `hist.probability()` is based) for more general plots.
#'
#' [plotquantiles()] to plot quantile ranges.
#'
#' @examples
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## calculate the probability, and its variability,
#' ## for the value 'Adelie' of the "species" variate
#' probs <- Pr(Y = data.frame(species = 'Adelie'), learnt = learnt, parallel = 1)
#' probs$values
#'
#' ## show the variability of this probability; equivalently show
#' ## the probability distribution for the relative frequency of
#' ## 'Adelie' penguins in the full population
#' hist(probs, legend = 'topright')
#'
#' @import grDevices
#'
#' @export
hist.probability <- function(
    x,
    subset = NULL,
    breaks = NULL,
    legend = 'top',
    lty = c(1, 2, 4, 3, 6, 5),
    lwd = 2,
    col = palette(),
    alpha.f = 1,
    fill.alpha.f = 0.125,
    showmean = TRUE,
    ##     c( ## Tol's colour-blind-safe scheme, or palette()
    ##     '#4477AA',
    ##     '#EE6677',
    ##     '#228833',
    ##     '#CCBB44',
    ##     '#66CCEE',
    ##     '#AA3377' #, '#BBBBBB'
    ## ),
    xlab = NULL,
    ylab = NULL,
    xlim = NULL,
    ylim = c(0, NA),
    main = NULL,
    grid = TRUE,
    add = FALSE,
    ...
){
    ## Replace object x keeping only values given in 'subset'
    if(!is.null(subset)){
        x <- prsubset(x, subset = subset)
    }

    ## Check that samples are available in the probability object
    if(is.null(x$samples)) {
        stop('The probability object does not contain any frequency samples')
        }
    pvar <- x$samples
    Ylen <- nrow(x$values)
    Xlen <- ncol(x$values)

    if(is.null(breaks)){n <- ceiling(sqrt(dim(pvar)[3])/2)} else {n <- NULL}

    ## Precompute histograms, to determine maximum y-value
    midslist <- densitylist <- list()
    i <- 0L
    for(xx in seq_len(Xlen)){ for(yy in seq_len(Ylen)){
        i <- i + 1L
        ff <- pvar[yy, xx, ]
        rg <- range(ff)
        if(diff(rg)==0){rg <- c(0, 1)}
        if(!is.null(n)){ breaks <- seq(rg[1], rg[2], length.out = n + 1) }
        hd <- graphics::hist(x = ff, breaks = breaks, plot = FALSE)
        midslist[[i]] <- hd$mids
        densitylist[[i]] <- hd$density
    } }

    if(is.null(xlab)){
        if(!is.null(x$Y)){
            xlab <- 'relative frequency'
        } else {
            xlab <- 'quantile'
        }
    }
    if(is.null(ylab)){ylab <- 'probability density'}
    if(isFALSE(fill.alpha.f) || !is.numeric(fill.alpha.f)){fill.alpha.f <- 0}

    if(missing(xlim)){xlim <- range(unlist(midslist))}
    if(is.na(ylim)[2]){ylim[2] <- max(unlist(densitylist))}

    i <- 0L
    for(xx in seq_len(Xlen)){ for(yy in seq_len(Ylen)){
        i <- i + 1L
        midx <- midslist[[i]]
        y <- densitylist[[i]]
        thiscol <- col[(i - 1) %% length(col) + 1]
        thislty <- lty[(i - 1) %% length(lty) + 1]
        if(alpha.f > 0){
            plotquantiles(x = midx, y = cbind(rep(0, length(y)), y),
                col = thiscol,
                alpha.f = fill.alpha.f,
                xlab = xlab, ylab = ylab,
                xlim = xlim, ylim = ylim,
                main = main,
                grid = grid,
                lty = 0,
                add = (add || i > 1),
                ...)
        }

        flexiplot(x = midslist[[i]], y = densitylist[[i]],
            xlab = xlab, ylab = ylab,
            xlim = xlim, ylim = ylim,
            main = main,
            col = thiscol,
            alpha.f = alpha.f,
            lty = thislty,
            lwd = lwd,
            grid = grid,
            xjitter = FALSE,
            yjitter = FALSE,
            add = (add || alpha.f >0 || i > 1),
            ...
        )
        if(isTRUE(showmean)){
            graphics::abline(v = x$values[yy, xx],
                col = adjustcolor(thiscol, alpha.f * 0.75),
                lty = thislty,
                lwd = lwd * 0.75)
        }
    } }

    ## Plot legends
    if(is.character(legend) &&
           (legend %in%
                 c("bottomright", "bottom", "bottomleft", "left", "topleft",
                     "top", "topright", "right", "center"))){
        if(!is.null(x$Y)){
        legs <- paste0(apply(x$Y, 1, function(xxx){
                paste0(paste0(names(xxx), ' = ', xxx), collapse = ', ')
        }))
        } else {
        legs <- paste0(apply(as.data.frame(x$pY), 1, function(xxx){
                paste0(paste0(xxx, '-qtl'), collapse = ', ')
        }))
        }
        if(!is.null(x$X)){
            legs <- c(outer( legs,
                paste0(' | ',
                    apply(x$X, 1, function(xxx){
                        paste0(paste0(names(xxx), ' = ', xxx), collapse = ', ')
                    })
                ),
                paste0))
        }

        graphics::legend(x = legend,
            legend = legs,
            bty = 'n',
            col = col,
            lty = lty,
            lwd = lwd,
            ...)
    }
    ## Return output of initial hist()
    invisible(hd)
}


#' Print an object of class "probability"
#'
#' @description
#' This [base::print()] method is a utility to display selected elements of a "probability" object obtained with [Pr()]; typically its posterior probabilies (element `$values`) and their variabilities (element `$quantiles`). If the `Y` or `X` variates are joint variates, this method also allow to display only selected values of them
#'
#' @param x Object of class "probability", obtained with [Pr()].
#' @param elements character or integer vector, or `NULL` (default): elements of the "probability" object to display. The syntax is the same as with [` [ `][base::Extract]. If `NULL`, the elements `$values` and `$quantiles` are displayed together in a special way.
#' @param subset Named list or named vector: which variate values to display. For the variates corresponding to the names in this list, only the vector of values corresponding to that variate is displayed.
#' @param digits positive number or `NULL` or `TRUE` (default): minimal number of significant digits, see [base::print.default()]. If value is `TRUE`, then the significant digits for elements `$values` and `$quantiles` are determined from their respective `$values.MCaccuracy` and `$quantiles.MCaccuracy` elements of the `probability` object, see [Pr()]; whereas `$samples` elements use 2 significant digits.
#' @param ... Other parameters to be passed to [base::print()].
#'
#' @return Its `x` argument, [invisibly][base::invisible()]; see [base::print()].
#'
#' @seealso
#' [Pr()] to calculate posterior probabilities and quantiles.
#'
#' [plot.probability()] to plot probabilities and quantiles calculated by `Pr()'.
#' [hist.probability()] to plot the variability of the probabilities as a distribution.
#'
#' @examples
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## Calculate the 3 x 2 probabilities for the 3 species
#' ## given bill-lengths of 43 mm and 44 mm
#'
#' Y <- data.frame(species = c('Adelie', 'Chinstrap', 'Gentoo'))
#' X <- data.frame(bill_len = c(43, 44))
#'
#' probs <- Pr(Y = Y, X = X, learnt = learnt, parallel = 1)
#'
#' ## display the values and variabilities of these probabilities
#' print(probs)
#'
#' ## diplay 'values' only, and only for the species value 'Gentoo'
#' print(probs, elements = 'values', subset = list(species = 'Gentoo'))
#'
#' @export
print.probability <- function(
    x,
    elements = NULL,
    subset = NULL,
    digits = TRUE,
    ...
){
    ## Replace object x keeping only values given in 'subset'
    if(!is.null(subset)){
        x <- prsubset(x, subset = subset)
    }

    if(is.null(elements)){
        ## rearrange and combine values and quantiles in a special way

        if(isTRUE(digits)){
            temp <- aperm(
                a = array(signif(x = c(x$values, x$quantiles),
                    digits = 1 + log10(
                        c(x$values / x$values.MCaccuracy,
                            x$quantiles / x$quantiles.MCaccuracy)
                    )),
                    dim = dim(x$quantiles) + c(0, 0, 1),
                    dimnames = c(
                        dimnames(x$values),
                        list(`prob. & vrb.` = c('value', paste0('Q', dimnames(x$quantiles)[[3]])))
                    ) ), perm = c(1,3,2))

            if(is.null(x$X)){temp <- temp[,,]}

            print(x = temp, ...)
        }else{
            temp <- aperm(
                a = array(c(x$values, x$quantiles),
                    dim = dim(x$quantiles) + c(0, 0, 1),
                    dimnames = c(
                        dimnames(x$values),
                        list(`prob. & vrb.` = c('value', paste0('Q', dimnames(x$quantiles)[[3]])))
                    ) ), perm = c(1,3,2))

            if(is.null(x$X)){temp <- temp[,,]}

            print(x = temp, digits = digits, ...)
        }
    } else {
        if(isTRUE(digits)){
            if('values' %in% elements){
                x$values <- signif(x = x$values,
                    digits = 1 + log10(c(x$values / x$values.MCaccuracy)) )
            }
            if('quantiles' %in% elements){
                x$quantiles <- signif(x = x$quantiles,
                    digits = 1 + log10(c(x$quantiles / x$quantiles.MCaccuracy)) )
            }
            if('samples' %in% elements){
                x$samples <- signif(x = x$samples, digits = 2)
            }
        print(x = x[elements], ...)
        } else {
        print(x = x[elements], digits = digits, ...)
        }
    }
    invisible(x)
}

#' Build augmented metadata file
#'
#' Builds an augmented metadata [data frame][base::data.frame()] from the metadata and data given to 'learn()'. This augmented metadata object is saved in the 'learnt' object produced by 'learn()'.
#'
#' In addition to the original metadata it contains info about transformed variates and their domains, estimated location- and scale-parameters, and similar metadata.
#'
#' Used in 'learn()'.
#'
#' @param data data.frame object
#' @param metadata data.frame object
#' @param Dthreshold Positive number: threshold of fraction
#'   of unique datapoints to total datapoints, to decide
#'   whether to treat a rounded variate as continuous
#' @param tscalefactor Positive number: scaling factor for variate conversion
#'
#' @return A [data frame][base::data.frame()] with auxmetadata.
#' @import stats
#' @keywords internal
buildauxmetadata <- function(data, metadata, Dthreshold = 1, tscalefactor = 4.266) {

    ## In the internal, rescaled representation,
    ## with the SD of the means equal to 3 and
    ## the rate of the variances equal to 1,
    ## the median of Q3 of the possible F distribs is 1.95
    ## and its mean is 2.55, while
    ## the median of the IQR is 3.35
    ## and its mean is 5.10.
    ## The MAD of the prior for the next unit is 3.67
    ## The IQR of the prior for the next unit is 4.96
    ## The prior median of the IQRs of the components is 1.349
    ## The prior quartiles of the IQRs of the components are 0.559 & 3.25
    ## The prior 0.89-quantile of the IQRs of the components is 7.72
    ## half-order of magnitude smaller than the IQR-median is 4.26585
    ## which is approx 80.5%-quantile of the component-IQR distribution

    Othreshold <- 10

    idR <- idC <- idD <- idB <- idO <- idN <- 1L
    posN <- posO <- 0L

#### Prepare empty metadata frame
    auxmetadata <- as.data.frame(list(
        name = NA, type = NA, mcmctype = NA, id = NA, # censored= NA,
        transform = NA,
        Nvalues = NA_integer_, # keep this as integer
        indexpos = NA_integer_, # keep this as integer
        halfstep = NA,
        domainmin = NA, domainmax = NA,
        minincluded = NA, maxincluded = NA,
        tdomainmin = NA, tdomainmax = NA,
        domainminplushs = NA, domainmaxminushs = NA,
        tdomainminplushs = NA, tdomainmaxminushs = NA,
        tlocation = NA, tscale = NA,
        plotmin = NA, plotmax = NA
    ))[-1, ]

    ## Correct for missing columns in metadata
    for(column in c('domainmin', 'domainmax', 'datastep',
        'minincluded', 'maxincluded',
        'V1', 'V2')){
        if(is.null(metadata[[column]])){ metadata[[column]] <- NA }
    }

    for (name in metadata$name) {
        minfo <- as.list(metadata[metadata$name == name, ])
        ## make sure 'type' is lowercase
        minfo$type <- tolower(minfo$type)
        transf <- 'identity' # temporary
        datavalues <- minfo[grep('^V[0-9]+$', names(minfo))]
        ndatavalues <- sum(!is.na(datavalues))
        domainmin <- as.numeric(minfo$domainmin)
        domainmax <- as.numeric(minfo$domainmax)
        domainminplushs <- NA
        tdomainminplushs <- NA
        domainmaxminushs <- NA
        tdomainmaxminushs <- NA
        Nvalues <- NA_integer_
        indexpos <- NA_integer_
        halfstep <- as.numeric(minfo$datastep) / 2
        if(!is.null(data)) {
            x <- data[[name]]
            x <- x[!is.na(x)]
        }
        ## Nvalues <- minfo$Nvalues
        ## plotmin <- minfo$plotmin
        ## plotmax <- minfo$plotmax

        ## Catch discrepancies in specification
        if(ndatavalues > 0 && (
            !is.na(domainmin) || !is.na(domainmax) || !is.na(halfstep)
        )) {
            stop('Variate "',
                name, '": Please specify the domain either as a list of values',
                '(fields "V1", "V2", ...)\nor as a combination of',
                '"domainmin", "domainmax", "datastep";',
                'not both.')
        }
        if(minfo$type %in% c('nominal', 'ordinal')) {
            if(ndatavalues == 0 && (
            is.na(domainmin) || is.na(domainmax) || is.na(halfstep)
            )) {
            stop('Variate "',
                name, '": Insufficient domain information.\n',
                'Please specify domain either as a list of values',
                '(fields "V1", "V2", ...)\nor as a combination of',
                '"domainmin", "domainmax", "datastep".')
            }
            if(ndatavalues > 0){
                Nvalues <- ndatavalues
            } else {
                Nvalues <- length(seq(domainmin, domainmax, by = 2 * halfstep))
            }
        }

        ## Identify binary variate
        if(minfo$type %in% c('nominal', 'ordinal') && Nvalues == 2) {
            mcmctype <- 'B'
            id <- idB
            idB <- idB + 1L
            if(ndatavalues != 2) {
                datavalues <- as.list(as.character(
                    c(domainmin, domainmax)
                ))
                names(datavalues) <- paste0('V', seq_len(2))
            }
            halfstep <- NA
            domainmin <- NA
            domainmax <- NA
            tdomainmin <- 0 # r/dbeta takes 0/1 inputs
            tdomainmax <- 1
            tlocation <- 0
            tscale <- 1
            plotmin <- NA
            plotmax <- NA
            minincluded <- NA
            maxincluded <- NA

        } else if (minfo$type == 'nominal') {
            ## nominal variate
            mcmctype <- 'N'
            id <- idN
            idN <- idN + 1L
            indexpos <- posN
            posN <- posN + Nvalues
            halfstep <- NA
            domainmin <- NA
            domainmax <- NA
            tdomainmin <- 1  # Nimble indexes categorical from 1
            tdomainmax <- Nvalues
            tlocation <- 0
            tscale <- 1
            plotmin <- NA
            plotmax <- NA
            minincluded <- NA
            maxincluded <- NA

        } else if (minfo$type == 'ordinal') {
            ## Ordinal variates can be specified in two different ways

### Consistency checks:
            if(ndatavalues > Othreshold) {
                stop('Variate "',
                    name, '": we cannot handle lists of values larger than ',
                    Othreshold, '.',
                    '\nPlease specifies its values as a numeric range')
            }

            ## Ordinal variates are internally handled in two different ways:
            ## if their values are less than Othreshold: type 'O', like nominal
            ## otherwise: type 'D', via a latent variable like rounded

### ordinal type 'O'
            if(Nvalues <= Othreshold) {
                mcmctype <- 'O'
                id <- idO
                idO <- idO + 1L
                indexpos <- posO
                posO <- posO + Nvalues
                ## convert domain specification to list
                if(ndatavalues == 0) {
                    datavalues <- as.list(as.character(
                        seq(domainmin, domainmax, length.out = Nvalues)
                    ))
                    names(datavalues) <- paste0('V', seq_len(Nvalues))
                }
                domainmin <- NA
                domainmax <- NA
                halfstep <- NA
                tdomainmin <- 1 # Nimble indexes categorical from 1
                tdomainmax <- Nvalues
                tlocation <- 0
                tscale <- 1
                plotmin <- NA
                plotmax <- NA
                minincluded <- TRUE
                maxincluded <- TRUE

### ordinal type 'D'
            } else {
                mcmctype <- 'D'
                id <- idD
                idD <- idD + 1L
                transf <- 'identity'
                if(!is.null(data)) {
                    ## set the location to the nearest centre of datastep interval
                    tempvalue <- quantile(x = x, probs = 0.5, type = 6, names = FALSE)
                    tlocation <- x[which.min(abs(x - tempvalue))]
                    tlocation <- tlocation +
                        round((tempvalue - tlocation) / (2 * halfstep)) *
                        2 * halfstep
                    tscale <- IQR(type = 6, na.rm = TRUE, x = x) / tscalefactor
                    if(tscale == 0){tscale <- 1 / tscalefactor}
                    plotmin <- min(x[is.finite(x)])
                    plotmin <- max(plotmin - IQR(x, type = 6) / 2,
                    (plotmin + domainmin) / 2, na.rm = TRUE)
                    plotmax <- max(x[is.finite(x)])
                    plotmax <- min(plotmax + IQR(x, type = 6) / 2,
                    (plotmax + domainmax) / 2, na.rm = TRUE)
                } else {
                    tlocation <- mean(domainmin, domainmax)
                    tscale <- ((domainmax - domainmin) / sqrt(12)) / tscalefactor
                    plotmin <- (31 * domainmin + domainmax) / 32
                    plotmax <- (31 * domainmax + domainmin) / 32
                }
                tdomainmin <- (domainmin - tlocation) / tscale
                tdomainmax <- (domainmax - tlocation) / tscale

                domainminplushs <- domainmin + halfstep
                domainmaxminushs <- domainmax - halfstep
                tdomainminplushs <- (domainminplushs - tlocation) / tscale
                tdomainmaxminushs <- (domainmaxminushs - tlocation) / tscale

                ## plotmin <- max(domainmin, plotmin)
                ## centre plotmin on a rounding bin
                tempvalue <- plotmin
                plotmin <- x[which.min(abs(x - tempvalue))]
                plotmin <- plotmin +
                    round((tempvalue - plotmin) / (2 * halfstep)) *
                    2 * halfstep

                ## plotmax <- min(domainmax, plotmax)
                ## centre plotmax on a rounding bin
                tempvalue <- plotmax
                plotmax <- x[which.min(abs(x - tempvalue))]
                plotmax <- plotmax +
                    round((tempvalue - plotmax) / (2 * halfstep)) *
                    2 * halfstep

                minincluded <- TRUE
                maxincluded <- TRUE
            }

#### Continuous variate (R,C,D)
        } else if (minfo$type == 'continuous') {
            Nvalues <- NA_integer_
            ## Rounded variates
            ## Empty datastep means 0
            if (length(halfstep) == 0 || is.null(halfstep) || is.na(halfstep)) {
                halfstep <- 0
            }
            ## Empty domainmin/max values mean -+Inf
            if(length(domainmin) == 0 || is.null(domainmin) || is.na(domainmin)) {
                domainmin <- -Inf
            }
            if(length(domainmax) == 0 || is.null(domainmax) || is.na(domainmax)) {
                domainmax <- +Inf
            }
            minincluded <- (tolower(minfo$minincluded) %in%
                                c('true', 't', 'yes', 'y', '1'))
            maxincluded <- (tolower(minfo$maxincluded) %in%
                                c('true', 't', 'yes', 'y', '1'))
            ## If the variate is rounded,
            ## we avoid a  latent-variable representation
            ## if the datapoints are "enough distinct" in higher dimension
            if(halfstep > 0 &&
               (is.null(data) ||
                nrow(unique(data)) / nrow(data) > Dthreshold)) {
                halfstep <- 0
            }

            if(!is.null(data)) {
                plotmin <- min(x[is.finite(x)])
                plotmax <- max(x[is.finite(x)])
            }

            ## Right now, some algorithms to set particular values
            ## are the same for all cases below,
            ## so they could be collected here
            ## But they are left in the individual cases
            ## in the event that they need a case-by-case tretment
            ## in the future

### Rounded continuous case
            if (halfstep > 0) {

                mcmctype <- 'D'
                id <- idD
                idD <- idD + 1L
                ##
                transf <- 'identity'
                if(!is.null(data)) {
                    ## set the location to the nearest centre of datastep interval
                    tempvalue <- quantile(x = x, probs = 0.5, type = 6, names = FALSE)
                    tlocation <- x[which.min(abs(x - tempvalue))]
                    tlocation <- tlocation +
                        round((tempvalue - tlocation) / (2 * halfstep)) *
                        2 * halfstep
                    tscale <- IQR(type = 6, na.rm = TRUE, x = x) / tscalefactor
                    if(tscale == 0){tscale <- 1 / tscalefactor}

                    plotmin <- max(plotmin - IQR(x, type = 6) / 2,
                    (plotmin + domainmin) / 2, na.rm = TRUE)
                    plotmax <- min(plotmax + IQR(x, type = 6) / 2,
                    (plotmax + domainmax) / 2, na.rm = TRUE)

                } else {
                    tlocation <- 0
                    tscale <- 1 / tscalefactor
                    plotmin <- max((31 * domainmin + domainmax) / 32,
                        -3 * tscale)
                    plotmax <- min((31 * domainmax + domainmin) / 32,
                        3 * tscale)
                }
                tdomainmin <- (domainmin - tlocation) / tscale
                tdomainmax <- (domainmax - tlocation) / tscale

                ## For a rounded variate it does not matter whether
                ## a domain boundary is included or excluded
                ## hence 'domainmin/max' are ignored
                domainminplushs <- domainmin + halfstep
                domainmaxminushs <- domainmax - halfstep
                ## if(minfo$minincluded){
                ## } else {
                ##     domainminplushs <- domainmin + halfstep * 3
                ## }
                ## if(minfo$maxincluded){
                ## } else {
                ##     domainmaxminushs <- domainmin - halfstep * 3
                ## }
                tdomainminplushs <- (domainminplushs - tlocation) / tscale
                tdomainmaxminushs <- (domainmaxminushs - tlocation) / tscale

                ## plotmin <- max(domainmin, plotmin)
                ## centre plotmin on a rounding bin
                tempvalue <- plotmin
                plotmin <- x[which.min(abs(x - tempvalue))]
                plotmin <- plotmin +
                    round((tempvalue - plotmin) / (2 * halfstep)) *
                    2 * halfstep

                ## plotmax <- min(domainmax, plotmax)
                ## centre plotmax on a rounding bin
                tempvalue <- plotmax
                plotmax <- x[which.min(abs(x - tempvalue))]
                plotmax <- plotmax +
                    round((tempvalue - plotmax) / (2 * halfstep)) *
                    2 * halfstep

            } else {
### Non-rounded continuous cases
                ## variate transformation to real-line domain
                if (is.finite(domainmin) && is.finite(domainmax) &&
                    minincluded && !maxincluded) {
                    ## doubly-bounded left-closed domain
                    mcmctype <- 'C'
                    id <- idC
                    idC <- idC + 1L
                    ##
                    transf <- 'logminus'
                    if(!is.null(data)) {
                        tlocation <- quantile(x = -log(domainmax - x), probs = 0.5,
                            type = 6, names = FALSE)
                        tscale <- IQR(type = 6, na.rm = TRUE,
                            x = -log(domainmax - x)) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- max(plotmin - IQR(x, type = 6) / 2,
                            domainmin, na.rm = TRUE)
                        plotmax <- min(plotmax + IQR(x, type = 6) / 2,
                        (plotmax + domainmax) / 2, na.rm = TRUE)

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- domainmin
                        plotmax <- (31 * domainmax + domainmin) / 32
                    }
                    tdomainmin <- (-log(domainmax - domainmin) -
                                                 tlocation) / tscale
                    tdomainmax <- +Inf


                } else if (is.finite(domainmin) && is.finite(domainmax) &&
                           !minincluded && maxincluded) {
                    ## doubly-bounded right-closed domain
                    mcmctype <- 'C'
                    id <- idC
                    idC <- idC + 1L
                    ##
                    transf <- 'log'
                    if(!is.null(data)) {
                        tlocation <- quantile(x = log(x - domainmin), probs = 0.5,
                            type = 6, names = FALSE)
                        tscale <- IQR(type = 6, na.rm = TRUE, x = log(x - domainmin)) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- max(plotmin - IQR(x, type = 6) / 2,
                            (plotmin + domainmin) / 2, na.rm = TRUE)
                        plotmax <- min(plotmax + IQR(x, type = 6) / 2,
                            domainmax, na.rm = TRUE)

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- (31 * domainmin + domainmax) / 32
                        plotmax <- domainmax
                    }
                    tdomainmin <- -Inf
                    tdomainmax <- (log(domainmax - domainmin) - tlocation) / tscale

                } else if (is.finite(domainmin) && is.finite(domainmax) &&
                           minincluded && maxincluded) {
                    ## doubly-bounded closed domain
                    mcmctype <- 'C'
                    id <- idC
                    idC <- idC + 1L
                    ##
                    transf <- 'identity'
                    if(!is.null(data)) {
                        tlocation <- quantile(x = x, probs = 0.5,
                            type = 6, names = FALSE)
                        tscale <- IQR(type = 6, na.rm = TRUE, x = x) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- max(plotmin - IQR(x, type = 6) / 2,
                            domainmin, na.rm = TRUE)
                        plotmax <- min(plotmax + IQR(x, type = 6) / 2,
                            domainmax, na.rm = TRUE)

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- domainmin
                        plotmax <- domainmax
                    }
                    tdomainmin <- (domainmin - tlocation) / tscale
                    tdomainmax <- (domainmax - tlocation) / tscale

                } else if (is.finite(domainmin) && minincluded) {
                    ## left-bounded left-closed domain
                    mcmctype <- 'C'
                    id <- idC
                    idC <- idC + 1L
                    ##
                    transf <- 'identity'
                    if(!is.null(data)) {
                        tlocation <- quantile(x = x, probs = 0.5,
                            type = 6, names = FALSE)
                        tscale <- IQR(type = 6, na.rm = TRUE, x = x) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- max(plotmin - IQR(x, type = 6) / 2,
                            domainmin, na.rm = TRUE)
                        plotmax <- plotmax + IQR(x, type = 6) / 2

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- domainmin
                        plotmax <- domainmin + 6
                    }
                    tdomainmin <- (domainmin - tlocation) / tscale
                    tdomainmax <- +Inf

                } else if (is.finite(domainmax) && maxincluded) {
                    ## right-bounded right-closed domain
                    mcmctype <- 'C'
                    id <- idC
                    idC <- idC + 1L
                    ##
                    transf <- 'identity'
                    if(!is.null(data)) {
                        tlocation <- quantile(x = x, probs = 0.5,
                            type = 6, names = FALSE)
                        tscale <- IQR(type = 6, na.rm = TRUE, x = x) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- plotmin - IQR(x, type = 6) / 2
                        plotmax <- min(plotmax + IQR(x, type = 6) / 2,
                            domainmax, na.rm = TRUE)

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- domainmax - 6
                        plotmax <- domainmax
                    }
                    tdomainmin <- -Inf
                    tdomainmax <- (domainmax - tlocation) / tscale

                } else if (is.finite(domainmin) && is.finite(domainmax) &&
                           !minincluded && !maxincluded) {
                    ## doubly-bounded open domain
                    mcmctype <- 'R'
                    id <- idR
                    idR <- idR + 1L
                    ##
                    transf <- 'Q'
                    if(!is.null(data)) {
                        tlocation <- median(
                            util_Q(0.5 +
                                       (x - (domainmax + domainmin) / 2) /
                                       (domainmax - domainmin)
                            ))
                        tscale <- IQR(type = 6, na.rm = TRUE, x =
                            util_Q(0.5 +
                                       (x - (domainmax + domainmin) / 2) /
                                       (domainmax - domainmin)
                            )) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- max(plotmin - IQR(x, type = 6) / 2,
                        (plotmin + domainmin) / 2, na.rm = TRUE)
                        plotmax <- min(plotmax + IQR(x, type = 6) / 2,
                        (plotmax + domainmax) / 2, na.rm = TRUE)

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- (31 * domainmin + domainmax) / 32
                        plotmax <- (31 * domainmax + domainmin) / 32

                    }
                    tdomainmin <- -Inf
                    tdomainmax <- +Inf

                } else if (is.finite(domainmin) && !minincluded) {
                    ## left-bounded left-open domain
                    mcmctype <- 'R'
                    id <- idR
                    idR <- idR + 1L
                    ##
                    transf <- 'log'
                    if(!is.null(data)) {
                        tlocation <- quantile(x = log(x - domainmin), probs = 0.5,
                            type = 6, names = FALSE)
                        tscale <- IQR(type = 6, na.rm = TRUE, x = log(x - domainmin)) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- max(plotmin - IQR(x, type = 6) / 2,
                        (plotmin + domainmin) / 2, na.rm = TRUE)
                        plotmax <- plotmax + IQR(x, type = 6) / 2

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- domainmin + 1
                        plotmax <- domainmin + 7
                    }
                    tdomainmin <- -Inf
                    tdomainmax <- +Inf

                } else if (is.finite(domainmax) && !maxincluded) {
                    ## right-bounded right-open domain
                    mcmctype <- 'R'
                    id <- idR
                    idR <- idR + 1L
                    ##
                    transf <- 'logminus'
                    if(!is.null(data)) {
                        tlocation <- quantile(x = -log(domainmax - x), probs = 0.5,
                            type = 6, names = FALSE)
                        tscale <- IQR(type = 6, na.rm = TRUE, x = -log(domainmax - x)) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- plotmin - IQR(x, type = 6) / 2
                        plotmax <- min(plotmax + IQR(x, type = 6) / 2,
                        (plotmax + domainmax) / 2, na.rm = TRUE)

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- domainmax - 7
                        plotmax <- domainmax - 1
                    }
                    tdomainmin <- -Inf
                    tdomainmax <- +Inf

                } else {
                    ## unbounded, open domain
                    mcmctype <- 'R'
                    id <- idR
                    idR <- idR + 1L
                    ##
                    transf <- 'identity'
                    if(!is.null(data)) {
                        tlocation <- quantile(x = x, probs = 0.5,
                            type = 6, names = FALSE)
                        tscale <- IQR(type = 6, na.rm = TRUE, x = x) / tscalefactor
                        if(tscale == 0){tscale <- 1 / tscalefactor}

                        plotmin <- plotmin - IQR(x, type = 6) / 2
                        plotmax <- plotmax + IQR(x, type = 6) / 2

                    } else {
                        tlocation <- 0
                        tscale <- 1 / tscalefactor
                        plotmin <- -3
                        plotmax <- 3
                    }
                    tdomainmin <- -Inf
                    tdomainmax <- +Inf
                }

                domainminplushs <- domainmin
                domainmaxminushs <- domainmax
                tdomainminplushs <- tdomainmin
                tdomainmaxminushs <- tdomainmax

            }
            ## end non-rounded case


### end continuous case
        } else {
            stop('ERROR: unknown variate type for "', name, '"')
        }
        ## Debugging
        ## print(auxmetadata[nrow(auxmetadata)])
        ## print(as.data.table(c(list(name=name, type=mcmctype, transform=transf,
        ## Nvalues=Nvalues, halfstep=halfstep,
        ## domainmin=domainmin, domainmax=domainmax,
        ## censormin=censormin, censormax=censormax, tlocation=tlocation,
        ## tscale=tscale, plotmin=plotmin, plotmax=plotmax, Q1=Q1, Q2=Q2, Q3=Q3),
        ## datavalues
        ## )))
        newrow <- c(
            list(
                name = name, type = minfo$type, mcmctype = mcmctype,
                id = id, # censored=cens,
                transform = transf,
                Nvalues = Nvalues, indexpos = indexpos,
                halfstep = halfstep,
                domainmin = domainmin, domainmax = domainmax,
                minincluded = minincluded, maxincluded = maxincluded,
                tdomainmin = tdomainmin, tdomainmax = tdomainmax,
                domainminplushs = domainminplushs,
                domainmaxminushs = domainmaxminushs,
                tdomainminplushs = tdomainminplushs,
                tdomainmaxminushs = tdomainmaxminushs,
                tlocation = tlocation, tscale = tscale,
                plotmin = plotmin, plotmax = plotmax
                ## Q1 = Q1, Q2 = Q2, Q3 = Q3, # not used in other scripts, possibly remove
                ## mctest1 = mctest1, mctest2 = mctest2, mctest3 = mctest3
            ),
            as.list(datavalues)
        )

        ## if(nrow(auxmetadata) > 0){
        ##     if(length(newrow) > ncol(auxmetadata)){
        ##         auxmetadata[,
        ##             setdiff(names(newrow), colnames(auxmetadata))] <-
        ##             NA_character_
        ##     } else if(length(newrow) < ncol(auxmetadata)){
        ##         newrow[setdiff(colnames(auxmetadata), names(newrow))] <-
        ##             NA_character_
        ##     }
        ## }
        ## auxmetadata <- rbind(auxmetadata, newrow)

        auxmetadata <- merge(auxmetadata, newrow, sort = FALSE, all = TRUE)
    }

    ## print(auxmetadata) # for debugging
    auxmetadata

    ## ## Used under package development:
    ## ## in case it's useful to print the 'auxmetadata' to file.
    ## 
    ## if (!missing(file) && file != FALSE) { # must save to file
    ##   if (is.character(file)) {
    ##     file <- paste0(sub('.rds$', '', file), '.rds')
    ##   } else {
    ##     file <- paste0('auxmetadata_', datafile)
    ##     file <- paste0(sub('.csv$', '', file), '.rds')
    ##   }
    ##   if (file.exists(file)) {
    ##     file.rename(from = file,
    ##                 to = paste0(sub('.rds$', '', file), '_bak',
    ##                             format(Sys.time(), '%y%m%dT%H%M%S'), '.rds'))
    ##   }
    ##   saveRDS(auxmetadata, file)
    ##   message('Saved proposal aux-metadata file as', file, '\n')
    ## } else {
    ##   auxmetadata
    ## }
}

#' Calculate mutual information between groups of joint variates
#'
#' @description This function calculates various entropic information measures between two grops of joint variates: the mutual information, the conditional entropies, and the entropies.
#'
#' @details If \eqn{Y_1} and \eqn{Y_2} are two variates, each of which can be a joint variate such as \eqn{Y_1 = (Y_{1,1}, Y_{1,2}, \dotsc)}, and \eqn{X} a third, also possibly join, variate, then the mutual information \eqn{\mathit{MI}} between \eqn{Y_1} and \eqn{Y_2}, conditional on \eqn{X = x}, is given by
#' \deqn{\mathit{MI}(Y_1, Y_2 \vert X = x) \mathrel{:=}
#' \sum_{y_1, y_2}
#' \mathrm{Pr}(Y_1 = y_1, Y_2 = y_2 \vert X = x, \text{data})
#' \log_2\frac{
#' \mathrm{Pr}(Y_1 = y_1, Y_2 = y_2 \vert X = x, \text{data})
#' }{
#' \mathrm{Pr}(Y_1 = y_1 \vert X = x, \text{data})
#' \cdot
#' \mathrm{Pr}(Y_2 = y_2 \vert X = x, \text{data})
#' } \, \mathrm{Sh}
#' }
#' an expression which can also be written in several other equivalent ways. It is an information-theoretic measure of association that is model-free, that is, does not depend on assumptions such as linearity, gaussianity, and similar. See `vignette('mutualinfo')` for discussion and example uses, and also the "References" section.  If \eqn{Y_1, Y_2} are *jointly gaussian variates*, then there is a mathematical correspondence between their mutual information and their Pearson correlation coefficient; see output `MI.rGauss` in the "Value" section.
#'
#' The conditional entropy of \eqn{Y_1} with respect to \eqn{Y_2}, conditional on \eqn{X = x}, is given by
#' \deqn{\mathit{CondEn12}(Y_1, Y_2 \vert X = x) \mathrel{:=}
#' -\sum_{y_1, y_2}
#' \mathrm{Pr}(Y_1 = y_1 \vert Y_2 = y_2, X = x, \text{data})
#' \log_2
#' \mathrm{Pr}(Y_1 = y_1 \vert Y_2 = y_2, X = x, \text{data})
#' \cdot
#' \mathrm{Pr}(Y_2 = y_2 \vert X = x, \text{data})
#' \, \mathrm{Sh}
#' }
#'
#' The (differential) entropy of \eqn{Y_1}, conditional on \eqn{X = x}, is given by
#' \deqn{\mathit{En1}(Y_1 \vert X = x) \mathrel{:=}
#' -\sum_{y_1}
#' \mathrm{Pr}(Y_1 = y_1 \vert X = x, \text{data})
#' \log_2
#' \mathrm{Pr}(Y_1 = y_1 \vert  X = x, \text{data})
#' \, \mathrm{Sh}
#' }
#'
#' see "References" section for discussions about entropy and conditional entropy.
#'
#' The function `mutualinfo()` calculates the quantities above for the joint variates specified in the arguments `Y1names` and `Y2names`, conditional on the values of the variates specified in the data frame `X`. If `X` is omitted or `NULL`, then the posterior probabilities \eqn{\mathrm{Pr}(Y_1 | \text{data})} etc. are used. Each variate in the argument `X` can be specified either as a point-value \eqn{X = x} or as a left-open interval \eqn{X \le x} or as a right-open interval \eqn{X \ge x}, through the argument `tails`.
#'
#' The computation of these quantities is done via Monte Carlo integration, using the samples produced by the [learn()] function. The present function also output the numerical error associated with this computation.
#'
#' @param Y1names Character vector: first group of joint variates
#' @param Y2names Character vector or `NULL`: second group of joint variates
#' @param X Matrix or data.frame or `NULL`: values of some variates conditional on which we want the probabilities.
#' @param learnt Either a character with the name of a directory or full path
#'   for an 'learnt.rds' object, or such an object itself.
#' @param tails Named vector or list, or `NULL` (default). The names must match some or all of the variates in arguments `X`. For variates in this list, the probability conditional is understood in an semi-open interval sense: \eqn{X \le x} or \eqn{X \ge x}, an so on. See analogous argument in [Pr()].
#' @param n Integer or `NULL` (default): number of samples from which to approximately calculate the mutual information. Default as many as Monte Carlo samples in `learnt`.
#' @param unit Either one of 'Sh' for *shannon* (default), 'Hart' for *hartley*, 'nat' for *natural unit*, or a positive real indicating the base of the logarithms to be used.
#' @param parallel Logical or positive integer or cluster object. `TRUE` (default): use roughly half of available cores; `FALSE`: use serial computation; integer: use this many cores. It can also be a cluster object previously created with [parallel::makeCluster()]; in this case the parallel computation will use this object.
#' @param verbose Logical, default `FALSE`: give messages about parallel processing?
#'
#' @return A list consisting of the following elements:
#'
#' - `MI`, a vector of `value` and `accuracy`: the mutual information between (joint) variates `Y1names` and (joint) variates `Y2names`.
#' - `CondEn12`, `CondEn21`, vectors of `value` and `accuracy`: the conditional entropy of the first variate given the second, and vice versa.
#' - `En1`, `En2`, vectors of `value` and `accuracy`: the (differential) entropies of the first and second variates.
#' - `MI.rGauss`, a vector of `value` and `accuracy`: the absolute value of the Pearson correlation coefficient \eqn{r} of a *multivariate Gaussian distribution* having mutual information `MI`; the two are related by \eqn{\mathrm{MI} = -\ln(1 - r^2)/2}. It may provide a vague intuition for the `MI` value for people more familiar with Pearson's correlation, but should be taken with a grain of salt.
#' - `unit`, `Y1names`, `Y1names`: same as the input arguments, included for the user's convenience.
#'
#' @seealso
#' [Pr()] to calculate probabilities and their variability.
#'
#' [learn()], which generates the `learnt` objects required by `mutualinfo()`.
#'
#' @examples
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## mutual information between variates 'species' and 'bill_len'
#' MI <- mutualinfo(Y1names = 'species', Y2names = 'bill_len',
#'   learnt = learnt, parallel = 1)
#'
#' paste0(MI$MI, ' ', MI$unit, collapse = ' +/- ')
#'
#' ## Shannon entropy of variate 'species'
#' paste0(MI$En1, ' ', MI$unit, collapse = ' +/- ')
#'
#'
#' \donttest{
#' ## Shannon entropy of variate 'species',
#' ## conditional on a bill length of 30 mm:
#' entr <- mutualinfo(
#'   Y1names = 'species',
#'   X = data.frame(bill_len = 30),
#'   learnt = learnt, parallel = 1
#' )
#'
#' paste0(entr$En1, ' ', entr$unit, collapse = ' +/- ')
#'
#' ## the entropy is now lower; indeed a penguin with a short bill length
#' ## is most probably of the 'Adelie' species:
#' probs <- Pr(
#'   Y = data.frame(species = c('Adelie', 'Gentoo', 'Chinstrap')),
#'   X = data.frame(bill_len = 30),
#'   learnt = learnt, parallel = 1
#' )
#'
#' print(probs)
#' }
#'
#' @importFrom extraDistr rcat
#' @importFrom extraDistr rbern
#' @import parallel
#' @import stats
#' @import utils
#'
#' @export
mutualinfo <- function(
    Y1names,
    Y2names = NULL,
    X = NULL,
    learnt,
    tails = NULL,
    n = NULL,
    unit = 'Sh',
    parallel = TRUE,
    verbose = FALSE
){
#### Mutual information and conditional entropy between Y2 and Y1
#### conditional on X, data, prior information
#### are calculated by Monte Carlo integration:
#### 0. adjusted component weights are calculated for conditioning on X:
####     all probabilities below are conditional on X & data & prior
#### 1. joint samples of Y1_i, Y2_i are drawn
#### 2. probabilities p(Y1|Y2) are calculated for each sample
####    the conditional entropy is Monte-Carlo approximated by
####    H(Y1|Y2) = - sum_{i} log p(Y1_i | Y2_i)
#### 3. probabilities p(Y1) are calculated for each sample
####    the entropy is Monte-Carlo approximated by
####    H(Y1) = - sum_{i} log p(Y1_i)
#### 4. the mutual info is Monte-Carlo approximated by
####    I(Y1|Y2) = sum_{i} [log p(Y1_i | Y2_i) - log p(Y1_i)]
####           = -H(Y1|Y2) + H(Y1)
####
#### For these computations it is not necessary to transform the Y1,Y2 variates
#### from the internal Monte Carlo representation to the original one

#### Requested parallel processing
    ## NB: doesn't make sense to have more cores than chains
    closeexit <- FALSE
    if ('cluster' %in% class(parallel)){
        ## user provides a cluster object
        cl <- parallel
    } else if (isTRUE(parallel)) {
        ## user wants us to register a parallel backend
        ## and to choose number of cores
        ncores <- max(1,
            floor(parallel::detectCores() / 2))
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
        if(verbose){message('Registered ', capture.output(print(cl)), '.')}
    } else if (isFALSE(parallel)) {
        ## user wants us not to use parallel cores
        ncores <- 1
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
    } else if (is.numeric(parallel) &&
                   is.finite(parallel) && parallel >= 1) {
        ## user wants us to register 'parallel' # of cores
        ncores <- parallel
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
        if(verbose){message('Registered ', capture.output(print(cl)), '.')}
    } else {
        stop("Unknown value of argument 'parallel'.")
    }

    ## Close parallel connections if any were opened
    if(closeexit) {
        closecoresonexit <- function(){
            if(verbose){message('Closing connections to cores.')}
            parallel::stopCluster(cl)
            ## parallel::setDefaultCluster(NULL)
        }
        on.exit(closecoresonexit())
    }

    ## Extract Monte Carlo output & aux-metadata
    ## If learnt is a string, check if it's a folder name or file name
    if (is.character(learnt)) {
        ## Check if 'learnt' is a folder containing learnt.rds
        if (file_test('-d', learnt) &&
                file.exists(file.path(learnt, 'learnt.rds'))) {
            learnt <- readRDS(file.path(learnt, 'learnt.rds'))
        } else {
            ## Assume 'learnt' the full path of learnt.rds
            ## possibly without the file extension '.rds'
            learnt <- paste0(sub('.rds$', '', learnt), '.rds')
            if (file.exists(learnt)) {
                learnt <- readRDS(learnt)
            } else {
                stop("The argument 'learnt' must be a folder containing learnt.rds, or the path to an rds-file containing the output from 'learn()'.")
            }
        }
    }
    ## Add check to see that learnt is correct type of object?
    auxmetadata <- learnt$auxmetadata
    learnt$auxmetadata <- NULL
    learnt$auxinfo <- NULL
    ncomponents <- nrow(learnt$W)
    nmcs <- ncol(learnt$W)

    if(is.null(n) || n == 0) {
        n <- 1 * nmcs
    } else if(n < 0) {
        n <- -n * nmcs
    }

    if(n <= nmcs) {
        sseq <- sort(sample.int(nmcs, n))
    } else {
        sseq <- c(rep(x = seq_len(nmcs), times = n %/% nmcs),
            seq_len(nmcs)[sort.int(sample.int(nmcs, n %% nmcs))])
    }

    if(all(is.na(X))){X <- NULL}
    if(!is.null(X)){
        X <- as.data.frame(X)
        if (nrow(X) > 1) {
            warning('Only the first row of X is considered')
            X <- X[1, , drop = FALSE]
        }
    }
    Xv <- colnames(X)

    if(!is.null(tails)){
        tails <- as.list(tails)
        if(is.null(names(tails))) {
            stop('Missing variate names in "tails"')
        }
    }
    tailsv <- names(tails)
    tailscentre <- list('==', 0, '0', NULL)
    tailsleft <- list('<=', -1, '-1', 'left')
    tailsright <- list('>=', 1, '+1', 'right')
    tailsvalues <- c(tailscentre, tailsleft, tailsright)

    ## Consistency checks
    if (unit == 'Sh') {
        lbase <- log(2)
    } else if (unit == 'Hart') {
        lbase <- log(10)
    } else if (unit == 'nat') {
        lbase <- 1
    } else if (is.numeric(unit) && unit > 0) {
        lbase <- log(unit)
    } else {
        stop("unit must be 'Sh', 'Hart', 'nat', or a positive real")
    }

    if(!is.character(Y1names) || any(is.na(Y1names))){
        stop('Y1names must be a vector of variate names')
    }
    if(!is.null(Y2names) && (!is.character(Y2names) || any(is.na(Y2names)))){
        stop('Y2names must be NULL or a vector of variate names')
    }

    ## More consistency checks
    if(!all(Y1names %in% auxmetadata$name)) {
        stop('unknown Y1 variates\n')
    }
    if(length(unique(Y1names)) != length(Y1names)) {
        stop('duplicate Y1 variates\n')
    }
    ##
    if(!is.null(Y2names) && !all(Y2names %in% auxmetadata$name)){
        stop('unknown Y2 variates\n')
    }
    if(length(unique(Y2names)) != length(Y2names)) {
        stop('duplicate Y2 variates\n')
    }


    if (!all(Xv %in% auxmetadata$name)) {
        stop('unknown X variates\n')
    }
    if (length(unique(Xv)) != length(Xv)) {
        stop('duplicate X variates\n')
    }
    ##
    if(any(Y1names %in% Y2names)) {
        stop('overlap in Y1 and Y2 variates\n')
    }
    if(any(Y1names %in% Xv)) {
        stop('overlap in Y1 and X variates\n')
    }
    if(any(Y2names %in% Xv)) {
        stop('overlap in Y2 and X variates\n')
    }

    if (!all(tailsv %in% Xv)) {
        warning('variate ',
            paste0(tailsv[!(tailsv %in% Xv)], collapse = ' '),
            ' not among X; ignored\n')
        tails <- tails[tailsv %in% Xv]
        tailsv <- names(tails)
    }
    if (length(unique(tailsv)) != length(tailsv)) {
        stop('duplicate "tails" variates\n')
    }
    if(!all(tails %in% tailsvalues)) {
        stop('"tails" values must be ',
            paste0(tailsvalues, collapse = ' '), '\n')
    }

    ## transform 'tails' to -1, +1
    ## +1: '<=',    -1: '>='
    ## this is opposite of the argument convention because
    ## interval probabilities are calculated with `lower.tail = TRUE`
    ## eg:
    ## pnorm(x, mean, sd, lower.tail = FALSE) ==
    ##     pnorm(-x, -mean, sd, lower.tail = TRUE)
    tails[tails %in% tailscentre] <- NULL
    cleft <- tails %in% tailsleft
    cright <- tails %in% tailsright
    tails[cleft] <- +1
    tails[cright] <- -1
    tails <- unlist(tails)


#### Step 0. Adjust component weights W for conditioning on X
    if(is.null(X)){
        lW <- log(learnt$W)
    } else {
        lpargs <- util_lprobsargsyx(
            x = X,
            auxmetadata = auxmetadata,
            learnt = learnt,
            tails = tails
        )

        lW <- util_lprobsbase(
            xVs = lpargs$xVs[[1]],
            params = lpargs$params,
            logW =  log(learnt$W)
        ) # rows=components, columns=samples

    } # end definition of lW if non-null X


#### Combine Y1,Y2 into single Y for speed
    Ynames <- c(Y1names, Y2names)

#### STEP 1. Draw samples of Ynames (that is, Y1names,Y2names)

    lWnorm <- util_denorm(lW)
    Ws <- extraDistr::rcat(n = n, prob = t(
        apply(X = lWnorm[, sseq, drop = FALSE], MARGIN = 2, FUN = function(xx){
            xx <- exp(xx)
            xx / sum(xx, na.rm = TRUE)
        }, simplify = TRUE)
    ))

    Yout <- NULL
    vYout <- NULL

    ## R
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'R'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        aux <- auxmetadata[toselect, ]
        vYout <- c(vYout, aux$name)
        totake <- cbind(rep.int(x = aux$id, times = rep(n, nvrt)), Ws, sseq)
        Yout <- c(Yout,
            rnorm(n = n * nvrt,
                mean = learnt$Rmean[totake],
                sd = learnt$Rsd[totake] )
        )
    }

    ## C
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'C'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        aux <- auxmetadata[toselect, ]
        vYout <- c(vYout, aux$name)
        totake <- cbind(rep.int(x = aux$id, times = rep(n, nvrt)), Ws, sseq)
        Yout <- c(Yout,
            rnorm(n = n * nvrt,
                mean = learnt$Cmean[totake],
                sd = learnt$Csd[totake] )
        )
    }

    ## D
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'D'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        aux <- auxmetadata[toselect, ]
        vYout <- c(vYout, aux$name)
        totake <- cbind(rep.int(x = aux$id, times = rep(n, nvrt)), Ws, sseq)
        Yout <- c(Yout,
            rnorm(n = n * nvrt,
                mean = learnt$Dmean[totake],
                sd = learnt$Dsd[totake] )
        )
    }

    ## O
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'O'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        vYout <- c(vYout, auxmetadata$name[toselect])
        for(i in toselect) {
            aux <- auxmetadata[i, ]
            totake <- cbind(Ws, sseq)
            Yout <- c(Yout,
                extraDistr::rcat(n = n,
                    prob = apply(
                        X = learnt$Oprob[aux$indexpos + seq_len(aux$Nvalues), ,],
                        MARGIN = 1, FUN = `[`, totake,
                        simplify = TRUE) )
            )
        }
    }

    ## N
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'N'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        vYout <- c(vYout, auxmetadata$name[toselect])
        for(i in toselect) {
            aux <- auxmetadata[i, ]
            totake <- cbind(Ws, sseq)
            Yout <- c(Yout,
                extraDistr::rcat(n = n,
                    prob = apply(
                        X = learnt$Nprob[aux$indexpos + seq_len(aux$Nvalues), ,],
                        MARGIN = 1, FUN = `[`, totake,
                        simplify = TRUE) )
            )
        }
    }

    ## B
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'B'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        aux <- auxmetadata[toselect, ]
        vYout <- c(vYout, aux$name)
        totake <- cbind(rep.int(x = aux$id, times = rep(n, nvrt)), Ws, sseq)
        Yout <- c(Yout,
            extraDistr::rbern(n = n * nvrt,
                prob = learnt$Bprob[totake])
        )
    }

    dim(Yout) <- c(n, length(Ynames))
    Yout <- Yout[, match(Ynames, vYout), drop = FALSE]
    colnames(Yout) <- Ynames

    Yout <- vtransform(Yout,
        auxmetadata = auxmetadata,
        Rout = 'original',
        Cout = 'original',
        Dout = 'original',
        Oout = 'original',
        Nout = 'original',
        Bout = 'original',
        logjacobianOr = NULL)

    Y1transf <- Yout[, Y1names, drop = FALSE]
    Y2transf <- Yout[, Y2names, drop = FALSE]
    rm(Yout)
    gc()


#### STEP 2. Calculate sum_i log2_p(Y1|Y2) for all samples
    lpargs1 <- util_lprobsargsyx(
        x = Y1transf,
        auxmetadata = auxmetadata,
        learnt = learnt,
        tails = NULL
    )

    lpargs2 <- util_lprobsargsyx(
        x = Y2transf,
        auxmetadata = auxmetadata,
        learnt = learnt,
        tails = NULL
    )

    out <- do.call(rbind,
        parallel::parLapply(cl = cl,
        X = mapply(c, lpargs1$xVs, lpargs2$xVs, SIMPLIFY = FALSE),
        fun = util_lprobsmi,
        params1 = lpargs1$params,
        params2 = lpargs2$params,
        lWnorm = lWnorm,
        lW = lW
        )
        )

    ## Jacobian factors
    logjacobians1 <- rowSums(
        as.matrix(vtransform(Y1transf,
            auxmetadata = auxmetadata,
            logjacobianOr = FALSE)),
        na.rm = TRUE)

    logjacobians2 <- rowSums(
        as.matrix(vtransform(Y2transf,
            auxmetadata = auxmetadata,
            logjacobianOr = FALSE)),
        na.rm = TRUE)

    out[, c('CondEn12', 'CondEn21', 'En1', 'En2')] <-
        out[, c('CondEn12', 'CondEn21', 'En1', 'En2')] -
        c(logjacobians1, logjacobians2)

    out <- unlist(apply(
        X = rbind(
            value = colMeans(x = out, na.rm = TRUE) / lbase,
            accuracy = signif(x = apply(
                X = out, MARGIN = 2, FUN = sd, na.rm = TRUE, simplify = TRUE
            ) / (sqrt(n) * lbase), digits = 2)
        ),
        MARGIN = 2, FUN = list, simplify = TRUE), recursive = FALSE)

    if(out$MI['value'] < 0){
        out$MI['accuracy'] <- out$MI['accuracy'] + out$MI['value']
        out$MI['value'] <- 0
    }


    ## ## generally there's no MI maximum for continous variates
    ## mmax <- paste0('En', which.min(c(out$En1['value'], out$En2['value'])) )
    rgauss <- sqrt(1 - exp(-2 * out$MI['value'] * lbase))

    c(out,
        list(#MImax = out[[mmax]],
            MI.rGauss = c(
                rgauss,
                signif(x = out$MI['accuracy'] *
                           exp(-2 * out$MI['value'] * lbase) / rgauss,
                    digits = 2)
            ),
            unit = unit, Y1names = Y1names, Y2names = Y2names
        )
    )
}

#' Calculate and save transformation function for ordinal variates
#'
#' It creates the interpolation functions 'util_Q', 'util_invQ', 'util_invDQ' and saves them into 'sysdata.rda'.
#'
#' Those three functions are used to transform variates having bounded domains into variates with unbounded domains. See <https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf>.
#'
#' NB: the functional form of this function does not depend on
#' the number of components, minalpha, and maxalpha parameters
#'
#' @import grDevices
#' @import graphics
#' @import stats
#' @keywords internal
createQfunction <- function(
    nint = 3600,
    nsamples = 2^24L,
    mean = 0,
    sd = 3,
    shapelo = 0.5,
    shapehi = 0.5,
    rate = 1,
    file = paste0('__Qfunction', nint, '_', sd),
#    save = TRUE, # save using usethis::use_data()
    plot = FALSE
) {
    ##
    seqnint <- (1:(nint - 1)) / nint
    means <- rnorm(nsamples, mean = mean, sd = sd)
    sds <- sqrt(
        nimble::rinvgamma(nsamples, shape = shapelo, rate = nimble::rinvgamma(nsamples, shape = shapehi, rate = rate))
        ## extraDistr::rbetapr(nsamples, shape1=shapehi, shape2=shapelo, scale=scale)
    )
    xsamples <- rnorm(nsamples,
        mean = means,
        sd = sds
    )
    ##
    thismad <- mad(xsamples, constant = 1)
    oquants <- c(
        NULL,
        quantile(x = xsamples, probs = seqnint, na.rm = TRUE,
            type = 6, names = FALSE),
        NULL
    )
    rm(xsamples)
    oquants <- (oquants - rev(oquants)) / 2
    util_Q <- approxfun(x = seqnint, y = oquants, yleft = -Inf, yright = +Inf)
    util_invQ <- approxfun(y = seqnint, x = oquants, yleft = 0, yright = 1)

    if (is.character(file)) {
        saveRDS(util_Q, paste0(file, '.rds'))
        saveRDS(util_invQ, paste0(file,'_inv', '.rds'))
    }
    ##
    xsamples <- util_Q(seqnint)
    cl <- parallel::makeCluster(8)
    oquants <- parallel::parSapply(cl, xsamples,
        function(x, mean, sd){ mean(dnorm(x, mean = mean, sd = sd)) },
        mean = means, sd = sds
        )
    stopCluster(cl)
    oquants <- (oquants + rev(oquants)) / 2
    ##
    ##
    util_invDQ <- approxfun(x = xsamples, y = oquants, yleft = 0, yright = 0)
    if (is.character(file)) {
        saveRDS(util_invDQ, paste0('D', file, '.rds'))
    }
    if (is.character(plot)) {
        ##
        nint <- 256
        xgrid <- seq(1 / nint, (nint - 1) / nint, length.out = nint - 1)
        pdf(file = paste0(sub('.pdf$', '', plot), '.pdf'),
            paper = 'special',
            height = floor(841 / sqrt(2)^(5)) / 25.4,
            width = floor(841 / sqrt(2)^(5 - 1)) / 25.4)
        flexiplot(
            x = xgrid, y = list(
                util_Q(xgrid), qnorm(xgrid, sd = thismad / qnorm(3 / 4)), qcauchy(xgrid, scale = thismad) # ,qlogis(xgrid,scale=1/qlogis(3/4))
            ),
            lwd = c(3, 2, 2, 5), lty = c(1, 2, 4, 3), alpha.f = c(1, rep(0.75, 3)),
            ylim = range(util_Q(xgrid)),
            ## xticks=c(0,0.25,0.5,0.75,1),xlabels=c(0,expression(italic(m)/4),expression(italic(m)/2),expression(3*italic(m)/4),expression(italic(m))),
            xlab = expression(italic(x)), ylab = expression(italic(Q)(italic(x))),
            mar = c(NA, 5, 1, 1)
        )
        legend('topleft', c(expression(italic(Q)), 'Gauss', 'Cauchy'), lwd = c(3, 2, 2, 5), lty = c(1, 2, 4, 3), col = c(1, 2, 3, 4), bty = 'n')
        dev.off()
    }

    ## ## This is only used during package development
    ## if(save){
    ##     usethis::use_data(util_Q, util_invQ, util_invDQ,
    ##         internal = TRUE, overwrite = TRUE)
    ## }

    list(util_Q = util_Q, util_invQ = util_invQ, util_invDQ = util_invDQ)
}

#' Calculate credibility quantiles on estimated quantile
#'
#' Calculates the lower and upper bound of a credibility interval, for various quantiles of the empirical distribution of a vector of MC samples.
#'
#' Tests show that it gives a very good approximation of the "true" credibility quantiles in the case of independent samples.
#'
#' Tested also on t-distributions with df=1.1 and Pareto with a=1.5 (mean exists, variance infinite).
#'
#' Used in 'workerfun()' in 'learn()'
#'
#' @param x A vector of MC samples
#' @param prob numeric vector of probabilities: quantiles whose error interval is being estimated.
#' @param Qpair vector of length two (further elements are ignored): lower and higher credibility-quantiles requested. Default yields a credibility interval of 68%, or one nominal normal standard deviation.
#'
#' @return A matrix with two rows and as many columns as elements in 'prob'. Forr each column, the first and second row determine the lower and upper bound of the credibility interval of width `Qpair[2] - Qpair[2]`.
#'
#' @import stats
#'
#' @keywords internal
funMCEQ <- function(x, prob = c(0.055, 0.945), Qpair = pnorm(c(-1, 1))){
    N <- length(x)
    straces <- sort(x)
    sapply(prob, function(aprob) {
        Xlo <- quantile(x = x, probs = aprob,
            na.rm = FALSE, names = FALSE, type = 6)
        ##
        essXlo <- funESS3(x <= Xlo)
        ##
        a <- qbeta(Qpair, essXlo * aprob + 1, essXlo * (1 - aprob) + 1)
        c(straces[max(round(a[1] * N), 1)], straces[min(round(a[2] * N), N)])
    })
}

#' Compute ESS
#'
#' Modified from 'rstan' <https://github.com/stan-dev/rstan/blob/develop/rstan/rstan/R/monitor.R>
#'
#' Used in 'workerfun()' in 'learn()', and in 'funMCEQ()'.
#'
#' @param x Vector of MC samples.
#'
#' @return Effective Sample Size.
#'
#' @keywords internal
funESS3 <- function(x){
  N <- length(x)
  if (N < 3L || any(!is.finite(x))) {
    return(NA_real_)
  }
  acov <- funAC(x)
  chain_mean <- mean(x)
  var_plus <- acov[1]
  mean_var <- var_plus * N / (N - 1)
  ##
  ## Geyer's initial positive sequence
  rho_hat_t <- rep.int(0, N)
  t <- 0
  rho_hat_even <- 1
  rho_hat_t[t + 1] <- rho_hat_even
  rho_hat_odd <- 1 - (mean_var - acov[t + 2]) / var_plus
  rho_hat_t[t + 2] <- rho_hat_odd
  while (t < length(acov) - 5 &&
             !is.nan(rho_hat_even + rho_hat_odd) &&
             (rho_hat_even + rho_hat_odd > 0)) {
                 t <- t + 2
                 rho_hat_even = 1 - (mean_var - acov[t + 1]) / var_plus
                 rho_hat_odd = 1 - (mean_var - acov[t + 2]) / var_plus
                 if ((rho_hat_even + rho_hat_odd) >= 0) {
                     rho_hat_t[t + 1] <- rho_hat_even
                     rho_hat_t[t + 2] <- rho_hat_odd
                 }
             }
  max_t <- t
  ## this is used in the improved estimate
  if (rho_hat_even>0){rho_hat_t[max_t + 1] <- rho_hat_even}
  ##
  ## Geyer's initial monotone sequence
  t <- 0
  while (t <= max_t - 4) {
    t <- t + 2
    if (rho_hat_t[t + 1] + rho_hat_t[t + 2] > rho_hat_t[t - 1] + rho_hat_t[t]) {
      rho_hat_t[t + 1] = (rho_hat_t[t - 1] + rho_hat_t[t]) / 2
      rho_hat_t[t + 2] = rho_hat_t[t + 1]
    }
  }
  ess <- N
  # Geyer's truncated estimate
  # tau_hat <- -1 + 2 * sum(rho_hat_t[1:max_t])
  # Improved estimate reduces variance in antithetic case
  tau_hat <- -1 + 2 * sum(rho_hat_t[1:max_t]) + rho_hat_t[max_t+1]
  # Safety check for negative values and with max ess equal to ess*log10(ess)
  tau_hat <- max(tau_hat, 1/log10(ess))
  ess <- ess / tau_hat
  ess
}


#' Calculate MC standard error using LaplacesDemon's batch means
#'
#' This function gives a good approximation of the "true" standard deviation in the case of independent samples. Multiply by `qnorm(x)` to obtain the `x`-quantile.
#'
#' Modified from <https://github.com/LaplacesDemonR/LaplacesDemon/blob/master/R/ESS.R>.
#'
#' Tested also on t-distributions with df=1.1 and Pareto with a=1.5 (mean exists, variance infinite).
#'
#' `sd() / sqrt(funESS3()` gives essentially identical results to `funMCSELD()`, but it's 20 times slower.
#'
#' Used in 'util_combineYX()' in 'Pr()'.
#'
#' @param x matrix, each row being a "trace", that is a set of MC samples, whose MCSE is to be estimated.
#'
#' @return MCSE estimates, one for each trace. Division by sqrt(N) is already performed.
#'
#' @keywords internal
funMCSELD <- function(x) {
    x <- as.matrix(x)
    N <- nrow(x)
    b <- floor(sqrt(N))
    a <- floor(N/b)
    i0 <- N - a * b
    Ys <- rbind(sapply(seq_len(a), function(k) {
        colMeans(x[((k - 1) * b + 1 + i0):(k * b + i0), , drop = FALSE])
    }))
    ## ## Note: this is 'w" in doi.org/10.1080/10618600.2015.1044092
    sqrt(b * rowSums((Ys - rowMeans(Ys))^2) / ((a - 1) * N))
}


#' Compute autocovariance
#'
#' Modified from rstan <https://github.com/stan-dev/rstan/blob/develop/rstan/rstan/R/monitor.R>.
#'
#' Used in 'funESS3()'.
#'
#' @param y Time series
#'
#' @return Autocovariances at different lags
#'
#' @import stats
#'
#' @keywords internal
funAC <- function(y) {
  N <- length(y)
  Mt2 <- 2 * fftNGS(N)
  yc <- y - mean(y)
  yc <- c(yc, rep.int(0, Mt2 - N))
  transform <- fft(yc)
  ac <- fft(Conj(transform) * transform, inverse = TRUE)
  # use "biased" estimate as recommended by Geyer (1992)
  ac <- Re(ac)[1:N] / (N^2 * 2)
  ac
}


#' Find optimal FFT size
#'
#' Modified from rstan <https://github.com/stan-dev/rstan/blob/develop/rstan/rstan/R/monitor.R>.
#'
#' Used in 'funAC()'.
#'
#' @param N Integer.
#'
#' @return Optimal FFT size
#'
#' @keywords internal
fftNGS <- function(N) {
  # Find the optimal next size for the FFT so that
  # a minimum number of zeros are padded.
  if (N <= 2)
    return(2)
  while (TRUE) {
    m <- N
    while ((m %% 2) == 0) m <- m / 2
    while ((m %% 3) == 0) m <- m / 3
    while ((m %% 5) == 0) m <- m / 5
    if (m <= 1)
      return(N)
    N <- N + 1
  }
}

#' Subset variates of an object of class "probability"
#'
#' An object of class "probability", obtained with the [Pr()] function, holds the probabilities for all possible combinations of values of a set of joint variates `Y` conditional on a set of joint variates `X`, together with the variabilities of these probabilities and some other information. In some cases one may wish to exclude some of the values of the `Y` or `X` variates. For instance `Y` in the probability-class object could include the variate "age" with values from 18 to 100, and one may want to retain the values from 60 to 80.
#'
#' @param x Object of class "probability", obtained with [Pr()].
#' @param subset Named list or named vector: variates to subset, given as list names, and corresponding values to subset.
#'
#' @return An object of class "probability", identical to the original object `x` except for a reduced range of values in some if its variates.
#'
## #' @seealso
## #' [Pr()], which generates probability objects.
## #'
## #' [plot.probability()] to plot probabilities and quantiles calculated by `Pr()'.
## #'
## #' [hist.probability()] to plot histograms of the probability distributions calculated by `Pr()`.
## #'
## #' @examples
## #' ## Load the example `learnt` object calculated from the "penguins" dataset;
## #' ## variates: 'species' and 'bill_len'
## #' learnt <- learntExample
## #'
## #' ## Calculate the probability object for the three values of variate 'species',
## #' ## given values 43 and 44 of variate 'bill_len';
## #' ## this object contains probabilities, quantiles, and other information
## #' probs <- Pr(
## #'   Y = data.frame(species = c('Adelie', 'Chinstrap', 'Gentoo')),
## #'   X = data.frame(bill_len = c(43, 44)),
## #'   learnt = learnt, parallel = 1
## #' )
## #'
## #' probs$values
## #'
## #' ## Subset by retaining the values 'Adelie' and 'Gentoo' for species,
## #' ## and 44 for bill length
## #' newprobs <- prsubset(
## #'   probs,
## #'   subset = list(species = c('Adelie', 'Gentoo'), bill_len = 43)
## #' )
## #'
## #' newprobs$values
## #'
## #' ## Plot these conditional probabilities and their variabilities
## #' plot(newprobs)
## #'
## #' hist(newprobs)
## #'
#' @keywords internal
prsubset <- function(
    x,
    subset
){
    Ynames <- names(x$Y)
    Xnames <- names(x$X)
    subset <- as.list(subset)
    vrtnames <- names(subset)

    if(!all(vrtnames %in% c(Ynames, Xnames))){
        stop("probability object does not contain some of the given variates")
    }

    ## subset Y
    for(vrt in vrtnames[vrtnames %in% Ynames]){
        selvals <- x$Y[[vrt]] %in% subset[[vrt]]
        x$values <- x$values[selvals, , drop = FALSE]
        if(!is.null(x$values.MCaccuracy)){
            x$values.MCaccuracy <- x$values.MCaccuracy[selvals, , drop = FALSE]
        }
        if(!is.null(x$quantiles)){
            x$quantiles <- x$quantiles[selvals, , , drop = FALSE]
        }
        if(!is.null(x$quantiles.MCaccuracy)){
            x$quantiles.MCaccuracy <- x$quantiles.MCaccuracy[selvals, , , drop = FALSE]
        }
        if(!is.null(x$samples)){
            x$samples <- x$samples[selvals, , , drop = FALSE]
        }
        x$Y <- x$Y[selvals, , drop = FALSE]
    }

    ## subset X
    for(vrt in vrtnames[vrtnames %in% Xnames]){
        selvals <- x$X[[vrt]] %in% subset[[vrt]]
        x$values <- x$values[, selvals, drop = FALSE]
        if(!is.null(x$values.MCaccuracy)){
            x$values.MCaccuracy <- x$values.MCaccuracy[, selvals, drop = FALSE]
        }
        if(!is.null(x$quantiles)){
            x$quantiles <- x$quantiles[, selvals, , drop = FALSE]
        }
        if(!is.null(x$quantiles.MCaccuracy)){
            x$quantiles.MCaccuracy <- x$quantiles.MCaccuracy[, selvals, , drop = FALSE]
        }
        if(!is.null(x$samples)){
            x$samples <- x$samples[, selvals, , drop = FALSE]
        }
        x$X <- x$X[selvals, , drop = FALSE]
    }

    x
}

#' Example metadata file
#'
#' @description
#' A [data frame][base::data.frame()] containing the prior information about the variates `species` and `bill_len` of the [datasets::penguins] dataset.
#'
#' @format ## `metadataExample`
#' A [data frame][base::data.frame()] with 2 rows and 10 columns.
#'
#' @returns No return value.
#'
#' @seealso
#' [metadatatemplate()] which helps producing this kind of metadata files from a given dataset.
#'
#' [learn()] which needs this kind of metadata files to "learn" from data.
"metadataExample"


#' Example `learnt` object produced by learn()
#'
#' @description
#' An example `learnt` object obtained by means of the [learn()] function, using the [datasets::penguins] dataset and the metadata in [metadataExample], according to the call
#'
#' ```
#' learn(data = penguins, metadata = metadataExample,
#'   nsamples = 225, nchains = 15)
#' ```
#'
#' It is a list that essentially contains posterior hyperparameters for drawing statistical inferences about the variates `species` and `bill_len`.
#'
#' **Note** that the [learn()] function that produced `learntExample` was called with the option to create only a limited number (225) of Monte Carlo samples, in order to reduce its memory size. Thus the numerical error associated with the Monte Carlo approximation is relatively in inferences drawn from the posterior hyperparameters saved in `learntExample`. It is only meant to be used for illustration purposes of the package's capabilities.
#'
#' @format ## `learntExample`
#' A list containing results from Markov-chain Monte Carlo computation, including diagnostics and variate metadata.
#'
#' @returns No return value.
#'
#' @seealso
#' [learn()], which produces this kind of object.
#'
#' [Pr()], [qPr()], [rPr()], [mutualinfo()]: functions that require this kind of object in order to calculate probabilities and quantiles, generate data points, and calculate mutual information.
"learntExample"

#' Prepare arguments for util_lprobsyx from data
#'
#' Used in 'Pr()', 'qPr()', 'rPr()', 'mutualinfo()'
#'
#' @keywords internal
util_lprobsargsyx <- function(
    x, auxmetadata, learnt, tails = NULL
) {
    Xv <- colnames(x)
    nX <- nrow(x)
    tailsv <- names(tails)

    xV0 <- xV1 <- xV2 <- xVN <- xVB <- matrix(NA_real_, 0, nX)

###
### point probability density
###
    nV0 <- FALSE
    V0mean <- V0sd <- NULL

### R-variates not in 'tails'
    toselect <- which((auxmetadata$name %in% Xv) &
                          !(auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'R'))
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV0 <- TRUE
        V0mean <- learnbind(V0mean,
            learnt$Rmean[aux$id, , , drop = FALSE])
        V0sd <- learnbind(V0sd,
            learnt$Rsd[aux$id, , , drop = FALSE])
        xV0 <- rbind(xV0,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Rout = 'normalized',
                logjacobianOr = NULL
            )))
        )
    }

### C-variates not in 'tails' and with some non-boundary value
    toselect <- which((auxmetadata$name %in% Xv) &
                          !(auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'C'))
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            xx <- x[,auxmetadata$name[i]]
            any(xx > auxmetadata$domainmin[i] & xx < auxmetadata$domainmax[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV0 <- TRUE
        V0mean <- learnbind(V0mean,
            learnt$Cmean[aux$id, , , drop = FALSE])
        V0sd <- learnbind(V0sd,
            learnt$Csd[aux$id, , , drop = FALSE])
        xV0 <- rbind(xV0,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Cout = 'boundisna',
                logjacobianOr = NULL
            )))
        )
    }

###
### tail probability
###
    nV1 <- FALSE
    V1mean <- V1sd <- NULL

### R-variates in 'tails'
    toselect <- which((auxmetadata$name %in% Xv) &
                          (auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'R'))
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV1 <- TRUE
        V1mean <- learnbind(V1mean,
                tails[aux$name] * learnt$Rmean[aux$id, , , drop = FALSE])
        V1sd <- learnbind(V1sd,
            learnt$Rsd[aux$id, , , drop = FALSE])
        xV1 <- rbind(xV1,
            tails[aux$name] *
                t(as.matrix(vtransform(
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Rout = 'normalized',
                    logjacobianOr = NULL
                )))
        )
    }

### C-variates in 'tails'
    toselect <- which((auxmetadata$name %in% Xv) &
                          (auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'C'))
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV1 <- TRUE
        V1mean <- learnbind(V1mean,
                tails[aux$name] * learnt$Cmean[aux$id, , , drop = FALSE])
        V1sd <- learnbind(V1sd,
            learnt$Csd[aux$id, , , drop = FALSE])
        xV1 <- rbind(xV1,
            tails[aux$name] *
                t(as.matrix(vtransform(
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Cout = 'boundnormalized',
                    logjacobianOr = NULL
                )))
        )
    }

### D-variates in 'tails'
    toselect <- which((auxmetadata$name %in% Xv) &
                          (auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'D'))
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV1 <- TRUE
        V1mean <- learnbind(V1mean,
                tails[aux$name] * learnt$Dmean[aux$id, , , drop = FALSE])
        V1sd <- learnbind(V1sd,
            learnt$Dsd[aux$id, , , drop = FALSE])
        xV1 <- rbind(xV1,
            tails[aux$name] *
                t(as.matrix(vtransform(
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Dout = 'boundnormalized',
                    logjacobianOr = NULL
                ))) +
                aux$halfstep / aux$tscale
        )
    }

### C-variates not in 'tails' and with left boundary values
    toselect <- which((auxmetadata$name %in% Xv) &
                          !(auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'C'))
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] <= auxmetadata$domainmin[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV1 <- TRUE
        V1mean <- learnbind(V1mean,
                learnt$Cmean[aux$id, , , drop = FALSE])
        V1sd <- learnbind(V1sd,
            learnt$Csd[aux$id, , , drop = FALSE])
        xV1 <- rbind(xV1,
                t(as.matrix(vtransform(
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Cout = 'leftbound',
                    logjacobianOr = NULL
                )))
        )
    }

### C-variates not in 'tails' and with right boundary values
    toselect <- which((auxmetadata$name %in% Xv) &
                          !(auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'C'))
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] >= auxmetadata$domainmax[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV1 <- TRUE
        V1mean <- learnbind(V1mean,
                - learnt$Cmean[aux$id, , , drop = FALSE]) # minus sign
        V1sd <- learnbind(V1sd,
            learnt$Csd[aux$id, , , drop = FALSE])
        xV1 <- rbind(xV1,
               - t(as.matrix(vtransform( # minus sign
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Cout = 'rightbound',
                    logjacobianOr = NULL
                )))
        )
    }

### D-variates not in 'tails' and with left boundary values
    toselect <- which((auxmetadata$name %in% Xv) &
                          !(auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'D'))
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] <= auxmetadata$domainminplushs[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV1 <- TRUE
        V1mean <- learnbind(V1mean,
                learnt$Dmean[aux$id, , , drop = FALSE])
        V1sd <- learnbind(V1sd,
            learnt$Dsd[aux$id, , , drop = FALSE])
        xV1 <- rbind(xV1,
                t(as.matrix(vtransform(
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Dout = 'leftbound',
                    logjacobianOr = NULL
                ))) +
                aux$halfstep / aux$tscale
        )
    }

### D-variates not in 'tails' and with right boundary values
    toselect <- which((auxmetadata$name %in% Xv) &
                          !(auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'D'))
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] >= auxmetadata$domainmaxminushs[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV1 <- TRUE
        V1mean <- learnbind(V1mean,
                - learnt$Dmean[aux$id, , , drop = FALSE]) # minus sign
        V1sd <- learnbind(V1sd,
            learnt$Dsd[aux$id, , , drop = FALSE])
        xV1 <- rbind(xV1,
               - t(as.matrix(vtransform( # minus sign
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Dout = 'rightbound',
                    logjacobianOr = NULL
                ))) +
                aux$halfstep / aux$tscale
        )
    }

###
### interval probability
###
    nV2 <- FALSE
    V2mean <- V2sd <- V2steps <- NULL

### D-variates not in 'tails'
    toselect <- which((auxmetadata$name %in% Xv) &
                          !(auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'D'))
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            xx <- x[,auxmetadata$name[i]]
            any(xx > auxmetadata$domainminplushs[i] &
                    xx < auxmetadata$domainmaxminushs[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nV2 <- TRUE
        V2mean <- learnbind(V2mean,
                learnt$Dmean[aux$id, , , drop = FALSE])
        V2sd <- learnbind(V2sd,
            learnt$Dsd[aux$id, , , drop = FALSE])
        V2steps <- aux$halfstep / aux$tscale
        xV2 <- rbind(xV2,
                t(as.matrix(vtransform(
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Dout = 'boundisna',
                    logjacobianOr = NULL
                )))
        )
    }

###
### discrete case
###
    nVN <- FALSE
    VNprobs <- NULL
    Nshift <- 0L
### O-variates not in 'tails'
    toselect <- which((auxmetadata$name %in% Xv) &
                          !(auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'O'))
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nVN <- TRUE
        ## Nindices <- unlist(lapply(seq_len(nrow(aux)), function(i) {
        ##     aux$indexpos[i] + seq_len(aux$Nvalues[i])
        ## }))
        Nindices <- unlist(mapply(FUN = function(i, n) {i + seq_len(n)},
            aux$indexpos, aux$Nvalues,
            SIMPLIFY = FALSE))
        VNprobs <- learnbind(VNprobs,
            learnt$Oprob[Nindices, , , drop = FALSE])
        xVN <- rbind(xVN,
                t(as.matrix(vtransform(
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Oout = 'numeric',
                    logjacobianOr = NULL
                ))) +
                    Nshift + c(0, cumsum(aux$Nvalues[-1]))
        )
        Nshift <- Nshift + length(Nindices)
    }
### N-variates
    toselect <- which((auxmetadata$name %in% Xv) &
                          (auxmetadata$mcmctype == 'N'))
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nVN <- TRUE
        ## Nindices <- unlist(lapply(seq_len(nrow(aux)), function(i) {
        ##     aux$indexpos[i] + seq_len(aux$Nvalues[i])
        ## }))
        Nindices <- unlist(mapply(FUN = function(i, n) {i + seq_len(n)},
            aux$indexpos, aux$Nvalues,
            SIMPLIFY = FALSE))
        VNprobs <- learnbind(VNprobs,
            learnt$Nprob[Nindices, , , drop = FALSE])
        xVN <- rbind(xVN,
                t(as.matrix(vtransform(
                    x[, aux$name, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Nout = 'numeric',
                    logjacobianOr = NULL
                ))) +
                    Nshift + c(0, cumsum(aux$Nvalues[-1]))
        )
        Nshift <- Nshift + length(Nindices)
    }

### O-variates in 'tails'
    toselect <- which((auxmetadata$name %in% Xv) &
                          (auxmetadata$name %in% tailsv) &
                          (auxmetadata$mcmctype == 'O'))
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nVN <- TRUE
        for(i in seq_len(nrow(aux))) {
            VNprobs <- learnbind(VNprobs,
                if(tails[aux$name[i]] < 0) {
                    rowcumsum(learnt$Oprob[
                        aux$indexpos[i] + seq_len(aux$Nvalues[i]), , , drop = FALSE
                    ])
                } else {
                    rowinvcumsum(learnt$Oprob[
                        aux$indexpos[i] + seq_len(aux$Nvalues[i]), , , drop = FALSE
                    ])
                }
            )
        }
        xVN <- rbind(xVN,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Oout = 'numeric',
                logjacobianOr = NULL
            ))) +
                Nshift + c(0, cumsum(aux$Nvalues[-1]))
        )
    }

###
### binary case
###
    nVB <- FALSE
    VBprobs <- NULL
### B-variates
    toselect <- which((auxmetadata$name %in% Xv) &
                          (auxmetadata$mcmctype == 'B'))
    if(length(toselect) > 0){
        aux <- auxmetadata[toselect, ]
        nVB <- TRUE
        VBprobs <- learnbind(VBprobs,
            learnt$Bprob[aux$id, , , drop = FALSE])
        xVB <- rbind(xVB,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Bout = 'numeric',
                logjacobianOr = NULL
            )))
        )
    }

    list(
        params = list(
            nV0 = nV0, V0mean = V0mean, V0sd = V0sd,
            nV1 = nV1, V1mean = V1mean, V1sd = V1sd,
            nV2 = nV2, V2mean = V2mean, V2sd = V2sd,
            V2steps = V2steps,
            nVN = nVN, VNprobs = VNprobs,
            nVB = nVB, VBprobs = VBprobs
        ),
        xVs = lapply(seq_len(nX), function(i){
            list(
                ii = i,
                xV0 = xV0[,i],
                xV1 = xV1[,i],
                xV2 = xV2[,i],
                xVN = xVN[,i],
                xVB = xVB[,i]
            )})
    )
}



#' Calculate collection of log-probabilities for different components and samples
#'
#' Used in 'Pr()', 'qPr()', 'rPr()', 'mutualinfo()', 'util_checkpoints()'.
#'
#' @return Matrix with as many rows as components and as many cols as samples
#' @keywords internal
util_lprobsbase <- function(
    xVs, params, logW,
    temporarydir = NULL, lab = ''
) {
    with(c(xVs, params), {
    out <- logW
    ## point probability density
    if(nV0) {
        out <- out + colSums(
            x = dnorm(x = xV0, mean = V0mean, sd = V0sd, log = TRUE),
            na.rm = TRUE, dims = 1)
    }
    ## tail probability
    if(nV1) {
        out <- out + colSums(
            x = pnorm(q = xV1, mean = V1mean, sd = V1sd,
                log.p = TRUE, lower.tail = TRUE),
            na.rm = TRUE, dims = 1)
    }
    ## interval probability
    if(nV2) {
        pright <- pnorm(q = xV2 + V2steps, mean = V2mean, sd = V2sd,
            log.p = TRUE, lower.tail = TRUE)
        ##
        out <- out + colSums(
            x = pright + log(-expm1(
                pnorm(q = xV2 - V2steps, mean = V2mean, sd = V2sd,
                    log.p = TRUE, lower.tail = TRUE) - pright
            )),
            na.rm = TRUE, dims = 1)
        ## ## this alternate form leads to infinities in some cases
        ## pleft <- pnorm(q = xV2 - V2steps, mean = V2mean, sd = V2sd,
        ##     log.p = TRUE, lower.tail = TRUE)
        ## ##
        ## out <- out + colSums(
        ##     x = pleft + log(expm1(
        ##         pnorm(q = xV2 + V2steps, mean = V2mean, sd = V2sd,
        ##             log.p = TRUE, lower.tail = TRUE) - pleft
        ##     )),
        ##     na.rm = TRUE, dims = 1)
    }
    ##
    if(nVN) {
        out <- out + colSums(
            x = log(VNprobs[xVN, , , drop = FALSE]),
            na.rm = TRUE, dims = 1)
    }
    ##
    if(nVB) {
        ## VBprob is the probability that x = 1
        out <- out + colSums(
            x = log(1 - xVB - VBprobs + 2 * xVB * VBprobs),
            na.rm = TRUE, dims = 1)
    }

    if(is.null(temporarydir)){
        out
    }else{
        saveRDS(out,
            file.path(temporarydir,
                paste0(lab, ii, '__.rds'))
        )
    }
    })
}



#' Calculate probabilities, quantiles, etc, for all Y and X combinations
#'
#' Used in 'Pr()'.
#'
#' @import stats
#'
#' @keywords internal
util_combineYX <- function(
    iyx,
    temporarydir, usememory = TRUE,
    doquantiles, quantiles,
    dosamples, nsamples,
    Qerror
) {

    if(usememory) {
        lprobX <- readRDS(file.path(temporarydir,
            paste0('__X', iyx['jx'], '__.rds')
        ))
        lprobY <- readRDS(file.path(temporarydir,
            paste0('__Y', iyx['jy'], '__.rds')
        ))
    }

    FF <- colSums(x = exp(lprobX + lprobY), na.rm = TRUE) /
        colSums(x = exp(lprobX), na.rm = TRUE)

    list(
        values = mean(x = FF, na.rm = TRUE),
        ##
        quantiles = if(doquantiles) {
            quantile(x = FF, probs = quantiles, type = 6,
                na.rm = TRUE, names = FALSE)
        },
        ##
        samples = if(dosamples) {
            FF <- FF[!is.na(FF)]
            FF[round(seq(1, length(FF), length.out = nsamples))]
        },
        ##
        values.MCaccuracy = funMCSELD(x = FF),
        ##
        quantiles.MCaccuracy = if(doquantiles) {
            temp <- funMCEQ(x = FF, prob = quantiles, Qpair = Qerror)
            (temp[2, ] - temp[1, ]) / 2
        }
        ##
        ## error = sd(FF, na.rm = TRUE)/sqrt(nmcsamples)
    )
}



#' Calculate pairs of log-probabilities for mutualinfo()
#'
#' Used in 'mutualinfo()'.
#'
#' @keywords internal
util_lprobsmi <- function(xVs, params1, params2, lWnorm, lW) {

    lprobY1 <- util_lprobsbase(xVs = xVs[1:6], params = params1, logW = 0)
    lprobY2 <- util_lprobsbase(xVs = xVs[7:12], params = params2, logW = 0)

    celWnorm <- colSums(exp(lWnorm))

### Construct probabilities from lprobY1, lprobY2
    lpY1and2 <- log(mean(
        colSums(exp(lprobY1 + lprobY2 + lWnorm)) / celWnorm,
        na.rm = TRUE))

    lpY1 <- log(mean(
        colSums(exp(lprobY1 + lWnorm)) / celWnorm,
        na.rm = TRUE))

    lpY2 <- log(mean(
        colSums(exp(lprobY2 + lWnorm)) / celWnorm,
        na.rm = TRUE))

    lprobnorm <- util_denorm(lprobY2 + lW)
    lpY1given2 <- log(mean(
        colSums(exp(lprobY1 + lprobnorm)) / colSums(exp(lprobnorm)),
        na.rm = TRUE))

    lprobnorm <- util_denorm(lprobY1 + lW)
    lpY2given1 <- log(mean(
        colSums(exp(lprobY2 + lprobnorm)) / colSums(exp(lprobnorm)),
        na.rm = TRUE))

    mi <- lpY1and2 - lpY1 - lpY2
    c(
        MI = mi,
        CondEn12 = -lpY1given2,
        CondEn21 = -lpY2given1,
        En1 = -lpY1,
        En2 = -lpY2
        ## MIalt = (mi + lpY1given2 - lpY1 + lpY2given1 - lpY2) / 3,
    )
}



#' Calculate quantiles for continuous Y by bisection
#'
#' Used in 'qPr()'.
#'
#' @import stats
#'
#' @keywords internal
util_qYXcont <- function(
    iyx,
    params1, params2,
    auxmetadata,
    temporarydir, usememory = TRUE,
    doquantiles, quantiles,
    dosamples, nsamples,
    Qerror,
    tol = .Machine$double.eps * 3
) {
    pY <- iyx['pY']

    if(usememory) {
        lprobX <- readRDS(file.path(temporarydir,
            paste0('__X', iyx['jx'], '__.rds')
        ))
    }
    sumlpX <- colSums(exp(lprobX), na.rm = TRUE)

    nmaxsamples <- ncol(params1)

#### Calculate quantile for the posterior probability distribution

    Yvals <- .Machine$double.xmax * c(-0.125, 0.125)

    values <- (Yvals[1] + Yvals[2]) / 2

    FF <- mean(colSums(exp(
        lprobX + pnorm(q = values,
            mean = params1, sd = params2,
            lower.tail = TRUE, log.p = TRUE)
    ), na.rm = TRUE) / sumlpX) - pY

    while(abs(FF) > tol && Yvals[2] - Yvals[1] > tol){
        Yvals[(FF > 0) + 1L] <- values
        values <- (Yvals[1] + Yvals[2]) / 2
        FF <- mean(colSums(exp(
            lprobX + pnorm(q = values,
                mean = params1, sd = params2,
                lower.tail = TRUE, log.p = TRUE)
        ), na.rm = TRUE) / sumlpX) - pY
    }

    values <- unname(unlist(vtransform(values,
        auxmetadata = auxmetadata,
        Rout = 'original',
        Cout = 'original',
        Dout = 'original',
        Oout = 'original',
        Nout = 'original',
        Bout = 'original',
        variates <- auxmetadata$name,
        logjacobianOr = NULL)))

#### Calculate quantile for the frequency samples
    if(doquantiles){
        selsamples <- TRUE
    } else if(dosamples) {
        nmaxsamples <- nsamples
        selsamples <- round(seq(1, ncol(params1), length.out = nsamples))
    }

    if(doquantiles || dosamples) {
        params1 <- t(params1[, selsamples])
        params2 <- t(params2[, selsamples])
        lprobX <- t(lprobX[, selsamples])

        Yvals <- rep(.Machine$double.xmax * c(-0.125, 0.125), each = nmaxsamples)
        dim(Yvals) <- c(nmaxsamples, 2)
        sampleseq <- 1:nmaxsamples

        samples <- (Yvals[, 1] + Yvals[, 2]) / 2
        FF <- rowSums(exp(
            lprobX + pnorm(q = samples,
                mean = params1, sd = params2,
                lower.tail = TRUE, log.p = TRUE)
        ), na.rm = TRUE) / sumlpX - pY

        tocheck <- abs(FF) > tol
        while(any(tocheck)) {
            choose <- c(sampleseq[tocheck], (FF[tocheck] > 0) + 1L)
            dim(choose) <- c(sum(tocheck), 2)
            Yvals[choose] <- samples[tocheck]
            samples[tocheck] <- (Yvals[tocheck, 1] + Yvals[tocheck, 2]) / 2
            FF <- rowSums(exp(
                lprobX + pnorm(q = samples,
                    mean = params1, sd = params2,
                    lower.tail = TRUE, log.p = TRUE)
            ), na.rm = TRUE) / sumlpX - pY
        tocheck <- abs(FF) > tol & Yvals[, 2] - Yvals[, 1] > tol
        }
        samples <- unname(unlist(vtransform(samples,
            auxmetadata = auxmetadata,
            Rout = 'original',
            Cout = 'original',
            Dout = 'original',
            Oout = 'original',
            Nout = 'original',
            Bout = 'original',
            variates <- auxmetadata$name,
            logjacobianOr = NULL)))
    }

    list(
        values = values,
        ##
        quantiles = if(doquantiles) {
            quantile(x = samples, probs = quantiles, type = 6,
                na.rm = TRUE, names = FALSE)
        },
        ##
        samples = if(dosamples) {
            samples[round(seq(1, length(samples), length.out = nsamples))]
        }
        ## values.MCaccuracy
        ## quantiles.MCaccuracy
)
}



#' Calculate quantiles for discrete Y by bisection
#'
#' Used in 'qPr()'.
#'
#' @import stats
#'
#' @keywords internal
util_qYXdiscr <- function(
    iyx,
    params1, params2,
    auxmetadata,
    temporarydir, usememory = TRUE,
    doquantiles, quantiles,
    dosamples, nsamples,
    Qerror = NULL,
    tol = NULL
){
    pY <- iyx['pY']

    if(usememory) {
        lprobX <- readRDS(file.path(temporarydir,
            paste0('__X', iyx['jx'], '__.rds')
        ))
    }
    sumlpX <- colSums(exp(lprobX), na.rm = TRUE)

    nmaxsamples <- dim(params1)[3]
    Nvalues <- auxmetadata$Nvalues

#### Calculate quantile for the posterior probability distribution

    values <- 1L

    FF <- mean(colSums(exp(
        lprobX + params1[values, ,]
    ), na.rm = TRUE) / sumlpX)

    while(FF < pY && values <= Nvalues){
        values <- values + 1L
        FF <- mean(colSums(exp(
            lprobX + params1[values, ,]
        ), na.rm = TRUE) / sumlpX)
    }
    values <- unname(unlist(vtransform(values,
        auxmetadata = auxmetadata,
        Rout = 'original',
        Cout = 'original',
        Dout = 'original',
        Oout = 'original',
        Nout = 'original',
        Bout = 'original',
        variates <- auxmetadata$name,
        logjacobianOr = NULL)))

#### Calculate quantile for the frequency samples
    if(doquantiles){
        selsamples <- TRUE
    } else if(dosamples) {
        nmaxsamples <- nsamples
        selsamples <- round(seq(1, ncol(params1), length.out = nsamples))
    }

    if(doquantiles || dosamples) {
        params1 <- aperm(a = params1[, , selsamples],
            perm = c(1, 3, 2), resize = TRUE)
        lprobX <- t(lprobX[, selsamples])

        samples <- rep(1L, nmaxsamples)

        i <- 1L

        FF <- rowSums(exp(
            lprobX + params1[i, ,]
        ), na.rm = TRUE) / sumlpX

        tocheck <- FF < pY
        while(any(tocheck)) {
            i <-  i + 1L
            samples[tocheck] <- i
            FF <- rowSums(exp(
            lprobX + params1[i, ,]
            ), na.rm = TRUE) / sumlpX
            tocheck <- FF < pY
        }
        samples <- unname(unlist(vtransform(samples,
            auxmetadata = auxmetadata,
            Rout = 'original',
            Cout = 'original',
            Dout = 'original',
            Oout = 'original',
            Nout = 'original',
            Bout = 'original',
            variates <- auxmetadata$name,
            logjacobianOr = NULL)))
    }

    list(
        values = values,
        ##
        quantiles = if(doquantiles) {
            quantile(x = samples, probs = quantiles, type = 6,
                na.rm = TRUE, names = FALSE)
        },
        ##
        samples = if(dosamples) {
            samples[round(seq(1, length(samples), length.out = nsamples))]
        }
        ## values.MCaccuracy
        ## quantiles.MCaccuracy
)
}



#' Utility function to avoid finite-precision accuracys
#'
#' Used in 'rPr()', 'mutualinfo()', 'util_lprobsmi()'.
#'
#' @keywords internal
util_denorm <- function(lprob) {
    apply(X = lprob, MARGIN = 2, FUN = function(xx) {
        xx - max(xx[is.finite(xx)])
    }, simplify = TRUE)
}


#' Cumulative sum along first dimension
#'
#' Used in 'util_lprobsargsyx()'.
#'
#' @keywords internal
rowcumsum <- function(x){
    for(i in 2:(dim(x)[1])){
        x[i,,] <- x[i,,] + x[i-1,,]
    }
    x
}


#' Inverse cumulative sum along first dimension
#'
#' Used in 'util_lprobsargsyx()'.
#'
#' @keywords internal
rowinvcumsum <- function(x){
    for(i in (dim(x)[1] - 1):1){
        x[i,,] <- x[i,,] + x[i+1,,]
    }
    x
}

#' Create a grid of values for a variate
#'
#' This function creates a set of values for a variate, based on the information from data and metadata stored in a `learnt` object, created by the [learn()] function. The set of values depends on the type of variate (nominal or continuous, rounded, and so on, see [metadata]). The range of values is chosen to include, and extend slightly beyond, the range observed in the data used in the [learn()] function. Variate domains are always respected.
#'
#' @param vrt Character: name of the variate, must match one of the names in the `metadata` file provided to the [learn()] function.
#' @param learnt Either a character with the name of a directory or full path for a 'learnt.rds' object, produced by the [learn()] function, or such an object itself.
#' @param length.out Numeric, positive (default 129): number of values to be created; used only for continuous, non-rounded variates (see [`metadata`]).
#'
#' @return A numeric or character vector of values.
#'
#' @seealso
#' [learn()], which generates the `learnt` objects required by `vrtgrid()`.
#'
#' [Pr()] to calculate probabilities and their variability.
#'
#' [plot.probability()] to plot probabilities and quantiles calculated by `Pr()`.
#'
#' @examples
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## set of values for the variate "species";
#' ## since this variate is of a nominal kind, all values are included
#' valuesSpecies <- vrtgrid(vrt = 'species', learnt = learnt)
#'
#' print(valuesSpecies)
#'
#' ## create a set of values for the variate "bill length";
#' ## this variate is continuous and rounded, only realistic values are included
#' valuesBill <- vrtgrid(vrt = 'bill_len', learnt = learnt)
#'
#' range(valuesBill)
#'
#' ## let's take a subset of these values, to speed up computation
#' valuesBill <- valuesBill[seq(to = length(valuesBill), length.out = 65)]
#'
#' ## calculate the conditional probabilities for the 'bill_len' values above,
#' ## given the values of 'species'
#' probs <- Pr(
#'   Y = data.frame(bill_len = valuesBill),
#'   X = data.frame(species = valuesSpecies),
#'   learnt = learnt, parallel = 1
#' )
#'
#' ## plot the conditional probability distributions, and their variability
#' plot(probs)
#'
#' @import utils
#'
#' @export
vrtgrid <- function(
    vrt,
    learnt,
    length.out = 129
){

    ## Extract auxmetadata
    ## If learnt is a string, check if it's a folder name or file name
    if (is.character(learnt)) {
        ## Check if 'learnt' is a folder containing learnt.rds
        if (file_test('-d', learnt) &&
                file.exists(file.path(learnt, 'learnt.rds'))) {
            learnt <- readRDS(file.path(learnt, 'learnt.rds'))
        } else {
            ## Assume 'learnt' the full path of learnt.rds
            ## possibly without the file extension '.rds'
            learnt <- paste0(sub('.rds$', '', learnt), '.rds')
            if (file.exists(learnt)) {
                learnt <- readRDS(learnt)
            } else {
                stop('The argument "learnt" must be a folder containing learnt.rds, or the path to an rds-file containing the output from "learn()".')
            }
        }
    }
    ## Add check to see that learnt is correct type of object?
    auxmetadata <- learnt$auxmetadata
    rm(learnt)

    ## Consistency checks
    if(!(vrt %in% auxmetadata$name)){
        stop('Variate "', vrt, '" not present in the list of variates.')
    }

    adata <- as.list(auxmetadata[auxmetadata$name == vrt, ])

        if(adata$mcmctype %in% c('R', 'C')){
            out <- seq(adata$plotmin, adata$plotmax, length.out = length.out)
        } else if(adata$mcmctype == 'D'){
            out <- seq(adata$plotmin, adata$plotmax, by = 2 * adata$halfstep)
            ## step <- as.integer(ceiling((adata$plotmax - adata$plotmin) /
            ##                                (2 * adata$halfstep)))
            ## seqstep <- seq_len(step)
            ## factors <- seqstep[!(step %% seqstep)]
            ## step <- factors[which.min(abs(factors - length.out + 1))] *
            ##     2 * adata$halfstep
            ## out <- seq(adata$plotmin, adata$plotmax,
            ##     length.out = 1 + round((adata$plotmax - adata$plotmin)/step))
        } else if(adata$mcmctype %in% c('B', 'O', 'N')){
            out <- unname(unlist(adata[paste0('V', seq_len(adata$Nvalues))]))
        } else {
            stop('Unknown variate type for "', vrt, '".')
        }
    out
}

#' Generate datapoints
#'
#' @description This function generates datapoints of chosen joint variates, according to posterior probabilities and posterior conditional probabilities.
#'
#' @details This function generates datapoints according to the posterior probability \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data})} or \eqn{\mathrm{Pr}(Y = y \vert X \le x, \text{data})} or combinations thereof, for the variates specified in the argument `Y`, and conditional on the variate values specified in the argument `X`. It is somewhat analogous to the `r`-variants of R distribution functions, such as [stats::rnorm()]. If `X` is omitted or `NULL`, then the posterior probability \eqn{\mathrm{Pr}(Y | \text{data})} is used. Each variate in the argument `X` can be specified either as a point-value \eqn{X = x} or as a left-open interval \eqn{X \le x} or as a right-open interval \eqn{X \ge x}, through the argument `tails`.
#'
#' @param n Positive integer: number of samples to draw.
#' @param Ynames Character vector: names of variates to draw jointly
#' @param X List or data.table or `NULL`: set of values of variates on which we want to condition the joint probability for `Y`. If `NULL` (default), no conditioning is made. Any rows beyond the first are discarded
#' @param learnt Either a character with the name of a directory or full path for a 'learnt.rds' object, produced by the [learn()] function, or such an object itself.
#' @param tails Named vector or list, or `NULL` (default). The names must match some or all of the variates in arguments `X`. For variates in this list, the probability conditional is understood in an semi-open interval sense: \eqn{X \le x} or \eqn{X \ge x}, an so on. See analogous argument in [Pr()].
#' @param mcsamples Vector of integers, or `'all'`, or `NULL` (default): which Monte Carlo samples calculated by the [learn()] function should be used to draw the variate values. The default is to choose a random subset if `n` is smaller than their number, otherwise to recycle them as necessary.
#' @param parallel Not used: this function does not use parallelization.
#'
#' @return A [data frame][base::data.frame()] of joint draws of the variates `Ynames` from the posterior distribution, conditional on `X`. The row names of the data frame report the Monte Carlo sample (from [learn()]) used for that draw, and the total number of draws from that sample so far.
#'
#' @seealso
#' [learn()], which generates the `learnt` objects required by `qPr()`.
#'
#' [Pr()] to calculate joint and conditional probabilities.
#'
#' [qPr()] to calculate quantiles.
#'
#' @examples
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## ## Example 1:
#' ## Generate 10 values of the 'species' variate,
#' ## according to the frequency distribution estimated from the data
#'
#' datapoints <- rPr(
#'   n = 10,
#'   Ynames = 'species',
#'   learnt = learnt
#' )
#'
#' c(datapoints)
#'
#'
#' ## ## Example 2:
#' ## Generate 5 joint values of the 'species' and 'bill_len' variates.
#'
#' datapoints <- rPr(
#'   n = 5,
#'   Ynames = c('species', 'bill_len'),
#'   learnt = learnt
#' )
#'
#' print(datapoints, row.names = FALSE) ## row names give MCMC information
#'
#'
#' ## ## Example 3:
#' ## Generate 5 values of the 'species' variate,
#' ## for the subpopulation of penguins having bill length shorter than 40 mm
#'
#' datapoints <- rPr(
#'   n = 5,
#'   Ynames = 'species',
#'   X = data.frame(bill_len = 40),
#'   tails = list(bill_len = -1),
#'   learnt = learnt
#' )
#'
#' c(datapoints)
#'
#' @importFrom extraDistr rcat
#' @importFrom extraDistr rbern
#' @import utils
#' @import stats
#'
#' @export
rPr <- function(
    n,
    Ynames,
    X = NULL,
    learnt,
    tails = NULL,
    mcsamples = NULL,
    parallel = NULL # unused
) {

    ## Extract Monte Carlo output & aux-metadata
    ## If learnt is a string, check if it's a folder name or file name
    if (is.character(learnt)) {
        ## Check if 'learnt' is a folder containing learnt.rds
        if (file_test('-d', learnt) &&
                file.exists(file.path(learnt, 'learnt.rds'))) {
            learnt <- readRDS(file.path(learnt, 'learnt.rds'))
        } else {
            ## Assume 'learnt' the full path of learnt.rds
            ## possibly without the file extension '.rds'
            learnt <- paste0(sub('.rds$', '', learnt), '.rds')
            if (file.exists(learnt)) {
                learnt <- readRDS(learnt)
            } else {
                stop("The argument 'learnt' must be a folder containing learnt.rds, or the path to an rds-file containing the output from 'learn()'.")
            }
        }
    }
    ## Add check to see that learnt is correct type of object?
    auxmetadata <- learnt$auxmetadata
    learnt$auxmetadata <- NULL
    learnt$auxinfo <- NULL
    ncomponents <- nrow(learnt$W)
    nmcsamples <- ncol(learnt$W)

    if(is.null(mcsamples) ||
           (is.character(mcsamples) && mcsamples == 'all') ||
           isTRUE(mcsamples)) {
        mcsamples <- seq_len(nmcsamples)
    } else if (any(!is.finite(mcsamples)) || any(mcsamples < 1)) {
        stop("'mcsamples' should be a list of positive integers or NULL")
    } else {
        mcsamples <- round(mcsamples[mcsamples <= nmcsamples])
    }

    nmcs <- length(mcsamples)
    if(n <= nmcs) {
        sseq <- mcsamples[sort.int(sample.int(nmcs, n))]
    } else {
        sseq <- c(rep(x = mcsamples, times = n %/% nmcs),
            mcsamples[sort.int(sample.int(nmcs, n %% nmcs))])
    }

    if(all(is.na(X))){X <- NULL}
    if(!is.null(X)){
        X <- as.data.frame(X)
        if (nrow(X) > 1) {
            warning('Only the first row of X is considered')
            X <- X[1, , drop = FALSE]
        }
    }
    Xv <- colnames(X)

    if(!is.null(tails)){
        tails <- as.list(tails)
        if(is.null(names(tails))) {
            stop('Missing variate names in "tails"')
        }
    }
    tailsv <- names(tails)
    tailscentre <- list('==', 0, '0', NULL)
    tailsleft <- list('<=', -1, '-1', 'left')
    tailsright <- list('>=', 1, '+1', 'right')
    tailsvalues <- c(tailscentre, tailsleft, tailsright)

    ## Consistency checks
    if(!is.character(Ynames) || any(is.na(Ynames))){
        stop('Ynames must be a vector of variate names')
    }
    if(!all(Ynames %in% auxmetadata$name)) {
        stop('unknown Y variates\n')
    }
    if(length(unique(Ynames)) != length(Ynames)) {
        stop('duplicate Y variates\n')
    }

    if (!all(Xv %in% auxmetadata$name)) {
        stop('unknown X variates\n')
    }
    if (length(unique(Xv)) != length(Xv)) {
        stop('duplicate X variates\n')
    }
    if(any(Ynames %in% Xv)) {
        stop('overlap in Y and X variates\n')
    }

    if (!all(tailsv %in% Xv)) {
        warning('"tails" variate ',
            paste0(tailsv[!(tailsv %in% Xv)], collapse = ' '),
            ' not among X; ignored\n')
        tails <- tails[tailsv %in% Xv]
        tailsv <- names(tails)
    }
    if (length(unique(tailsv)) != length(tailsv)) {
        stop('duplicate "tails" variates\n')
    }
    if(!all(tails %in% tailsvalues)) {
        stop('"tails" values must be ',
            paste0(tailsvalues, collapse = ' '), '\n')
    }

    ## transform 'tails' to -1, +1
    ## +1: '<=',    -1: '>='
    ## this is opposite of the argument convention because
    ## interval probabilities are calculated with `lower.tail = TRUE`
    ## eg:
    ## pnorm(x, mean, sd, lower.tail = FALSE) ==
    ##     pnorm(-x, -mean, sd, lower.tail = TRUE)
    tails[tails %in% tailscentre] <- NULL
    cleft <- tails %in% tailsleft
    cright <- tails %in% tailsright
    tails[cleft] <- +1
    tails[cright] <- -1
    tails <- unlist(tails)


#### Adjust component weights W for conditioning on X
    if(is.null(X)){
        lW <- log(learnt$W)
    } else {
        lpargs <- util_lprobsargsyx(
            x = X,
            auxmetadata = auxmetadata,
            learnt = learnt,
            tails = tails
        )

        lW <- util_lprobsbase(
            xVs = lpargs$xVs[[1]],
            params = lpargs$params,
            logW =  log(learnt$W)
        ) # rows=components, columns=samples

    } # end definition of lW if non-null X


    ## Utility function to avoid finite-precision errors
    ## now externally defined
    ## denorm <- function(lprob) {
    ##     apply(X = lprob, MARGIN = 2, FUN = function(xx) {
    ##         xx - max(xx[is.finite(xx)])
    ##     }, simplify = TRUE)
    ## }

#### Draw samples of Ynames

    lWnorm <- util_denorm(lW[, sseq, drop = FALSE])
    Ws <- extraDistr::rcat(n = n, prob = t(
        apply(X = lWnorm, MARGIN = 2, FUN = function(xx){
            xx <- exp(xx)
            xx/sum(xx, na.rm = TRUE)
        }, simplify = TRUE)
    ))

    Yout <- NULL
    vYout <- NULL

    ## R
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'R'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        aux <- auxmetadata[toselect, ]
        vYout <- c(vYout, aux$name)
        totake <- cbind(rep.int(x = aux$id, times = rep(n, nvrt)), Ws, sseq)
        Yout <- c(Yout,
            rnorm(n = n * nvrt,
                mean = learnt$Rmean[totake],
                sd = learnt$Rsd[totake] )
        )
    }

    ## C
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'C'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        aux <- auxmetadata[toselect, ]
        vYout <- c(vYout, aux$name)
        totake <- cbind(rep.int(x = aux$id, times = rep(n, nvrt)), Ws, sseq)
        Yout <- c(Yout,
            rnorm(n = n * nvrt,
                mean = learnt$Cmean[totake],
                sd = learnt$Csd[totake] )
        )
    }

    ## D
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'D'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        aux <- auxmetadata[toselect, ]
        vYout <- c(vYout, aux$name)
        totake <- cbind(rep.int(x = aux$id, times = rep(n, nvrt)), Ws, sseq)
        Yout <- c(Yout,
            rnorm(n = n * nvrt,
                mean = learnt$Dmean[totake],
                sd = learnt$Dsd[totake] )
        )
    }

    ## O
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'O'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        vYout <- c(vYout, auxmetadata$name[toselect])
        for(i in toselect) {
            aux <- auxmetadata[i, ]
            totake <- cbind(Ws, sseq)
            Yout <- c(Yout,
                extraDistr::rcat(n = n,
                    prob = apply(
                        X = learnt$Oprob[aux$indexpos + seq_len(aux$Nvalues), ,],
                        MARGIN = 1, FUN = `[`, totake,
                        simplify = TRUE) )
            )
        }
    }

    ## N
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'N'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        vYout <- c(vYout, auxmetadata$name[toselect])
        for(i in toselect) {
            aux <- auxmetadata[i, ]
            totake <- cbind(Ws, sseq)
            Yout <- c(Yout,
                extraDistr::rcat(n = n,
                    prob = apply(
                        X = learnt$Nprob[aux$indexpos + seq_len(aux$Nvalues), ,],
                        MARGIN = 1, FUN = `[`, totake,
                        simplify = TRUE) )
            )
        }
    }

    ## B
    toselect <- which((auxmetadata$name %in% Ynames) &
                          (auxmetadata$mcmctype == 'B'))
    nvrt <- length(toselect)
    if(nvrt > 0){
        aux <- auxmetadata[toselect, ]
        vYout <- c(vYout, aux$name)
        totake <- cbind(rep.int(x = aux$id, times = rep(n, nvrt)), Ws, sseq)
        Yout <- c(Yout,
            extraDistr::rbern(n = n * nvrt,
                prob = learnt$Bprob[totake])
        )
    }

    dim(Yout) <- c(n, length(Ynames))
    Yout <- Yout[, match(Ynames, vYout), drop = FALSE]
    colnames(Yout) <- Ynames

    Yout <- vtransform(Yout,
        auxmetadata = auxmetadata,
        Rout = 'original',
        Cout = 'original',
        Dout = 'original',
        Oout = 'original',
        Nout = 'original',
        Bout = 'original',
        logjacobianOr = NULL)

    ## row-name scheme: 'mcsample.draw'
    rownames(Yout) <- paste0(sseq, '_', ((seq_len(n) - 1L) %/% nmcs) + 1L)

    Yout
}

#' Format datapoints used for MCMC monitoring
#'
#' Used in 'util_Pcheckpoints()' within 'learn()'.
#'
#' @param x Datapoints to be used for checking MCMC progress
#' @param auxmetadata auxmetadata object
#' @param pointsid Id of datapoints
#'
#' @return some arguments to be repeatedly used in util_Pcheckpoints
#'
#' @keywords internal
util_prepPcheckpoints <- function(
    x, auxmetadata, pointsid = NULL
){

    nX <- nrow(x)

    auxV0a <- auxV0b <- auxV1a <- auxV1b <- auxV1c <- auxV1d <- NULL
    auxV2 <- auxVNO <- auxVNN <- auxVB <- NULL
    V2steps <- NULL

    nV0 <- nV1 <- nV2 <- nVN <- nVB <- FALSE

    xV0 <- xV1 <- xV2 <- xVN <- xVB <- matrix(data = NA_real_,
        nrow = 0, ncol = nX, dimnames = NULL)

###
### point probability density
###

### R-variates not in 'cumul'
    toselect <- which(auxmetadata$mcmctype == 'R')
    if(length(toselect) > 0){
        nV0 <- TRUE
        aux <- auxmetadata[toselect, ]
        auxV0a <- aux$id
        xV0 <- rbind(xV0,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Rout = 'normalized',
                logjacobianOr = NULL
            )))
        )
    }

### C-variates not in 'cumul' and with some non-boundary value
    toselect <- which(auxmetadata$mcmctype == 'C')
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] > auxmetadata$domainmin[i] &
                    x[,auxmetadata$name[i]] < auxmetadata$domainmax[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        nV0 <- TRUE
        aux <- auxmetadata[toselect, ]
        auxV0b <- aux$id
        xV0 <- rbind(xV0,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Cout = 'boundisna',
                logjacobianOr = NULL
            )))
        )
    }

###
### tail probability
###

### C-variates not in 'cumul' and with left boundary values
    toselect <- which(auxmetadata$mcmctype == 'C')
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] <= auxmetadata$domainmin[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        nV1 <- TRUE
        aux <- auxmetadata[toselect, ]
        auxV1a <- aux$id
        xV1 <- rbind(xV1,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Cout = 'leftbound',
                logjacobianOr = NULL
            )))
        )
    }

### C-variates not in 'cumul' and with right boundary values
    toselect <- which(auxmetadata$mcmctype == 'C')
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] >= auxmetadata$domainmax[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        nV1 <- TRUE
        aux <- auxmetadata[toselect, ]
        auxV1b <- aux$id
        xV1 <- rbind(xV1,
            - t(as.matrix(vtransform( # minus sign
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Cout = 'rightbound',
                logjacobianOr = NULL
            )))
        )
    }

### D-variates not in 'cumul' and with left boundary values
    toselect <- which(auxmetadata$mcmctype == 'D')
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] <= auxmetadata$domainminplushs[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        nV1 <- TRUE
        aux <- auxmetadata[toselect, ]
        auxV1c <- aux$id
        xV1 <- rbind(xV1,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Dout = 'leftbound',
                logjacobianOr = NULL
            )))
        )
    }

### D-variates not in 'cumul' and with right boundary values
    toselect <- which(auxmetadata$mcmctype == 'D')
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] >= auxmetadata$domainmaxminushs[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        nV1 <- TRUE
        aux <- auxmetadata[toselect, ]
        auxV1d <- aux$id
        xV1 <- rbind(xV1,
            - t(as.matrix(vtransform( # minus sign
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Dout = 'rightbound',
                logjacobianOr = NULL
            )))
        )
    }

###
### interval probability
###

### D-variates not in 'cumul'
    toselect <- which(auxmetadata$mcmctype == 'D')
    if(length(toselect) > 0) {
        toselect <- toselect[sapply(toselect, function(i){
            any(x[,auxmetadata$name[i]] > auxmetadata$domainminplushs[i] &
                    x[,auxmetadata$name[i]] < auxmetadata$domainmaxminushs[i],
                na.rm = TRUE)
        })]
    }
    if(length(toselect) > 0){
        nV2 <- TRUE
        aux <- auxmetadata[toselect, ]
        auxV2 <- aux$id
        V2steps <- aux$halfstep / aux$tscale
        xV2 <- rbind(xV2,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Dout = 'boundisna',
                logjacobianOr = NULL
            )))
        )
    }

###
### discrete case
###
    Nshift <- 0L
### O-variates not in 'cumul'
    toselect <- which(auxmetadata$mcmctype == 'O')
    if(length(toselect) > 0){
        nVN <- TRUE
        aux <- auxmetadata[toselect, ]
        auxVNO <- aux$id
        Nindices <- unlist(mapply(FUN = function(i, n) {i + seq_len(n)},
            aux$indexpos, aux$Nvalues,
            SIMPLIFY = FALSE))
        xVN <- rbind(xVN,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Oout = 'numeric',
                logjacobianOr = NULL
            ))) +
                Nshift + c(0, cumsum(aux$Nvalues[-1]))
        )
        Nshift <- Nshift + length(Nindices)
    }
### N-variates
    toselect <- which(auxmetadata$mcmctype == 'N')
    if(length(toselect) > 0){
        nVN <- TRUE
        aux <- auxmetadata[toselect, ]
        auxVNN <- aux$id
        Nindices <- unlist(mapply(FUN = function(i, n) {i + seq_len(n)},
            aux$indexpos, aux$Nvalues,
            SIMPLIFY = FALSE))
        xVN <- rbind(xVN,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Nout = 'numeric',
                logjacobianOr = NULL
            ))) +
                Nshift + c(0, cumsum(aux$Nvalues[-1]))
        )
    }

###
### binary case
###

### B-variates
    toselect <- which(auxmetadata$mcmctype == 'B')
    if(length(toselect) > 0){
        nVB <- TRUE
        aux <- auxmetadata[toselect, ]
        auxVB <- aux$id
        xVB <- rbind(xVB,
            t(as.matrix(vtransform(
                x[, aux$name, drop = FALSE],
                auxmetadata = auxmetadata,
                Bout = 'numeric',
                logjacobianOr = NULL
            )))
        )
    }

    list(
        nV0 = nV0, nV1 = nV1, nV2 = nV2, nVN = nVN, nVB = nVB,
        ## xV0 = xV0, xV1 = xV1, xV2 = xV2, xVN = xVN, xVB = xVB,
        auxV0a = auxV0a, auxV0b = auxV0b,
        auxV1a = auxV1a, auxV1b = auxV1b, auxV1c = auxV1c, auxV1d = auxV1d,
        auxV2 = auxV2,
        auxVNO = auxVNO, auxVNN = auxVNN,
        auxVB = auxVB,
        V2steps = V2steps,
        pointsid = pointsid,
        ##
        xVs = lapply(seq_len(nX), function(i){
            list(
                xV0 = xV0[,i],
                xV1 = xV1[,i],
                xV2 = xV2[,i],
                xVN = xVN[,i],
                xVB = xVB[,i]
            )})
    )
}




#' Calculate joint frequencies for MCMC-monitoring checkpoints
#'
#' Used in 'learn()'.
#'
#' @param testdata List of objects calculated with util_prepPcheckpoints
#' @param learnt mcsamples object
#'
#' @keywords internal
#'
#' @return The joint frequencies of Y corresponding to the Monte Carlo samples
util_Pcheckpoints <- function(
    testdata, learnt
) {
    with(c(testdata, learnt), {

        nsamples <- ncol(W)
        ncomponents <- nrow(W)


###
### point probability density
###
        V0mean <- V0sd <- NULL

### R-variates not in 'cumul'
        if(length(auxV0a) > 0){
            V0mean <- learnbind(V0mean, Rmean)
            V0sd <- learnbind(V0sd, sqrt(Rvar)) # still variance
        }

### C-variates not in 'cumul' and with some non-boundary value
        if(length(auxV0b) > 0) {
            V0mean <- learnbind(V0mean,
                Cmean[auxV0b, , , drop = FALSE])
            V0sd <- learnbind(V0sd,
                sqrt(Cvar[auxV0b, , , drop = FALSE])) # still variance
        }

###
### tail probability
###
        V1mean <- V1sd <- NULL

### C-variates not in 'cumul' and with left boundary values
        if(length(auxV1a) > 0){
            V1mean <- learnbind(V1mean,
                Cmean[auxV1a, , , drop = FALSE])
            V1sd <- learnbind(V1sd,
                sqrt(Cvar[auxV1a, , , drop = FALSE])) # still variance
        }

### C-variates not in 'cumul' and with right boundary values
        if(length(auxV1b) > 0){
            V1mean <- learnbind(V1mean,
                - Cmean[auxV1b, , , drop = FALSE]) # minus sign
            V1sd <- learnbind(V1sd,
                sqrt(Cvar[auxV1b, , , drop = FALSE])) # still variance
        }

### D-variates not in 'cumul' and with left boundary values
        if(length(auxV1c) > 0){
            V1mean <- learnbind(V1mean,
                Dmean[auxV1c, , , drop = FALSE])
            V1sd <- learnbind(V1sd,
                sqrt(Dvar[auxV1c, , , drop = FALSE])) # still variance
        }

### D-variates not in 'cumul' and with right boundary values
        if(length(auxV1d) > 0){
            V1mean <- learnbind(V1mean,
                - Dmean[auxV1d, , , drop = FALSE]) # minus sign
            V1sd <- learnbind(V1sd,
                sqrt(Dvar[auxV1d, , , drop = FALSE])) # still variance
        }

###
### interval probability
###
        V2mean <- V2sd <- NULL

### D-variates not in 'cumul'
        if(length(auxV2) > 0){
            V2mean <- Dmean
            V2sd <- sqrt(Dvar) # still variance
        }

###
### discrete case
###
        VNprobs <- NULL

### O-variates not in 'cumul'
        if(length(auxVNO) > 0){
            VNprobs <- learnbind(VNprobs,
                Oprob)
        }
### N-variates
        if(length(auxVNN) > 0){
            VNprobs <- learnbind(VNprobs,
                Nprob)
        }

###
### binary case
###
        VBprobs <- NULL

### B-variates
        if(length(auxVB) > 0){
            VBprobs <- Bprob
        }


        lprobX <- log(W)
        do.call(cbind,
            lapply(
                X = xVs,
                FUN = function(xx){
                    lprobY <- util_lprobsbase(
                        xVs = xx,
                        params = list(
                            nV0 = nV0,
                            V0mean = V0mean,
                            V0sd = V0sd,
                            nV1 = nV1,
                            V1mean = V1mean,
                            V1sd = V1sd,
                            nV2 = nV2,
                            V2mean = V2mean,
                            V2sd = V2sd,
                            V2steps = V2steps,
                            nVN = nVN,
                            VNprobs = VNprobs,
                            nVB = nVB,
                            VBprobs = VBprobs
                            ),
                        logW = lprobX,
                        temporarydir = NULL
                    )
                   ## Output: rows=components, columns=samples
                    colSums(exp(lprobY)) / colSums(W)
                })
            )
    }) # Output: rows=components, columns=samples
}

#' Monte Carlo computation of posterior probability distribution
#'
#' @description
#' Compute the posterior joint probability distribution of the variates conditional on the given data, by means of Markov-chain Monte Carlo, using the package **Nimble**.
#'
#' @details
#' This function takes as main inputs a set of data and metadata, and computes the full joint probability distribution for new data, including its variability. From this full joint distribution any other distributions of interest can subsequently be computed; see [Pr()] and related functions. This computation can also be interpreted as an estimation of the full joint frequency distribution of the variates in the *whole population*, beyond the sample data, together with its uncertainty. The computation allows for the use of datapoints with partially missing variables: imputation is automatically made. This imputation is *principled*, made according to the rules of probability theory.
#'
#' The output is a "learnt" object, typically saved in a `learnt.rds` file, which is used in all subsequent probabilistic computations. Other information about the computation is provided in logs and plots, saved in a directory specified by the user.
#'
#' See `vignette('intro')` for introductory examples.
#'
#' The computation is "non-parametric": probability or frequency distributions are not assumed to be Gaussian or of any other specific shape; no "model" is assumed. The mathematical representation of the space of joint frequency distributions follows ideas of Dunson & Bhattacharya (2011); see [technical manual](https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf) for details.
#'
#' The computation is done via Markov-chain Monte Carlo, using the package [**Nimble**](https://cran.r-project.org/package=nimble). "Convergence" of the Monte Carlo computation is automatically assessed with methods described in Vehtari & al. (2021) and Kwon & al. (2025); see [technical manual](https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf) for details. The default values for convergence require that all of the following three conditions be fulilled:
#'
#' - The computation's numerical error (Monte-Carlo Standard Error) for the posterior probability must be smaller than 4.7% of the standard deviation of the posterior's variability.
#'  - The computation's numerical error for the 0.055- and 0.945-percentiles of the posterior's variability should be smaller than 4.7% of the distance between them.
#'
#' Typically this requirement leads to final results obtained with the [Pr()] function having at least two significant digits.
#'
#' The `learn()` function can take hours or even days to perform its computations, depending on the size of the dataset, number of variates, and the (initially unknown) "shape" of the underlying probability distribution. For this reason it is typically called within an R script, executed via [utils::Rscript]. For example, a script 'myscript.R' could have the following structure:
#'
#' ```
#' library('prova')
#'
#' learn(
#'   data = 'filename_with_data.csv', # CSV file containing the dataset
#'   metadata = 'filename_with_metadata.csv', # CSV file containing the metadata
#'   outputdir = 'some_directory', # path to output directory
#'   parallel = 8 # machine has more than 8 cores, so we use 8
#'   ## possibly other arguments to learn()
#' )
#' ```
#'
#' and then be called on a bash terminal with
#'
#' ```
#' $ Rscript myscript.R > learnoutput.log 2>&1 &
#' ```
#'
#' with such a call, the file 'learnoutput.log' will contain information about how the computation is proceeding and the estimated end time.
#'
#' @param data A dataset, given as a [base::data.frame()] or as a file path to a CSV file. If missing or `NULL`, then the prior probability distribution is calculated.
#' @param metadata [metadata] about the dataset's variates, given either as a [data frame][base::data.frame()] or as a file path to a CSV file.
#' @param auxdata A larger dataset, given as a data frame or as a file path to a CSV file. Such a dataset would be too large to use in the Monte Carlo sampling, but can still be used to help estimate some hyperparameters.
#' @param outputdir `NULL` (default) or `NA` or character: path to folder where output information and diagnostics should be saved. If `NULL`, a directory is created in the temporary-directory space given by [base::tempdir()]. If `NA`, a directory is created in the current working directory given by [base::getwd()]. If character, this is taken to be the output directory; it should of course be writable by the user.
#' @param valueislearnt Logical or `NULL`: should the `VALUE` returned be the `learnt` object containing the results from the Monte Carlo computation? Default `TRUE`. If `FALSE`, then `VALUE` is the output directory name. If `NULL`, then `VALUE` is `NULL`.
#' @param nsamples Integer, default 3600: number of desired, *approximately independent* Monte Carlo samples. If this argument is changed, the user is also required to explicitly give either `nchains` or `nsamplesperchain`, but not both; the remaining third argument is determined from \eqn{\mathrm{nsamples} = \mathrm{nchains} \times \mathrm{nsamplesperchain}}.
#' @param nchains Integer, default 8: number of Monte Carlo chains. If this argument is changed, the user is also required to explicitly give either `nsamples` or `nsamplesperchain`, but not both; the remaining third argument is determined from \eqn{\mathrm{nsamples} = \mathrm{nchains} \times \mathrm{nsamplesperchain}}.
#' @param nsamplesperchain Integer, default 450: number of *approximately independent* Monte Carlo samples per chain. If this argument is changed, the user is also required to explicitly give either `nsamples` or `nchains`, but not both; the remaining third argument is determined from \eqn{\mathrm{nsamples} = \mathrm{nchains} \times \mathrm{nsamplesperchain}}.
#' @param parallel Logical or positive integer or cluster object. `TRUE` (default): use roughly half of available cores; `FALSE` (default): use serial computation; integer: use this many cores. It can also be a cluster object previously created with [parallel::makeCluster()]; in this case the parallel computation will use this object.
#' @param seed Integer or `NULL` (default): use this seed for the random number generator. If `NULL`, do not set the seed.
#' @param cleanup Logical, default `TRUE`: remove diagnostic files at the end of the computation?
#' @param appendinfo Logical, default `TRUE`: append information about number of variates ('V'), number of data points ('D'), number of Monte Carlo samples ('S'), and timestamp, to the name of the output directory `outputdir`? The appended string has the format 'Vn_Dn_Sn_YYMMDDTHHMMSS'.
#' @param subsampledata Integer or `NULL` (default): if integer, use only that many datapoints from the original dataset in the `data` argument.
#' @param prior Logical: Calculate the prior distribution? Default is `FALSE` unless `data` argument is missing or `NULL`.
#' @param startupMCiterations Integer, default 3600: number of initial Monte Carlo iterations.
#' @param minMCiterations Integer, default 0: minimum number of Monte Carlo iterations to be doneby a chain.
#' @param maxMCiterations Integer, default `Inf`: Do at most this many Monte Carlo iterations per chain.
#' @param maxhours Numeric, default `Inf`: approximate time limit, in hours, for the Monte Carlo computation to last.
#' @param ncheckpoints Integer or `NULL`, default 12: number of datapoints (per chain) to use for checking when the Monte Carlo computation should end. If `NULL`, this is equal to number of variates + 2. If Inf, use all datapoints.
#' @param maxrelMCSE Numeric positive, default `+Inf`: desired maximal *relative Monte Carlo Standard Error* of calculated probabilities with respect to their variability with new data. The default `+Inf` means that `minESS` is used instead. `maxrelMCSE` is related to `minESS` by \eqn{\mathrm{maxrelMCSE} = 1/\sqrt{\mathrm{minESS} + \mathrm{initES}}}.
#' @param minESS Numeric positive or `NULL`, default 450: desired minimal Monte Carlo *Expected Sample Size*. If `NULL`, it is equal to the final `nsamplesperchain`. `minESS` is related to `maxrelMCSE` by \eqn{\mathrm{minESS} = 1/\mathrm{maxrelMCSE}^2 - \mathrm{initES}}.
#' @param initES Numeric positive, default 2: number of initial "burn-in" samples, separated by the Expected Sample Size, to be discarded. Note that the Monte Carlo chain typically starts in a high-probability region, so there is no reason to discard many initial samples.
#' @param thinning Integer or `NULL` (default): thin out the Monte Carlo samples by this value. If `NULL`: let the diagnostics decide the thinning value.
#' @param verbose Logical, default `TRUE`: output the progress to terminal? If `FALSE`, the progress is outputted to the file `'main.log'` in the `outputdir` directory.
#' @param plottraces Logical, default `TRUE`: save plots of the Monte Carlo traces of diagnostic values?
#' @param showKtraces Logical, default `FALSE`: save plots of the Monte Carlo traces of the K parameter?
#' @param showAlphatraces Logical, default `FALSE`: save plots of the Monte Carlo traces of the Alpha parameter?
#' @param hyperparams List: hyperparameters of the hyperprior; see values in "Usage".
#'
#' @returns A "learnt" object, or name of directory containing such an object and other output files, or `NULL`, depending on argument `valueislearnt`.
#'
#' `learn()` saves several files in a directory. By default this output directory is a temporary directory within the one used by [base::tempdir()], but an alternative one can be chosen with the argument `outputdir =`. The output directory contain several diagnostic files for the Monte Carlo computation; in particular:
#'
#' - `MCtraces.pdf`: shows several trace plots of the Monte Carlo sampling; the correspondin data are in the file `MCtraces.rds`.
#' - `plotsamples_learnt.pdf`, `plotquantiles_learnt.pdf`: show the marginal posterior distributions of each individual variate, together with their variability (as samples or quantiles).
#' - `log-1.log`, `log-2.log`, ... one for each parallel core; report the progress of each parallel Monte Carlo computation and notes about it.
#' - `rng_seed.rds`: the state of the pseudorandom seed (see [base::Random]) when `learn()` was called.
#' - `metadata.csv`: a copy of the metadata.
#'
#' It is recommended that you give an explicit argument `outputdir =` and save the directory with the files above for future reference. In particular, the `MCtraces.pdf` plot and `MCtraces.rds` data can be useful to report Monte Carlo convergence in any work of yours that used **Prova**.
#'
#' @references
#'
#' For the mathematical representation of the frequency space:
#'
#' - Dunson, Bhattacharya (2011): *Nonparametric Bayes regression and classification through mixtures of product kernels* <doi:10.1093/acprof:oso/9780199694587.003.0005>.
#' - Ishwaran, Zarepour (2002): *Exact and approximate sum representations for the Dirichlet process* <doi:10.2307/3315951>.
#' - Porta Mana <https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf>.
#'
#' About Bayesian inference under exchangeability ("population inference"):
#'
#' - Lindley, Novick (1981): *The role of exchangeability in inference*, <doi:10.1214/aos/1176345331>.
#' - Bernardo, Smith (2000): *Bayesian Theory*. Wiley <doi:10.1002/9780470316870>.
#' - Porta Mana <https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf>.
#'
#' About nonparametrics:
#'
#' - Müller et al. (2015): *Nonparametric Bayesian inference*. IMS <doi:10.1007/978-3-319-18968-0>.
#' - Hjort et al. (2010): *Bayesian Nonparametrics*. Cambridge University Press <doi:10.1017/CBO9780511802478>.
#'
#' About Markov-chain Monte Carlo and "convergence":
#'
#' - de Valpine, Paciorek, Turek, & al. (2026): *NIMBLE: MCMC, Particle Filtering, and Programmable Hierarchical Modeling* <doi:10.5281/zenodo.1211190>, <https://cran.r-project.org/package=nimble>.
#' - Kwon & al. (2025): *MCMC stopping rules in latent variable modelling* <doi:10.1111/bmsp.12357>.
#' - Vehtari & al. (2021): *Rank-normalization, folding, and localization: an improved R-hat for assessing convergence of MCMC* <doi:10.1214/20-BA1221>.
#' - Roy (2020): *Convergence diagnostics for Markov chain Monte Carlo* <doi:10.1146/annurev-statistics-031219-041300>.
#' - Gilks & al. (1998): *Markov Chain Monte Carlo in Practice*. Chapman & Hall/CRC <doi:10.1201/b14835>.
#' - D. J. C. MacKay (2005): *Information Theory, Inference, and Learning Algorithms*. Cambridge University Press <https://www.inference.org.uk/itila/book.html>.
#' - Porta Mana <https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf>.
#'
#'
#' @seealso
#' [metadatatemplate()] to help writing metadata files.
#'
#' [Pr()] to calculate probabilities, and [qPr()] to calculate quantiles, given the data processed by `learn()`.
#'
#' [rPr()] to generate datapoints similar to the data processed by `learn()`.
#'
#' [mutualinfo()] to calculate mutual information given the data processed by `learn()`.
#'
#' [pread.csv()] and [pwrite.csv()] to read and write CSV files in the format used by `learn()`.
#'
#' @examples
#' \donttest{
#' ### WARNING: the following example, if run, might even take a minute or more.
#'
#' ## Create dataset with 3 points of variate 'V' for demonstration:
#' dataset <- data.frame(V = rnorm(n = 3))
#'
#' ## Create metadata file:
#' metadata <- data.frame(name = 'V', type = 'continuous')
#'
#' ## Learn from the data:
#' learnt <- learn(
#'   data = dataset, metadata = metadata,
#'       ## the following parameters are unrealistic
#'       ## only used to reduce computation time for this example
#'   nsamples = 10, nchains = 1,
#'   startupMCiterations = 10, maxMCiterations = 10,
#'   minESS = 0, initES = 0
#' )
#'
#' ## Check structure of `learnt` object:
#' str(learnt)
#' }
#'
#' @import grDevices
#' @import graphics
#' @import parallel
#' @import stats
#' @import utils
#'
#' @export
## #' @rawNamespace import(nimble, except = rcat)
learn <- function(
    data,
    metadata,
    auxdata = NULL,
    outputdir = NULL,
    nsamples = 3600,
    nchains = 8,
    nsamplesperchain = 450,
    parallel = TRUE,
    seed = NULL,
    cleanup = TRUE,
    appendinfo = TRUE,
    valueislearnt = TRUE,
    subsampledata = NULL,
    prior = missing(data) || is.null(data),
    startupMCiterations = 3600,
    minMCiterations = 0,
    maxMCiterations = +Inf,
    maxhours = +Inf,
    ncheckpoints = 12,
    maxrelMCSE = +Inf,
    minESS = 450,
    initES = 2,
    thinning = NULL,
    verbose = TRUE,
    plottraces = !cleanup,
    showKtraces = FALSE,
    showAlphatraces = FALSE,
    hyperparams = list(
        ncomponents = 64,
        minalpha = -4,
        maxalpha = 4,
        byalpha = 1,
        Rshapelo = 0.5,
        Rshapehi = 0.5,
        Rvarm1 = 3^2,
        Cshapelo = 0.5,
        Cshapehi = 0.5,
        Cvarm1 = 3^2,
        Dshapelo = 0.5,
        Dshapehi = 0.5,
        Dvarm1 = 3^2,
        Bshapelo = 1,
        Bshapehi = 1,
        Dthreshold = 1,
        tscalefactor = 4.266,
        Oprior = 'Hadamard',
        Nprior = 'Hadamard',
        avoidzeroW = NULL, # NULL: Turek's, TRUE: eps non-conj., FALSE: conj.
        initmethod = 'datacentre',
        Qerror = pnorm(c(-1, 1))
        ## Qerror = c(0.055, 0.945) # pnorm(c(-1, 1))
    )
) {
    ## Collect preliminary messages that will be printed to
    ## the main log file in 'outputdir' (and to terminal if verbose = TRUE)
    ## They cannot yet be printed because the 'outputdir' name
    ## may depend on the output of the messages themselves
    predirmsgs <- list(paste0('Prova v', packageVersion('prova'), '.'))


    ## Handles R CMD check:
    ## these are variables or functions handled in a special way by Nimble
    ## but R CMD check erroneously sees them as "rogue"
    if(FALSE){
        Dthreshold <- tscalefactor <- minalpha <- maxalpha <- byalpha <- ncomponents <- Rvarm1 <- Rshapelo <- Rshapehi <- Cvarm1 <- Cshapelo <- Cshapehi <- Dvarm1 <- Dshapelo <- Dshapehi <- Oprior <- Nprior <- Bshapelo <- Bshapehi <- avoidzeroW <- initmethod <- Qerror <- NULL
    }

##################################################
#### Various internal parameters
##################################################

#### Hyperparameters and other internal parameters
    hyperparamsDefaults <- list(
        ncomponents = 64,
        minalpha = -4,
        maxalpha = 4,
        byalpha = 1,
        Rshapelo = 0.5,
        Rshapehi = 0.5,
        Rvarm1 = 3^2,
        Cshapelo = 0.5,
        Cshapehi = 0.5,
        Cvarm1 = 3^2,
        Dshapelo = 0.5,
        Dshapehi = 0.5,
        Dvarm1 = 3^2,
        Bshapelo = 1,
        Bshapehi = 1,
        Dthreshold = 1,
        tscalefactor = 4.266,
        Oprior = 'Hadamard',
        Nprior = 'Hadamard',
        avoidzeroW = NULL,
        initmethod = 'datacentre',
        Qerror = pnorm(c(-1, 1))
        ## Qerror = c(0.055, 0.945) # pnorm(c(-1, 1))
    )

    ## Reference credibility-interval width
    ## maybe in future versions there will be arguments
    ## for these parameters
    Qlo <- 0.055 # (100 - 89) / 200 # 0.055
    Qhi <- 0.945 # (100 + 89) / 200 # 0.945

    ## Allow user to specify hyperparameters only partially:
    ## missing ones are given default values above
    ## Check for unknown names first
    if(length(
        setdiff(names(hyperparams), names(hyperparamsDefaults))
    ) > 0){
        stop('Unknown hyperparameters:',
            setdiff(names(hyperparams), names(hyperparamsDefaults))
        )
    }

    for(aname in names(hyperparamsDefaults)){
        if(is.null(hyperparams[[aname]])){
            hyperparams[[aname]] <- hyperparamsDefaults[[aname]]
        }
        assign(aname, hyperparams[[aname]])
    }

#### Other options
    Alphatoslice <- FALSE # FALSE typically leads to underflow
    Ktoslice <- FALSE # FALSE typically leads to underflow
    RWtoslice <- FALSE
    changeSamplerOrder <- TRUE
    ##
    showsamples <- 100 # number of samples to show.
    plotDisplayedQuantiles <- c(5.5, 94.5)/100 # c(1, 31) / 32 # quantiles to show
    ncomponentsamples <- 128 # number of samples of Alpha and K
    showsamplertimes <- FALSE ##
    family <- 'Palatino' # font family in plots


#### Start timer
    timestart0 <- Sys.time()

    ## Set the RNG seed if given by user, or if no seed already exists
    if (!is.null(seed) || !exists('.Random.seed')) {
        set.seed(seed)
    }
    currentseed <- .Random.seed

##################################################
#### Argument-consistency checks
##################################################

#### Consistency checks for numbers of samples, chains, cores
    ## The defaults are 3600 samples from 8 chains, 450 samples per chain
    ## The user can choose any two
    ## nsamples = nchains * nsamplesperchain

    if(!missing(nchains) && !missing(nsamplesperchain) &&
           missing(nsamples)){
        ## given: nchains, nsamplesperchain
        nsamples <- nchains * nsamplesperchain
    } else if (missing(nchains) && !missing(nsamplesperchain) &&
                   !missing(nsamples)){
        ## given: nsamples, nsamplesperchain
        nchains <- ceiling(nsamples / nsamplesperchain)
        if(nsamples != nchains * nsamplesperchain){
            nsamples <- nchains * nsamplesperchain
            predirmsgs <- c(predirmsgs, list(
                paste0('Increasing number of samples to ', nsamples,
                    ' to comply with given "nsamplesperchain".')
            ))
        }
    } else if (!missing(nchains) && missing(nsamplesperchain) &&
                   !missing(nsamples)){
        ## given: nsamples, nchains
        nsamplesperchain <- ceiling(nsamples / nchains)
        if(nsamples != nchains * nsamplesperchain){
            nsamples <- nchains * nsamplesperchain
            predirmsgs <- c(predirmsgs, list(
                paste0('Increasing number of samples to ', nsamples,
                    ' to comply with given "nchains".')
            ))
        }
    } else if (!(missing(nchains) && missing(nsamplesperchain) &&
                     missing(nsamples))){
        stop('Please specify exactly two among "nsamples", "nchains", "nsamplesperchain"')
    }


#### Requested parallel processing

    ## Parallellisation is done in any case,
    ## so that objects from the Monte Carlo simulation are not left in memory.

    ## Number of cores is reduced if larger than 'nchains':
    ## doesn't make sense to have more cores than chains
    closeexit <- FALSE
    if ('cluster' %in% class(parallel)){
        ## user provides a cluster object
        cl <- parallel
    } else if (isTRUE(parallel)) {
        ## user wants us to register a parallel backend
        ## and to choose number of cores
        ncores <- max(1,
            min(nchains, floor(parallel::detectCores() / 2)))
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
        predirmsgs <- c(predirmsgs, list(
            paste0('Registered ', capture.output(print(cl)), '.')
        ))
    } else if (isFALSE(parallel)) {
        ## user wants us not to use parallel cores
        ncores <- 1
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
    } else if (is.numeric(parallel) &&
                   is.finite(parallel) && parallel >= 1) {
        ## user wants us to register 'parallel' # of cores
        ncores <- min(nchains, parallel)
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
        predirmsgs <- c(predirmsgs, list(
            paste0('Registered ', capture.output(print(cl)), '.')
        ))
    } else {
        stop("Unknown value of argument 'parallel'.")
    }

    ## Close parallel connections if any were opened
    if(closeexit) {
        closecoresonexit <- function(){
            parallel::stopCluster(cl)
            gc(full = TRUE)
            ## parallel::setDefaultCluster(NULL)
        }
        on.exit(closecoresonexit())
    }

    minchainspercore <- nchains %/% ncores
    coreswithextrachain <- nchains %% ncores

    if (is.numeric(thinning) && thinning > 0) {
        thinning <- ceiling(thinning)
    } else if (!is.null(thinning)) {
        stop('Invalid "thinning" argument.')
    }

    ## Make sure 'startupMCiterations' is at least equal to 2
    startupMCiterations <- max(2, startupMCiterations)

    ## maxrelMCSE and minESS
    maxrelMCSE <- min(maxrelMCSE, 1 / sqrt(minESS + initES))
    minESS <- (1 / maxrelMCSE)^2 - initES

    if(all(!is.finite(c(maxrelMCSE, maxMCiterations)))) {
        stop('Impossible stopping rule')
    }

##################################################
#### Read and process data and metadata
##################################################


#### Read metadata
    ## Check whether argument 'metadata' is a string for a file name
    ## otherwise we assume it's a data.frame or similar object
    if (is.character(metadata) && file.exists(metadata)) {
        metadata <- read.csv(metadata,
            na.strings = '', stringsAsFactors = FALSE,
            colClasses=c(
                name = 'character',
                type = 'character',
                domainmin = 'numeric',
                domainmax = 'numeric',
                datastep = 'numeric',
                minincluded = 'character',
                maxincluded = 'character'
            ))
    }
    metadata <- as.data.frame(metadata)

    ## eliminate possible empty V-columns
    for(i in intersect(paste0('V', 11:3), colnames(metadata))){
        if(all(is.na(metadata[, i]))){
            metadata <- metadata[, -which(colnames(metadata) == i), drop=FALSE]
        }
    }

#### Dataset
    ## Check if 'data' is given
    if(!(missing(data) || is.null(data))) {
        ## Check if 'data' argument is an existing file
        ## otherwise we assume it is an object
        datafile <- NULL
        if (is.character(data)) {
            ## add '.csv' if missing
            datafile <- paste0(sub('.csv$', '', data), '.csv')
            if (file.exists(datafile)) {
                data <- read.csv(datafile,
                    na.strings = '', stringsAsFactors = FALSE, tryLogical = FALSE)
            } else {
                stop('Cannot find data file')
            }
        }
        data <- as.data.frame(data)
        rownames(data) <- NULL

        ## convert factors to strings if necessary
        if(any(sapply(data, is.factor))){
            ## message('Converting factors to characters.')
            . <- sapply(data, is.factor)
            data[, .] <- lapply(data[, ., drop = FALSE], as.character)
        }

        ## Consistency checks for data
        ## They should be moved to an external function

        ## Are data missing variates?
        if (!all(metadata[['name']] %in% colnames(data))) {
            stop('Missing variates in data. Check data- and metadata-files.')
        }

        ## Drop variates in data that are not in the metadata file
        if (!all(colnames(data) %in% metadata[['name']])) {
            predirmsgs <- c(predirmsgs, list(
                'Data have additional variates. Dropping them.'
            ))
            subvar <- intersect(colnames(data), metadata[['name']])
            data <- data[, subvar, drop = FALSE]
            rm(subvar)
        }

        ## Remove empty datapoints
        tokeep <- which(apply(data, 1, function(xx) { !all(is.na(xx)) }))
        if(length(tokeep) == 0 && !prior) {
            stop('Data are given but empty')
        } else if(length(tokeep) < nrow(data)) {
            message('Warning: data contain empty datapoints. Dropping them.')
            data <- data[tokeep, , drop = FALSE]
        }
        rm(tokeep)

        ## Check if the user wants to use a subset of the dataset
        if (is.numeric(subsampledata)) {
            ## @@TODO: find faster and memory-saving subsetting
            predirmsgs <- c(predirmsgs, list(
                'Subsampling data, as requested.'
            ))
            data <- data[sample(seq_len(nrow(data)),
                min(subsampledata, nrow(data)),
                replace = FALSE), ]
        }

        npoints <- nrow(data)

    } else {
        ## data not given: we assume user wants prior calculation
        predirmsgs <- c(predirmsgs, list('Missing data'))
        prior <- TRUE
        npoints <- 0
    }

#### Auxiliary dataset
    ## used to extract information about hyperparameters
    ## Check if 'auxdata' is given
    if(!(missing(auxdata) || is.null(auxdata))) {
        ## Check if 'auxdata' argument is an existing file
        ## otherwise we assume it is an object
        auxdatafile <- NULL
        if (is.character(auxdata)) {
            ## add '.csv' if missing
            auxdatafile <- paste0(sub('.csv$', '', auxdata), '.csv')
            if (file.exists(auxdata)) {
                auxdata <- read.csv(auxdatafile,
                    na.strings = '', stringsAsFactors = FALSE, tryLogical = FALSE)
            } else {
                stop('Cannot find auxdata file')
            }
        }
        auxdata <- as.data.frame(auxdata)

        ## Consistency checks for auxdata
        ## They should be moved to an external function

        ## Are auxdata missing variates?
        if (!all(metadata[['name']] %in% colnames(auxdata))) {
            stop('Missing variates in auxdata. Check auxdata- and metadata-files.')
        }

        ## Drop variates in auxdata that are not in the metadata file
        if (!all(colnames(auxdata) %in% metadata[['name']])) {
            predirmsgs <- c(predirmsgs, list(
                '"auxdata" have additional variates. Dropping them.'
            ))
            subvar <- intersect(colnames(auxdata), metadata[['name']])
            auxdata <- auxdata[, subvar, drop = FALSE]
            rm(subvar)
        }

        ## Remove empty datapoints
        tokeep <- which(apply(auxdata, 1, function(xx) { !all(is.na(xx)) }))
        if(length(tokeep) == 0 && !prior) {
            stop('Auxdata are given but empty')
        } else if(length(tokeep) < nrow(auxdata)) {
            predirmsgs <- c(predirmsgs, list(
                '"auxdata" contain empty datapoints. Dropping them.'
            ))
            auxdata <- auxdata[tokeep, , drop = FALSE]
        }
        rm(tokeep)

    } else {
        ## auxdata not given
        auxdata <- NULL
    }

    ## Build auxiliary metadata object; we'll save it later
    ## message('Calculating auxiliary metadata.')
    auxmetadata <- buildauxmetadata(
        data = (if (is.null(auxdata)) {data} else {auxdata}),
        metadata = metadata,
        Dthreshold = Dthreshold,
        tscalefactor = tscalefactor
    )
    ## print(auxmetadata) # for debugging

    predirmsgs <- c(predirmsgs, list(
        paste0('Learning from ', npoints, ' datapoints, ',
            nrow(auxmetadata), ' variates.')
    ))

#### Output-folder setup
    prefix <- paste0(
        'V', nrow(auxmetadata),
        '_D',
        (if (npoints == 1 && all(is.na(data))) {
            0
        } else {
            npoints
        }),
        ## '-K', ncomponents, # unimportant for user
        '_S', nsamples,
        '_', format(Sys.time(), '%y%m%dT%H%M%S')
    )
    if(is.null(outputdir)){
        if(!isTRUE(valueislearnt)) {
            warning('With the chosen "outputdir" and "valueislearnt" arguments, results are not saved to a persistent directory and not outputted; they will likely be lost.')
        }
        ## Use a temporary output directory
        dirname <- tempfile(pattern = paste0('prova-', prefix, '_'))
    } else if (is.na(outputdir)) {
        ## Create unique output directory within current directory
        dirname <- tempfile(pattern = paste0('prova-', prefix, '_'),
            tmpdir = getwd())
    } else if (is.character(outputdir)) {
        ## Create directory specified by user
        dirname <- outputdir
        if (appendinfo){ dirname <- paste0(dirname, '-', prefix) }
    } else {
        stop('Invalid "outputdir" argument')
    }
    dir.create(dirname)

    ## This is in case we need to add some extra specifier to the  files
    ## all 'dashnameroot' can be deleted in a final version
    dashnameroot <- NULL

    ## Create main log file in output directory
    ## also keep printing to terminal if verbose = TRUE
    mainlog <- file.path(dirname, 'main.log')
    maincon0 <- file(mainlog, open = 'w')
    closecons0 <- function(){
        ## Close output to main log file
        ## wrap in try() in case errors happen during parallel processing,
        ## during which the connection is closed
        try(flush(maincon0), silent = TRUE)
        try(close(maincon0), silent = TRUE)
    }
    on.exit(closecons0(), add = TRUE)

    if(verbose){
        print2user <- function(...){
            ## print to console
            message(...)
            ## print to main log file
            writeLines(text = paste0(...), con = maincon0, sep = '\n')
            flush(maincon0)
        }
    }else{
        print2user <- function(...){
            ## print only to main log file
            writeLines(text = paste0(...), con = maincon0, sep = '\n')
            flush(maincon0)
        }
    }

    ## Print information about output directory
    print2user('\nSaving output in directory\n', dirname, '\n')

    ## Print accumulated messages
    for(msg in predirmsgs){print2user(msg)}

    ## Save copy of metadata and auxmetadata in directory
    write.csv(metadata, file = file.path(dirname, 'metadata.csv'),
        row.names = FALSE, quote = FALSE, na = '')
    saveRDS(auxmetadata,
        file = file.path(dirname, paste0('___auxmetadata', dashnameroot, '.rds')))

    ## Save initial RNG seed in case needed by user
    saveRDS(currentseed,
        file = file.path(dirname, paste0('rng_seed', dashnameroot, '.rds')))
    parallel::clusterSetRNGStream(cl = cl, currentseed)
#### number of checkpoints for Monte Carlo stopping rule
    if(is.null(ncheckpoints)) {
        ncheckpoints <- nrow(auxmetadata) + 1
    }
    ncheckpoints <- round(ncheckpoints)
    if (ncheckpoints < 1) {
        stop('"ncheckpoints" must be > 0')
    }

    ## if data is not empty, we use it to create data for likelihood
    if(!is.null(data)) {
        ## testdata is created and saved outside of the parallel processes
        ## so that the dataset does not need to be exported to them
        ## (using extra memory)
        ## Each chain uses a different set of testdata
        for(achain in 0:nchains) {
            pointsid <- sort(sample(seq_len(npoints), min(ncheckpoints, npoints)))
            testdata <- util_prepPcheckpoints(
                x = data[pointsid, , drop = FALSE],
                auxmetadata = auxmetadata,
                pointsid = pointsid
            )
            saveRDS(testdata,
                file = file.path(dirname, paste0('___testdata_', achain, '.rds')))
        }
        rm(testdata)
        rm(pointsid)
    }

#### Check if user wants to calculate prior
    if (prior) {
        print2user('CALCULATING PRIOR DISTRIBUTION')
        if(is.null(data)) {
            ## no data available: construct one datapoint from the metadata info
            testdata <- as.matrix(
                as.data.frame(
                    lapply(seq_len(nrow(auxmetadata)),
                        function(ii){
                            if(auxmetadata[ii, 'mcmctype'] %in%
                                   c('R', 'C', 'D', 'L')) {
                                rnorm(n = ncheckpoints, mean = 0, sd = 2)
                            } else {
                                sample(seq_len(auxmetadata[ii, 'Nvalues']),
                                    ncheckpoints,
                                    replace = TRUE) +
                                    auxmetadata[ii, 'indexpos']
                            }
                        }
                    ),
                    row.names = seq_len(ncheckpoints),
                    col.names = auxmetadata$name)
            )
            ## ## original alternative way. Delete soon
            ## testdata <- data.frame(1:3)[,-1]
            ## for(xx in seq_len(nrow(auxmetadata))) {
            ##   xx <- as.list(auxmetadata[xx, ])
            ##   toadd <- unlist(xx[paste0('mctest', 1:3)])
            ##   if (xx[['mcmctype']] %in% c('B', 'N', 'O')) {
            ##     toadd <- unlist(xx[paste0('V', toadd)])
            ##   }
            ##   testdata <- cbind(testdata, toadd)
            ## }
            ## colnames(testdata) <- auxmetadata[['name']]
            for(achain in 0:nchains) {
                saveRDS(testdata,
                    file = file.path(dirname, paste0('___testdata_', achain, '.rds')))
            }
            rm(testdata)
        }
        ## create empty dataset: Monte Carlo sampling is non-Markov
        data <- as.data.frame(
            matrix(NA, nrow = 1, ncol = nrow(metadata),
                dimnames = list(NULL, metadata$name))
        )
    }

    ## #### Select and save data for loglikelihood
    ##   ## Find which datapoints have not all missing entries
    ##   dataNoNa <- which(apply(data, 1, function(xx) { !all(is.na(xx)) }))
    ##   ndataNoNa <- length(dataNoNa)
    ##
    ##   if (ndataNoNa > 0) {
    ##     ## if "testdata" is moved into the for-loop,
    ##     ## then each chain uses a different set of testdata
    ##     for(achain in seq_len(nchains)) {
    ##       testdata <- data[sort(sample(dataNoNa, min(ncheckpoints, ndataNoNa))),]
    ##       saveRDS(testdata,
    ##               file = file.path(dirname, paste0('___testdata_', achain, '.rds')))
    ##     }
    ##   } else {
    ##       ## no data available: construct one datapoint from the metadata info
    ##       testdata <- data.table(sapply(seq_len(nrow(auxmetadata)), function(xx) {
    ##       # consider making this function separate
    ##       xx <- auxmetadata[xx, ]
    ##       toadd <- xx[, paste0('mctest', 1:3), with = FALSE]
    ##       if (xx[['mcmctype']] %in% c('B', 'N', 'O)) {
    ##         toadd <- xx[1, paste0('V', toadd), with = FALSE]
    ##       }
    ##       toadd
    ##     }))
    ##     colnames(testdata) <- auxmetadata[['name']]
    ##     for(achain in seq_len(nchains)) {
    ##       saveRDS(testdata,
    ##           file = file.path(dirname, paste0('___testdata_', achain, '.rds')))
    ##     }
    ##     rm(testdata)
    ##     }




##################################################
#### Define functions
##################################################

    ## Load auxiliary functions from external files
    ## source('samplesFDistribution.R')
    ## source('plotFsamples.R')
    ## source('tplotfunctions.R')
    ## source('vtransform.R')
    ## ## These are used within the foreach core loop
    ## source('proposeburnin.R')
    ## source('proposethinning.R')
    ## source('mcsubset.R')



#####################################################
#### Prepare arguments for Nimble model
#####################################################

    ## R: continuous open domain
    ## C: continuous closed domain (or censored)
    ## D: continuous rounded
    ## L: latent
    ## O: ordinal
    ## N: nominal
    ## B: binary
    ## number and names of variates of each type
    vn <- vnames <- list(R=NULL, C=NULL, D=NULL, O=NULL, N=NULL, B=NULL)

    for (atype in names(vn)) {
        vnames[[atype]] <- auxmetadata[auxmetadata$mcmctype == atype, 'name']
        vn[[atype]] <- length(vnames[[atype]])
    }

    ## ## REMOVE soon: previous version
    ## vn <- list() # How many variates of each type
    ## vnames <- list() # The names for variates of each type
    ## for (atype in c('R', 'C', 'D', 'L', 'O', 'N', 'B')) {
    ##   vnames[[atype]] <- auxmetadata[auxmetadata$mcmctype == atype, 'name']
    ##   vn[[atype]] <- length(vnames[[atype]])
    ## }

#### CONSTANTS OF NIMBLE MODEL
    ## These constants are available in the Nimble environment
    ## They don't have to be accessed by constants$varname
    ## vn$R + vn$C + vn$D + vn$L
    nalpha <- length(seq(minalpha, maxalpha, by = byalpha))
    probalpha0 <- (1:nalpha)^2.25
    probalpha0 <- probalpha0/sum(probalpha0)
    constants <- c(
        list(
            ncomponents = ncomponents,
            npoints = npoints,
            nalpha = nalpha,
            alphabase = sqrt(2),
            probalpha0 = probalpha0,
            dirchalphas = rep((2^(minalpha - 0.5)) / ncomponents, ncomponents)
        ),
        if (vn$R > 0) { # continuous open domain
            list(
                Rn = vn$R,
                Rmean1 = rep(0, 1),
                Rvarm1 = rep(Rvarm1, 1),
                Rvar1 = rep(1, 1),
                Rshapelo = rep(Rshapelo, 1),
                Rshapehi = rep(Rshapehi, 1)
            )
        },
        if (vn$C > 0) { # continuous closed domain
            list(
                Cn = vn$C,
                Cmean1 = rep(0, 1),
                Cvarm1 = rep(Cvarm1, 1),
                Cvar1 = rep(1, 1),
                Cshapelo = rep(Cshapelo, 1),
                Cshapehi = rep(Cshapehi, 1),
                Cleft = as.matrix(vtransform(data[, vnames$C, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Cout = 'left', logjacobianOr = NULL)),
                Cright = as.matrix(vtransform(data[, vnames$C, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Cout = 'right', logjacobianOr = NULL)),
                Clatinit = as.matrix(vtransform(data[, vnames$C, drop = FALSE],
                    auxmetadata,
                    Cout = 'init', logjacobianOr = NULL))
            )
            ## Cleft & Cright are as many as the datapoints
            ## so we do not create copies outside of Nimble
            ## to save RAM
        },
        if (vn$D > 0) { # continuous rounded
            list(
                Dn = vn$D,
                Dmean1 = rep(0, 1),
                Dvarm1 = rep(Dvarm1, 1),
                Dvar1 = rep(1, 1),
                Dshapelo = rep(Dshapelo, 1),
                Dshapehi = rep(Dshapehi, 1),
                Dleft = as.matrix(vtransform(data[, vnames$D, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Dout = 'left', logjacobianOr = NULL)),
                Dright = as.matrix(vtransform(data[, vnames$D, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Dout = 'right', logjacobianOr = NULL)),
                Dlatinit = as.matrix(vtransform(data[, vnames$D, drop = FALSE],
                    auxmetadata,
                    Dout = 'init', logjacobianOr = NULL))
            )
        },
        ## if (vn$L > 0) { # latent
        ##     list(
        ##         Ln = vn$L,
        ##         Lmean1 = rep(0, 1),
        ##         Lvarm1 = rep(Lvarm1, 1),
        ##         Lvar1 = rep(1, 1),
        ##         Lshapelo = rep(Lshapelo, 1),
        ##         Lshapehi = rep(Lshapehi, 1),
        ##         Lleft = as.matrix(vtransform(data[, vnames$L, drop = FALSE],
        ##             auxmetadata = auxmetadata,
        ##             Lout = 'left')),
        ##         Lright = as.matrix(vtransform(data[, vnames$L, drop = FALSE],
        ##             auxmetadata = auxmetadata,
        ##             Lout = 'right')),
        ##         Llatinit = as.matrix(vtransform(data[, vnames$L, drop = FALSE],
        ##             auxmetadata,
        ##             Lout = 'init'))
        ##     )
        ## },
        if (vn$O > 0) { # ordinal
            Ocards <- auxmetadata[auxmetadata$name %in% vnames$O, 'Nvalues']
            Omaxn <- sum(Ocards)
            Oi <- auxmetadata[auxmetadata$name %in% vnames$O, 'indexpos'] + 1L
            ## if(vn$O == 1){Oi <- c(Oi, NA_integer_)}
            Of <- Oi + Ocards - 1L
            ## ## we choose a flatter hyperprior for ordinal variates
            ## we choose a Hadamard-like hyperprior for nominal variates
            Oalpha0 <- if(Oprior == 'Hadamard') {
                as.vector(unlist(sapply(Ocards, function(acard){
                    rep(1 / acard, acard)
                })))
            } else {
                as.vector(unlist(sapply(Ocards, function(acard){
                    rep(1, acard)
                })))
            }
            ##
            list(
                On = vn$O,
                Ocards = Ocards,
                Oi = Oi,
                Of = Of,
                Oalpha0 = Oalpha0
            )
        },
        if (vn$N > 0) { # nominal
            Ncards <- auxmetadata[auxmetadata$name %in% vnames$N, 'Nvalues']
            Nmaxn <- sum(Ncards)
            Ni <- auxmetadata[auxmetadata$name %in% vnames$N, 'indexpos'] + 1L
            ## if(vn$N == 1){Ni <- c(Ni, NA_integer_)}
            Nf <- Ni + Ncards - 1L
            ## ## we choose a flatter hyperprior for ordinal variates
            ## we choose a Hadamard-like hyperprior for nominal variates
            Nalpha0 <- if(Nprior == 'Hadamard') {
                as.vector(unlist(sapply(Ncards, function(acard){
                    rep(1 / acard, acard)
                })))
            } else {
                as.vector(unlist(sapply(Ncards, function(acard){
                    rep(1, acard)
                })))
            }
            ##
            list(
                Nn = vn$N,
                Ncards = Ncards,
                Ni = Ni,
                Nf = Nf,
                Nalpha0 = Nalpha0
            )
        },
        if (vn$B > 0) { # binary
            list(
                Bn = vn$B,
                Bshapelo = rep(Bshapelo, 1),
                Bshapehi = rep(Bshapehi, 1)
            )
        },
        if(isTRUE(avoidzeroW)){
            list(epsd = .Machine$double.xmin)
        }
    ) # End constants

    saveRDS(constants,
        file = file.path(dirname, paste0('___constants', dashnameroot, '.rds')))

#### DATAPOINTS
    ## for each list element: rows: data; cols: variates of that element
    datapoints <- c(
        if (vn$R > 0) { # continuous open domain
            list(
                Rdata = as.matrix(vtransform(data[, vnames$R, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Rout = 'normalized', logjacobianOr = NULL))
            )
        },
        if (vn$C > 0) { # continuous closed domain
            list(
                Caux = as.matrix(vtransform(data[, vnames$C, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Cout = 'aux', logjacobianOr = NULL)),
                Clat = as.matrix(vtransform(data[, vnames$C, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Cout = 'lat', logjacobianOr = NULL))
            )
        },
        if (vn$D > 0) { # continuous rounded
            list(
                Daux = as.matrix(vtransform(data[, vnames$D, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Dout = 'aux', logjacobianOr = NULL))
            )
        },
        ## if (vn$L > 0) { # latent
        ##     list(
        ##         Laux = as.matrix(vtransform(data[, vnames$L, drop = FALSE],
        ##             auxmetadata = auxmetadata,
        ##             Lout = 'aux'))
        ##     )
        ## },
        if (vn$O > 0) { # nominal
            list(
                Odata = as.matrix(vtransform(data[, vnames$O, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Oout = 'numeric', logjacobianOr = NULL))
            )
        },
        if (vn$N > 0) { # nominal
            list(
                Ndata = as.matrix(vtransform(data[, vnames$N, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Nout = 'numeric', logjacobianOr = NULL))
            )
        },
        if (vn$B > 0) { # binary
            list(
                Bdata = as.matrix(vtransform(data[, vnames$B, drop = FALSE],
                    auxmetadata = auxmetadata,
                    Bout = 'numeric', logjacobianOr = NULL))
            )
        }
    ) # End datapoints

    saveRDS(datapoints,
        file = file.path(dirname, paste0('___datapoints', dashnameroot, '.rds')))

#### Output information to user
    print2user('Starting Monte Carlo sampling of ', nsamples, ' samples by ',
        nchains, ' chains')

    samplespacedims <- (vn$R * 2 + vn$C * 2 + vn$D * 2 + # means, vars
                            (if(vn$O > 0){Omaxn - vn$O}else{0}) +
                            (if(vn$N > 0){Nmaxn - vn$N}else{0}) +
                            vn$B + # independent probs
                            1) * ncomponents - 1 # component weights

    samplespacexdims <- 1 + # Alpha
        (vn$R + vn$C + vn$D) * ncomponents + # rates
        npoints + # K
        (vn$C + vn$D) * npoints * ncomponents + # latents
        sum(is.na(data)) # missing data


    print2user('in a space of ', samplespacedims, ' (effectively ',
        samplespacedims + samplespacexdims, ') dimensions.')

    print2user('Using ', ncores, ' cores: ',
        nsamplesperchain, ' samples per chain, max ',
        minchainspercore + (coreswithextrachain > 0), ' chains per core.')
    print2user('Requested:   ESS ', round(minESS),
        '   rel.MCSE ', signif(maxrelMCSE, 3), '.')
    print2user('Core logs are being saved in individual files.')

    ## function  to format printing of time
    printtimediff <- function(tim) {
        paste0(signif(tim, 2), ' ', attr(tim, 'units'))
    }

    ## Close output to main log file
    close(maincon0)
    
#####################################################
#### BEGINNING OF PARALLEL LOOP OVER CORES
#####################################################
    ## Parallel execution over cores
    ## parallel:clusterExport(cl = cl,
    ##     varlist = c(
    ##         'dirname',
    ##         'dashnameroot',
    ##         'avoidzeroW',
    ##         'constants',
    ##         'initmethod'
    ##     ),
    ##     envir = .GlobalEnv)

    chaininfo <- parallel::parLapply(
        cl = cl,
        X = 1:ncores,
        fun = workerfun,
        ## arguments to workerfun
        dirname = dirname,
        dashnameroot = dashnameroot,
        avoidzeroW = avoidzeroW,
        initmethod = initmethod,
        constants = constants,
        datapoints = datapoints,
        vn = vn,
        showAlphatraces = showAlphatraces,
        Alphatoslice = Alphatoslice,
        Ktoslice = Ktoslice,
        RWtoslice = RWtoslice,
        changeSamplerOrder = changeSamplerOrder,
        minchainspercore = minchainspercore,
        coreswithextrachain = coreswithextrachain,
        nchains = nchains,
        maxhours = maxhours,
        timestart0 = timestart0,
        showsamplertimes = showsamplertimes,
        startupMCiterations = startupMCiterations,
        maxMCiterations = maxMCiterations,
        showKtraces = showKtraces,
        ncomponents = ncomponents,
        plottraces = plottraces,
        Qlo = Qlo,
        Qhi = Qhi,
        Qerror = Qerror,
        minESS = minESS,
        initES = initES,
        nsamplesperchain = nsamplesperchain,
        minMCiterations = minMCiterations,
        printtimediff = printtimediff,
        family = family,
        mainlog = mainlog,
        verbose = verbose,
        ##
        chunk.size = NULL
    )
    chaininfo <- do.call(what = rbind, args = chaininfo)

    ## Resume output to main log file
    maincon0 <- file(mainlog, open = 'a')

############################################################
#### END OF PARALLEL LOOP OVER CORES
############################################################

    maxusedcomponents <- max(chaininfo[, 'maxusedcomponents'])
    maxiterations <- max(chaininfo[, 'maxiterations'])
    nonfinitechains <- sum(chaininfo[, 'nonfinitechains'])
    stoppedchains <- sum(chaininfo[, 'stoppedchains'])
    headertime <- as.difftime(mean(chaininfo[, 'headertime']), units = 'secs')
    MCtime <- as.difftime(sum(chaininfo[, 'MCtime'])/nchains, units = 'secs')
    maxusedmem <- max(chaininfo[, 'usedmem'])
    totusedmem <- sum(chaininfo[, 'usedmem'])
############################################################
#### End of all MCMC
############################################################


############################################################
#### Join chains
############################################################

    headertimestart <- Sys.time()
    ## Save random seeds used in the parallel processing
    if (!is.null(attr(chaininfo, 'rng'))) { # parallel processing
        saveRDS(attr(chaininfo, 'rng'),
            file = file.path(dirname,
                paste0('rng_parallelseeds', dashnameroot, '.rds')
            ))
    }

    ## Read the samples saved by each chain and concatenate them
    ## could be moved to a separate file?
    joinmc <- function(mc1, mc2) {
        mapply(
            function(xx, yy) {
                temp <- c(xx, yy)
                dx <- dim(yy)[-length(dim(yy))]
                dim(temp) <- c(dx, length(temp) / prod(dx))
                temp
            },
            mc1, mc2,
            SIMPLIFY = FALSE
        )
    }

## mcsamples0 <- foreach(chainnumber = 1:nchains,
##         .combine = joinmc, .multicombine = FALSE) %do% {
##             padchainnumber <- sprintf(paste0('%0', nchar(nchains), 'i'), chainnumber)
##
##             readRDS(file = file.path(dirname,
##                 paste0('___mcsamples', dashnameroot, '--',
##                     padchainnumber, '.rds')
##             ))
##         }

    for(chainnumber in 1:nchains){
        padchainnumber <- sprintf(paste0('%0', nchar(nchains), 'i'), chainnumber)
        if(chainnumber == 1){
            mcsamples <- readRDS(file = file.path(dirname,
                paste0('___mcsamples', dashnameroot, '--',
                    padchainnumber, '.rds')
            ))
        } else {
            mcsamples <- joinmc(mcsamples,
                readRDS(file = file.path(dirname,
                    paste0('___mcsamples', dashnameroot, '--',
                        padchainnumber, '.rds')
                ))
            )
        }
    }

    ## chainnumber <- 1
    ## padchainnumber <- sprintf(paste0('%0', nchar(nchains), 'i'), chainnumber)
    ## mcsamples <- readRDS(file = file.path(dirname,
    ##             paste0('___mcsamples', dashnameroot, '--',
    ##                 padchainnumber, '.rds')
    ## ))
    ## if(nchains > 1){
    ##     for(chainnumber in 2:nchains){
    ##         padchainnumber <- sprintf(paste0('%0', nchar(nchains), 'i'), chainnumber)
    ##         mcsamples <- joinmc(mcsamples,
    ##             readRDS(file = file.path(dirname,
    ##             paste0('___mcsamples', dashnameroot, '--',
    ##                 padchainnumber, '.rds')
    ##             ))
    ##         )
    ##     }
    ## }

#### Save all final parameters together with some aux metadata in one file
    learnt <- c(mcsamples, list(
        auxmetadata = auxmetadata,
        auxinfo = list(
            nchains = nchains,
            npoints = npoints,
            hyperparams = hyperparams,
            maxiterations = maxiterations,
            maxusedcomponents = maxusedcomponents,
            nonfinitechains = nonfinitechains,
            stoppedchains = stoppedchains
        )
    ))
    rm(mcsamples)
    ## The 'learnt' object is preliminarily saved as soon as possible,
    ## in case something crashes during subsequent diagnostics
    saveRDS(learnt,
        file = file.path(dirname, paste0('___learnt', dashnameroot, '.rds'))
    )

    ## This is to delete the last 'estimated end time'
    print2user('\r                                                               \n')

    print2user('Finished Monte Carlo sampling.')

    print2user(
        'Highest number of Monte Carlo iterations across chains: ',
        maxiterations, '.',
        '\nHighest number of used mixture components: ',
        maxusedcomponents, '.'
    )
    if (maxusedcomponents > ncomponents - 5) {
        print2user('\nTOO MANY MIXTURE COMPONENTS USED!\n',
            'Consider re-running with increased "ncomponents" parameter.')
        warning('\nTOO MANY MIXTURE COMPONENTS USED!\n',
            'Consider re-running with increased "ncomponents" parameter.\n')
    }

    if (nonfinitechains > 0) {
        print2user('\nNote: ', nonfinitechains, ' chains had some non-finite outputs.')
    }

    if (stoppedchains > 0) {
        print2user('\nWarning: ', stoppedchains,
            ' chains were stopped before reaching required precision\n',
            'in order to meet the required time constraints.')
        ## Important to warn the user about this
            warning(stoppedchains,
                ' chains were stopped before reaching required precision\n',
                'in order to meet the required time constraints.')
        }



############################################################
#### Final joint diagnostics
############################################################

    testdata <- readRDS(file = file.path(dirname,
        paste0('___testdata_', 0, '.rds')))
    print2user('\nChecking test data\n(',
        paste0('#', testdata$pointsid, collapse = ' '), '):')

    oktraces <- util_Pcheckpoints(
        testdata = testdata,
        learnt = learnt
    )

    ## Substitute variances for standard deviations
    if(!is.null(learnt$Rvar)){
        learnt$Rvar <- sqrt(learnt$Rvar)
        names(learnt)[which(names(learnt) == 'Rvar')] <- 'Rsd'
    }
    if(!is.null(learnt$Cvar)){
        learnt$Cvar <- sqrt(learnt$Cvar)
        names(learnt)[which(names(learnt) == 'Cvar')] <- 'Csd'
    }
    if(!is.null(learnt$Dvar)){
        learnt$Dvar <- sqrt(learnt$Dvar)
        names(learnt)[which(names(learnt) == 'Dvar')] <- 'Dsd'
    }

    ## Save 'learnt' object
    saveRDS(learnt,
        file = file.path(dirname, paste0('learnt', dashnameroot, '.rds'))
    )


    oktraces <- cbind(
        exp(rowMeans(log(oktraces), na.rm = TRUE)), # geometric mean
        oktraces
    )

    oktraces <- oktraces[apply(oktraces, 1, function(x) { all(is.finite(x)) }), ,
        drop = FALSE]

    colnames(oktraces) <- c('gmean', testdata$pointsid)

    saveRDS(oktraces, file = file.path(dirname,
        paste0('MCtraces', dashnameroot, '.rds')
    ))

    ## ##############################
    ## ## MCSE, ESS, THINNING, ETC ##
    ## ##############################

    N <- nrow(oktraces)

    jointdiagn <- apply(oktraces, 2, function(atrace) {
### same as within cores

        ## Calculate empirical quantiles of the vector of MC samples,
        ## having 'Qlo' and 'Qhi' probabilities
        Xlo <- quantile(x = atrace, probs = Qlo,
            na.rm = FALSE, names = FALSE, type = 6)
        Xhi <- quantile(x = atrace, probs = Qhi,
            na.rm = FALSE, names = FALSE, type = 6)
        ## quantile width
        qwidth <- Xhi - Xlo

        ## Calculate credibility intervals (proxy for MCSE)
        ## for the empirical quantiles above
        ## as in Vehtari et al. 2021
        temp <- funMCEQ(x = atrace, prob = c(Qlo, Qhi), Qpair = Qerror)
        wXlo <- temp[2, 1] - temp[1, 1]
        wXhi <- temp[2, 2] - temp[1, 2]

        ## For each monitored point,
        ## transform its posterior samples to normalized ranks,
        ## then calculate the ESSs of their means,
        ## as in Vehtari et al. 2021
        ## and Blom 1958 sect 6.10 eqn (6.10.5)
        ## (alternative (6.10.3) gives roughly same results
        ## but is computationally slightly more expensive)
        essnrmean <- funESS3(qnorm(
        (rank(atrace, na.last = NA, ties.method = 'average') - 0.5) / N
        ))

        ## We check the relative MCSE of the following quantities:
        ## - the mean of the MC samples, transformed to normalized ranks
        ## - the lower quantile 'Qlo'
        ## - the upper quantile 'Qhi'
        ## so as to ensure the stability of the mean value
        ## (representing the "long-run frequency")
        ## and of its 'Qlo'- and 'Qhi'-quantiles
        ## (representing the uncertainty about the freq. estimate)
        relmcse <- c(1 / sqrt(essnrmean), wXlo / qwidth, wXhi / qwidth)

        autothinning <- N * max(relmcse)^2

### same as within cores

        c(max(relmcse[-1]), essnrmean, autothinning, qwidth)
    })

    ## Output available diagnostics
    toprint <- list(
        'rel. quantile error' = jointdiagn[1, ],
        'ESS' = jointdiagn[2, ],
        'needed thinning' = jointdiagn[3, ],
        'average' = colMeans(oktraces),
        'quantile width' = jointdiagn[4, ]
    )
    learnt$auxinfo <- c(learnt$auxinfo, toprint)
    saveRDS(learnt,
        file = file.path(dirname, paste0('learnt', dashnameroot, '.rds'))
    )

####
    for(i in names(toprint)) {
        thisdiagn <- toprint[[i]]
        print2user(i, ': ',
            if(length(thisdiagn) > 1){
                paste(signif(range(thisdiagn), 3), collapse = ' to ')
            } else {
                signif(thisdiagn, 3)
            }
        )
    }


    ## Plot various info and traces
    print2user('\nPlotting final Monte Carlo traces and marginal samples...\n')

    graphics.off()
    pdf(file.path(dirname,
        paste0('MCtraces', dashnameroot, '.pdf')
    ), height = 8.27, width = 11.69)
    ## Traces of likelihood and cond. probabilities
    for (avar in 1:ncol(oktraces)) {
        ## Do not join separate chains in the plot
        division <- (if(N > nchains) nchains else 1)
        flexiplot(
            y = 10*log10(oktraces[is.finite(oktraces[, avar]), avar]),
            ## x = matrix(seq_len(nsamples), ncol = division),
            ## y = matrix(10*log10(oktraces[, avar]), ncol = division),
            type = 'l', lty = 1, col = 1,
            ## col = 1:6, # to evidence consecutive chains
            main = paste0('#', colnames(oktraces)[avar], ': ',
                paste(
                    names(toprint),
                    sapply(toprint, function(xx){
                        signif(xx[avar], 3)
                    }),
                    collapse = ' | ', sep = ': '
                )),
            cex.main = 1.25,
            ylab = paste0('log_F(#',
                colnames(oktraces)[avar],
                ')/dHart'),
            xlab = 'sample', family = family
            ## mar = c(NA, 6, NA, NA)
        )
        abline(v = seq(from = 1, to = nsamples, by = nsamplesperchain)[-1] - 0.5,
            lty = 2, col = 2, lwd = 1)
    }

    plotFsamples(
        filename = file.path(dirname,
            paste0('plotsamples_learnt', dashnameroot)),
        learnt = learnt,
        data = data,
        plotvariability = 'samples',
        nFsamples = showsamples, plotprobability = TRUE,
        datahistogram = TRUE, datascatter = TRUE,
        parallel = cl
    )

    plotFsamples(
        filename = file.path(dirname,
            paste0('plotquantiles_learnt', dashnameroot)),
        learnt = learnt,
        data = data,
        plotvariability = 'quantiles',
        nFsamples = plotDisplayedQuantiles, plotprobability = TRUE,
        datahistogram = TRUE, datascatter = TRUE,
        parallel = cl
    )

    totalfinaltime <- difftime(Sys.time(), timestart0, units = 'auto')
    print2user(
        'Total computation time: ', printtimediff(totalfinaltime),
        '\nAverage preparation & finalization time: ',
        printtimediff(
            difftime(Sys.time() + headertime, headertimestart, units = 'auto')
        ), '.',
        '\nAverage Monte Carlo time per chain: ',
        printtimediff(
            difftime(headertimestart + MCtime, headertimestart, units = 'auto')
        ), '.',
        '\nMax total memory used: approx ', signif(totusedmem, 2), 'MB.',
        '\nMax memory used per core: approx ', signif(maxusedmem, 2), 'MB.'
    )

#### Remove partial files if required
    ## Partial files are identified by at last three initial underscores "___"
    ## Should we leave the plots of partial traces?
    ## maybe create an additional argument to let the user decide?
    if (cleanup) {
        print2user('Removing temporary output files.')
        file.remove(dir(dirname,
            pattern = paste0('^___.*\\..*$'),
            full.names = TRUE
        ))
    }

#### WHAT MIGHT BE ADDED:
    ## - when 'showKtraces' is true:
    ## a histogram over number of components over all chains
    ## (for the moment there's one plot per chain)

    asterisks <- paste0(rep('*', max(nchar(dirname), 26)), collapse = '')
    print2user('Finished.\n\n', asterisks,
        '\nOutput saved in directory\n', dirname, '\n',
        asterisks, '\n')


    ## What should we output? how about the full name of the output dir?
    if (isTRUE(valueislearnt)) {
        learnt
    } else if (isFALSE(valueislearnt)) {
        dirname
    }
}



#' Worker function called by learn()
#'
#' This worker function is defined outside of learn() in order to avoid import of spurious objects into the parallel workers, and from the parallel workes into the main R session, with waste of memory.
#'
#' @import grDevices
#' @import graphics
#' @import stats
#' @import utils
#'
#' @keywords internal
workerfun <- function(
    acore,
    dirname,
    dashnameroot,
    avoidzeroW,
    initmethod,
    constants,
    datapoints,
    vn,
    showAlphatraces,
    Alphatoslice,
    Ktoslice,
    RWtoslice,
    changeSamplerOrder,
    minchainspercore,
    coreswithextrachain,
    nchains,
    maxhours,
    timestart0,
    showsamplertimes,
    startupMCiterations,
    maxMCiterations,
    showKtraces,
    ncomponents,
    plottraces,
    Qlo,
    Qhi,
    Qerror,
    minESS,
    initES,
    nsamplesperchain,
    minMCiterations,
    printtimediff,
    family,
    mainlog,
    verbose
){
    ## Handles R CMD check:
    ## these are variables or functions handled in a special way by Nimble
    ## but R CMD check erroneously sees them as "rogue"
    if(FALSE){
nimbleFunction <- sampler_BASE <- extractControlElement <- model <- target <- Ndeps <- depNodes <- rdirch <- eps <- model <- calcNodes <- nimCopy <- mvSaved <- nimbleCode <- dirchalphas <- alphabase <- Alpha <- W <- epsd <- Rn <- Cn <- Dn <- On <- Nn <- Bn <- npoints <- probalpha0 <- Nmaxn <- Bshapelo <- Bshapehi <- Oalpha0 <- Oi <- Of <- Nalpha0 <- Ni <- Nf <- nimbleModel <- compileNimble <- configureMCMC <- buildMCMC <- ncomponentsamples <- nalpha <- minalpha <- maxalpha <- byalpha <- NULL
    }

    ## functions to format printing of time
    printtimeend <- function(tim) {
        format(Sys.time() + tim, format='%Y-%m-%d %H:%M')
    }

#### NOTA BENE:
#### All cat() and print() commands within this function print only to a file,
#### in a directory selected by the user (or within tempdir()),
#### and they do so only within a temporary, non-interactive R session
#### spawned by parallel.
#### The user therefore never sees them and has no need to suppress them.
#### The information thus saved in the file can be important
#### in the case of publications and similar.
####
#### Use of sink() together with cat() is forced by the fact that
#### Nimble would otherwise send diagnostic messages to the main session/console
#### as it uses cat() internally.

    ## Create log file
    ## Redirect diagnostics and service messages there
    outcon <- file(file.path(dirname,
        paste0('log', dashnameroot,
            '-', acore, '.log')
    ), open = 'w')
    sink(file = outcon, type = 'output', split = FALSE)
    sink(file = outcon, type = 'message', split = FALSE)
    maincon <- file(mainlog, open = 'a')
    ## Restore sink in case of exit or error
    closecons <- function(){
        ## Close output to core log files
        flush(outcon)
        sink(file = NULL, type = 'output')
        sink(file = NULL, type = 'message')
        close(outcon)
        ## Close output to main log file & console
        flush(maincon)
        close(maincon)
    }
    on.exit(closecons())

    ## We need to send some messages to individual core-log files,
    ## others to the main log file (and to terminal if verbose = TRUE)
    ## This is done by changing output sink:
    if(verbose){
        print2user <- function(...){
            flush(outcon)
            ## print to console
            sink(file = NULL, type = 'message')
            message(..., appendLF = FALSE)
            ## print to main log file
            sink(file = maincon, type = 'message')
            message(..., appendLF = FALSE)
            flush(maincon)
            ## redirect to core log file
            sink(file = outcon, type = 'message')
        }
    }else{
        print2user <- function(...){
            flush(outcon)
            ## print only to main log file
            sink(file = maincon, type = 'message')
            message(..., appendLF = FALSE)
            flush(maincon)
            ## redirect to core log file
            sink(file = outcon, type = 'message')
        }
    }

    usedmem <- sum(gc(full = TRUE)[,6])

    ## Timer
    headertimestart <- Sys.time()
    cat('Log core', acore)
    cat(' - Current time:',
        strftime(as.POSIXlt(headertimestart), '%Y-%m-%d %H:%M:%S'), '\n')

    ## Check if Nimble package is installed and load it

#### NB: Nimble *must* be attached (not just loaded)
#### with 'require()' or 'library()', see
#### <https://groups.google.com/g/nimble-users/c/dDNE3L_sPxI/m/IyRgmXOaBwAJ>.
#### This situation was approved by CRAN reviewer at submission of prova v0.8.0
    loadnimble <- require
    if (!loadnimble('nimble', quietly = TRUE)) {
        stop("Package 'nimble' must be installed.", call. = FALSE)
    }

    if (acore == 1) {
        print2user(paste0(
        'C-compiling samplers appropriate to the variates (Nimble v',
        packageVersion('nimble'), ')\n',
        'this can take tens of minutes. Please wait...\n'
        ))
    }

    ## suppressPackageStartupMessages(library('nimble'))
    cat('Loaded Nimble', paste0('v', packageVersion('nimble')), '\n')
    ## requireNamespace("nimble", quietly = TRUE)
    ##library('nimble')


#### Alternative Dirichlet sampler, to deal with zero-W preserving conjugacy
#### from Daniel Turek 2025-06-29, nimble-users group
    if(is.null(avoidzeroW)) {
        conjugate_dirch_plus_eps <- nimbleFunction(
            contains = sampler_BASE,
            setup = function(model, mvSaved, target, control) {
                ## control list extraction
                eps <- extractControlElement(control,
                    'eps', .Machine$double.xmin)
                ## node list generation
                target <- model$expandNodeNames(target)
                calcNodes <- model$getDependencies(target)
                depNodes <- model$getDependencies(target, self = FALSE)
                ## numeric value generation
                d <- length(model[[target]])
                Ndeps <- length(depNodes)
                ## checks
                if(length(target) > 1) stop('conjugate_dirch_plus_eps only operates on a single node')
                if(model$getDistribution(target) != 'ddirch') stop('conjugate_dirch_plus_eps target node must have ddirch distribution')
                depDists <- sapply(depNodes, function(x) model$getDistribution(x))
                if(!all(depDists == 'dcat')) stop('conjugate_dirch_plus_eps all dependencies should have dcat distributions')
            },
            run = function() {
                posterior_alpha <- model$getParam(target, 'alpha')
                for(i in 1:Ndeps) {
                    depValue <- model$getParam(depNodes[i], 'value')
                    posterior_alpha[depValue] <- posterior_alpha[depValue] + 1
                }
                newTargetValue <- rdirch(1, posterior_alpha)
                newTargetValue <- newTargetValue + eps
                model[[target]] <<- newTargetValue
                model$calculate(calcNodes)
                nimCopy(from = model, to = mvSaved, row = 1, nodes = calcNodes, logProb = TRUE)
            },
            methods = list(
                reset = function() {}
            )
        )
    }

#### COMPONENT REPRESENTATION OF FREQUENCY SPACE
#### Dirichlet-process mixture of product-kernels

    ## hierarchical probability structure
    finitemix <- nimbleCode({
        ## Component weights
        Alpha ~ dcat(prob = probalpha0[1:nalpha])
        alphas[1:ncomponents] <- dirchalphas[1:ncomponents] * alphabase^Alpha
        W[1:ncomponents] ~ ddirch(alpha = alphas[1:ncomponents])
        if(isTRUE(avoidzeroW)){
            W0[1:ncomponents] <- W[1:ncomponents] + epsd
        }
        ## Probability density for the parameters of the components
        for (k in 1:ncomponents) {
            ## Probability distributions of parameters
            ## of the different variate types
            if (vn$R > 0) { # continuous open domain
                for (v in 1:Rn) {
                    Rmean[v, k] ~ dnorm(mean = Rmean1, var = Rvarm1)
                    Rrate[v, k] ~ dinvgamma(shape = Rshapehi, rate = Rvar1)
                    Rvar[v, k] ~ dinvgamma(shape = Rshapelo, rate = Rrate[v, k])
                }
            }
            if (vn$C > 0) { # continuous closed domain
                for (v in 1:Cn) {
                    Cmean[v, k] ~ dnorm(mean = Cmean1, var = Cvarm1)
                    Crate[v, k] ~ dinvgamma(shape = Cshapehi, rate = Cvar1)
                    Cvar[v, k] ~ dinvgamma(shape = Cshapelo, rate = Crate[v, k])
                }
            }
            if (vn$D > 0) { # continuous rounded
                for (v in 1:Dn) {
                    Dmean[v, k] ~ dnorm(mean = Dmean1, var = Dvarm1)
                    Drate[v, k] ~ dinvgamma(shape = Dshapehi, rate = Dvar1)
                    Dvar[v, k] ~ dinvgamma(shape = Dshapelo, rate = Drate[v, k])
                }
            }
            ## if (vn$L > 0) { # latent
            ##     for (v in 1:Ln) {
            ##         Lmean[v, k] ~ dnorm(mean = Lmean1, var = Lvarm1)
            ##         Lrate[v, k] ~ dinvgamma(shape = Lshapehi, rate = Lvar1)
            ##         Lvar[v, k] ~ dinvgamma(shape = Lshapelo, rate = Lrate[v, k])
            ##     }
            ## }
            if (vn$O == 1) { # ordinal, one only
                Oprob[Oi:Of, k] ~ ddirch(alpha = Oalpha0[Oi:Of])
            }
            if (vn$O > 1) { # ordinal, many
                for (v in 1:On) {
                    Oprob[Oi[v]:Of[v], k] ~ ddirch(alpha = Oalpha0[Oi[v]:Of[v]])
                }
            }
            if (vn$N == 1) { # nominal, one only
                    Nprob[Ni:Nf, k] ~ ddirch(alpha = Nalpha0[Ni:Nf])
            }
            if (vn$N > 1) { # nominal, many
                for (v in 1:Nn) {
                    Nprob[Ni[v]:Nf[v], k] ~ ddirch(alpha = Nalpha0[Ni[v]:Nf[v]])
                }
            }
            if (vn$B > 0) { # binary
                for (v in 1:Bn) {
                    Bprob[v, k] ~ dbeta(shape1 = Bshapelo, shape2 = Bshapehi)
                }
            }
        }
        ## Probability of data
        for (d in 1:npoints) {
            if(isTRUE(avoidzeroW)){
                K[d] ~ dcat(prob = W0[1:ncomponents])
            } else {
                K[d] ~ dcat(prob = W[1:ncomponents])
            }
            ##
            if (vn$R > 0) { # continuous open domain
                for (v in 1:Rn) {
                    Rdata[d, v] ~ dnorm(mean = Rmean[v, K[d]], var = Rvar[v, K[d]])
                }
            }
            if (vn$C > 0) { # continuous closed domain
                for (v in 1:Cn) {
                    Caux[d, v] ~ dconstraint(Clat[d, v] >= Cleft[d, v] &
                                                 Clat[d, v] <= Cright[d, v])
                    Clat[d, v] ~ dnorm(mean = Cmean[v, K[d]], var = Cvar[v, K[d]])
                }
            }
            if (vn$D > 0) { # continuous rounded
                for (v in 1:Dn) {
                    Daux[d, v] ~ dconstraint(Dlat[d, v] >= Dleft[d, v] &
                                                 Dlat[d, v] < Dright[d, v])
                    Dlat[d, v] ~ dnorm(mean = Dmean[v, K[d]], var = Dvar[v, K[d]])
                }
            }
            ## if (vn$L > 0) { # latent
            ##     for (v in 1:Ln) {
            ##         Laux[d, v] ~ dconstraint(Llat[d, v] >= Lleft[d, v] &
            ##                                  Llat[d, v] < Lright[d, v])
            ##         Llat[d, v] ~ dnorm(mean = Lmean[v, K[d]], var = Lvar[v, K[d]])
            ##     }
            ## }
            if (vn$O == 1) { # ordinal, one only
                Odata[d, 1] ~ dcat(prob = Oprob[Oi:Of, K[d]])
            }
            if (vn$O > 1) { # ordinal, many
                for (v in 1:On) {
                    Odata[d, v] ~ dcat(prob = Oprob[Oi[v]:Of[v], K[d]])
                }
            }
            if (vn$N == 1) { # nominal, one only
                Ndata[d, 1] ~ dcat(prob = Nprob[Ni:Nf, K[d]])
            }
            if (vn$N > 1) { # nominal, many
                for (v in 1:Nn) {
                    Ndata[d, v] ~ dcat(prob = Nprob[Ni[v]:Nf[v], K[d]])
                }
            }
            if (vn$B > 0) { # binary
                for (v in 1:Bn) { # Bprob is the probability that Bdata=1
                    Bdata[d, v] ~ dbern(prob = Bprob[v, K[d]])
                }
            }
        }
    }) # end finitemix NimbleCode


#### INITIAL-VALUE FUNCTION
    ## init functions defined in 'util_mcmcinit.R'
    if(initmethod == 'precluster'){
        ## pre-clustering, k-mean style
        initsfn <- function() {
            ## Create components centres
            ## distance function
            ## NB: all variances will be initialized to 1
            lpnorm <- function(xx){abs(xx)}
            distances <- matrix(0, nrow = constants$npoints,
                ncol = constants$ncomponents)
            if (vn$R > 0) { # continuous open domain
                Rmeans <- matrix(rnorm(
                    n = vn$R * constants$ncomponents,
                    mean = constants$Rmean1,
                    sd = sqrt(constants$Rvarm1)
                ), nrow = vn$R, ncol = constants$ncomponents)
                Rvars <- matrix(1, nrow = vn$R, ncol = constants$ncomponents)
                ## distances from datapoints
                distances <- distances + apply(Rmeans, 2, function(ameans){
                    colSums(lpnorm(t(datapoints$Rdata) - ameans), na.rm = TRUE)
                })
            }
            if (vn$C > 0) { # continuous closed domain
                Cmeans <- matrix(rnorm(
                    n = vn$C * constants$ncomponents,
                    mean = constants$Cmean1,
                    sd = sqrt(constants$Cvarm1)
                ), nrow = vn$C, ncol = constants$ncomponents)
                Cvars <- matrix(1, nrow = vn$C, ncol = constants$ncomponents)
                ## distances from datapoints
                distances <- distances + apply(Cmeans, 2, function(ameans){
                    colSums(lpnorm(t(datapoints$Clat) - ameans), na.rm = TRUE)
                })
            }
            if (vn$D > 0) { # discrete
                Dmeans <- matrix(rnorm(
                    n = vn$D * constants$ncomponents,
                    mean = constants$Dmean1,
                    sd = sqrt(constants$Dvarm1)
                ), nrow = vn$D, ncol = constants$ncomponents)
                Dvars <- matrix(1, nrow = vn$D, ncol = constants$ncomponents)
                ## distances from datapoints
                distances <- distances + apply(Dmeans, 2, function(ameans){
                    colSums(lpnorm(t(constants$Dlatinit) - ameans), na.rm = TRUE)
                })
            }
            ## if (vn$L > 0) { #
            ##     Lmeans <- matrix(rnorm(
            ##         n = vn$L * constants$ncomponents,
            ##         mean = constants$Lmean1,
            ##         sd = sqrt(constants$Lvarm1)
            ##     ), nrow = vn$L, ncol = constants$ncomponents)
            ##     ## distances from datapoints
            ##     distances <- distances + apply(Lmeans, 2, function(ameans){
            ##         colSums(lpnorm(t(constants$Llatinit) - ameans), na.rm = TRUE)
            ##     })
            ## }
            ## if (vn$B > 0) {
            ##     Bprobs <- matrix(rbeta(
            ##         n = vn$B * constants$ncomponents,
            ##         shape1 = Bshapelo,
            ##         shape2 = Bshapehi,
            ##         ), nrow = vn$B, ncol = constants$ncomponents)
            ##     ## distances from datapoints
            ##     distances <- distances + apply(Bprobs, 2, function(ameans){
            ##         colSums(lpnorm(t(datapoints$Bdata) - ameans), na.rm = TRUE)
            ##     })
            ## }

            ## assign datapoints to component with closest centre
            K <- apply(distances, 1, which.min)
            occupied <- unique(K)

            ## recalculate components centres according to their points
            if (vn$R > 0) {
                Rmeans[, occupied] <- sapply(occupied, function(acomponent){
                    colMeans(
                        datapoints$Rdata[which(K == acomponent), , drop = FALSE],
                        na.rm = TRUE)
                })
                Rmeans[, -occupied] <- 0
                ## Rvars[, occupied] <- sapply(occupied, function(acomponent){
                ##     apply(
                ##         datapoints$Rdata[which(K == acomponent), , drop = FALSE],
                ##         2, sd, na.rm = TRUE)
                ## })^2
                ## Rvars[is.na(Rvars)] <- 1
            }
            if (vn$C > 0) {
                Cmeans[, occupied] <- sapply(occupied, function(acomponent){
                    colMeans(
                        datapoints$Clat[which(K == acomponent), , drop = FALSE],
                        na.rm = TRUE)
                })
                Cmeans[, -occupied] <- 0
                ## Cvars[, occupied] <- sapply(occupied, function(acomponent){
                ##     apply(
                ##         datapoints$Clat[which(K == acomponent), , drop = FALSE],
                ##         2, sd, na.rm = TRUE)
                ## })^2
                ## Cvars[is.na(Cvars)] <- 1
            }
            if (vn$D > 0) {
                Dmeans[, occupied] <- sapply(occupied, function(acomponent){
                    colMeans(
                        constants$Dlatinit[which(K == acomponent), , drop = FALSE],
                        na.rm = TRUE)
                })
                Dmeans[, -occupied] <- 0
                ## Dvars[, occupied] <- sapply(occupied, function(acomponent){
                ##     apply(constants$Dlatinit[which(K == acomponent), , drop = FALSE],
                ##         2, sd, na.rm = TRUE)
                ## })^2
                ## Dvars[is.na(Dvars)] <- 1
            }
            ## if (vn$L > 0) { # continuous open domain
            ##     Lmeans[, occupied] <- sapply(occupied, function(acomponent){
            ##         colMeans(constants$Llatinit[which(K == acomponent), , drop = FALSE],
            ##             na.rm = TRUE)
            ##     })
            ##     Lmeans[, -occupied] <- 0
            ## }
            ## if (vn$B > 0) {
            ##     Bprobs[, occupied] <- sapply(occupied, function(acomponent){
            ##         colMeans(datapoints$Bdata[which(K == acomponent), , drop = FALSE],
            ##             na.rm = TRUE)
            ##     })
            ##     Bprobs[, -occupied] <- 0.5
            ## }
            ## Alpha <- sample(1:constants$nalpha, 1, prob = constants$probalpha0, replace = TRUE)
            ## W <- c(rep(rempoints, miconstants$npoints), rep(1, constants$ncomponents - miconstants$npoints))
            ## W <- W/sum(W)

            outlist <- list(
                Alpha = round(constants$nalpha/2),
                W = rep(1/constants$ncomponents, constants$ncomponents),
                ## ## Assign every point to the closest component centre
                K = K
                ## ## Other assignment methods:
                ## ## A. assign all points to an unsystematically chosen component
                ## K = rep(sample(rep(which(W > 0), 2), 1), nepoints)
                ## ## B. distribute points unsystematically among components
                ## K = sample(rep(which(W > 0), 2), constants$npoints, replace = TRUE)
                ## ## or:
                ## ## C. assign all points to the most probable component
                ## K = rep(which.max(W), constants$npoints)
                ## ## or:
                ## ## D. assign all points to the least probable component
                ## K = rep(which(W == min(W[W > 0]))[1], constants$npoints)
                ## ## or:
                ## ## E. distribute points unsystematically among M=2 components
                ## K = sample(sample(rep(which(W > 0), 2), 2, replace = TRUE),
                ##           constants$npoints, replace = TRUE)
                ## ## F. mix methods A. and B.
                ## K = (if(achain %% 2 == Ksample) {
                ##        ## ## assign all points to an unsystematically chosen component
                ##        rep(sample(rep(which(W > 0), 2), 1), constants$npoints)
                ##      } else {
                ##        ## distribute points unsystematically among components
                ##        sample(rep(which(W > 0), 2), constants$npoints, replace = TRUE)
                ##      })
            )
            ##
            if (vn$R > 0) { # continuous open domain
                outlist <- c(
                    outlist,
                    list(
                        Rmean = Rmeans,
                        Rrate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Rshapehi,
                                rate = constants$Rvar1),
                            nrow = vn$R, ncol = constants$ncomponents
                        ),
                        Rvar = Rvars
                    )
                )
            }
            if (vn$C > 0) { # continuous closed domain
                outlist <- c(
                    outlist,
                    list(
                        Cmean = Cmeans,
                        Crate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Cshapehi,
                                rate = constants$Cvar1),
                            nrow = vn$C, ncol = constants$ncomponents
                        ),
                        Cvar = Cvars,
                        ## for data with boundary values
                        Clat = constants$Clatinit
                        ## Clat = vtransform(data[, vnames$C, with = FALSE],
                        ##   auxmetadata, Cout = 'init')
                    )
                )
            }
            if (vn$D > 0) { # continuous rounded
                outlist <- c(
                    outlist,
                    list(
                        Dmean = Dmeans,
                        Drate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Dshapehi,
                                rate = constants$Dvar1),
                            nrow = vn$D, ncol = constants$ncomponents
                        ),
                        Dvar = Dvars,
                        ## for data with boundary values
                        Dlat = constants$Dlatinit
                        ## Dlat = vtransform(data[, vnames$D, with = FALSE],
                        ##   auxmetadata, Dout = 'init')
                    )
                )
            }
            ## if (vn$L > 0) { # latent
            ##     outlist <- c(
            ##         outlist,
            ##         list(
            ##             Lmean = Lmeans,
            ##             Lrate = matrix(
            ##                 nimble::qinvgamma(p = 0.5,
            ##                     shape = constants$Lshapehi,
            ##                     rate = constants$Lvar1),
            ##                 nrow = vn$L, ncol = constants$ncomponents
            ##             ),
            ##             Lvar = matrix(1,
            ##                 nrow = vn$L, ncol = constants$ncomponents),
            ##             ## for data with boundary values
            ##             Llat = constants$Llatinit
            ##             ## Llat = vtransform(data[, vnames$L, with = FALSE],
            ##             ##   auxmetadata, Lout = 'init')
            ##         )
            ##     )
            ## }
            if (vn$O > 0) { # ordinal
                outlist <- c(
                    outlist,
                    list(
                        Oprob = matrix(unlist(sapply(constants$Ocards, function(acard){
                            rep(1 / acard, acard)
                        })), nrow = sum(constants$Ocards), ncol = ncomponents)
                    )
                )
            }
            if (vn$N > 0) { # nominal
                outlist <- c(
                    outlist,
                    list(
                        Nprob = matrix(unlist(sapply(constants$Ncards, function(acard){
                            rep(1 / acard, acard)
                        })), nrow = sum(constants$Ncards), ncol = ncomponents)
                    )
                )
            }
            if (vn$B > 0) { # binary
                outlist <- c(
                    outlist,
                    list(
                        ## Bprob = Bprobs
                        Bprob = matrix(0.5, nrow = vn$B, ncol = constants$ncomponents)
                    )
                )
            }
            ##
            outlist
        } # end precluster

    } else if(initmethod == 'datacentre'){
        ## pre-clustering, centering on data
        initsfn <- function() {
            ## assign each cluster to a datapoint
            iK <- sample(seq_len(constants$npoints), constants$ncomponents,
                replace = (constants$npoints < constants$ncomponents))

            ## Create components centres
            ## distance function
            ## NB: all variances will be initialized to 1
            lpnorm <- function(xx){abs(xx)}
            distances <- matrix(0, nrow = constants$npoints,
                ncol = constants$ncomponents)
            if (vn$R > 0) { # continuous open domain
                Rmeans <- t(datapoints$Rdata[iK, , drop = FALSE])
                ## distances from datapoints
                distances <- distances + apply(Rmeans, 2, function(ameans){
                    colSums(lpnorm(t(datapoints$Rdata) - ameans), na.rm = TRUE)
                })
            }
            if (vn$C > 0) { # continuous closed domain
                Cmeans <- t(datapoints$Clat[iK, , drop = FALSE])
                ## distances from datapoints
                distances <- distances + apply(Cmeans, 2, function(ameans){
                    colSums(lpnorm(t(datapoints$Clat) - ameans), na.rm = TRUE)
                })
            }
            if (vn$D > 0) { # discrete
                Dmeans <- t(constants$Dlatinit[iK, , drop = FALSE])
                ## distances from datapoints
                distances <- distances + apply(Dmeans, 2, function(ameans){
                    colSums(lpnorm(t(constants$Dlatinit) - ameans), na.rm = TRUE)
                })
            }
            ## if (vn$L > 0) { #
            ##     Lmeans <- matrix(rnorm(
            ##         n = vn$L * constants$ncomponents,
            ##         mean = constants$Lmean1,
            ##         sd = sqrt(constants$Lvarm1)
            ##     ), nrow = vn$L, ncol = constants$ncomponents)
            ##     ## distances from datapoints
            ##     distances <- distances + apply(Lmeans, 2, function(ameans){
            ##         colSums(lpnorm(t(constants$Llatinit) - ameans), na.rm = TRUE)
            ##     })
            ## }
            ## if (vn$B > 0) {
            ##     Bprobs <- matrix(rbeta(
            ##         n = vn$B * constants$ncomponents,
            ##         shape1 = Bshapelo,
            ##         shape2 = Bshapehi,
            ##         ), nrow = vn$B, ncol = constants$ncomponents)
            ##     ## distances from datapoints
            ##     distances <- distances + apply(Bprobs, 2, function(ameans){
            ##         colSums(lpnorm(t(datapoints$Bdata) - ameans), na.rm = TRUE)
            ##     })
            ## }

            ## assign datapoints to component with closest centre
            K <- apply(distances, 1, which.min)
            occupied <- unique(K)

            ## recalculate components centres according to their points
            if (vn$R > 0) {
                Rmeans[, occupied] <- sapply(occupied, function(acomponent){
                    colMeans(datapoints$Rdata[which(K == acomponent), , drop = FALSE],
                        na.rm = TRUE)
                })
                Rmeans[, -occupied] <- 0
            }
            if (vn$C > 0) {
                Cmeans[, occupied] <- sapply(occupied, function(acomponent){
                    colMeans(datapoints$Clat[which(K == acomponent), , drop = FALSE],
                        na.rm = TRUE)
                })
                Cmeans[, -occupied] <- 0
            }
            if (vn$D > 0) {
                Dmeans[, occupied] <- sapply(occupied, function(acomponent){
                    colMeans(constants$Dlatinit[which(K == acomponent), , drop = FALSE],
                        na.rm = TRUE)
                })
                Dmeans[, -occupied] <- 0
            }
            ## if (vn$L > 0) { # continuous open domain
            ##     Lmeans[, occupied] <- sapply(occupied, function(acomponent){
            ##         colMeans(constants$Llatinit[which(K == acomponent), , drop = FALSE],
            ##             na.rm = TRUE)
            ##     })
            ##     Lmeans[, -occupied] <- 0
            ## }
            ## if (vn$B > 0) {
            ##     Bprobs[, occupied] <- sapply(occupied, function(acomponent){
            ##         colMeans(datapoints$Bdata[which(K == acomponent), , drop = FALSE],
            ##             na.rm = TRUE)
            ##     })
            ##     Bprobs[, -occupied] <- 0.5
            ## }
            ## Alpha <- sample(1:constants$nalpha, 1, prob = constants$probalpha0, replace = TRUE)
            ## W <- c(rep(rempoints, miconstants$npoints), rep(1, constants$ncomponents - miconstants$npoints))
            ## W <- W/sum(W)

            outlist <- list(
                Alpha = round(constants$nalpha/2),
                W = rep(1/constants$ncomponents, constants$ncomponents),
                ## ## Assign every point to the closest component centre
                K = K
                ## ## Other assignment methods:
                ## ## A. assign all points to an unsystematically chosen component
                ## K = rep(sample(rep(which(W > 0), 2), 1), nepoints)
                ## ## B. distribute points unsystematically among components
                ## K = sample(rep(which(W > 0), 2), constants$npoints, replace = TRUE)
                ## ## or:
                ## ## C. assign all points to the most probable component
                ## K = rep(which.max(W), constants$npoints)
                ## ## or:
                ## ## D. assign all points to the least probable component
                ## K = rep(which(W == min(W[W > 0]))[1], constants$npoints)
                ## ## or:
                ## ## E. distribute points unsystematically among M=2 components
                ## K = sample(sample(rep(which(W > 0), 2), 2, replace = TRUE),
                ##           constants$npoints, replace = TRUE)
                ## ## F. mix methods A. and B.
                ## K = (if(achain %% 2 == Ksample) {
                ##        ## ## assign all points to an unsystematically chosen component
                ##        rep(sample(rep(which(W > 0), 2), 1), constants$npoints)
                ##      } else {
                ##        ## distribute points unsystematically among components
                ##        sample(rep(which(W > 0), 2), constants$npoints, replace = TRUE)
                ##      })
            )
            ##
            if (vn$R > 0) { # continuous open domain
                outlist <- c(
                    outlist,
                    list(
                        Rmean = Rmeans,
                        Rrate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Rshapehi,
                                rate = constants$Rvar1),
                            nrow = vn$R, ncol = constants$ncomponents
                        ),
                        Rvar = matrix(1,
                            nrow = vn$R, ncol = constants$ncomponents)
                    )
                )
            }
            if (vn$C > 0) { # continuous closed domain
                outlist <- c(
                    outlist,
                    list(
                        Cmean = Cmeans,
                        Crate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Cshapehi,
                                rate = constants$Cvar1),
                            nrow = vn$C, ncol = constants$ncomponents
                        ),
                        Cvar = matrix(1,
                            nrow = vn$C, ncol = constants$ncomponents),
                        ## for data with boundary values
                        Clat = constants$Clatinit
                        ## Clat = inferno:::vtransform(data[, vnames$C, with = FALSE],
                        ##   auxmetadata, Cout = 'init')
                    )
                )
            }
            if (vn$D > 0) { # continuous rounded
                outlist <- c(
                    outlist,
                    list(
                        Dmean = Dmeans,
                        Drate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Dshapehi,
                                rate = constants$Dvar1),
                            nrow = vn$D, ncol = constants$ncomponents
                        ),
                        Dvar = matrix(1,
                            nrow = vn$D, ncol = constants$ncomponents),
                        ## for data with boundary values
                        Dlat = constants$Dlatinit
                        ## Dlat = inferno:::vtransform(data[, vnames$D, with = FALSE],
                        ##   auxmetadata, Dout = 'init')
                    )
                )
            }
            ## if (vn$L > 0) { # latent
            ##     outlist <- c(
            ##         outlist,
            ##         list(
            ##             Lmean = Lmeans,
            ##             Lrate = matrix(
            ##                 nimble::qinvgamma(p = 0.5,
            ##                     shape = constants$Lshapehi,
            ##                     rate = constants$Lvar1),
            ##                 nrow = vn$L, ncol = constants$ncomponents
            ##             ),
            ##             Lvar = matrix(1,
            ##                 nrow = vn$L, ncol = constants$ncomponents),
            ##             ## for data with boundary values
            ##             Llat = constants$Llatinit
            ##             ## Llat = inferno:::vtransform(data[, vnames$L, with = FALSE],
            ##             ##   auxmetadata, Lout = 'init')
            ##         )
            ##     )
            ## }
            if (vn$O > 0) { # ordinal
                outlist <- c(
                    outlist,
                    list(
                        Oprob = matrix(unlist(sapply(constants$Ocards, function(acard){
                            rep(1 / acard, acard)
                        })), nrow = sum(constants$Ocards), ncol = ncomponents)
                    )
                )
            }
            if (vn$N > 0) { # nominal
                outlist <- c(
                    outlist,
                    list(
                        Nprob = matrix(unlist(sapply(constants$Ncards, function(acard){
                            rep(1 / acard, acard)
                        })), nrow = sum(constants$Ncards), ncol = ncomponents)
                    )
                )
            }
            if (vn$B > 0) { # binary
                outlist <- c(
                    outlist,
                    list(
                        ## Bprob = Bprobs
                        Bprob = matrix(0.5,
                            nrow = vn$B, ncol = constants$ncomponents)
                    )
                )
            }
            ##
            outlist
        } # end datacentre

    } else if(initmethod == 'prior'){
        ## values chosen from prior
        initsfn <- function() {
            Alpha <- sample(1:constants$nalpha, 1, prob = probalpha0[1:constants$nalpha])
            W <- nimble::rdirch(n = 1,
                alpha = constants$dirchalphas[1:constants$ncomponents] *
                    constants$alphabase^Alpha)
            outlist <- list(
                Alpha = Alpha,
                W = W,
                K = sample(rep(which(W > 0), 2), constants$npoints, replace = TRUE)
            )
            ##
            if (vn$R > 0) { # continuous open domain
                Rrate <- matrix(
                    nimble::rinvgamma(
                        n = vn$R * constants$ncomponents,
                        shape = constants$Rshapehi,
                        rate = constants$Rvar1),
                    nrow = vn$R, ncol = constants$ncomponents
                )
                outlist <- c(
                    outlist,
                    list(
                        Rmean = matrix(rnorm(
                            n = vn$R * constants$ncomponents,
                            mean = constants$Rmean1,
                            sd = sqrt(constants$Rvarm1)
                        ), nrow = vn$R, ncol = constants$ncomponents),
                        Rrate = Rrate,
                        Rvar = matrix(
                            nimble::rinvgamma(
                                n = vn$R * constants$ncomponents,
                                shape = constants$Rshapelo,
                                rate = Rrate),
                            nrow = vn$R, ncol = constants$ncomponents
                        )
                    )
                )
            }
            if (vn$C > 0) { # continuous closed domain
                Crate <- matrix(
                    nimble::rinvgamma(
                        n = vn$C * constants$ncomponents,
                        shape = constants$Cshapehi,
                        rate = constants$Cvar1),
                    nrow = vn$C, ncol = constants$ncomponents
                )
                outlist <- c(
                    outlist,
                    list(
                        Cmean = matrix(rnorm(
                            n = vn$C * constants$ncomponents,
                            mean = constants$Cmean1,
                            sd = sqrt(constants$Cvarm1)
                        ), nrow = vn$C, ncol = constants$ncomponents),
                        Crate = Crate,
                        Cvar = matrix(
                            nimble::rinvgamma(
                                n = vn$C * constants$ncomponents,
                                shape = constants$Cshapelo,
                                rate = Crate),
                            nrow = vn$C, ncol = constants$ncomponents
                        ),
                        Clat = constants$Clatinit
                    )
                )
            }
            if (vn$D > 0) { # continuous rounded
                Drate <- matrix(
                    nimble::rinvgamma(
                        n = vn$D * constants$ncomponents,
                        shape = constants$Dshapehi,
                        rate = constants$Dvar1),
                    nrow = vn$D, ncol = constants$ncomponents
                )
                outlist <- c(
                    outlist,
                    list(
                        Dmean = matrix(rnorm(
                            n = vn$D * constants$ncomponents,
                            mean = constants$Dmean1,
                            sd = sqrt(constants$Dvarm1)
                        ), nrow = vn$D, ncol = constants$ncomponents),
                        Drate = Drate,
                        Dvar = matrix(
                            nimble::rinvgamma(
                                n = vn$D * constants$ncomponents,
                                shape = constants$Dshapelo,
                                rate = Drate),
                            nrow = vn$D, ncol = constants$ncomponents
                        ),
                        Dlat = constants$Dlatinit
                    )
                )
            }
            if (vn$O > 0) { # ordinal
                outlist <- c(
                    outlist,
                    list(
                        Oprob = matrix(unlist(sapply(constants$Ocards, function(acard){
                            rep(1 / acard, acard)
                        })), nrow = sum(constants$Ocards), ncol = ncomponents)
                    )
                )
            }
            if (vn$N > 0) { # nominal
                outlist <- c(
                    outlist,
                    list(
                        Nprob = aperm(array(sapply(1:vn$N, function(avar) {
                            sapply(1:constants$ncomponents, function(aclus) {
                                constants$Nalpha0[avar, ]/sum(constants$Nalpha0[avar, ])
                                ## nimble::rdirch(n = 1, alpha = constants$Nalpha0[avar, ])
                            })
                        }), dim = c(Nmaxn, constants$ncomponents, vn$N)))
                    )
                )
            }
            if (vn$B > 0) { # nominal
                outlist <- c(
                    outlist,
                    list(
                        Bprob = matrix(
                            rbeta(n = vn$B * constants$ncomponents,
                                shape1 = Bshapelo, shape2 = Bshapehi),
                            nrow = vn$B, ncol = constants$ncomponents)
                    )
                )
            }
            ##
            outlist
        } # end prior


    } else if(initmethod == 'allcentre'){
        ## all components equal, all points to first component
        initsfn <- function() {
            initvar <- 6
            outlist <- list(
                Alpha = round(constants$nalpha/2),
                W = rep(1/constants$ncomponents, constants$ncomponents),
                ## ## Assign every point to first component
                K = rep(1, constants$npoints)
            )
            ##
            if (vn$R > 0) { # continuous open domain
                outlist <- c(
                    outlist,
                    list(
                        Rmean =  matrix(0, nrow = vn$R, ncol = constants$ncomponents),
                        Rrate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Rshapehi,
                                rate = constants$Rvar1),
                            nrow = vn$R, ncol = constants$ncomponents
                        ),
                        Rvar = matrix(initvar, nrow = vn$R, ncol = constants$ncomponents)
                    )
                )
            }
            if (vn$C > 0) { # continuous closed domain
                outlist <- c(
                    outlist,
                    list(
                        Cmean = matrix(0, nrow = vn$C, ncol = constants$ncomponents),
                        Crate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Cshapehi,
                                rate = constants$Cvar1),
                            nrow = vn$C, ncol = constants$ncomponents
                        ),
                        Cvar = matrix(initvar, nrow = vn$C, ncol = constants$ncomponents),
                        ## for data with boundary values
                        Clat = constants$Clatinit
                    )
                )
            }
            if (vn$D > 0) { # continuous rounded
                outlist <- c(
                    outlist,
                    list(
                        Dmean = matrix(0, nrow = vn$D, ncol = constants$ncomponents),
                        Drate = matrix(
                            nimble::qinvgamma(p = 0.5,
                                shape = constants$Dshapehi,
                                rate = constants$Dvar1),
                            nrow = vn$D, ncol = constants$ncomponents
                        ),
                        Dvar = matrix(initvar, nrow = vn$D, ncol = constants$ncomponents),
                        ## for data with boundary values
                        Dlat = constants$Dlatinit
                    )
                )
            }
            if (vn$O > 0) { # ordinal
                outlist <- c(
                    outlist,
                    list(
                        Oprob = matrix(unlist(sapply(constants$Ocards, function(acard){
                            rep(1 / acard, acard)
                        })), nrow = sum(constants$Ocards), ncol = ncomponents)
                    )
                )
            }
            if (vn$N > 0) { # nominal
                outlist <- c(
                    outlist,
                    list(
                        Nprob = matrix(unlist(sapply(constants$Ncards, function(acard){
                            rep(1 / acard, acard)
                        })), nrow = sum(constants$Ncards), ncol = ncomponents)
                    )
                )
            }
            if (vn$B > 0) { # binary
                outlist <- c(
                    outlist,
                    list(
                        Bprob = matrix(0.5, nrow = vn$B, ncol = constants$ncomponents)
                    )
                )
            }
            ##
            outlist
        } # end allcentre

    } else if(initmethod == 'allinone'){
        ## Components from prior, data all in one
        initsfn <- function() {
            outlist <- list(
                Alpha = round(constants$nalpha/2),
                W = rep(1/constants$ncomponents, constants$ncomponents),
                ## ## Assign every point to first component
                K = rep(1, constants$npoints)
            )
            ##
            if (vn$R > 0) { # continuous open domain
                Rrate <- matrix(
                    nimble::rinvgamma(
                        n = vn$R,
                        shape = constants$Rshapehi,
                        rate = constants$Rvar1),
                    nrow = vn$R, ncol = constants$ncomponents
                )
                outlist <- c(
                    outlist,
                    list(
                        Rmean = matrix(rnorm(
                            n = vn$R,
                            mean = constants$Rmean1,
                            sd = sqrt(constants$Rvarm1)
                        ), nrow = vn$R, ncol = constants$ncomponents),
                        Rrate = Rrate,
                        Rvar = matrix(
                            nimble::rinvgamma(
                                n = vn$R,
                                shape = constants$Rshapelo,
                                rate = Rrate),
                            nrow = vn$R, ncol = constants$ncomponents
                        )
                    )
                )
            }
            if (vn$C > 0) { # continuous closed domain
                Crate <- matrix(
                    nimble::rinvgamma(
                        n = vn$C,
                        shape = constants$Cshapehi,
                        rate = constants$Cvar1),
                    nrow = vn$C, ncol = constants$ncomponents
                )
                outlist <- c(
                    outlist,
                    list(
                        Cmean = matrix(rnorm(
                            n = vn$C,
                            mean = constants$Cmean1,
                            sd = sqrt(constants$Cvarm1)
                        ), nrow = vn$C, ncol = constants$ncomponents),
                        Crate = Crate,
                        Cvar = matrix(
                            nimble::rinvgamma(
                                n = vn$C,
                                shape = constants$Cshapelo,
                                rate = Crate),
                            nrow = vn$C, ncol = constants$ncomponents
                        ),
                        Clat = constants$Clatinit
                    )
                )
            }
            if (vn$D > 0) { # continuous rounded
                Drate <- matrix(
                    nimble::rinvgamma(
                        n = vn$D,
                        shape = constants$Dshapehi,
                        rate = constants$Dvar1),
                    nrow = vn$D, ncol = constants$ncomponents
                )
                outlist <- c(
                    outlist,
                    list(
                        Dmean = matrix(rnorm(
                            n = vn$D,
                            mean = constants$Dmean1,
                            sd = sqrt(constants$Dvarm1)
                        ), nrow = vn$D, ncol = constants$ncomponents),
                        Drate = Drate,
                        Dvar = matrix(
                            nimble::rinvgamma(
                                n = vn$D,
                                shape = constants$Dshapelo,
                                rate = Drate),
                            nrow = vn$D, ncol = constants$ncomponents
                        ),
                        Dlat = constants$Dlatinit
                    )
                )
            }
            if (vn$O > 0) { # ordinal
                outlist <- c(
                    outlist,
                    list(
                        Oprob = matrix(unlist(sapply(seq_along(constants$Ocards),
                            function(v){nimble::rdirch(n = 1,
                                alpha = Oalpha0[Oi[v]:Of[v]])
                            })), nrow = sum(constants$Ocards), ncol = ncomponents)
                    )
                )
            }
            if (vn$N > 0) { # nominal
                outlist <- c(
                    outlist,
                    list(
                        Nprob = matrix(unlist(sapply(seq_along(constants$Ncards),
                            function(v){nimble::rdirch(n = 1,
                                alpha = Nalpha0[Ni[v]:Nf[v]])
                            })), nrow = sum(constants$Ncards), ncol = ncomponents)
                    )
                )
            }
            if (vn$B > 0) { # nominal
                outlist <- c(
                    outlist,
                    list(
                        Bprob = matrix(
                            rbeta(n = vn$B,
                                shape1 = Bshapelo, shape2 = Bshapehi),
                            nrow = vn$B, ncol = constants$ncomponents)
                    )
                )
            }
            ##
            outlist
        } # end allinone

    } else {stop('Unknown "initmethod"')}

#################################################
#### NIMBLE SETUP
##################################################

    finitemixnimble <- nimbleModel(
        code = finitemix,
        name = 'finitemixnimble1',
        constants = constants,
        data = datapoints,
        ## dimensions = c(
        ##     ##     if (vn$R > 0) {
        ##     ##         list(
        ##     ##             Rdata = c(npoints, vn$R)
        ##     ##             )
        ##     ##     },
        ##     ##     if (vn$C > 0) {
        ##     ##         list(
        ##     ##             Caux = c(npoints, vn$C),
        ##     ##             Clat = c(npoints, vn$C),
        ##     ##             Cleft = c(npoints, vn$C),
        ##     ##             Cright = c(npoints, vn$C),
        ##     ##             Clatinit = c(npoints, vn$C)
        ##     ##             )
        ##     ##     },
        ##     ##     if (vn$D > 0) {
        ##     ##         list(
        ##     ##             Daux = c(npoints, vn$D),
        ##     ##             Dleft = c(npoints, vn$D),
        ##     ##             Dright = c(npoints, vn$D),
        ##     ##             Dlatinit = c(npoints, vn$D)
        ##     ##             )
        ##     ##     },
        ##     ## if (vn$O > 0) {
        ##     ##     list(
        ##     ##         Oi = length(Oi),
        ##     ##         Of = length(Of)
        ##     ##         ##             Odata = c(npoints, vn$O)
        ##     ##     )
        ##     ## },
        ##     ## if (vn$N > 0) {
        ##     ##     list(
        ##     ##         Ni = length(Ni),
        ##     ##         Nf = length(Nf)
        ##     ##         ##             Ndata = c(npoints, vn$N)
        ##     ##     )
        ##     ## }
        ##     ##     if (vn$B > 0) {
        ##     ##         list(
        ##     ##             Bdata = c(npoints, vn$B)
        ##     ##             )
        ##     ##     }
        ## ),
        inits = initsfn()
    )

    Cfinitemixnimble <- compileNimble(finitemixnimble,
        showCompilerOutput = FALSE)
    usedmem <- max(usedmem, sum(gc(full = TRUE)[,6])) #garbage collection

    confnimble <- configureMCMC(
        Cfinitemixnimble, # nodes = NULL
        monitors = c(
            'W',
            if (vn$R > 0) {
                c('Rmean', 'Rvar')
            },
            if (vn$C > 0) {
                c('Cmean', 'Cvar')
            },
            if (vn$D > 0) {
                c('Dmean', 'Dvar')
            },
            ## if (vn$L > 0) {
            ##     c('Lmean', 'Lvar')
            ## },
            if (vn$O > 0) {
                c('Oprob')
            },
            if (vn$N > 0) {
                c('Nprob')
            },
            if (vn$B > 0) {
                c('Bprob')
            }
        ),
        ## It is necessary to monitor K to see if all components were used
        ## if 'showAlphatraces' is true,
        ## then the Alpha-parameter trace is also recorded and shown
        monitors2 = c(if (showAlphatraces) { 'Alpha' },
            'K')
    )

    ## ## Uncomment to debug Nimble (in case of Nimble updates)
    ## print(confnimble$getUnsampledNodes())
    ## cat('\nEX1\n')
    ## confnimble$printSamplers(executionOrder=TRUE)

    targetslist <- sapply(confnimble$getSamplers(), function(xx) xx$target)
    nameslist <- sapply(confnimble$getSamplers(), function(xx) xx$name)
    ## cat('\nNAMESLIST', nameslist, '\n')

    ## replace W's sampler with the one from Daniel Turek
    if(is.null(avoidzeroW)) {
        confnimble$replaceSampler('W', conjugate_dirch_plus_eps)
    }

    ## replace Alpha's cat-sampler with slice
    if (Alphatoslice &&
            !('Alpha' %in% targetslist[nameslist == 'posterior_predictive'])) {
        confnimble$replaceSampler(target='Alpha', type='slice')
        ## ## Old replacement method, didn't work in previous Nimble
        ## confnimble$removeSamplers('Alpha')
        ## confnimble$addSampler(target = 'Alpha', type = 'slice')
    }

    ## replace K's cat-sampler with slice
    if (Ktoslice) {
        for (asampler in grep('^K\\[', targetslist, value = TRUE)) {
            if (!(asampler %in% targetslist[nameslist == 'posterior_predictive'])) {
                confnimble$replaceSampler(target=asampler, type='slice')
                ## ## Old replacement method, didn't work in previous Nimble
                ## confnimble$removeSamplers(asampler)
                ## confnimble$addSampler(target = asampler, type = 'slice')
            }
        }
    }

    ## replace all RW samplers with slice
    ## Should find a way to do this faster
    ## testreptime <- Sys.time()
    if (RWtoslice) {
        for (asampler in targetslist[nameslist == 'RW']) {
            ## ## New replacement method, didn't work in previous Nimble
            confnimble$replaceSampler(target=asampler, type='slice')
            ## ## Old replacement method:
            ## confnimble$removeSamplers(asampler)
            ## confnimble$addSampler(target = asampler, type = 'slice')
        }
    }

    ## ## Uncomment when debugging Nimble
    ## print(confnimble$getUnsampledNodes())
    ## cat('\nEX2\n')
    ## confnimble$printSamplers(executionOrder=TRUE)

#### change execution order for some variates
    if (changeSamplerOrder) {
        ## call this to do a first reordering of the samplers
        ## it places posterior-predictive nodes last
        mcsampler <- buildMCMC(confnimble)

        samplerorder <- c(
            'K',
            'W',
            'Alpha',
            if (vn$R > 0) {
                c('Rmean', 'Rrate', 'Rvar')
            },
            if (vn$C > 0) {
                c('Cmean', 'Crate', 'Cvar')
            },
            if (vn$D > 0) {
                c('Dmean', 'Drate', 'Dvar')
            },
            ## if (vn$L > 0) {
            ##     c('Lmean', 'Lrate', 'Lvar')
            ## },
            if (vn$O > 0) {
                c('Oprob')
            },
            if (vn$N > 0) {
                c('Nprob')
            },
            if (vn$B > 0) {
                c('Bprob')
            }
        )
        ##
        neworder <- unlist(lapply(samplerorder, function(var){
            grep(
                paste0('^', var, '(\\[.+\\])*$'),
                sapply(confnimble$getSamplers(), function(x) {
                    if (!(x$name == 'posterior_predictive')) {
                        x$target
                    } else {
                        NULL
                    }
                })
            )
        }))
        ## ## Uncomment for debugging
        ## cat('\nNEW ORDER',neworder,'\n')
        confnimble$setSamplerExecutionOrder(c(setdiff(
            confnimble$getSamplerExecutionOrder(),
            neworder
        ), neworder))
        ## cat('\nEX3\n')
        ## confnimble$printSamplers(executionOrder=TRUE)

    }

#### Compile Monte Carlo sampler
    print(confnimble)
    mcsampler <- buildMCMC(confnimble)
    ## print(confnimble$getUnsampledNodes())
    Cmcsampler <- compileNimble(mcsampler, resetFunctions = TRUE)

    cat('\nSetup time',
        printtimediff(difftime(Sys.time(), headertimestart, units = 'auto')),
        '\n')
    headertime <- difftime(Sys.time(), headertimestart, units = 'secs')


    ## Inform user that compilation is done, if core 1:
    if (acore == 1) {
        print2user(paste0('\rCompiled core ', acore, '. ',
            'Number of samplers: ',
            length(confnimble$samplerExecutionOrder), '.             \n',
            'Estimating remaining time, please be patient...'))
    }

    ## cat('Loop over chains')
##################################################
#### LOOP OVER CHAINS (WITHIN ONE CORE)
##################################################
    ## Start timer
    MCtimestart <- Sys.time()

    maxusedcomponents <- 0
    maxiterations <- 0
    ## keep count of chains having non-finite outputs
    nonfinitechains <- 0L
    ## keep count of chains stopped prematurely
    stoppedchains <- 0L
    usedmem <- max(usedmem, sum(gc(full = TRUE)[,6])) #garbage collection

#### LOOP OVER CHAINS IN CORE
    nchainsperthiscore <- minchainspercore + (acore <= coreswithextrachain)
    ## print2user(paste0('\ncore ',acore,': ',nchainsperthiscore,'\n'))

    for (achain in 1:nchainsperthiscore) {

        chainnumber <- achain + minchainspercore * (acore - 1) +
            min(coreswithextrachain, acore - 1)
        padchainnumber <- sprintf(paste0('%0', nchar(nchains), 'i'), chainnumber)

        ## calculate how the remaining time is allotted to remaining chains
        timeleft <- (maxhours -
                         as.double(Sys.time() - timestart0, units = 'hours')) /
            (nchainsperthiscore - achain + 1)

        showsamplertimes0 <- showsamplertimes && (achain == 1)
        ## showAlphatraces0 <- showAlphatraces && (achain==1)
        niter <- min(startupMCiterations, maxMCiterations)
        ## ## Experimental: decrease number of iterations based on previous chain
        ## if(achain == 1){
        ##   niter <- startupMCiterations
        ## }else{
        ##  niter <- max(min(startupMCiterations,remainiter*2), 128)
        ##  }

        ## ## nitertot: total number of MC samples from chain start
        nitertot <- 0L
        remainiter <- +Inf
        reset <- TRUE
        allmcsamples <- NULL
        ## allmcsamplesKA <- list(Alpha = NULL, K = NULL)
        allmcsamplesKA <- NULL
        flagll <- FALSE
        flagnonfinite <- FALSE
        cat('\n###########################################################\nChain #',
            chainnumber,
            '(chain', achain, 'of', nchainsperthiscore, 'for this core)\n'
        )
        ## Read data to be used in log-likelihood
        testdata <- readRDS(file = file.path(dirname,
            paste0('___testdata_', chainnumber, '.rds')))
        cat('\nDatapoints for testing chain behaviour:\n',
            paste0('#', testdata$pointsid), '\n')

        ## will contain the MC traces of the test points
        traces <- matrix(NA_real_, nrow = 0, ncol = 1,
            dimnames = list(NULL, 'gmean'))


        ## Initial values for this chain
        ## random seed is taken care of by %doRNG%
        Cfinitemixnimble$setInits(initsfn())

        subiter <- 1L
        savedchunks <- 0L
#### WHILE-LOOP CONTINUING UNTIL STOP CRITERION MET
        while (remainiter > 0) {
            cat('\nIterations:', niter, '\n')

            ## MONTE-CARLO CALL
            ## If reporting Alpha or K traces,
            ## then save them more frequently
            ## Otherwise just save the last value
            Cmcsampler$run(
                niter = niter,
                thin = 1,
                thin2 = (if (showAlphatraces || showKtraces) {
                    max(1, round(niter / ncomponentsamples))
                } else {
                    max(2, floor(niter / 2))
                }),
                nburnin = 0,
                time = showsamplertimes0,
                reset = reset,
                resetMV = TRUE
            )

            ## iterationAsLastIndex: See sect 7.7 of Nimble manual
            mcsamples <- as.list(Cmcsampler$mvSamples,
                iterationAsLastIndex = TRUE)
            mcsamplesKA <- as.list(Cmcsampler$mvSamples2,
                iterationAsLastIndex = FALSE)

            ## 'mcsamplesKA$K' contains the component identity
            ## of each training datapoint, but we only want
            ## the number of distinct components used:
            mcsamplesKA$K <- apply(mcsamplesKA$K, 1,
                function(xx){length(unique(xx))})

            if (showAlphatraces) {
                dim(mcsamplesKA$Alpha) <- NULL # from matrix to vector
            }

            cat('\nCurrent time:',
                format(Sys.time(), '%Y-%m-%d %H:%M:%S'))
            cat('\nMCMC time',
                printtimediff(difftime(Sys.time(), MCtimestart, units = 'auto')),
                '\n')

#### Remove iterations with non-finite values
            if(any(!is.finite(unlist(mcsamples)))) {
                toRemove <- sort(unique(unlist(lapply(mcsamples, function(xx) {
                    temp <- which(!is.finite(xx), arr.ind = TRUE)
                    temp[, ncol(temp)]
                }))))

                cat('\nWARNING:', length(toRemove), 'NON-FINITE SAMPLES\n')
                ##
                flagnonfinite <- TRUE
                nonfinitechains <- TRUE
                saveRDS(mcsamples, file = file.path(dirname,
                    paste0('NONFINITEmcsamples',
                        dashnameroot, '--', padchainnumber,
                        '_', achain, '-',
                        acore, '-i', nitertot, '.rds')
                ))
                if (length(toRemove) == ncol(mcsamples$W)) {
                    cat('\n...TOO MANY NON-FINITE OUTPUTS!\n')
                    ## ## registerDoSEQ()
                    ## if(exists('cl')){ parallel::stopCluster(cl) }
                    ## stop('...TOO MANY NON-FINITE OUTPUTS. ABORTING')
                    mcsamples <- NULL
                    niter <- 0
                } else {
                    mcsamples <- mcsubset(mcsamples, -toRemove)
                    niter <- niter - length(toRemove)
                }
            }

            nitertot <- nitertot + niter

            ##
            if (showsamplertimes0) {
                samplertimes <- Cmcsampler$getTimes()
                names(samplertimes) <- sapply(
                    confnimble$getSamplers(),
                    function(x) x$target
                )
                sprefixes <- unique(sub(
                    '^([^[]+)(\\[.*\\])', '\\1',
                    names(samplertimes)
                ))
                cat('\nSampler times:\n')
                print(sort(sapply(sprefixes,
                    function(x) sum(samplertimes[grepl(x, names(samplertimes))])),
                    decreasing = TRUE
                ))
            }

            ## Check how many components were used at the last step
            ## usedcomponents <- mcsamplesKA$K[length(mcsamplesKA$K)]
            usedcomponents <- max(mcsamplesKA$K)
            cat('\nUSED COMPONENTS:', usedcomponents, 'OF', ncomponents, '\n')

#### Update traces
            ## Log-likelihood
            diagntime <- Sys.time()
            ##

            ll <- util_Pcheckpoints(
                testdata = testdata,
                learnt = mcsamples
            )

            ll <- cbind(
                exp(rowMeans(log(ll), na.rm = TRUE)), # geometric mean
                ll
            )
            colnames(ll) <- c('gmean', testdata$pointsid)

            if(plottraces){
                saveRDS(ll,
                    file = file.path(dirname,
                        paste0('____tempPtraces-',
                            padchainnumber, '-',
                            subiter, '.rds'))
                )
            }

            traces <- rbind(traces, ll[, 1, drop = FALSE])

            toRemove <- which(!is.finite(traces))
            if (length(toRemove) > 0) {
                flagll <- TRUE
                oktraces <- traces[-unique(toRemove), , drop = FALSE]
            } else {
                oktraces <- traces
            }

            ## ## version before geom.mean
            ## toRemove <- which(!is.finite(traces), arr.ind = TRUE)
            ## if (length(toRemove) > 0) {
            ##     flagll <- TRUE
            ##     oktraces <- traces[-unique(toRemove[, 1]), , drop = FALSE]
            ## } else {
            ##     oktraces <- traces
            ## }

            ## ## for debugging
            ## saveRDS(traces,
            ##     file = file.path(dirname,
            ##         paste0('___traces',
            ##             dashnameroot, '--',
            ##             padchainnumber,'-',subiter, '.rds')
            ##     ))
            ## saveRDS(oktraces,
            ##     file = file.path(dirname,
            ##         paste0('___oktraces',
            ##             dashnameroot, '--',
            ##             padchainnumber,'-',subiter, '.rds')
            ##     ))


            ## ##############################
            ## ## MCSE, ESS, THINNING, ETC ##
            ## ##############################

            N <- nrow(oktraces)

            ## Calculate empirical quantiles of the vector of MC samples,
            ## having 'Qlo' and 'Qhi' probabilities
            Xlo <- quantile(x = oktraces, probs = Qlo,
                na.rm = FALSE, names = FALSE, type = 6)
            Xhi <- quantile(x = oktraces, probs = Qhi,
                na.rm = FALSE, names = FALSE, type = 6)
            ## quantile width between empirical quantiles
            qwidth <- Xhi - Xlo

            ## Calculate credibility intervals (proxy for MCSE)
            ## for the empirical quantiles above
            ## as in Vehtari et al. 2021
            temp <- funMCEQ(x = oktraces, prob = c(Qlo, Qhi), Qpair = Qerror)
            wXlo <- temp[2, 1] - temp[1, 1]
            wXhi <- temp[2, 2] - temp[1, 2]

            ## For each monitored point,
            ## transform its posterior samples to normalized ranks,
            ## then calculate the ESSs of their means,
            ## as in Vehtari et al. 2021
            ## and Blom 1958 sect 6.10 eqn (6.10.5)
            ## (alternative (6.10.3) gives roughly same results
            ## but is computationally slightly more expensive)
            essnrmean <- funESS3(qnorm(
            (rank(oktraces, na.last = NA, ties.method = 'average') - 0.5) / N
            ))

            ## We check the relative MCSE of the following quantities:
            ## - the mean of the MC samples, transformed to normalized ranks
            ## - the lower quantile 'Qlo'
            ## - the upper quantile 'Qhi'
            ## so as to ensure the stability of the mean value
            ## (representing the "long-run frequency")
            ## and of its 'Qlo'- and 'Qhi'-quantiles
            ## (representing the uncertainty about the freq. estimate)
            relmcse <- c(1 / sqrt(essnrmean), wXlo / qwidth, wXhi / qwidth)

            autothinning <- N * max(relmcse)^2

            ## Output available diagnostics
            toprint <- list(
                'rel. quantile error' = max(relmcse[-1]),
                'ESS from rel. qt. error' = (1/max(relmcse[-1]))^2,
                'ESS' = essnrmean,
                'needed thinning' = autothinning,
                'average' = mean(oktraces),
                'quantile width' = qwidth
            )

####
            for(i in names(toprint)) {
                thisdiagn <- toprint[[i]]
                cat(paste0('\n', i, ':'),
                    if(length(thisdiagn) > 1){
                        paste(signif(range(thisdiagn), 3), collapse = ' to ')
                    } else {
                        signif(thisdiagn, 3)
                    }
                )
            }

            rm(oktraces)

            cat('\nDiagnostics time',
                printtimediff(difftime(Sys.time(), diagntime, units = 'auto')),
                '\n')

#### concatenate samples with those of previous chunk, if existing
            ## add overall iteration index `MCindex` to mcsamples
            mcsamples$MCindex <- seq(to = nitertot,
                length.out = ncol(mcsamples$W))
            ## need to add 1D to behave well with mcsubset()
            dim(mcsamples$MCindex) <- ncol(mcsamples$W)

            ## concatenate samples
                allmcsamples <- mcjoin(allmcsamples, mcsamples)

            if (is.null(allmcsamplesKA)) {
                allmcsamplesKA <- mcsamplesKA
            } else {
                ## Concatenate samples of K and Alpha
                ## if (showAlphatraces || showKtraces) {
                allmcsamplesKA <- mapply(
                    function(xx, yy) {
                        c(xx, yy)
                    },
                    allmcsamplesKA, mcsamplesKA,
                    SIMPLIFY = FALSE
                )
                ## }
            }

            ## ######################################
            ## ## CHECK IF CHAIN MUST BE CONTINUED ##
            ## ######################################

            lowess <- (1 / max(relmcse))^2

            neededsamples <- max(ceiling(autothinning * (minESS + initES)),
                nsamplesperchain)

            ## if(relmcse > minrelMCSE)
            if(lowess < minESS + initES) {
                ## too small ESS
                reqiter <- neededsamples - nitertot
            } else {
                reqiter <- 0
            }

            ## respect min and max iterations chosen by user
            remainiter <- max(minMCiterations - nitertot, reqiter)
            remainiter <- min(maxMCiterations - nitertot, remainiter)

            cat('\nTotal iterations', nitertot,
                '- require further', reqiter,
                '- continue for', remainiter, '\n')

            if(as.double(Sys.time() - timestart0, units = 'hours') >= timeleft &&
                   remainiter > 0 && nitertot >= nsamplesperchain
            ) {
                cat('but stopping chain because of lack of time\n')
                stoppedchains <- stoppedchains + 1L
                remainiter <- 0
            }

            ## Save cumulated mcsamples to save memory
            ## Only saved when we have at least
            ## 'startupMCiterations' samples
            if(ncol(allmcsamples$W) > startupMCiterations/2 ||
                   remainiter <= 0){
                savedchunks <- savedchunks + 1L
                saveRDS(
                    allmcsamples,
                    file = file.path(dirname,
                        paste0('____tempmcsamples-',
                            padchainnumber, '-',
                            savedchunks, '.rds'))
                )
                allmcsamples <- NULL
            }
            gc(full = TRUE)

            if (remainiter > 0) { # This chain is going to continue

                ## limit number of iterations per loop, to save memory
                niter <- min(remainiter + 1L, startupMCiterations)
                subiter <- subiter + 1L
                cat('\nChain #', chainnumber, '- chunk', subiter,
                    '(chain', achain, 'of', nchainsperthiscore,
                    'for this core): increasing by', niter, '\n'
                )
            }

            ## ###########
            ## ## PLOTS ##
            ## ###########

#### Plot diagnostic traces of current chain
            if (plottraces) {
                cat('\nPlotting traces and samples.\n')

                ## Plot various info and traces
                ## colpalette <- 1:6 #seq_len(ncol(traces))
                ## names(colpalette) <- colnames(traces)[1:min(6, ncol(traces))]
                graphics.off()
                pdf(file.path(dirname,
                    paste0('___mcpartialtraces', dashnameroot, '--',
                        padchainnumber, '_', achain, '-', acore, '.pdf')
                ), height = 8.27, width = 11.69)

                ## Summary stats
                matplot(1:2, type = 'l', col = 'white',
                    main = paste0('Stats chain ', achain),
                    axes = FALSE, ann = FALSE)
                ## Legends
                ## legendpositions <- c('topleft', 'topright', 'bottomleft', 'bottomright')
                ## for (alegend in seq_along(grouplegends)) {
                ##     legend(x = legendpositions[alegend], bty = 'n', cex = 1.5,
                ##         legend = grouplegends[[alegend]]
                ##     )
                ## }
                legend(x = 'left', bty = 'n', cex = 0.75,
                    legend = c(
                        paste0('Chain ', chainnumber, ' - ',
                            achain, ' of core ', acore),
                        ##
                        paste0('Test points ',
                            paste0('#', testdata$pointsid, collapse=' ')
                        ),
                        ##
                        paste0('Iterations: ', nitertot),
                        ##
                        paste0('Used components: ', usedcomponents,
                            ' of ', ncomponents),
                        ##
                        'NOTES:',
                        if (flagnonfinite) {
                            'some non-finite MC outputs'
                        },
                        if (usedcomponents > ncomponents - 5) {
                            'too many components used'
                        },
                        if (flagll) {
                            'non-finite values in diagnostics'
                        }
                    )
                )

                ## Traces of likelihood and cond. probabilities
                ## Save user's par()
                ## oldpar <- par(mfrow = c(1, 1))
                ## oldpar$new <- NULL
                ## on.exit(par(oldpar), add = TRUE)

                for (avar in 1:ncol(traces)) {
                    flexiplot(
                        y = 10*log10(traces[is.finite(traces[, avar]), avar]),
                        type = 'l', lty = 1, col = 1,
                        main = paste0('#', colnames(traces)[avar], ': ',
                            paste(
                                names(toprint),
                                sapply(toprint, function(xx){
                                    signif(xx[avar], 3)
                                }),
                                collapse = ' | ', sep = ': '
                            )),
                        cex.main = 1.25,
                        ylab = paste0('log_F(#',
                            colnames(traces)[avar],
                            ')/dHart'),
                        xlab = 'Monte Carlo sample',
                        family = family
                        ## mar = c(NA, 6, NA, NA)
                    )
                }
                dev.off()
            }

#### Plot Alpha and component occupation, if required
            if (showAlphatraces || showKtraces) {
                cat('Plotting component and Alpha information.\n')
                pdf(file = file.path(dirname,
                    paste0('hyperparams_traces', dashnameroot, '--',
                        padchainnumber, '_', achain, '-', acore, '.pdf')),
                    height = 8.27, width = 11.69)

                if (showKtraces) {
                    cat('\nSTATS USED COMPONENTS:\n')
                    print(summary(allmcsamplesKA$K))
                    ##
                    flexiplot(y = allmcsamplesKA$K, ylab = 'used components',
                        xlab = 'iteration', ylim = c(0, ncomponents))
                    flexiplot(x = 0:ncomponents,
                        y = tabulate(allmcsamplesKA$K + 1, nbins = ncomponents + 1),
                        type = 'l', xlab = 'used components', ylab = NA,
                        ylim = c(0, NA))
                }
                if (showAlphatraces) {
                    cat('\nSTATS alpha:\n')
                    print(summary(allmcsamplesKA$Alpha, na.rm = TRUE))
                    flexiplot(y = allmcsamplesKA$Alpha,
                        ylab = bquote(alpha), xlab = 'iteration',
                        ylim = c(1, nalpha))
                    flexiplot(x = seq(minalpha, maxalpha, by = byalpha),
                        y = tabulate(allmcsamplesKA$Alpha, nbins = nalpha),
                        type = 'l', xlab = bquote(alpha), ylab = '',
                        ylim = c(0, NA))
                }
                dev.off()
            }

            ## ###############
            ## ## END PLOTS ##
            ## ###############

#### Print estimated end time
            fracchain <- achain - remainiter / (remainiter + nitertot)
            endTime <- Sys.time() + 180 +
                ( (nchainsperthiscore + (acore > coreswithextrachain) - fracchain) *
                      difftime(Sys.time(), MCtimestart) / fracchain )
            print2user(paste0(
                '\rSampling. Core ', acore, ' estimated end time: ',
                format(endTime, format='%Y-%m-%d %H:%M'), '   '
                ))

            reset <- FALSE
        }
#### END WHILE-LOOP OVER CHUNKS OF ONE CHAIN


        ## #####################################
        ## ## BUILD AND SAVE CHAIN QUANTITIES ##
        ## #####################################

#### Determine which MC samples should be saved
        cat('\nKeeping ', nsamplesperchain, '\n')

        tokeep <- round(seq(from = autothinning * initES,
            to = nitertot,
            length.out = nsamplesperchain
        ))
        if(length(tokeep) > length(unique(tokeep))){
            cat('\nWARNING: reduced thinning owing to time constraints\n')
            tokeep <- round(seq(from = max(1, nitertot - nsamplesperchain + 1L),
                to = nitertot,
                length.out = nsamplesperchain))
        }

        ## Read, subset, combine saved samples
        for(chunk in seq_len(savedchunks)){
            tempmcsamples <- readRDS(file = file.path(
                dirname,
                paste0('____tempmcsamples-',
                    padchainnumber, '-',
                    chunk, '.rds')
            ))
            tempmcsamples <- mcsubset(tempmcsamples,
                which(tempmcsamples$MCindex %in% tokeep)
            )

                allmcsamples <- mcjoin(allmcsamples, tempmcsamples)
        }

        ## Save thinned total chain
        saveRDS(allmcsamples,
            file = file.path(dirname,
                paste0('___mcsamples',
                    dashnameroot, '--',
                    padchainnumber, '.rds'))
        )

        saveRDS(allmcsamplesKA,
            file = file.path(dirname,
                paste0('___hyperparams_traces',
                    dashnameroot, '--',
                    padchainnumber, '.rds'))
        )

        ## put 'tokeep' in first slot if saving only nsamplesperchain
        saveRDS(traces[ , , drop = FALSE],
            file = file.path(dirname,
                paste0('___mcpartialtraces',
                    dashnameroot, '--',
                    padchainnumber, '.rds')
            ))

        ## possibly increase the count of chains with non-finite outputs
        nonfinitechains <- nonfinitechains + flagnonfinite

        cat('\nCurrent time:', format(Sys.time(), '%Y-%m-%d %H:%M:%S'))
        cat('\nMCMC + diagnostics time',
            printtimediff(difftime(Sys.time(), MCtimestart, units = 'auto')),
            '\n')

        maxusedcomponents <- max(maxusedcomponents, usedcomponents)
        maxiterations <- max(maxiterations, nitertot)
    }
#### END LOOP OVER CHAINS (WITHIN ONE CORE)

    ##
    cat('\nCurrent time:',
        format(Sys.time(), '%Y-%m-%d %H:%M:%S'))

    cat('\nTotal time',
        printtimediff(difftime(Sys.time(), headertimestart, units = 'auto')),
        '\n')

#### Inform user that this core has finished
    ## ## timing text, for white spaces:
    ## ## "Sampling. Core 2 estimated end time: xxxx-xx-xx xx:xx   "
    print2user(paste0('\rCore ', acore,
        ' finished.                                        \nSampling. Estimating remaining time...'))

    ## output information from a core,
    ## passed to the originally calling process
    cbind(
        maxusedcomponents = maxusedcomponents,
        maxiterations = maxiterations,
        nonfinitechains = nonfinitechains,
        stoppedchains = stoppedchains,
        headertime = headertime,
        MCtime = difftime(Sys.time(), MCtimestart, units = 'secs'),
        usedmem = max(usedmem, sum(gc()[,6]))
    )
}

#' Eliminate samples from mcsamples object
#'
#' Used in 'learn()'.
#'
#' @keywords internal
mcsubset <- function(learnt, subsamples) {
    lapply(learnt, function(xx) {
        do.call('[', c(
            list(xx),
            rep(TRUE, length(dim(xx)) - 1),
            list(subsamples),
            list(drop = FALSE))
        )
    })
}


#' Concatenate mcsample objects
#'
#' Used in 'learn()'.
#'
#' @keywords internal
mcjoin <- function(x, y){
    if(is.null(x)){
        y
    } else {
        mapply(
            function(xx, yy) {
                temp <- c(xx, yy)
                dx <- dim(yy)[-length(dim(yy))]
                dim(temp) <- c(dx, length(temp) / prod(dx))
                temp
            },
            x, y,
            SIMPLIFY = FALSE
        )
    }
}


#' Bind 3D arrays by first dimension
#'
#' Used in 'util_checkpoints()' within 'learn()', and in various functions in 'util_lprobs.R'.
#'
#' NB: the following variant is slower:
#'
#' ```
#' function(x, y) {
#'     out <- c(aperm(x), aperm(y))
#'     dim(out) <- c(rev(dim(x)[-1]), dim(x)[1] + dim(y)[1])
#'     aperm(out)
#' }
#' ```
#'
#' @keywords internal
learnbind <- function(x, y) {
    if(is.null(x)) {
        y
    } else {
        nrx <- dim(x)[1]
        nry <- dim(y)[1]
        out <- array(data = NA, dim = c(nrx + nry, dim(x)[-1]),
            dimnames = NULL)
        out[seq_len(nrx), ,] <- x
        out[nrx + seq_len(nry), ,] <- y
        out
    }
}

#' Transforms variates to different representations
#'
#' See <https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf>.
#'
#' @param x data.table object containing data to be transformed
#' @param auxmetadata auxmetadata object
#' @param Rout Character, output of R-type variate, with possible values:
#'  'normalized': for internal MCMC use
#'  'mi': for use in mutualinfo()
#'  'original': original representation
#' @param Cout Character, output of C-type variate, with possible values:
#'  'init': for internal MCMC use (init input)
#'  'left', 'right': for internal MCMC use
#'  'aux', 'lat': for internal MCMC use
#'  'boundnormalized': for sampling functions
#'  'boundisinf': for sampling functions
#'  'mi': for use in mutualinfo()
#'  'original': original representation
#' @param Dout Character, output of D-type variate, with possible values:
#'  'init': for internal MCMC use (init input)
#'  'left', 'right': for internal MCMC use
#'  'aux': for internal MCMC use
#'  'boundisinf': for sampling functions
#'  'normalized': for sampling functions
#'  'mi': for use in mutualinfo()
#'  'original': original representation
#' @param Oout Character, output of O-type variate, with possible values:
#'  'numeric': for internal MCMC use, values 1,2,...
#'  'original': original representation
#' @param Nout Character, output of N-type variate, with possible values:
#'  'numeric': for internal MCMC use, values 1,2,...
#'  'original': original representation
#' @param Bout Character, output of B-type variate, with possible values:
#'  'numeric': for internal MCMC use, values 0,1
#'  'original': original representation
#' @param variates Character vector, names of variates
#'   corresponding to columns of x (in case x misses column names)
#' @param logjacobianOr Logical or `NULL`: output is the log-Jacobian in orginal or transformed domain? `NULL` (default) means do not calculate the log-Jacobians
#'
#' @return A [data frame][base::data.frame()] of transformed variates, or their log-Jacobians.
#'
#' @keywords internal
vtransform <- function(
    x,
    auxmetadata,
    Rout = NULL,
    Cout = NULL,
    Dout = NULL,
    Bout = NULL,
    Oout = NULL,
    Nout = NULL,
    variates = NULL,
    logjacobianOr = NULL
) {
    useLquantiles <- TRUE

    if(!is.data.frame(x)){
        if(is.vector(x)){
            dim(x) <- c(length(x), 1)
        }
        x <- as.data.frame(x)
    }
    if (!is.null(variates)) {
        colnames(x) <- variates
    }
    as.data.frame(lapply(colnames(x), function(v) {
        datum <- x[[v]]
        with(data = as.list(auxmetadata[auxmetadata$name == v, ]),
        {

            if (is.null(logjacobianOr)) {
#### Transformation to internal value for MCMC

#### Continuous, open domain
                if (mcmctype == 'R') {

                    if (Rout == 'normalized') {
                        ## used for MCMC
                        if (transform == 'log') {
                            datum <- log(datum - domainmin)
                        } else if (transform == 'logminus') {
                            datum <- -log(domainmax - datum)
                        } else if (transform == 'Q') {
                            datum <- util_Q(0.5 +
                                                (datum - (domainmin + domainmax)/2) /
                                                (domainmax - domainmin)
                            )
                        }
                        datum <- (datum - tlocation) / tscale

                    } else if (Rout == 'original') {
                        ## transformation from MCMC representation
                        ## to original domain
                        datum <- datum * tscale + tlocation
                        if (transform == 'log') {
                            datum <- exp(datum) + domainmin
                        } else if (transform == 'logminus') {
                            datum <- domainmax - exp(-datum)
                        } else if (transform == 'Q') {
                            datum <- (util_invQ(datum) - 0.5) * (domainmax - domainmin) +
                                (domainmin + domainmax)/2
                        }

                    } else if (Rout != 'mi'){
                        ## used in mutualinfo()
                        stop('Unknown transformation for variate ', v)
                    }

#### Continuous, closed domain
                } else if (mcmctype == 'C') {

                    if (Cout == 'init') {
                        ## init value used for MCMC
                        ## output:
                        ## for datapoints with known value within domain
                        ## it must be set to NA (will be ignored by Nimble)
                        ## because it's already given as data;
                        ## for datapoints with unknown value
                        ## we set it to 0;
                        ## for datapoints at boundaries
                        ## we set it slightly outside the boundaries.
                        selna <- is.na(datum)
                        selmin <- !is.na(datum) & (datum <= domainmin)
                        selmax <- !is.na(datum) & (datum >= domainmax)
                        selmid <- !is.na(datum) &
                            (datum > domainmin) & (datum < domainmax)
                        datum[selna] <- 0L
                        datum[selmin] <- tdomainmin - 0.125
                        datum[selmax] <- tdomainmax + 0.125
                        datum[selmid] <- NA

                    } else if (Cout == 'left') {
                        ## used for MCMC
                        sel <- is.na(datum) | (datum < domainmax)
                        datum[sel] <- -Inf
                        datum[!sel] <- tdomainmax

                    } else if (Cout == 'right') {
                        ## used for MCMC
                        sel <- is.na(datum) | (datum > domainmin)
                        datum[sel] <- +Inf
                        datum[!sel] <- tdomainmin

                    } else if (Cout == 'lat') {
                        ## latent variable used for MCMC
                        ## Boundary values are not fixed, free to roam
                        datum[(datum >= domainmax) |
                                  (datum <= domainmin)] <- NA

                        ## Non-boundary values are fixed data
                        if (transform == 'log') {
                            datum <- log(datum - domainmin)
                        } else if (transform == 'logminus') {
                            datum <- -log(domainmax - datum)
                        } else if (transform == 'Q') {
                            datum <- util_Q(0.5 +
                                                (datum -
                                                     (domainmin + domainmax)/2) /
                                                (domainmax - domainmin))
                        }
                        ##
                        datum <- (datum - tlocation) / tscale

                    } else if (Cout == 'aux') {
                        ## aux, indicator variable used for MCMC
                        sel <- is.na(datum)
                        datum[sel] <- NA
                        datum[!sel] <- 1L

                    } else if (Cout == 'boundisna') {
                        ## used in sampling functions
                        ## points outside or on boundaries are moved to infinity
                        datum[datum <= domainmin] <- NA
                        datum[datum >= domainmax] <- NA
                        ##
                        if (transform == 'log') {
                            datum <- log(datum - domainmin)
                        } else if (transform == 'logminus') {
                            datum <- -log(domainmax - datum)
                        } else if (transform == 'Q') {
                            datum <- util_Q(0.5 +
                                                (datum -
                                                     (domainmin + domainmax)/2) /
                                                (domainmax - domainmin))
                        }
                        ##
                        datum <- (datum - tlocation) / tscale

                    } else if (Cout == 'leftbound') {
                        ## used in Pr()
                        ## non-boundary points are st to NA
                        ## boundaries are enforced
                        datum[datum > domainmin] <- NA
                        datum[datum < domainmin] <- domainmin
                        ##
                        if (transform == 'log') {
                            datum <- log(datum - domainmin)
                        } else if (transform == 'logminus') {
                            datum <- -log(domainmax - datum)
                        } else if (transform == 'Q') {
                            datum <- util_Q(0.5 +
                                                (datum -
                                                     (domainmin + domainmax)/2) /
                                                (domainmax - domainmin))
                        }
                        ##
                        datum <- (datum - tlocation) / tscale

                    } else if (Cout == 'rightbound') {
                        ## used in Pr()
                        ## non-boundary points are st to NA
                        ## boundaries are enforced
                        datum[datum < domainmax] <- NA
                        datum[datum > domainmax] <- domainmax
                        ##
                        if (transform == 'log') {
                            datum <- log(datum - domainmin)
                        } else if (transform == 'logminus') {
                            datum <- -log(domainmax - datum)
                        } else if (transform == 'Q') {
                            datum <- util_Q(0.5 +
                                                (datum -
                                                     (domainmin + domainmax)/2) /
                                                (domainmax - domainmin))
                        }
                        ##
                        datum <- (datum - tlocation) / tscale

                    } else if (Cout == 'boundisinf') {
                        ## used in sampling functions
                        ## points outside or on boundaries are moved to infinity
                        selmax <- (datum >= domainmax)
                        selmin <- (datum <= domainmin)
                        ##
                        if (transform == 'log') {
                            datum <- log(datum - domainmin)
                        } else if (transform == 'logminus') {
                            datum <- -log(domainmax - datum)
                        } else if (transform == 'Q') {
                            datum <- util_Q(0.5 +
                                                (datum -
                                                     (domainmin + domainmax)/2) /
                                                (domainmax - domainmin))
                        }
                        ##
                        datum <- (datum - tlocation) / tscale
                        datum[selmax] <- +Inf
                        datum[selmin] <- -Inf

                    } else if (Cout == 'boundnormalized') {
                        ## used in Pr()
                        ## points outside boundaries are moved to boundaries
                        datum[datum < domainmin] <- domainmin
                        datum[datum > domainmax] <- domainmax
                        ##
                        if (transform == 'log') {
                            datum <- log(datum - domainmin)
                        } else if (transform == 'logminus') {
                            datum <- -log(domainmax - datum)
                        } else if (transform == 'Q') {
                            datum <- util_Q(0.5 +
                                                (datum -
                                                     (domainmin + domainmax)/2) /
                                                (domainmax - domainmin))
                        }
                        ##
                        datum <- (datum - tlocation) / tscale

                    } else if (Cout == 'miboundisna') {
                        ## used in mutualinfo
                        datum[datum >= tdomainmax] <- NA
                        datum[datum <= tdomainmin] <- NA

                    } else if (Cout == 'mileftbound') {
                        ## used in mutualinfo
                        datum[datum > tdomainmin] <- NA
                        datum[datum < tdomainmin] <- tdomainmin

                    } else if (Cout == 'mirightbound') {
                        ## used in mutualinfo
                        datum[datum < tdomainmax] <- NA
                        datum[datum > tdomainmax] <- tdomainmax

                    } else if (Cout == 'original') {
                        ## transformation from MCMC representation
                        ## to original domain
                        datum <- datum * tscale + tlocation
                        if (transform == 'log') {
                            datum <- exp(datum) + domainmin
                        } else if (transform == 'logminus') {
                            datum <- domainmax - exp(-datum)
                        } else if (transform == 'Q') {
                            datum <- (util_invQ(datum) - 0.5) *
                                (domainmax - domainmin) +
                                (domainmin + domainmax) / 2
                        }
                        datum[datum < domainmin] <- domainmin
                        datum[datum > domainmax] <- domainmax

                    } else {
                        stop('Unknown transformation for variate ', v)
                    }

#### Continuous rounded
                } else if (mcmctype == 'D') {
                    ## NOTA BENE:
                    ## except for the 'original' (inverse-transformation) case,
                    ## we expect D-data to be given with correctly rounded values

                    if (Dout == 'init') {
                        ## init value used for MCMC
                        ## output:
                        ## for datapoints with known value within domain
                        ## we set it to their transformed value;
                        ## for datapoints with unknown value
                        ## we set it to 0;
                        ## for datapoints at boundaries
                        ## we set it slightly outside the boundaries.
                        selna <- is.na(datum)
                        datum <- (datum - tlocation) / tscale
                        datum[selna] <- 0L

                    } else if (Dout == 'left') {
                        ## used for MCMC
                        datum <- datum - halfstep
                        selmin <- is.na(datum) | (datum <= domainmin)
                        selmax <- (datum >= domainmaxminushs)
                        datum <- (datum - tlocation) / tscale
                        datum[selmin] <- -Inf
                        datum[selmax] <- tdomainmaxminushs

                    } else if (Dout == 'right') {
                        ## used for MCMC
                        datum <- datum + halfstep
                        selmin <- (datum <= domainminplushs)
                        selmax <- is.na(datum) | (datum >= domainmax)
                        datum <- (datum - tlocation) / tscale
                        datum[selmin] <- tdomainminplushs
                        datum[selmax] <- +Inf

                    } else if (Dout == 'aux') {
                        ## aux variable used for MCMC
                        sel <- is.na(datum)
                        datum[sel] <- NA
                        datum[!sel] <- 1L

                        ## ## not used anymore
                        ## } else if (Dout == 'boundisinf') {
                        ##     ## used in sampling functions
                        ##     selmin <- (datum - halfstep <= domainmin)
                        ##     selmax <- (datum + halfstep >= domainmax)
                        ##     ##
                        ##     datum <- (datum - tlocation) / tscale
                        ##     datum[selmin] <- -Inf
                        ##     datum[selmax] <- +Inf

                    } else if (Dout == 'boundnormalized') {
                        ## used in sampling functions
                        datum[datum < domainmin] <- domainmin
                        datum[datum > domainmax] <- domainmax
                        ##
                        datum <- (datum - tlocation) / tscale

                    } else if (Dout == 'boundisna') {
                        ## used in Pr()
                        datum[datum <= domainminplushs] <- NA
                        datum[datum >= domainmaxminushs] <- NA
                        datum <- (datum - tlocation) / tscale

                    } else if (Dout == 'leftbound') {
                        ## used in Pr()
                        datum[datum > domainminplushs] <- NA
                        datum[datum < domainmin] <- domainmin
                        datum <- (datum - tlocation) / tscale

                    } else if (Dout == 'rightbound') {
                        ## used in Pr()
                        datum[datum < domainmaxminushs] <- NA
                        datum[datum > domainmax] <- domainmax
                        datum <- (datum - tlocation) / tscale

                    } else if (Dout == 'miboundisna') {
                        ## used in mutualinfo
                        datum[datum <= tdomainminplushs] <- NA
                        datum[datum >= tdomainmaxminushs] <- NA
                        datum <- round((datum * tscale) / (2 * halfstep)) *
                            2 * halfstep / tscale

                    } else if (Dout == 'mileftbound') {
                        ## used in mutualinfo
                        datum[datum > tdomainminplushs] <- NA
                        datum[datum < tdomainmin] <- tdomainmin
                        datum <- round((datum * tscale) / (2 * halfstep)) *
                            2 * halfstep / tscale

                    } else if (Dout == 'mirightbound') {
                        ## used in mutualinfo
                        datum[datum < tdomainmaxminushs] <- NA
                        datum[datum > tdomainmax] <- tdomainmax
                        datum <- round((datum * tscale) / (2 * halfstep)) *
                            2 * halfstep / tscale

                    } else if (Dout == 'original') {
                        ## transformation from MCMC representation
                        ## to original domain
                        datum <- round((datum * tscale) / (2 * halfstep)) *
                            2 * halfstep + tlocation
                        datum[datum < domainmin] <- domainmin
                        datum[datum > domainmax] <- domainmax
                        ## datum[datum <= domainmin] <- domainmin
                        ## datum[datum >= domainmax] <- domainmax

                    } else {
                        stop('Unknown transformation for variate ', v)
                    }

                    ## Ordinal
                } else if (mcmctype == 'O') {

                    if (Oout == 'numeric') {
                        bvalues <- seq_len(Nvalues)
                        names(bvalues) <- sapply(bvalues, function(x)get(paste0('V', x)))
                        ## names(bvalues) <- unlist(xinfo[paste0('V', bvalues)])
                        datum <- bvalues[as.character(datum)]

                    } else if (Oout == 'index') {
                        bvalues <- seq_len(Nvalues)
                        names(bvalues) <- sapply(bvalues, function(x)get(paste0('V', x)))
                        ## names(bvalues) <- unlist(xinfo[paste0('V', bvalues)])
                        datum <- bvalues[as.character(datum)] + indexpos

                    } else if (Oout == 'original') {
                        bvalues <- seq_len(Nvalues)
                        names(bvalues) <- sapply(bvalues, function(x)get(paste0('V', x)))
                        ## names(bvalues) <- unlist(xinfo[paste0('V', bvalues)])
                        datum <- names(bvalues[datum])

                    } else if (Oout != 'mi') {
                        stop('Unknown transformation for variate ', v)
                    }

                    ## Nominal
                } else if (mcmctype == 'N') {

                    if (Nout == 'numeric') {
                        bvalues <- seq_len(Nvalues)
                        names(bvalues) <- sapply(bvalues, function(x)get(paste0('V', x)))
                        ## names(bvalues) <- unlist(xinfo[paste0('V', bvalues)])
                        if(!all(as.character(datum[!is.na(datum)]) %in% names(bvalues))) {
                            stop('Unknown values for variate ', v)
                        }
                        datum <- bvalues[as.character(datum)]

                    } else if (Nout == 'index') {
                        bvalues <- seq_len(Nvalues)
                        names(bvalues) <- sapply(bvalues, function(x)get(paste0('V', x)))
                        ## names(bvalues) <- unlist(xinfo[paste0('V', bvalues)])
                        datum <- bvalues[as.character(datum)] + indexpos

                    } else if (Nout == 'original') {
                        bvalues <- seq_len(Nvalues)
                        names(bvalues) <- sapply(bvalues, function(x)get(paste0('V', x)))
                        ## names(bvalues) <- unlist(xinfo[paste0('V', bvalues)])
                        datum <- names(bvalues[datum])

                    } else if (Nout != 'mi') {
                        stop('Unknown transformation for variate ', v)
                    }

                    ## Binary
                } else if (mcmctype == 'B') {
                    ## V1 corresponds to 0
                    ## V2 corresponds to 1
                    ## in the MC sampler, Bprob is the probability of V2
                    if (Bout == 'numeric') {
                        bvalues <- 0:1
                        names(bvalues) <- sapply(bvalues + 1L,
                            function(x)get(paste0('V', x)))
                        if(!all(as.character(datum[!is.na(datum)]) %in% names(bvalues))) {
                            stop('Unknown values for variate ', v)
                        }
                        datum <- bvalues[as.character(datum)]

                    } else if (Bout == 'original') {
                        bvalues <- 0:1
                        names(bvalues) <- sapply(bvalues + 1L,
                            function(x)get(paste0('V', x)))
                        datum <- names(bvalues[datum + 1L])

                    } else if (Bout != 'mi') {
                        stop('Unknown transformation for variate ', v)
                    }
                }

#### Calculation of reciprocal log-Jacobian factors
            } else if (logjacobianOr) {
                ## 'datum' is given in the original domain

                if (mcmctype %in% c('R', 'C')) {
                    if (mcmctype %in% c('C')) {
                        selma <- datum <= domainminplushs | datum >= domainmaxminushs
                    }
                    if (transform == 'log') {
                        datum <- -log(datum - domainmin) - log(tscale)
                    } else if (transform == 'logminus') {
                        datum <- -log(domainmax - datum) - log(tscale)
                    } else if (transform == 'Q') {
                        datum <- util_Q(0.5 +
                                            (datum - (domainmin + domainmax)/2) /
                                            (domainmax - domainmin)
                        )
                        datum <- -log(util_invDQ(datum)) -
                            log(tscale) - log(domainmax - domainmin)
                    } else if (transform == 'identity') {
                        datum[!is.na(datum)] <- -log(tscale)
                    } else {
                        stop('Unknown transformation for variate ', v)
                    }
                    if (mcmctype %in% c('C')) {
                        datum[selma] <- 0
                    }
                    datum[is.na(datum)] <- 0
                } else {
                    datum <- numeric(length(datum))
                }

            } else if (!logjacobianOr) {

                ## 'datum' is given in the transformed domain

                if (mcmctype %in% c('R', 'C')) {
                    if (mcmctype %in% c('C')) {
                        selma <- datum <= tdomainminplushs | datum >= tdomainmaxminushs
                    }
                    if (transform == 'log') {
                        datum <- tscale * datum + tlocation - log(tscale)
                    } else if (transform == 'logminus') {
                        datum <- -tscale * datum - tlocation - log(tscale)
                    } else if (transform == 'Q') {
                        datum <- -log(util_invDQ(tscale * datum + tlocation)) -
                            log(tscale) - log(domainmax - domainmin)
                    } else if (transform == 'identity') {
                        datum[!is.na(datum)] <- -log(tscale)
                    } else {
                        stop('Unknown transformation for variate ', v)
                    }
                    if (mcmctype %in% c('C')) {
                        datum[selma] <- 0
                    }
                    datum[is.na(datum)] <- 0
                } else {
                    datum <- numeric(length(datum))
                }

            } else {
                stop('Unknown "logjacobianOr" value')
            }

            datum
        })
    }), row.names = NULL,  col.names = colnames(x))
}

#' Plot one-dimensional posterior probabilities
#'
#' Used in 'learn()' to plot diagnostics.
#'
#' @param filename Character: name of plot output file
#' @param learnt Either a character with the name of a directory or full
#'   path for an 'learnt.rds' object, or such an object itself
#' @param data data.table object or filepath: datapoints
#' @param plotprobability Logical: plot the resulting probability curve
#' @param plotvariability Character, either 'samples' or 'quantiles':
#'   how to plot the variability of the probability distribution with new samples
#' @param nFsamples Positive number: if plotvariability='samples', then
#'   number of samples of representative frequency distributions to display
#'   as variability;
#'   if plotvariability='quantiles', then
#'   the quantiles (in range 0 to 0.5) to show
#' @param datahistogram Logical: plot the data as histogram?
#' @param datascatter Logical: plot the data as scatterplot along the x-axis?
#' @param parallel Logical or numeric: whether to use pre-existing parallel
#'   workers, or how many to create and use
#'
#' @return No return value.
#'
#' @import grDevices
#' @import graphics
#' @import utils
#'
#' @keywords internal
plotFsamples <- function(
    filename,
    learnt,
    data,
    plotprobability = TRUE,
    plotvariability = 'samples',
    nFsamples = NULL,
    datahistogram = !(missing(data) || is.null(data)),
    datascatter = !(missing(data) || is.null(data)),
    parallel = TRUE
) {
    fontfamily <- 'Palatino'
    ## Tol colour-blind-friendly palette
    black <- 'black'
    red <- '#EE6677'
    blue <- '#4477AA'
    green <- '#228833'
    yellow <- '#CCBB44'
    purple <- '#AA3377'
    cyan <- '#66CCEE'
    grey <- '#BBBBBB'
    midgrey <- '#888888'

    ## Extract Monte Carlo output & auxmetadata
    ## If 'learnt' is a string, check if it's a folder name or file name
    if (is.character(learnt)) {
        ## Check if 'learnt' is a folder containing learnt.rds
        if (file_test('-d', learnt) &&
                file.exists(file.path(learnt, 'learnt.rds'))) {
            learnt <- readRDS(file.path(learnt, 'learnt.rds'))
        } else {
            ## Assume 'learnt' the full path of learnt.rds
            ## possibly without the file extension '.rds'
            learnt <- paste0(sub('.rds$', '', learnt), '.rds')
            if (file.exists(learnt)) {
                learnt <- readRDS(learnt)
            } else {
                stop('The argument "learnt" must be a folder containing learnt.rds, or the path to an rds-file containing the output from "learn".')
            }
        }
    }

    auxmetadata <- learnt$auxmetadata
    ## learnt$auxmetadata <- NULL
    nsamples <- ncol(learnt$W)

    nodata <- missing(data) || is.null(data) || isFALSE(data)
    if (datahistogram && nodata) {
        datahistogram <- FALSE
        message('NOTE while plotting: "datahistogram" is TRUE but there is no data')
    }
    if (datascatter && nodata) {
        datascatter <- FALSE
        message('NOTE while plotting: "datascatter" is TRUE but there is no data')
    }

    addylab <- ''
    if (plotvariability == 'quantiles') {
        if(is.null(nFsamples)) {nFsamples <- c(5.5, 94.5) / 100}
        ## plotprobability <- TRUE
        if (any(nFsamples <= 0 | nFsamples >= 1)) {
            nFsamples <- c(5.5, 94.5) / 100
        }
        quants <- sort(unique(round(c(nFsamples, 1 - nFsamples), 6)))
        nmcsamples <- NULL
        addylab <- paste0(' & ', ceiling(diff(quants) * 100), '% variability')
    } else {
        if(is.null(nFsamples)) {nFsamples <- 100}
        if (nFsamples == 'all') {
            nFsamples <- nsamples
        }
        quants <- NULL
        nmcsamples <- abs(nFsamples)
    }

    addplot <- FALSE

    graphics.off()
    ## make sure graphics device is closed also in case of error
    on.exit(dev.off())

    pdf(file = paste0(filename, '.pdf'), family = fontfamily,
        height = 8.27, width = 11.69)
    ## Save user's par()
    ## oldpar <- par(mfrow = c(1, 1))
    ## oldpar$new <- NULL
    ## on.exit(par(oldpar))
    ## on.exit(dev.off(), add = TRUE)

    for (name in auxmetadata[['name']]) {
        ## Check if we have proper data for this variate
        if (datahistogram || datascatter) {
            theresdata <- (sum(!is.na(data[[name]])) > 1)
        }

        ## save the possibly existing V-values for this variate
        datavalues <- as.list(auxmetadata[auxmetadata$name == name,
            grep('^V[0-9]+$', names(auxmetadata))
        ])
        datavalues <- unlist(datavalues[!is.na(datavalues)])
        nvaluelist <- length(datavalues)

        with(as.list(auxmetadata[auxmetadata$name == name, ]),
        {

            if(mcmctype == 'R') {
                Xgrid <- cbind(seq(plotmin, plotmax, length.out = 256))
                colnames(Xgrid) <- name
                probabilities <- Pr(Y = Xgrid, X = NULL,
                    learnt = learnt,
                    quantiles = quants,
                    nsamples = nmcsamples,
                    parallel = parallel,
                    verbose = FALSE,
                    keepYX = FALSE)

                dim(probabilities$values) <- NULL

                ## Find appropriate plot height across plots
                if (plotvariability == 'samples') {
                    dim(probabilities$samples) <- dim(probabilities$samples)[-2]
                    ymax <- quantile(x = apply(probabilities$samples, 2,
                        function(x) {
                            quantile(x = c(x), probs = 31 / 32, type = 6,
                                na.rm = TRUE, names = FALSE)
                        }
                    ), probs = 31 / 32, type = 6, na.rm = TRUE, names = FALSE)
                } else {
                    dim(probabilities$quantiles) <- dim(probabilities$quantiles)[-2]
                    ymax <- max(probabilities$quantiles[
                        is.finite(probabilities$quantiles)])
                    ##     apply(probabilities[, , drop = FALSE], 1,
                    ##     function(x) {
                    ##         quantile(x = x, max(quants), type = 6, na.rm = TRUE)
                    ##     }
                    ## )
                    ## ymax <- max(ymax[is.finite(ymax)])
                }

                ## prepare info for histogram data plots if required
                if (datahistogram && theresdata) {
                    datum <- data[[name]]
                    datum <- datum[!is.na(datum)]
                    ## try to guess optimal bin size
                    nh <- max(32, round(length(datum) / 64))

                    ## histo <- thist(datum, n = nh, extendbreaks = FALSE)
                    histo <- hist(datum, breaks = nh, plot = FALSE)
                    hmax <- max(histo$density)

                    if(ymax / hmax < 0.3 | ymax / hmax > 3) {
                        histo$density <- histo$density / max(histo$density) * ymax
                    } else {
                        ymax <- max(ymax, histo$density)
                    }
                }

                ## If required, plot frequency samples
                if (plotvariability == 'samples') {
                    ## tplot(
                    ##     x = Xgrid, y = probabilities$samples,
                    ##     xlim = range(Xgrid), ylim = c(0, ymax),
                    ##     type = 'l', lty = 1, lwd = 2,
                    ##     col = , alpha = 1/8,
                    ##     xlab = name,
                    ##     ylab = paste0('probability density', addylab),
                    ##     family = fontfamily
                    ## )
                    flexiplot(
                        x = Xgrid, y = probabilities$samples,
                        xlim = range(Xgrid), ylim = c(0, ymax),
                        type = 'l', lty = 1, lwd = 2,
                        col = adjustcolor(cyan, 1 / 8),
                        xlab = name,
                        ylab = paste0('probability density', addylab),
                        family = fontfamily
                    )
                    addplot <- TRUE # new plots must keep this one
                } else if (plotvariability == 'quantiles') {
                    ## marguncertainty <- t(apply(probabilities, 1, function(x) {
                    ##     quantile(x = x, probs = quants, type = 6, na.rm = TRUE)
                    ## }))

                    plotquantiles(
                        x = Xgrid,
                        y = probabilities$quantiles,
                        ## y = marguncertainty[, , drop = FALSE],
                        col = cyan, alpha.f = 0.25,
                        xlim = range(Xgrid), ylim = c(0, ymax),
                        xlab = name,
                        ylab = paste0('probability density', addylab),
                        family = fontfamily,
                        add = addplot
                    )
                    addplot <- TRUE # new plots must keep this one
                }

                ## If required, plot probability
                if (plotprobability) {
                    ## tplot(
                    ##     x = Xgrid,
                    ##     y = probabilities$values,
                    ##     ## y = rowMeans(probabilities[, , drop = FALSE], na.rm = TRUE),
                    ##     xlim = range(Xgrid), ylim = c(0, ymax),
                    ##     type = 'l', cex = 0.5, lty = 1, lwd = 4,
                    ##     col = 1, alpha = 0.75,
                    ##     xlab = name,
                    ##     ylab = paste0('probability density', addylab),
                    ##     family = fontfamily,
                    ##     add = addplot
                    ## )
                    flexiplot(
                        x = Xgrid,
                        y = probabilities$values,
                        ## y = rowMeans(probabilities[, , drop = FALSE], na.rm = TRUE),
                        xlim = range(Xgrid), ylim = c(0, ymax),
                        type = 'l', cex = 0.5, lty = 1, lwd = 4,
                        col = adjustcolor(blue, 0.75),
                        xlab = name,
                        ylab = paste0('probability density', addylab),
                        family = fontfamily,
                        add = addplot
                    )
                }

                ## If required and possible, plot data histogram
                if (datahistogram && theresdata) {
                    ## tplot(
                    ##     x = histo$mids, y = histo$density,
                    ##     xlim = range(Xgrid), ylim = c(0, ymax),
                    ##     type = 'l', cex = 0.5, lty = 1, lwd = 2,
                    ##     col = 4, alpha = 0.5,
                    ##     border = '#555555', border.alpha = 1 / 4,
                    ##     xlab = name,
                    ##     ylab = paste0('probability density', addylab),
                    ##     family = fontfamily,
                    ##     add = addplot
                    ## )
                    flexiplot(
                        x = histo$mids, y = histo$density,
                        xlim = range(Xgrid), ylim = c(0, ymax),
                        type = 'l', cex = 0.5, lty = 1, lwd = 2,
                        col = adjustcolor(yellow, 0.5),
                        xlab = name,
                        ylab = paste0('probability density', addylab),
                        family = fontfamily,
                        add = addplot
                    )
                }
                ## End R case

            } else if(mcmctype == 'C') {
                ## C-type variates are peculiar because
                ## their boundary values have a finite probability,
                ## not a probability density.
                ## These values need different handling and representation

                Xgrid <- cbind(seq(plotmin, plotmax, length.out = 256))
                ## }
                colnames(Xgrid) <- name

                xin <- Xgrid > domainminplushs & Xgrid < domainmaxminushs

                probabilities <- Pr(Y = Xgrid, X = NULL,
                    learnt = learnt,
                    quantiles = quants,
                    nsamples = nmcsamples,
                    parallel = parallel,
                    verbose = FALSE,
                    keepYX = FALSE)

                dim(probabilities$values) <- NULL

                ## Find appropriate plot height across plots
                if (plotvariability == 'samples') {
                    dim(probabilities$samples) <- dim(probabilities$samples)[-2]
                    ymax <- quantile(x = apply(
                        probabilities$samples[xin, , drop = FALSE], 2,
                        function(x) {
                            quantile(x = c(x), probs = 31 / 32, type = 6,
                                na.rm = TRUE, names = FALSE)
                        }
                    ), probs = 31 / 32, type = 6, na.rm = TRUE, names = FALSE)
                    ## ymax <- quantile(x = apply(
                    ##     probabilities[xin, subsamples, drop = FALSE],
                    ##     2, function(x) {
                    ##         quantile(x = x, probs = 31 / 32, type = 6, na.rm = TRUE)
                    ##     }
                    ## ), 31 / 32, type = 6, na.rm = TRUE)
                } else {
                    dim(probabilities$quantiles) <- dim(probabilities$quantiles)[-2]
                    temp <- probabilities$quantiles[xin, ]
                    ymax <- max(temp[is.finite(temp)])
                    ## ymax <- apply(probabilities[xin, , drop = FALSE], 1,
                    ##     function(x) {
                    ##         quantile(x, max(quants), type = 6, na.rm = TRUE)
                    ##     }
                    ## )
                    ## ymax <- max(ymax[is.finite(ymax)])
                }

                ## prepare info for histogram data plots if required
                if (datahistogram && theresdata) {
                    datum <- data[[name]]
                    datum <- datum[!is.na(datum)]

                    din <- datum > domainminplushs & datum < domainmaxminushs
                    ## try to guess optimal bin size
                    nh <- max(32, round(length(datum[din]) / 64))

                    ## histo <- thist(datum[din], n = nh, extendbreaks = FALSE)
                    histo <- hist(datum, breaks = nh, plot = FALSE)
                    hmax <- max(histo$density)

                    hleft <- sum(datum <= domainminplushs) / length(datum)
                    hright <- sum(datum >= domainmaxminushs) / length(datum)

                    if(ymax / hmax < 0.3 | ymax / hmax > 3) {
                        histo$density <- histo$density / max(histo$density) * ymax
                    } else {
                        ymax <- max(ymax, histo$density)
                    }
                }

                ## If required, plot frequency samples
                if (plotvariability == 'samples') {
                    if (any(xin)) {
                        ## tplot(
                        ##     x = Xgrid[xin],
                        ##     y = probabilities$samples[xin, , drop = FALSE],
                        ##     xlim = range(Xgrid), ylim = c(0, ymax),
                        ##     type = 'l', lty = 1, lwd = 2,
                        ##     col = 5, alpha = 1/8,
                        ##     xlab = name,
                        ##     ylab = paste0('probability density', addylab),
                        ##     family = fontfamily
                        ## )
                        flexiplot(
                            x = Xgrid[xin],
                            y = probabilities$samples[xin, , drop = FALSE],
                            xlim = range(Xgrid), ylim = c(0, ymax),
                            type = 'l', lty = 1, lwd = 2,
                            col = adjustcolor(cyan, 1 / 8),
                            xlab = name,
                            ylab = paste0('probability density', addylab),
                            family = fontfamily
                        )
                        addplot <- TRUE
                    }

                    ## Boundary points
                    if (any(!xin)) {
                        ## tplot(
                        ##     x = Xgrid[!xin],
                        ##     y = probabilities$values[!xin] * ymax,
                        ##     type = 'p', pch = 2, cex = 2,
                        ##     col = 5, alpha = 1/8,
                        ##     family = fontfamily,
                        ##     add = addplot
                        ## )
                        flexiplot(
                            x = Xgrid[!xin],
                            y = probabilities$values[!xin] * ymax,
                            type = 'p', pch = 2, cex = 2,
                            col = adjustcolor(cyan, 1 / 8),
                            family = fontfamily,
                            add = addplot
                        )
                        addplot <- TRUE
                    }

                } else if (plotvariability == 'quantiles') {
                    ## marguncertainty <- t(apply(probabilities, 1, function(x) {
                    ##     quantile(x, quants, type = 6, na.rm = TRUE)
                    ## }))

                    if (any(xin)) {
                        plotquantiles(
                            x = Xgrid[xin],
                            y = probabilities$quantiles[xin, , drop = FALSE],
                            col = cyan, alpha.f = 0.25,
                            xlim = range(Xgrid), ylim = c(0, ymax),
                            xlab = name,
                            ylab = paste0('probability density', addylab),
                            family = fontfamily,
                            add = addplot
                        )
                        addplot <- TRUE
                    }

                    ## Boundary points
                    if (any(!xin)) {
                        ## tplot(
                        ##     x = matrix(Xgrid[!xin],
                        ##         nrow = 2, ncol = sum(!xin), byrow = TRUE),
                        ##     y = t(probabilities$quantiles[!xin, , drop = FALSE]) * ymax,
                        ##     type = 'l', pch = 2, cex = 2,
                        ##     col = 5, alpha = 0.25,
                        ##     lty = 1, lwd = 16,
                        ##     add = addplot
                        ## )
                        flexiplot(
                            x = matrix(Xgrid[!xin],
                                nrow = 2, ncol = sum(!xin), byrow = TRUE),
                            y = probabilities$quantiles[!xin, , drop = FALSE] *
                                ymax,
                            type = 'l', pch = 2, cex = 2,
                            col = adjustcolor(cyan, 0.25),
                            lty = 1, lwd = 16,
                            add = addplot
                        )
                        addplot <- TRUE
                    }
                }

                ## If required, plot probability
                if (plotprobability) {
                    if (any(xin)) {
                        ## tplot(
                        ##     x = Xgrid[xin],
                        ##     y = probabilities$values[xin],
                        ##     xlim = range(Xgrid), ylim = c(0, ymax),
                        ##     type = 'l', cex = 0.5, lty = 1, lwd = 4,
                        ##     col = 1, alpha = 0.75,
                        ##     xlab = name,
                        ##     ylab = paste0('probability density', addylab),
                        ##     family = fontfamily,
                        ##     add = addplot
                        ## )
                        flexiplot(
                            x = Xgrid[xin],
                            y = probabilities$values[xin],
                            xlim = range(Xgrid), ylim = c(0, ymax),
                            type = 'l', cex = 0.5, lty = 1, lwd = 4,
                            col = adjustcolor(blue, 0.75),
                            xlab = name,
                            ylab = paste0('probability density', addylab),
                            family = fontfamily,
                            add = addplot
                        )
                        addplot <- TRUE
                    }

                    ## Boundary points
                    if (any(!xin)) {
                        ## tplot(
                        ##     x = Xgrid[!xin],
                        ##     y = probabilities$values[!xin] * ymax,
                        ##     type = 'p', pch = 2, cex = 2,
                        ##     col = 1, alpha = 0.75,
                        ##     lty = 1, lwd = 3,
                        ##     add = addplot
                        ## )
                        flexiplot(
                            x = Xgrid[!xin],
                            y = probabilities$values[!xin] * ymax,
                            type = 'p', pch = 2, cex = 2,
                            col = adjustcolor(blue, 0.75),
                            lty = 1, lwd = 3,
                            add = addplot
                        )
                        addplot <- TRUE
                    }
                }

                ## If required and possible, plot data histogram
                if (datahistogram && theresdata) {
                    if(hleft + hright < 1) {
                        ## tplot(
                        ##     x = histo$mids, y = histo$density,
                        ##     xlim = range(Xgrid), ylim = c(0, ymax),
                        ##     type = 'l', col = 4, alpha = 0.5,
                        ##     border = '#555555', border.alpha = 1/4,
                        ##     xlab = name,
                        ##     ylab = paste0('probability density', addylab),
                        ##     family = fontfamily,
                        ##     add = addplot
                        ## )
                        flexiplot(
                            x = histo$mids, y = histo$density,
                            xlim = range(Xgrid), ylim = c(0, ymax),
                            type = 'l', cex = 0.5, lty = 1, lwd = 2,
                            col = adjustcolor(yellow, 0.5),
                            ## border = '#555555', border.alpha = 1/4,
                            xlab = name,
                            ylab = paste0('probability density', addylab),
                            family = fontfamily,
                            add = addplot
                        )
                        addplot <- TRUE
                    }

                    ## Boundary points
                    if (hleft + hright > 0) {
                        ## tplot(
                        ##     x = c(if(hleft > 0){domainminplushs},
                        ##         if(hright > 0){domainmaxminushs}),
                        ##     y = c(if(hleft > 0){hleft},
                        ##         if(hright > 0){hright}) * ymax,
                        ##     type = 'p', pch = 0, cex = 2,
                        ##     col = 4, alpha = 0.5,
                        ##     lty = 1, lwd = 5,
                        ##     family = fontfamily,
                        ##     add = addplot
                        ## )
                        flexiplot(
                            x = c(if(hleft > 0){domainminplushs},
                                if(hright > 0){domainmaxminushs}),
                            y = c(if(hleft > 0){hleft},
                                if(hright > 0){hright}) * ymax,
                            type = 'p', pch = 0, cex = 2,
                            col = adjustcolor(yellow, 0.5),
                            lty = 1, lwd = 5,
                            family = fontfamily,
                            add = addplot
                        )
                        addplot <- TRUE
                    }
                }
                ## End C case

            } else if (mcmctype %in% c('D', 'O', 'N', 'B')) {
                ## These variate types all have finite probabilities
                if(nvaluelist > 0) {
                    ## Xgrid <-  as.matrix(
                    ##     vtransform(x = datavalues,
                    ##         variates = name,
                    ##         auxmetadata = auxmetadata,
                    ##         Oout = 'numeric',
                    ##         Nout = 'numeric',
                    ##         Bout = 'numeric',
                    ##         logjacobianOr = NULL
                    ##     ))
                    Xgrid <- cbind(datavalues)
                    colnames(Xgrid) <- name
                    rownames(Xgrid) <- datavalues

                } else {
                    ## we must construct an X-grid
                    Xgrid <- cbind(seq(plotmin, plotmax, by = halfstep * 2))
                    colnames(Xgrid) <- name
                    rownames(Xgrid) <- NULL
                    Xticks <- NULL
                }

                probabilities <- Pr(Y = Xgrid, X = NULL,
                    learnt = learnt,
                    quantiles = quants,
                    nsamples = nmcsamples,
                    parallel = parallel,
                    verbose = FALSE,
                    keepYX = FALSE)

                dim(probabilities$values) <- NULL


                ## Find appropriate plot height across plots
                if (plotvariability == 'samples') {
                    dim(probabilities$samples) <- dim(probabilities$samples)[-2]
                    ymax <- quantile(x = apply(probabilities$samples, 2,
                        function(x) {
                            quantile(x = x, probs = 31 / 32, type = 6,
                                na.rm = TRUE, names = FALSE)
                        }
                    ), probs = 31 / 32, type = 6, na.rm = TRUE, names = FALSE)
                    ## ymax <- quantile(apply(
                    ##     probabilities[, subsamples, drop = FALSE],
                    ##     2, function(x) {
                    ##         quantile(x, 31 / 32, type = 6, na.rm = TRUE)
                    ##     }
                    ## ), 31 / 32, type = 6, na.rm = TRUE)
                } else {
                    dim(probabilities$quantiles) <- dim(probabilities$quantiles)[-2]
                    temp <- probabilities$quantiles
                    ymax <- max(temp[is.finite(temp)])
                    ## ymax <- apply(probabilities[, , drop = FALSE], 1,
                    ##     function(x) {
                    ##         quantile(x, max(quants), type = 6, na.rm = TRUE)
                    ##     }
                    ## )
                    ## ymax <- max(ymax[is.finite(ymax)])
                }

                ## prepare info for histogram data plots if required
                if (datahistogram && theresdata) {
                    datum <- data[[name]]
                    datum <- datum[!is.na(datum)]

                    if(nvaluelist > 0) {
                        histo <- as.vector(table(factor(datum,
                            levels = rownames(Xgrid)))) / length(datum)
                    } else {
                        ## histo <- thist(datum,
                        ##     n = c(Xgrid - halfstep, max(Xgrid) + halfstep),
                        ##     extendbreaks = FALSE)$counts / length(datum)
                        histo <- hist(datum,
                            breaks = c(Xgrid - halfstep, max(Xgrid) + halfstep),
                            plot = FALSE)$counts / length(datum)
                    }

                    ymax <- max(ymax, histo)
                }

                ## If required, plot frequency samples
                if (plotvariability == 'samples') {
                    ## tplot(
                    ##     x = Xgrid,
                    ##     y = probabilities$samples,
                    ##     xlim = range(Xgrid), ylim = c(0, ymax),
                    ##     xticks = Xticks, xlabels = rownames(Xgrid),
                    ##     type = 'l', lty = 1, lwd = 2,
                    ##     col = 5, alpha = 1/8,
                    ##     xlab = name,
                    ##     ylab = paste0('probability', addylab),
                    ##     family = fontfamily
                    ## )
                    flexiplot(
                        x = Xgrid,
                        y = probabilities$samples,
                        ## xlim = range(Xgrid),
                        ylim = c(0, ymax),
                        ## xticks = Xticks, xlabels = rownames(Xgrid),
                        type = 'o', lty = 1, lwd = 2,
                        col = adjustcolor(cyan, 1 / 8),
                        xlab = name,
                        ylab = paste0('probability', addylab),
                        family = fontfamily
                    )
                    addplot <- TRUE

                } else if (plotvariability == 'quantiles') {
                    ## marguncertainty <- t(apply(probabilities, 1, function(x) {
                    ##     quantile(x, quants, type = 6, na.rm = TRUE)
                    ## }))
                    plotquantiles(
                        x = Xgrid,
                        y = probabilities$quantiles,
                        col = cyan, alpha.f = 0.25,
                        ## xlim = range(Xgrid),
                        ylim = c(0, ymax),
                        ## xticks = Xticks, xlabels = rownames(Xgrid),
                        xlab = name,
                        ylab = paste0('probability', addylab),
                        family = fontfamily,
                        add = addplot
                    )
                    addplot <- TRUE
                }

                ## If required, plot probability
                if (plotprobability) {
                    ## tplot(
                    ##     x = Xgrid,
                    ##     y = probabilities$values,
                    ##     xlim = range(Xgrid), ylim = c(0, ymax),
                    ##     xticks = Xticks, xlabels = rownames(Xgrid),
                    ##     type = 'b', cex = 0.5, lty = 1, lwd = 4,
                    ##     col = 1, alpha = 0.75,
                    ##     xlab = name,
                    ##     ylab = paste0('probability', addylab),
                    ##     family = fontfamily,
                    ##     add = addplot
                    ## )
                    flexiplot(
                        x = Xgrid,
                        y = probabilities$values,
                        xlim = range(Xgrid), ylim = c(0, ymax),
                        ## xticks = Xticks, xlabels = rownames(Xgrid),
                        type = 'o', pch = 16, cex = 0.5, lty = 1, lwd = 4,
                        col = adjustcolor(blue, 0.75),
                        xlab = name,
                        ylab = paste0('probability', addylab),
                        family = fontfamily,
                        add = addplot
                    )
                    addplot <- TRUE
                }

                ## If required and possible, plot data histogram
                if (datahistogram && theresdata) {
                    ## tplot(
                    ##     x = Xgrid, y = histo,
                    ##     xlim = range(Xgrid), ylim = c(0, ymax),
                    ##     xticks = Xticks, xlabels = rownames(Xgrid),
                    ##     type = 'l', col = 4, alpha = 0.5,
                    ##     border = '#555555', border.alpha = 1/4,
                    ##     xlab = name,
                    ##     ylab = paste0('probability', addylab),
                    ##     family = fontfamily,
                    ##     add = addplot
                    ## )
                    flexiplot(
                        x = Xgrid, y = histo,
                        xlim = range(Xgrid), ylim = c(0, ymax),
                        ## xticks = Xticks, xlabels = rownames(Xgrid),
                        type = 'o', pch = 1, cex = 0.5, lty = 1, lwd = 2,
                        col = adjustcolor(yellow, 0.5),
                        ## border = '#555555', border.alpha = 1/4,
                        xlab = name,
                        ylab = paste0('probability', addylab),
                        family = fontfamily,
                        add = addplot
                    )
                    addplot <- TRUE
                }
                ## End D, O, N, B case
            } else {
                message('NOTE while plotting: unknown MC type for variate ', name)
            }

            ## Scatter plot of data if required
            if (datascatter && theresdata) {
                datum <- data[[name]]
                datum <- datum[!is.na(datum)]
                if (mcmctype %in% c('O', 'N', 'B')) {
                    datum <- as.matrix(
                        vtransform(x = datum,
                            variates = name,
                            auxmetadata = auxmetadata,
                            Oout = 'numeric',
                            Nout = 'numeric',
                            Bout = 'numeric',
                            logjacobianOr = NULL
                        ))
                }
                if(mcmctype == 'B'){datum <- datum + 1}
                rug(x = jitter(datum, amount = 0), side = 1,
                    col = adjustcolor(yellow,
                        alpha.f = exp((-length(datum) + 1) / 128)),
                    quiet = TRUE)
                ## scatteraxis(
                ##     side = 1, n = NA, alpha = 0.75, ext = 5,
                ##     x = datum + runif(length(datum),
                ##         min = -min(diff(sort(c(par('usr')[1:2],
                ##             unique(datum))))) / 1.5,
                ##         max = min(diff(sort(c(par('usr')[1:2],
                ##             unique(datum))))) / 1.5
                ##     ),
                ##     col = 4
                ## )
                ## ## These lines plot quartiles
                ## if (vtype %in% c('R', 'D', 'C', 'L')) {
                ## fiven <- fivenum(datum)
                ## abline(v = fiven, col = paste0(palette()[c(2, 4, 5, 4, 2)], '44'),
                ##     lwd = 5, lty = 2)
                ## }
            }
        }) # End with
    }
    invisible()
}

#' Metadata and helper function for metadata
#'
#' Metadata and helper function to create a template metadata file or object.
#'
#' The [learn()] function needs metadata about the variates present in the data. Such metadata can be provided either as a `csv` file or as a [base::data.frame()]. The function `buildmetadata` creates a template metadata csv-file, or outputs a metadata data.frame, by trying to *guess* metadata information from the dataset.The guesses may be very incorrect (as already said, metadata is information not contained in the data, so no algorithm can exist that extracts it from the data). **The user *must* modify and correct this template, using it as a starting point to prepare the correct metadata information.**
#'
#' @param data A dataset, given as a [data frame][base::data.frame()]
#' or as a file path to a csv file.
#' @param file Character or `NULL` (default): name of csv file where the metadata should be saved; if `NULL`: output metadata as `VALUE`.
#' @param includevrt Character or `NULL`: name of variates in dataset to be included.
#' @param excludevrt Character or `NULL`: name of variates in dataset to be excluded.
#' @param addsummary2metadata Logical: also output some diagnostic statistics
#'    in the metadata? Default `FALSE`.
#' @param backupfiles Logical: rename previous metadata file if it exists?
#' Default `TRUE`.
#' @param verbose Logical: output heuristics for each variate? Default `TRUE`.
#'
#' @return A preliminary [data frame][base::data.frame()] containing the metadata, [invisibly][base::invisible()] if `file = NULL`. If argument `file` is a character, a preliminary metadata file is also created with that name or path.
#'
#' @section Metadata information and format:
#'
#' In order to correctly learn from a dataset, the [learn()] function needs information that is not contained in the data themeselves; that is, it needs *meta*data. Metadata are provided either as a `csv` file or as a [base::data.frame()].
#'
#' A metadata file or data.frame must contain one row for each simple variate in the given inference problem, and the following fields (columns), even if some of them may be empty:
#'
#' `name`, `type`, `domainmin`, `domainmax`, `datastep`, `minincluded`, `maxincluded`, `V1`, `V2`, (possibly additional `V`-fields, sequentially numbered)
#'
#' The `type` field has three possible values: `nominal`, `ordinal`, `continuous`. The remaining fields that must be filled in depend on the `type` field. Here is a list of requirements:
#'
#' - **`nominal`** and **`ordinal`**: require *either* `V1`, `V2`, ... fields *or* `domainmin`, `domainmax`, `datastep` (all three) fields. No other fields are required.
#'
#' - **`continuous`**: requires `domainmin`, `domainmax`, `datastep`, `minincluded`, `maxincluded`.
#'
#' Here are the meanings and possible values of the fields:
#'
#' **`name`**: The name of the variate. This must be the same character string as it appears in the dataset (be careful about upper- and lower-case).
#'
#' **`type`**: The data type of variate `name`. Possible values are `nominal`, `ordinal`, `continuous`.
#'
#' - A *nominal* (also called *categorical*) variate has a discrete, finite number of possible values which have no intrinsic ordering. Examples could be a variate related to colour, with values "red", "green", "blue", and so on; or a variate related to cat breeds, with values "Siamese", "Abyssinian", "Persian", and so on. The possible values of the variate must be given in the fields `V1`, `V2`, and so on. It is important to include values that are possible but are *not* present in the dataset. A variate having only two possible values (binary variate), for example "yes" and "no", can be specified as nominal.
#'
#' - An *ordinal* variate has a discrete, finite number of possible values which do have an intrinsic ordering. Examples could be a Likert-scaled variate for the results of a survey, with values "very dissatisfied", "dissatisfied", "satisfied", "very satisfied"; or a variate related to the levels of some quantities, with values "low", "medium", "high"; or a variate having a numeric scale with values from 1 to 10. Whether a variate is nominal or ordinal often depends on the context. The possible values of the variate but be given in either one (but not both) or two ways: (1) in the fields `V1`, `V2`, ..., as for nominal variates; (2) as the fields `domainmin`, `domainmax`, `datastep`. Option (2) only works with numeric, equally spaced values: it assumes that the first value is `domainmin`, the second is `domainmin`+`datastep`, the third is `domainmin`+2*`datastep`, and so on up to the last value, `domainmax`.
#'
#' - A *continuous* variate has a continuum of values with an intrinsic ordering. Examples could be a variate related to the width of an object; or to the age of a person; or one coordinate of an object in a particular reference system. A continuous variate requires specification of the fields `domainmin`, `domainmax`, `datastep`, `minincluded`, `maxincluded`. Some naturally continuous variates are often rounded to a given precision; for instance, the age of a person might be reported as rounded to the nearest year (25 years, 26 years, and so on); or the length of an object might be reported to the nearest centimetre (1 m, 1.01 m, 1.02 m, and so on). The minimum distance between such rounded values **must** be reported in the `datastep` field; this would be `1` in the age example and `0.01` in the length example above. See below for further explanation of why reporting such rounding is important.
#'
#' **`domainmin`**: The minimum value that the variate (ordinal or continuous) can take on. Possible values are a real number or an empty value, which is then interpreted as `-Inf` (explicit values like `-Inf`, `-inf`, `-infinity` should also work). Some continuous variates, like age or distance or temperature, are naturally positive, and therefore have `domainmin` equal `0`. But in other contexts the minimum value could be different. For instance, if a given inference problem only involves people of age 18 or more, then `domainmin` would be set to `18`.
#'
#' **`domainmax`**: The maximum value that the variate (ordinal or continuous) can take on. Possible values are a real number, or an empty value, which is then interpreted as `+Inf` (explicit values like `Inf`, `inf`, `infinity` should also work). As with `domainmin`, the maximum value depends on the context. An age-related variate could theoretically have `domainmax` equal to infinity (empty value in the metadata file); but if a given study categorizes some people as "90 years old or older", then `domainmax` should be set to `90`.
#'
#' **`datastep`**: The minimum distance between the values of a variate (ordinal or continuous). Possible values are a positive real number or an empty value, which is then interpreted as 0 (the explicit value `0` is also accepted). For a numeric ordinal variate, `datastep` is the step between consecutive values. For a continuous *rounded* variate, `datastep` is the minimum distance between different values that occurs because of rounding; see the examples given above. The function `buildmetadata` has some heuristics to determine whether the variate is rounded or not. See further details under the section Rounding below.
#'
#' **`minincluded`**, **`maxincluded`**: Whether the minimum (`domainmin`) and maximum(`domainmax`) values of a *continuous* variate can really appear in the data or not. Possible values are `true` (or `t` or `yes`) or `false` (or `f`, `no`, or an empty field); upper- or lower-case is irrelevant. Here are some examples about the meaning of these fields. (a) A continuous *unrounded* variate such as temperature has 0 as a minimum possible value `domainmin`, but this value itself is physically impossible and can never appear in data; in this case `minincluded` is empty (or set to `false` or `no`). (b) A variate related to the *unrounded* length, in metres, of some objects may take on any positive real value; but suppose that all objects of length 5 or less are grouped together under the value `5`. It is then possible for several datapoints to have value `5`: one such datapoint could originally have the value 3.782341...; another the value 4.929673..., and so on. In this case `domainmin` is set to `5`, and `minincluded` is set to `true` (or `yes`). Similarly for the maximum value of a variate and `maxincluded`. Note that if `domainmin` is minus-infinity (empty value in the metadata file), then `minincluded` is automatically empty (that is, `false`), and similarly for `maxincluded` if `domainmax` is infinity.
#'
#' 
## #' @section Necessity of metadata:
## #'
## #' To be written.
#'
#' @seealso
#' [learn()], which generates the information necessary to calculate posterior probabilities, based on data and metadata.
#'
#' @examples
#' ## Create a preliminary data frame of metadata for the `penguins` dataset
#' metadata <- metadatatemplate(data = datasets::penguins, file = NULL)
#'
#' ## Note how the preliminary data frame includes additional spots
#' ## for values of nominal and ordinal variates
#' ## which could be missing from the data
#' print(metadata)
#'
#'
#' ## Create a preliminary data frame of metadata for the `penguins` dataset,
#' ## including only the 'species' and 'bill_len' variates:
#' metadata2 <- metadatatemplate(
#'   data = datasets::penguins, file = NULL,
#'   includevrt = c('species', 'bill_len')
#' )
#'
#' print(metadata2)
#'
#'
#' ## Create a preliminary data frame of metadata for the `penguins` dataset,
#' ## excluding the 'year' variate:
#' metadata3 <- metadatatemplate(
#'   data = datasets::penguins, file = NULL,
#'   excludevrt = 'year'
#' )
#'
#' print(metadata3)
#'
#' ## Generate 10 points for a continuous variate in (0, 1)
#' dataset <- runif(10)
#'
#' ## `metadatatemplate` correctly guesses the variate minimum,
#' ## but not the maximum (`NA` is equivalent to `+Inf`)
#' metadata <- metadatatemplate(data = dataset, file = NULL)
#' print(metadata)
#'
#' @aliases metadata metadatatemplate
#'
#' @import stats
#' @import utils
#'
#' @export
metadatatemplate <- function(
    data,
    file = NULL,
    includevrt = NULL,
    excludevrt = NULL,
    addsummary2metadata = FALSE,
    backupfiles = FALSE,
    verbose = TRUE
) {

    ## Greater Common Denominator function
    ## used to calculate multiplicity of values of continuous variates
    ## and guess whether they are rounded
    gcd <- function(...) {
        suppressWarnings(Reduce(function(a, b) {
            if (b == 0) a else Recall(b, a %% b)
        }, c(...)))
    }

    ## keywords to recognize ordinal variates
    ordinalkeywords <- c('low', 'high', 'very', 'medium', 'strong',
        'poor', 'good', 'always')

    ## collect warning messages to be given at the end
    warninglist <- NULL

    ## Read datafile if it exists
    datafile <- NULL
    if (is.character(data) && file.exists(data)) {
        datafile <- data
        data <- read.csv(datafile,
            na.strings = '', stringsAsFactors = FALSE, tryLogical = FALSE)
    }
    data <- as.data.frame(data)

    ## convert factors to strings if necessary
    if(any(sapply(data, is.factor))){
    if(verbose){ message('Converting factors to characters') }
        . <- sapply(data, is.factor)
        data[, .] <- lapply(data[, ., drop = FALSE], as.character)
    }

#### Prepare empty metadata frame
    metadata <- as.data.frame(
        c(list(name = NA,
            type = NA,
            domainmin = NA,
            domainmax = NA,
            datastep = NA,
            minincluded = NA,
            maxincluded = NA
            ## lowvalue = NA,
            ## centralvalue = NA,
            ## highvalue = NA,
            ## plotmin = NA,
            ## plotmax = NA
        ),
        setNames(as.list(rep(NA, 11)), paste0('V', 1:11)),
            if (addsummary2metadata) {
                list(datamin = NA,
                    datamax = NA,
                    datamaxrep = NA,
                    dataNvalues = NA)
            }
        )
    )[-1, ]

    ## build list of variates, including only those in 'includevrt'
    ## and excluding those in 'excludevrt'
    variatelist <- colnames(data)
    if(!is.null(includevrt)) {
        variatelist <- intersect(variatelist, includevrt)
    }
    if(!is.null(excludevrt)) {
        variatelist <- setdiff(variatelist, excludevrt)
    }

    if(verbose){message('Analyzing', length(variatelist), 'variates for',
        nrow(data), 'datapoints.') }

    ## Loop over variates (columns) in data frame
    for (name in variatelist) {
        if(verbose){ message('* "', name, '" variate:') }
        ## remove missing values
        x <- data[[name]]
        x <- x[!is.na(x)]

        ## ## Factors are now dealt with at the beginning
        ## ## check if it's a 'factor' object and transform
        ## if(is.factor(x)){
        ##     x <- as.character(x)
        ## }

        unx <- sort(unique(x))
        uniquex <- length(unx)

        ## if this variate has only one value, give a special warning
        if (uniquex < 2) {
            ## if(verbose){
            ##     message('  Only one value present.')
            ## }
            warninglist <- c(warninglist,
                paste0('\n* "', name, '" variate',
                    ' does not have at least two distinct values.',
                    ## '\nDiscarded because non-informative.',
                    '\nPlease insert its characteristics by hand in the metadata file.'))
            ## next
        }

        ## If the values are strings,
        ## check whether they can be reinterpreted as numeric
        if(!is.numeric(unx) &&
               !any(is.na(suppressWarnings(as.numeric(unx))))) {
            x <- as.numeric(x)
            unx <- sort(as.numeric(unx))
        }

        ## If the variate is numeric and has more than two values,
        ## calculate several characteristics used later
        if (is.numeric(unx) && uniquex > 2) {
            ## these two are used for diagnostic purposes
            datamin <- min(x)
            datamax <- max(x)
            ## values strictly within the interior of the domain
            ix <- x[!(x %in% range(x))]
            maxrep <- max(table(ix)) # max of repeated inner values
            meanrep <- mean(table(ix)) # average of repeated inner values
            ## differences between consecutive unique values
            jumpsx <- unique(signif(diff(unx), 3))
            rangex <- diff(range(x))
            multi <- 10^(-min(floor(log10(jumpsx))))
            ## jumps between unique values are integer multiples
            ## of 'jumpquantum'
            jumpquantum <- gcd(jumpsx * multi) / multi
            datavalues <- NULL
        } else {
            datavalues <- as.character(unx)
            unx <- as.character(unx)
        }

### Check whether the variate has been first rounded
### and then transformed to logarithmic scale.
### This may lead to problems
        if(is.numeric(unx) && datamin > 0 && uniquex > 2) {
            for(base in c(2, exp(1), 10)) {
                ljumpsx <- unique(signif(diff(log(unx, base)), 3))
                lrangex <- diff(log(range(x), base))
                lmulti <- 10^(-min(floor(log10(ljumpsx))))
                ## jumps between unique values are integer multiples
                ## of 'jumpquantum'
                ljumpquantum <- gcd(ljumpsx * lmulti) / lmulti
                if (!(
                    ljumpquantum / lrangex < 1e-5 || # no visible gaps between values
                        (ljumpquantum / lrangex < 1e-3 &&
                             maxrep <= 100)  # visible gaps, but not many value repetitions
                )) {
                    warninglist <- c(warninglist,
                        paste0('\n* "', name, '" variate',
                            ' appears to have been rounded\n',
                            'and then transformed to logarithmic scale.\n',
                            'This may lead to problems in the inference.\n',
                            'Preferably, transform it back to non-logarithmic scale.'))
                    break
                }
            }
        }

#### Now make educated guess for the type of variate

        ## #### For the moment binary variates counted within nominal ones
        ## if (uniquex == 2) {
        ##     ## Binary variate? (has only two unique values)
        ##     type <- 'binary'
        ##     Nvalues <- 2
        ##     datastep <- NA
        ##     domainmin <- NA
        ##     domainmax <- NA
        ##     minincluded <- NA
        ##     maxincluded <- NA
        ##     ## lowvalue <- NA
        ##     ## centralvalue <- NA
        ##     ## highvalue <- NA
        ##     plotmin <- NA
        ##     plotmax <- NA
        ##     names(datavalues) <- paste0('V', 1:2)
        ##     ##
        ##     if(verbose){
        ##         message('  Two different values detected:\n',
        ##             paste0('"', datavalues, '"', collapse=', '))
        ##         message('  Assuming variate to be BINARY.')
        ##     }
        ##
        ## } else
            if (uniquex <= 2 || (!is.numeric(x) &&
                       !any(sapply(ordinalkeywords, function(keyw){
                           grepl(keyw, datavalues, fixed = TRUE)
                       })))) {
            ## Nominal variate? (non-numeric values)
            type <- 'nominal'
            Nvalues <- uniquex
            domainmin <- NA
            domainmax <- NA
            datastep <- NA
            minincluded <- NA
            maxincluded <- NA
            ## lowvalue <- NA
            ## centralvalue <- NA
            ## highvalue <- NA
            ## plotmin <- NA
            ## plotmax <- NA
            names(datavalues) <- paste0('V', 1:Nvalues)
            ##
            if(verbose){
                message('  - ', Nvalues, ' different values detected:\n',
                    paste0('"', datavalues, '"', collapse = ', '))
                message('  which do not seem to refer to an ordered scale.')
                message('  Assuming variate to be NOMINAL.')
            }

        } else if (!is.numeric(x)) {
            ## Nominal variate? (non-numeric values)
            ## Ordinal variate with few numeric values?
            type <- 'ordinal'
            Nvalues <- uniquex
            domainmin <- NA
            domainmax <- NA
            datastep <- NA
            minincluded <- NA
            maxincluded <- NA
            ## lowvalue <- NA
            ## centralvalue <- NA
            ## highvalue <- NA
            ## plotmin <- NA
            ## plotmax <- NA
            names(datavalues) <- paste0('V', 1:Nvalues)
            ##
            if(verbose){
                message('  - ', Nvalues, ' different non-numeric values detected:\n',
                    paste0('"', datavalues, '"', collapse = ', '))
                message('  which seem to refer to an ordered scale.')
                message('  Assuming variate to be ORDINAL.')
                message('  Please appropriately reorder its values in metadata file.')
            }
            warninglist <- c(warninglist,
                paste0('\n* "', name, '" variate',
                    ' appears to be ordinal.',
                    '\nPlease appropriately reorder its values in metadata file.'))


#### Now we know that the variate is numeric

        } else if (uniquex <= 10 &&
                       (all(x == round(x)) || jumpquantum == round(jumpquantum))) {
            ## Ordinal variate with few numeric values?
            type <- 'ordinal'
            Nvalues <- uniquex
            ## if the values are spaced by an integer,
            ## no need to use the 'V' fields
            if(jumpquantum == round(jumpquantum)) {
                datastep <- jumpquantum
                domainmin <- datamin
                domainmax <- datamax
                datavalues <- NULL
            } else {
                domainmin <- NA
                domainmax <- NA
                datastep <- NA
                datavalues <- as.character(unx)
                names(datavalues) <- paste0('V', 1:Nvalues)
            }
            minincluded <- NA
            maxincluded <- NA
            ## lowvalue <- NA
            ## centralvalue <- NA
            ## highvalue <- NA
            ## plotmin <- NA
            ## plotmax <- NA
            ##
            if(verbose){
                message('  - Only', Nvalues, ' different numeric values detected:\n')
                if(jumpquantum == round(jumpquantum)) {
                    message('from ', domainmin, ' to ', domainmax,
                        ' in steps of ', jumpquantum)
                } else {
                    message(paste0('"', datavalues, '"', collapse = ', '))
                }
                message('  Assuming variate to be ORDINAL.')
            }

            ## #### This categorization is not used in the present version
            ## } else if (uniquex <= 10 && ) {
            ##     ## Ordinal variate with many numeric values?
            ##     type <- 'ordinal'
            ##     Nvalues <- uniquex
            ##     datastep <- jumpquantum
            ##     domainmin <- datamin
            ##     domainmax <- datamax
            ##     minincluded <- NA
            ##     maxincluded <- NA
            ##     ## lowvalue <- NA
            ##     ## centralvalue <- NA
            ##     ## highvalue <- NA
            ##     plotmin <- NA
            ##     plotmax <- NA
            ##     datavalues <- NULL
            ##     ##
            ##     if(verbose){
            ##         message(' - ', Nvalues, ' different numeric values detected')
            ##         message('  distance between datapoints is a multiple of integer',
            ##             jumpquantum)
            ##         message('  Assuming variate to be ORDINAL.')
            ##     }
            ##     warninglist <- c(warninglist,
            ##         paste0('* "', name, '" variate',
            ##             ' appears to be ordinal,',
            ##             '\nbut it could also be a rounded continous variate'))


        } else {
            ## The variate seems continuous
            type <- 'continuous'
            Nvalues <- Inf
            ## preliminary values, possibly modified below
            datastep <- NA
            domainmin <- NA # signif(datamax - 3 * rangex, 1)
            domainmax <- NA # signif(datamax + 3 * rangex, 1)

            ## Q1 <- quantile(x = x, probs = 0.25, type = 6)
            ## centralvalue <- quantile(x = x, probs = 0.5, type = 6)
            ## Q3 <- quantile(x = x, probs = 0.75, type = 6)
            ## ## Borderline case if the first and second quartile have the same value
            ## if (lowvalue == highvalue) {
            ##   lowvalue <- if (sum(x < Q1) > 0) {
            ##                 max(x[x < Q1])
            ##               } else {
            ##                 max(x[x <= Q1])
            ##               }
            ##   highvalue <- if (sum(x > Q3) > 0) {
            ##                  min(x[x > Q3])
            ##                } else {
            ##                  min(x[x >= Q3])
            ##                }
            ## }
            ## plotmin <- datamin - IQR(x, type=6) / 2
            ## plotmax <- datamax + IQR(x, type=6) / 2
            datavalues <- NULL
            ##
            if(verbose){
                message('  - Numeric values between ',
                    datamin, ' and ', datamax)
                message('  Assuming variate to be CONTINUOUS.')
            }

            roundedflag <- FALSE
#### Consider subcases for datastep
            ## for rounded variates it doesn't matter whether the boundaries
            ## are included in the domain or not
            if (!(
                jumpquantum / rangex < 1e-5 || # no visible gaps between values
                    (jumpquantum / rangex < 1e-3 &&
                         maxrep <= 100)  # visible gaps, but not many value repetitions
            )) {
                roundedflag <- TRUE
                ## The variate seems to have quantized differences
                ## between unique values
                ## hence it might be rounded or integer/ordinal

                ## if(rangex / datastep > 256) {
                ##   ## The variate seems rounded,
                ##   ## but no latent-variable representation is needed
                ##   message('\nNOTE: variate ', name, ' is probably rounded,\n',
                ##           'but is treated as not rounded ',
                ##           'owing to its large range of values.\n')
                ##   datastep <- 0
                ## } else {
                datastep <- jumpquantum
                minincluded <- maxincluded <- NA
                ##
                if(verbose){
                    message('  - Distance between datapoints is a multiple of ',
                        jumpquantum)
                    message('  Assuming variate to be ROUNDED.')
                }
                if(jumpquantum >=1) {
                    warninglist <- c(warninglist,
                        paste0('\n* "', name, '" variate',
                            ' appears to be continuous and rounded,',
                            '\nbut it could also be an ordinal variate'))
                }
                ## domainmin <- signif(datamin - 4 * rangex, 1)
                ## domainmax <- signif(datamax + 4 * rangex, 1)

            } # End rounded case


#### Consider subcases for closedness of domain
            if(!roundedflag){
                minincluded <- NA
                maxincluded <- NA
            }
            if (sum(x == datamin) > 1) {
                ## seems to have left-closed domain
                domainmin <- datamin
                ## plotmin <- datamin
                if(!roundedflag){
                    minincluded <- 'true'
                }
                ##
                if(verbose){
                    message('  - Several datapoints have minimum value ',
                        datamin)
                    message('  Assuming "domainmin" to be this minimum observed value')
                    if(!roundedflag){
                        message('  and to be included in the domain (singular probabilities there).')
                    }
                }
            } else if (all(unx > 0)) {
                ## seems to be strictly positive
                domainmin <- 0
                ## plotmin <- max((domainmin + datamin) / 2, plotmin)
                if(!roundedflag){
                    minincluded <- NA
                }
                ##
                if(verbose){
                    message('  - All values are positive')
                    message('  Assuming "domainmin" to be 0')
                    if(!roundedflag){
                        message('  with 0 excluded from domain.')
                    }
                }
            } else if (all(unx >= 0)) {
                ## seems to be non-negative
                domainmin <- 0
                ## plotmin <- max((domainmin + datamin) / 2, plotmin)
                if(!roundedflag){
                    minincluded <- 'true'
                }
                ##
                if(verbose){
                    message('  - All values are non-negative')
                    message('  Assuming "domainmin" to be 0')
                    if(!roundedflag){
                        message('  with 0 included in the domain.')
                    }
                }
            }

            if (sum(x == datamax) > meanrep) {
                ## seems to right-closed domain
                domainmax <- datamax
                ## plotmax <- datamax
                if(!roundedflag){
                    maxincluded <- 'true'
                }
                ##
                if(verbose){
                    message('  - Many datapoints have maximum value ',
                        datamax)
                    message('  Assuming "domainmax" to be this maximum observed value')
                    if(!roundedflag){
                        message('  and to be included in the domain (singular probabilities there).')
                    }
                }
            }

        } # End continuous-variate case

#### Create metadata object
        metadata <- merge(metadata,
            c(
                list(name = name,
                    type = type,
                    ## Nvalues = Nvalues,
                    domainmin = domainmin,
                    domainmax = domainmax,
                    datastep = datastep,
                    minincluded = minincluded,
                    maxincluded = maxincluded
                    ## lowvalue = signif(lowvalue,3),
                    ## centralvalue = signif(centralvalue,3),
                    ## highvalue = signif(highvalue,3),
                    ## plotmin = signif(plotmin,2),
                    ## plotmax = signif(plotmax,2)
                ),
                if (addsummary2metadata) {
                    list(datamin = datamin,
                        datamax = datamax,
                        datamaxrep = max(table(x)),
                        dataNvalues = uniquex)
                },
                as.list(datavalues)
            ),
            sort = FALSE, all = TRUE)
    } # End loop over variates

    ## Print warnings
    if(!is.null(warninglist)){
        message(
            '=========',
            '\nWARNINGS',
            ' - please make sure to check these variates in the metadata file:')
        for(awarning in warninglist){message(awarning)}
        message('=========')
    }

    ## Save to file if the file parameter is set
    if (!(is.null(file) || isFALSE(file))) {
        ## If file is a string, name the metadata file that
        if (is.character(file)) {
            file <- paste0(sub('.csv$', '', file), '.csv')
        }
        ## If file is not a string, and the data is read from a file,
        ## save the metadata as metadata_<datafile>.csv
        else if (!is.null(datafile)) {
            file <- paste0(sub('.csv$', '', file), '_metadata.csv')
        }
        ## If the file already exists, rename the old file as a backup file
        ## and name this metadata file as intended
        if (backupfiles) {
            if (file.exists(file)) {
                file.rename(from = file, to = paste0(sub('.csv$', '', file), '_bak', format(Sys.time(), '%y%m%dT%H%M%S'), '.csv'))
            }
        }
        ## Save the file and print invisibly
        pwrite.csv(metadata, file)
        message('\nSaved proposal metadata file as ',
            paste0('"', file, '"'))
        invisible(metadata)
    } else {
        ## Else just print to console
        metadata
    }
}

#' Calculate posterior probabilities
#'
#' @description This function calculates posterior probability densities, cumulative posterior probabilities, and mixtures thereof. It also outputs the variability of such probabilities if more training data were available, and the Monte Carlo Standard Error for the calculated posterior probabilities.
#'
#' @details This function calculates the posterior probability \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data})}, where \eqn{Y = y} and \eqn{X = x} are two (non overlapping) sets of joint variate values, inputted as [data frame][base::data.frame()] arguments `Y` and `X`. It is somewhat analogous to the `d`-variants and `p`-variantes of R distribution functions, such as [stats::dnorm()] and [stats::pnorm()]. If `X` is omitted or `NULL`, then the posterior probability \eqn{\mathrm{Pr}(Y = y \vert \text{data})} is calculated.
#'
#' For some variates in `Y` or `X`, tail values can also be prescribed, so that this function calculates mixed probabilities such as \deqn{\mathrm{Pr}(Y_1 = y_1, Y_2 \le y_2, \dotsc \vert X_1 = x_1, X_2 \ge x_2, \dotsc, \text{data})\ .} Tail values are inputted via the `'tails'` argument; see "Usage".
#'
#' This function also outputs the variability of the posterior probabilities above, that is, probabilities such as \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{new\,data}, \text{data})} that we could have if more learning data were provided, as well as a number of samples of the possible values of such probability. This variability can be outputted in two ways; the user can choose either, or both, or none:
#'
#' - As samples (default 3600 samples, depending on the 'nsamples' argument given to the [learn()] function) of the alternative values that the posterior probability could have.
#' - As quantiles (default 5.5%, 25%, 75%, 94.5%) of the possible variability.
#'
#' If several joint values are given for `Y` or `X`, the function will create a 2D grid of results for all possible combinations of the given `Y` and `X` values.
#'
#' This function also allows for base-rate or other prior-probability corrections: If a prior (for instance, a base rate) for `Y` is given, the function will calculate the probability \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data}, \text{prior})} from \eqn{\mathrm{Pr}(X = x \vert Y = y, \text{data})} and the prior, by means of Bayes's theorem.
#'
#' Each variate in each argument `Y`, `X` can be specified either as a point-value \eqn{Y = y} or as a left-open interval \eqn{Y \le y} or as a right-open interval \eqn{Y \ge y}, through the argument `tails`.
#'
#' See `vignette('intro')` for example uses.
#'
#' @param Y Matrix or data.table: set of values of variates of which we want
#'   the joint probability of. One variate per column, one set of values per row.
#' @param X Matrix or data.table or `NULL` (default): set of values of variates on which we want to condition the joint probability of `Y`. If `NULL`, no conditioning is made (except for conditioning on the learning dataset and prior assumptions). One variate per column, one set of values per row.
#' @param learnt Either a character with the name of a directory or full path for a 'learnt.rds' object, produced by the [learn()] function, or such an object itself.
#' @param tails Named vector or list, or `NULL` (default). The names must match some or all of the variates in arguments `Y` and `X`. For variates in this list, the probability arguments are understood in an semi-open interval sense: \eqn{Y \le y} or \eqn{Y \ge y}, an so on. This is true for `Y` and `X` variates (on the left and on the right of the conditional sign \eqn{\,\vert\,}). A left-open interval \eqn{Y \le y} is indicated by `'<='` or `'left'` or `-1`; a right-open interval \eqn{Y \ge y} is indicated by `'>='` or `'right'` or `+1`. Values `NULL`, `'=='`, `0` indicate that a point value `Y = y` (not an interval) should be calculated. **NB**: the semi-open intervals *always* include the given value; this is important for ordinal or rounded variates. For instance, if \eqn{Y} is an integer variate, then to calculate  \eqn{\mathrm{Pr}(Y < 3)} you should require \eqn{\mathrm{Pr}(Y \le 2)}; for this reason we also have that \eqn{\mathrm{Pr}(Y \le 2)} and  \eqn{\mathrm{Pr}(Y \ge 2)} generally add up to *more* than 1.
#' @param priorY Numeric vector with the same length as the rows of `Y`, or `TRUE`, or `NULL` (default): prior probabilities or base rates for the `Y` values. If `TRUE`, the prior probabilities are assumed to be all equal.
#' @param nsamples Integer or `NULL` or `'all'` (default): desired number of samples of the variability of the probability for `Y`. If `NULL`, no samples are reported. If `'all'` (or `Inf`), all samples obtained by the [learn()] function are used.
#' @param quantiles Numeric vector, between 0 and 1, or `NULL`: desired quantiles of the variability of the probability for `Y`. Default `c(0.055, 0.25, 0.75, 0.945)`, that is, the 5.5%, 25%, 75%, 94.5% quantiles. These are typical quantile values in the Bayesian literature: they give 50% and 89% credibility intervals, which correspond to 1 shannons and 0.5 shannons of uncertainty (see <doi:10.5281/zenodo.17072199>). If `NULL`, no quantiles are calculated.
#' @param parallel Logical or positive integer or cluster object. `TRUE` (default): use roughly half of available cores; `FALSE`: use serial computation; integer: use this many cores. It can also be a cluster object previously created with [parallel::makeCluster()]; in this case the parallel computation will use this object.
#' @param sep character, default `','`: character to separate variate names and values
#' @param solidus character, default `'|'`: character prepended to names of the variates in the conditional (typically the `X` variates).
#' @param verbose Logical, default `FALSE`: give messages about parallel processing?
#' @param keepYX Logical, default `TRUE`: keep a copy of the `Y` and `X` arguments in the output? This is used for the plot method.
#'
#' @return An object of class "probability", effectively a list consisting of the following elements:
#'
#' - `values`: a matrix with the probabilities \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data})}, for all joint values \eqn{y} of the \eqn{Y}-variates (rows) and  all joint values \eqn{x} of the \eqn{X}-variates (columns).
#' - `quantiles` (possibly `NULL`): an array with the variability quantiles (3rd dimension of the array) for such probabilities.
#' - `samples` (possibly `NULL`): an array with the variability samples (3rd dimension of the array) for such probabilities.
#' - `values.MCaccuracy`, `quantiles.MCaccuracy`: arrays with the numerical accuracies (roughly speaking a standard deviation) of the Monte Carlo calculations for the `values` and `quantiles` elements.
#' - `Y`, `X`: copies of the `Y` and `X` arguments.
#'
#' @references
#'
#' - Lindley, Novick (1981): *The role of exchangeability in inference*, <doi:10.1214/aos/1176345331>.
#' - Bernardo, Smith (2000): *Bayesian Theory*. Wiley <doi:10.1002/9780470316870>.
#' - Jaynes (2003): *Probability Theory: The Logic of Science*. Cambridge University Press <doi:10.1017/CBO9780511790423>.
#' - MacKay (2005): *Information Theory, Inference, and Learning Algorithms*. Cambridge University Press <https://www.inference.org.uk/itila/book.html>.
#' - Porta Mana (2025): *What's special about 89% credibility intervals?* <doi:10.5281/zenodo.17072199>.
#'
#' @seealso
#' [learn()], which generates the `learnt` objects required by `Pr()`.
#'
#' [plot.probability()] to plot probabilities and quantiles calculated by `Pr()`.
#'
#' [hist.probability()] to plot histograms of the probability distributions calculated by `Pr()`.
#'
#' [print.probability()] to print the main elements of the probabilities calculated by `Pr()`.
#'
#' [qPr()] to calculate quantiles for a specific variate, that is, the variate values having given probabilities.
#'
#' [rPr()] to generate datapoints.
#'
#' @examples
#' ## Load the example `learnt` object calculated from the "penguins" dataset;
#' ## variates: 'species' and 'bill_len'
#' learnt <- learntExample
#'
#' ## ## Example 1:
#' ## Calculate the probability that an unknown penguin from this population
#' ## is of species 'Adelie'
#'
#' probs <- Pr(
#'   Y = data.frame(species = 'Adelie'),
#'   learnt = learnt, parallel = 1
#' )
#'
#' ## display the probability value
#' probs$values
#'
#' ## the full-population frequency of 'Adelie' penguins is unknown;
#' ## display the 5.5%- and 94.5%-probability values
#' ## for such frequency
#' probs$quantiles[, , c('5.5%', '94.5%')]
#'
#' ## we can also plot the probability distribution for this full-population frequency
#' hist(probs, legend = 'topright')
#'
#'
#' ## ## Example 2:
#' ## Calculate the 3 probabilities that an unknown penguin from this population
#' ## is of species 'Adelie', 'Chinstrap', 'Gentoo'
#'
#' probs <- Pr(
#'   Y = data.frame(species = c('Adelie', 'Chinstrap', 'Gentoo')),
#'   learnt = learnt, parallel = 1
#' )
#'
#' ## display the 3 probability values
#' probs$values
#'
#' ## the full-population frequencies of the three species are unknown;
#' ## display the 5.5%- and 94.5%-probability values
#' ## for such frequencies
#' probs$quantiles[, , c('5.5%', '94.5%')]
#'
#' ## plot the probabilities and quantiles
#' plot(probs)
#'
#' ## plot the probability distribution for the full-population frequency
#' ## of each species
#' hist(probs)
#'
#' ## ## Example 3:
#' ## Calculate the probability that an unknown penguin is of species 'Adelie'
#' ## GIVEN that its bill length is 43 mm
#'
#' probs <- Pr(
#'   Y = data.frame(species = 'Adelie'),
#'   X = data.frame(bill_len = 43),
#'   learnt = learnt, parallel = 1
#' )
#'
#' ## display the probability value
#' probs$values
#'
#' ## the full-subpopulation frequency of 'Adelie' penguins,
#' ## among penguins having bill length of 43 mm, is unknown;
#' ## display the 5.5%- and 94.5%-probability values
#' ## for such conditional frequency
#' probs$quantiles[, , c('5.5%', '94.5%')]
#'
#'
#' ## ## Example 4:
#' ## Calculate the probability that
#' ## an unknown penguin is of species 'Adelie' AND its bill length is 43 mm
#'
#' probs <- Pr(
#'   Y = data.frame(species = 'Adelie', bill_len = 43),
#'   learnt = learnt, parallel = 1
#' )
#'
#' ## display the probability value
#' probs$values
#'
#' ## display the 5.5%- and 94.5%-probability values
#' ## for the full-population frequency of 'Adelie' penguins with 43 mm bills
#' probs$quantiles[, , c('5.5%', '94.5%')]
#'
#'
#' ## ## Example 5:
#' ## Calculate the 3 x 2 probabilities for the 3 species
#' ## GIVEN bill-lengths of 43 mm and 44 mm
#'
#' Y <- data.frame(species = c('Adelie', 'Chinstrap', 'Gentoo'))
#'
#' X <- data.frame(bill_len = c(43, 44))
#'
#' probs <- Pr(Y = Y, X = X, learnt = learnt, parallel = 1)
#'
#' ## display the 3 x 2 probability values
#' probs$values
#'
#' ## display the 5.5%- and 94.5%-probability values
#' ## for the full-population joint frequencies
#' probs$quantiles[, , c('5.5%', '94.5%')]
#'
#' ## plot the probabilities and quantiles
#' plot(probs)
#'
#'
#' ## ## Example 6:
#' ## Calculate the 3 x 2 joint probabilities for the 3 species
#' ## AND bill-lengths of 43 mm and 44 mm
#'
#' Y <- expand.grid(
#'   species = c('Adelie', 'Chinstrap', 'Gentoo'),
#'   bill_len = c(43, 44)
#' )
#'
#' probs <- Pr(Y = Y, learnt = learnt, parallel = 1)
#'
#' ## display the 6 joint-probability values
#' probs$values
#'
#' ## display the 5.5%- and 94.5%-probability values
#' ## for the full-population joint frequencies
#' probs$quantiles[, , c('5.5%', '94.5%')]
#'
#'
#' @import parallel
#' @import stats
#' @import utils
#'
#' @export
Pr <- function(
    Y,
    X = NULL,
    learnt,
    tails = NULL,
    priorY = NULL,
    nsamples = 'all',
    quantiles = c(0.055, 0.25, 0.75, 0.945),
    parallel = TRUE,
    sep = ',',
    solidus = '|',
    verbose = FALSE,
    keepYX = TRUE
) {
    ## #' @param usememory Logical, default `TRUE`: save partial results to disc, to avoid excessive RAM use. (For the moment only possible value is `TRUE`.)
    usememory <- TRUE

    Qerror <- pnorm(c(-1, 1))

#### Requested parallel processing
    ## NB: doesn't make sense to have more cores than chains
    closeexit <- FALSE
    if ('cluster' %in% class(parallel)){
        ## user provides a cluster object
        cl <- parallel
    } else if (isTRUE(parallel)) {
        ## user wants us to register a parallel backend
        ## and to choose number of cores
        ncores <- max(1,
            floor(parallel::detectCores() / 2))
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
        if(verbose){message('Registered ', capture.output(print(cl)), '.')}
    } else if (isFALSE(parallel)) {
        ## user wants us not to use parallel cores
        ncores <- 1
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
    } else if (is.numeric(parallel) &&
                   is.finite(parallel) && parallel >= 1) {
        ## user wants us to register 'parallel' # of cores
        ncores <- parallel
        cl <- parallel::makeCluster(ncores)
        closeexit <- TRUE
        if(verbose){message('Registered ', capture.output(print(cl)), '.')}
    } else {
        stop("Unknown value of argument 'parallel'.")
    }

    ## Close parallel connections if any were opened
    if(closeexit) {
        closecoresonexit <- function(){
            if(verbose){message('Closing connections to cores.')}
            parallel::stopCluster(cl)
            ## parallel::setDefaultCluster(NULL)
        }
        on.exit(closecoresonexit())
    }

    ## Extract Monte Carlo output & auxmetadata
    ## If learnt is a string, check if it's a folder name or file name
    if (is.character(learnt)) {
        ## Check if 'learnt' is a folder containing learnt.rds
        if (file_test('-d', learnt) &&
                file.exists(file.path(learnt, 'learnt.rds'))) {
            learnt <- readRDS(file.path(learnt, 'learnt.rds'))
        } else {
            ## Assume 'learnt' the full path of learnt.rds
            ## possibly without the file extension '.rds'
            learnt <- paste0(sub('.rds$', '', learnt), '.rds')
            if (file.exists(learnt)) {
                learnt <- readRDS(learnt)
            } else {
                stop('The argument "learnt" must be a folder containing learnt.rds, or the path to an rds-file containing the output from "learn()".')
            }
        }
    }
    ## Add check to see that learnt is correct type of object?
    auxmetadata <- learnt$auxmetadata
    learnt$auxmetadata <- NULL
    learnt$auxinfo <- NULL
    ncomponents <- nrow(learnt$W)
    nmcsamples <- ncol(learnt$W)

    if(is.numeric(nsamples)){
        if(is.na(nsamples) || nsamples < 1) {
            nsamples <- NULL
        } else if(!is.finite(nsamples)) {
            nsamples <- nmcsamples
        }
    } else if (is.character(nsamples) && nsamples == 'all'){
        nsamples <- nmcsamples
    }


    Y <- as.data.frame(Y)
    Yv <- colnames(Y)

    if(all(is.na(X))){X <- NULL}
    if(!is.null(X)){X <- as.data.frame(X)}
    Xv <- colnames(X)

    if(!is.null(tails)){
        tails <- as.list(tails)
        if(is.null(names(tails))) {
            stop('Missing variate names in "tails"')
        }
    }
    tailscentre <- list('==', 0, '0', NULL)
    tailsleft <- list('<=', -1, '-1', 'left')
    tailsright <- list('>=', 1, '+1', 'right')
    tailsvalues <- c(tailscentre, tailsleft, tailsright)

    ## Consistency checks

    if (!all(Yv %in% auxmetadata$name)) {
        stop('unknown Y variate ',
            paste0(Yv[!(Yv %in% auxmetadata$name)], collapse = ' '),
            '\n')
    }
    if (length(unique(Yv)) != length(Yv)) {
        stop('duplicate Y variates\n')
    }

    if (!all(Xv %in% auxmetadata$name)) {
        stop('unknown X variate ',
            paste0(Xv[!(Xv %in% auxmetadata$name)], collapse = ' '),
            '\n')
    }
    if (length(unique(Xv)) != length(Xv)) {
        stop('duplicate X variates\n')
    }

    if (any(Yv %in% Xv)) {
        stop('overlap in Y and X variates\n')
    }

    tailsv <- names(tails)
    if (!all(tailsv %in% c(Yv, Xv))) {
        warning('"tails" variate ',
            paste0(tailsv[!(tailsv %in% c(Yv, Xv))], collapse = ' '),
            ' not among Y and X; ignored\n')
        tails <- tails[tailsv %in% c(Yv, Xv)]
        tailsv <- names(tails)
    }
    if (length(unique(tailsv)) != length(tailsv)) {
        stop('duplicate "tails" variates\n')
    }
    if(!all(tails %in% tailsvalues)) {
        stop('"tails" values must be ',
            paste0(tailsvalues, collapse = ' '), '\n')
    }

    ## transform 'tails' to -1, +1
    ## +1: '<=',    -1: '>='
    ## this is opposite of the argument convention because
    ## interval probabilities are calculated with `lower.tail = TRUE`
    ## eg:
    ## pnorm(x, mean, sd, lower.tail = FALSE) ==
    ##     pnorm(-x, -mean, sd, lower.tail = TRUE)
    tails[tails %in% tailscentre] <- NULL
    cleft <- tails %in% tailsleft
    cright <- tails %in% tailsright
    tails[cleft] <- +1
    tails[cright] <- -1
    tails <- unlist(tails)

    ## Check if a prior for Y is given, in that case Y and X will be swapped
    if (isFALSE(priorY) || is.null(priorY)) {
        priorY <- NULL
        doquantiles <- !is.null(quantiles)
    } else {
        if(is.null(X)){ stop('X must be non-null if priorY is given') }

        if(!isTRUE(priorY) && length(priorY) != nrow(Y)) {
            stop('priorY must have as many elements a the rows of Y')
        }

        doquantiles <- FALSE
        if(!is.null(quantiles)) {
            ## message('For the moment, ',
            ##     'computation of quantiles is affected by a larger error',
            ##     'if "priorY" is specified.')
            nsamples0 <- nsamples
            nsamples <- max(nsamples0, 1200L)
        }

        ## if priorY is TRUE, the user wants a uniform prior distribution
        if(isTRUE(priorY)){
            priorY <- 1 + numeric(nrow(Y))
        } else {
            ## Check for invalid probabilities
            if (!is.numeric(priorY) || any(priorY < 0)) {
                stop('priorY contains invalid probabilities')
            }
        }

        ## Swap X and Y, to use Bayes's theorem
        . <- Y
        Y <- X
        X <- .
        rm(.)
        . <- Yv
        Yv <- Xv
        Xv <- .
        rm(.)
    }

    nY <- nrow(Y)
    nX <- max(nrow(X), 1L)

    dosamples <- !is.null(nsamples)


#### tmp dir where to save X and Y objects
    temporarydir <- tempdir()


#### Calculate and save arrays for X values:
    if (is.null(X)) {
        lprobX <- log(learnt$W)
        saveRDS(lprobX,
            file.path(temporarydir,
                paste0('__X', 1, '__.rds'))
        )
    } else {
        ## Construction of the arguments for util_lprobs, X argument
        lpargs <- util_lprobsargsyx(
            x = X,
            auxmetadata = auxmetadata,
            learnt = learnt,
            tails = tails
        )

        invisible(parallel::parLapply(cl = cl,
            X = lpargs$xVs,
            fun = util_lprobsbase,
            params = lpargs$params,
            logW = c(log(learnt$W)),
            temporarydir = temporarydir,
            lab = '__X'
        ))
    }


#### Calculate and save arrays for Y values:

    ## Construction of the arguments for util_lprobs, Y argument
    ## jacobians <- exp(-rowSums(
    ##     log(vtransform(Y,
    ##         auxmetadata = auxmetadata,
    ##         invjacobian = TRUE)),
    ##     na.rm = TRUE
    ## ))

    lpargs <- util_lprobsargsyx(
        x = Y,
        auxmetadata = auxmetadata,
        learnt = learnt,
        tails = tails
    )

    invisible(parallel::parLapply(cl = cl,
        X = lpargs$xVs,
        fun = util_lprobsbase,
        params = lpargs$params,
        logW = 0,
        temporarydir = temporarydir,
        lab = '__Y'
    ))

    ## Calculation with all Y and X combinations
    keys <- c('values', 'quantiles', 'samples', 'values.MCaccuracy', 'quantiles.MCaccuracy')
    ##
    combfnr <- function(...){setNames(do.call(mapply,
        c(FUN = `rbind`, lapply(X = ..., FUN = `[`, keys, drop = FALSE))),
        keys)}
    ## combfnc <- function(...){setNames(do.call(mapply, c(FUN=cbind, lapply(list(...), `[`, keys))), keys)}

    out <- combfnr(parallel::parApply(cl = cl,
            X = expand.grid(jy = seq_len(nY), jx = seq_len(nX)),
            MARGIN = 1,
            FUN = util_combineYX,
            temporarydir = temporarydir, usememory = usememory,
            doquantiles = doquantiles, quantiles = quantiles,
            dosamples = dosamples, nsamples = nsamples,
            Qerror = Qerror
        ))

    ## clean temp files
    if(usememory) {
        unlink(x = c(
            sapply(seq_len(nX), function(jx){
                file.path(temporarydir, paste0('__X', jx, '__.rds'))
            }),
            sapply(seq_len(nY), function(jy){
                file.path(temporarydir, paste0('__Y', jy, '__.rds'))
            })
        ))
    }


    if(is.null(priorY)){
        y <- Y
        y[, colnames(Y) %in% tailsv] <- NA
        jacobians <- exp(rowSums(
            as.matrix(vtransform(y,
                auxmetadata = auxmetadata,
                logjacobianOr = TRUE)),
            na.rm = TRUE
        ))
        rm(y)
    }

    ## transform to grid
    ## in the output-list elements the Y & X values are the rows
    dim(out$values.MCaccuracy) <- dim(out$values) <- c(nY, nX)

    ## dimension & value names for variates
    if(!is.null(X)){
        Xnames <- setNames(object = list(
            apply(X = X, MARGIN = 1, FUN = paste0, collapse = sep,
                simplify = TRUE)),
            nm = paste0(if(is.null(priorY)){solidus},
                paste0(colnames(X), collapse = sep)) )
    } else {
        Xnames <- list(NULL)
    }
    if(!is.null(Y)){
        Ynames <- setNames(object = list(
            apply(X = Y, MARGIN = 1, FUN = paste0, collapse = sep,
                simplify = TRUE)),
            nm = paste0(if(!is.null(priorY)){solidus},
                paste0(colnames(Y), collapse = sep)) )
    } else {
        Ynames <- list(NULL)
    }

    if(is.null(priorY)){
        ## multiply by jacobian factors
        out$values <- out$values * jacobians
        out$values.MCaccuracy <- signif(x = out$values.MCaccuracy * jacobians,
            digits = 2)

        dimnames(out$values) <- c(Ynames, Xnames)
        dimnames(out$values.MCaccuracy) <- c(Ynames, Xnames)

    } else {
        ## Bayes's theorem
        out$values <- t(priorY * t(out$values))
        out$values.MCaccuracy <- NULL
        normf <- rowSums(out$values, na.rm = TRUE)
        out$values <- t(out$values/normf)

        dimnames(out$values) <- c(Xnames, Ynames)
    }

    if(dosamples){
        temp <- list(sample = round(seq(1, nmcsamples, length.out = nsamples)))

        ## transform to grid
        dim(out$samples) <- c(nY, nX, nsamples)

        if(is.null(priorY)){
            ## multiply by jacobian factors
            out$samples <- out$samples * jacobians

            dimnames(out$samples) <- c(Ynames, Xnames, temp)

        } else {
            ## Bayes's theorem
            out$samples <- priorY * aperm(a = out$samples, perm = c(2, 1, 3),
                resize = TRUE)
            normf <- c(t(colSums(out$samples, na.rm = TRUE)))
            out$samples <- aperm(a = aperm(a = out$samples, perm = NULL,
                resize = TRUE) / normf, perm = NULL,
                resize = TRUE)

            dimnames(out$samples) <- c(Xnames, Ynames, temp)
        }
    }

    if(doquantiles){
        temp <- list(Q = names(quantile(x = 1, probs = quantiles, names = TRUE)))
        if(is.null(priorY)){
            ## transform to grid
            dim(out$quantiles) <- c(nY, nX, length(quantiles))
            dim(out$quantiles.MCaccuracy) <- c(nY, nX, length(quantiles))
            ## multiply by jacobian factors
            out$quantiles <- out$quantiles * jacobians
            out$quantiles.MCaccuracy <- signif(x = out$quantiles.MCaccuracy * jacobians,
                digits = 2)

            dimnames(out$quantiles) <- c(Ynames, Xnames, temp)
            dimnames(out$quantiles.MCaccuracy) <- c(Ynames, Xnames, temp)

        } else {
            ## calculate quantiles from samples
            out$quantiles <- aperm(
                a = apply(X = out$samples, MARGIN = c(1, 2), FUN = quantile,
                    probs = quantiles, type = 6,
                    na.rm = TRUE, names = FALSE,
                    simplify = TRUE),
                perm = c(2, 3, 1), resize = FALSE)

            ## adjust number of samples as originally requested
            if(is.null(nsamples0)) {
                out$samples <- NULL
            } else if(nsamples0 < nsamples) {
                out$samples <-out$samples[ , ,
                    round(seq(1, nsamples, length.out = nsamples0))]
            }

            dimnames(out$quantiles) <- c(Xnames, Ynames, temp)
            dimnames(out$quantiles.MCaccuracy) <- c(Xnames, Ynames, temp)
        }

    }

    if(isTRUE(keepYX)){
        ## save Y and X values in the output; useful for plotting methods
        if(is.null(priorY)){
            out$Y <- Y
            out$X <- X
        } else {
            out$Y <- X
            out$X <- Y
            }
    }

    class(out) <- 'probability'
    out
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_checkpoints.R
\name{util_prepPcheckpoints}
\alias{util_prepPcheckpoints}
\title{Format datapoints used for MCMC monitoring}
\usage{
util_prepPcheckpoints(x, auxmetadata, pointsid = NULL)
}
\arguments{
\item{x}{Datapoints to be used for checking MCMC progress}

\item{auxmetadata}{auxmetadata object}

\item{pointsid}{Id of datapoints}
}
\value{
some arguments to be repeatedly used in util_Pcheckpoints
}
\description{
Used in 'util_Pcheckpoints()' within 'learn()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{util_qYXcont}
\alias{util_qYXcont}
\title{Calculate quantiles for continuous Y by bisection}
\usage{
util_qYXcont(
  iyx,
  params1,
  params2,
  auxmetadata,
  temporarydir,
  usememory = TRUE,
  doquantiles,
  quantiles,
  dosamples,
  nsamples,
  Qerror,
  tol = .Machine$double.eps * 3
)
}
\description{
Used in 'qPr()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prova-package.R
\docType{package}
\name{prova-package}
\alias{prova}
\alias{prova-package}
\title{prova: Nonparametric Probabilistic-Statistical Variate Analysis with Automated Markov-Chain Monte Carlo}
\description{
Calculate posterior joint and conditional probabilities, probability distributions of population frequencies, and information-theoretic measures, by means of Bayesian nonparametric methods. Data imputation is automatic and done in a principled way. Markov-chain Monte Carlo calculations are automatically handled and do not require user supervision. Applications range from statistical estimation and probabilistic hypothesis testing to evidence-based inference and decision making, in a wide range of disciplines from astrophysics to medicine. For more details and examples see for instance Porta Mana et al. (2026) \doi{10.31219/osf.io/8nr56}, Dunson & Bhattacharya (2011) \doi{10.1093/acprof:oso/9780199694587.003.0005}, Lindley & Novick (1981) \doi{10.1214/aos/1176345331}, Bernardo & Smith (2000) \doi{10.1002/9780470316870}, Müller et al. (2015) \doi{10.1007/978-3-319-18968-0}. Requires the packages 'Nimble', 'parallel', 'extraDistr'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://pglpm.github.io/prova/}
  \item \url{https://github.com/pglpm/prova/}
}

}
\author{
\strong{Maintainer}: PierGianLuca Porta Mana \email{pgl@portamana.org} (\href{https://orcid.org/0000-0002-6070-0784}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item PierGianLuca Porta Mana \email{pgl@portamana.org} (\href{https://orcid.org/0000-0002-6070-0784}{ORCID}) [copyright holder]
}

Other contributors:
\itemize{
  \item Aurora Grefsrud \email{agre@hvl.no} (\href{https://orcid.org/0009-0005-9063-4131}{ORCID}) [contributor]
  \item Håkon Mydland \email{haakon.mydland@gmail.com} (\href{https://orcid.org/0009-0004-9558-8894}{ORCID}) [contributor]
  \item Maksim Ohvrill \email{maksimohvrill@hotmail.com} [contributor]
  \item Simen Hesthamar Hauge \email{simen@hhnet.no} (\href{https://orcid.org/0009-0005-6158-2846}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_datacheck.R
\name{prova.data}
\alias{prova.data}
\alias{pwrite.csv}
\alias{pread.csv}
\title{Write and read CSV files in \strong{Prova}}
\usage{
pwrite.csv(x, file, ...)

pread.csv(file, ...)
}
\arguments{
\item{x}{The object to be written. Preferably a matrix or data frame; if not, it is attempted to coerce \code{x} to a \link[base:data.frame]{data frame}. See \code{\link[utils:write.csv]{utils::write.csv()}}.}

\item{file}{Either a character naming a file or a connection open for writing or reading. See \code{\link[utils:write.csv]{utils::write.csv()}} and \code{\link[utils:read.csv]{utils::read.csv()}}.}

\item{...}{Other arguments to be passed to \code{\link[utils:write.csv]{utils::write.csv()}} or \code{\link[utils:read.csv]{utils::read.csv()}}. Arguments 'row.names', 'quote', 'na', 'na.strings', 'tryLogical', 'sep', 'dec' are not allowed.}
}
\value{
\code{pread.csv} returns a \link[base:data.frame]{data frame} containing a representation of the data in the file; see \code{\link[utils:read.csv]{utils::read.csv()}}. \verb{pwrite.csv' returns }NULL` \link[base:invisible]{invisibly}.
}
\description{
Utility functions to read and write CSV files in the format required by \strong{Prova}
}
\details{
The functions \code{\link[=learn]{learn()}} and \code{\link[=metadatatemplate]{metadatatemplate()}} accept CSV files formatted as follows:
\itemize{
\item Decimal values should be separated by a \emph{dot}; no comma should be used to separate thousands etc. Example: \code{86342.75}.
\item Character and names should be quoted in single or double quotes. Example: \code{"female"}.
\item Values should be separated by \emph{commas}, not by tabs or semicolons.
\item Missing values should be simply \emph{empty}, not denoted by "NA", "missing", "-", or similar.
\item Preferably there should not be factors (see \link[base:factor]{base::factor}); use character names instead.
}

The utility functions \code{pwrite.csv()} and \code{pread.csv()} are wrappers to \code{\link[utils:write.csv]{utils::write.csv()}} and \code{\link[utils:read.csv]{utils::read.csv()}} that set appropriate default parameters according to the formatting rules above.
}
\examples{
## Save the 'penguins' dataset in a (temporary) file
filename <- tempfile(fileext = '.csv')

pwrite.csv(penguins, file = filename)

## check first few lines of the raw file
writeLines(readLines(filename, n = 10))

}
\seealso{
\code{\link[=metadatatemplate]{metadatatemplate()}} to help writing metadata files.

\code{\link[=learn]{learn()}}, which needs a metadata data-frame or CSV file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qPr.R
\name{qPr}
\alias{qPr}
\title{Calculate quantiles}
\usage{
qPr(
  p = c(0.25, 0.5, 0.75),
  Yname,
  X = NULL,
  learnt,
  tails = NULL,
  priorY = NULL,
  nsamples = "all",
  quantiles = c(0.055, 0.5, 0.945),
  parallel = TRUE,
  sep = ",",
  solidus = "|",
  verbose = FALSE,
  keepYX = TRUE,
  tol = .Machine$double.eps * 10
)
}
\arguments{
\item{p}{Numeric vector of probability levels. Default: \code{c(0.25, 0.5, 0.75)}.}

\item{Yname}{Character vector: name of variate whose quantiles will be computed.}

\item{X}{Matrix or data.table or \code{NULL} (default): set of values of variates on which we want to condition. If \code{NULL}, no conditioning is made (except for conditioning on the learning dataset and prior assumptions). One variate per column, one set of values per row.}

\item{learnt}{Either a character with the name of a directory or full path for a 'learnt.rds' object, produced by the \code{\link[=learn]{learn()}} function, or such an object itself.}

\item{tails}{Named vector or list, or \code{NULL} (default). The names must match some or all of the variates in arguments \code{X}. For variates in this list, the probability conditional is understood in a semi-open interval sense: \eqn{X \le x} or \eqn{X \ge x}, an so on. See analogous argument in \code{\link[=Pr]{Pr()}}.}

\item{priorY}{Reserved for use in future versions of the package.}

\item{nsamples}{Integer or \code{NULL} or \code{'all'} (default): desired number of samples of the variability of the quantile for \code{Y}. If \code{NULL}, no samples are reported. If \code{'all'} (or \code{Inf}), all samples obtained by the \code{\link[=learn]{learn()}} function are used.}

\item{quantiles}{Numeric vector, between 0 and 1, or \code{NULL}: desired quantiles of the variability of the quantile for \code{Y}. Default \code{c(0.055, 0.25, 0.75, 0.945)}, that is, the 5.5\%, 25\%, 75\%, 94.5\% quantiles (these are typical quantile values in the Bayesian literature: they give 50\% and 89\% credibility intervals, which correspond to 1 shannons and 0.5 shannons of uncertainty). If \code{NULL}, no quantiles are calculated.}

\item{parallel}{Logical or positive integer or cluster object. \code{TRUE} (default): use roughly half of available cores; \code{FALSE}: use serial computation; integer: use this many cores. It can also be a cluster object previously created with \code{\link[parallel:makeCluster]{parallel::makeCluster()}}; in this case the parallel computation will use this object.}

\item{sep}{character, default \code{','}: character to separate variate names and values}

\item{solidus}{character, default \code{'|'}: character prepended to names of the variates in the conditional (typically the \code{X} variates).}

\item{verbose}{Logical, default \code{FALSE}: give messages about parallel processing?}

\item{keepYX}{Logical, default \code{TRUE}: keep a copy of the \code{Y} and \code{X} arguments in the output? This is used for the plot method.}

\item{tol}{numeric positive: tolerance in the calculation of quantiles. Default: \code{.Machine$double.eps * 10} (typically \code{2.22045e-15}).}
}
\value{
A list of the following elements:
\itemize{
\item \code{values}: a matrix with the requested \eqn{Y}-quantiles \code{p} conditional on the requested \eqn{X}-values in \code{X}, for all combinations of \code{p} (rows) and \code{X} (columns).
\item \code{quantiles} (possibly \code{NULL}): an array with the variability quantiles (3rd dimension of the array) for the quantiles of the \code{value} element.
\item \code{samples} (possibly \code{NULL}): an array with the variability samples (3rd dimension of the array) for such quantiles.
\item \code{Y}, \code{X}: copies of the \code{Y} and \code{X} arguments.
}
}
\description{
This function calculates the quantiles of posterior probabilities and posterior conditional probabilities. It also outputs the variability of such quantiles if more training data were available.
}
\details{
This function calculates the quantiles of \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data})} or of \eqn{\mathrm{Pr}(Y = y \vert X \le x, \text{data})} or combinations thereof, at specified cumulative-probability levels. In other words, it calculates the values of \eqn{Y} having specified cumulative probabilities or conditional probabilities. It also calculates the variability of those quantiles if more learning data were provided. It is somewhat analogous to the \code{q}-variants of R distribution functions, such as \code{\link[stats:qnorm]{stats::qnorm()}}. The variability can be expressed in the form of quantiles, samples, or both, as in the \code{\link[=Pr]{Pr()}} function. If several joint values are given for the probability levels and for \code{X}, the function creates a 2D grid of results for all possible combinations of the given probability levels and \code{X} values. Each variate in the argument \code{X} can be specified either as a point-value \eqn{X = x} or as a left-open interval \eqn{X \le x} or as a right-open interval \eqn{X \ge x}, through the argument \code{tails}.
}
\examples{
### WARNING: the following examples, if run, might even take a minute or more.

\donttest{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## ## Example 1:
## Calculate the 5.5\%-, 50\%-, and 94.5\%-quantiles for the variate "bill lengt",
## that is, the values of "bill length" having such cumulative probabilities

quants <- qPr(
  Yname = 'bill_len',
  learnt = learnt, parallel = 1
)

## display the quantile values
quants$values

## verify these values using Pr():
probs <- Pr(
  Y = data.frame(bill_len = c(quants$values)),
  tails = list(bill_len = -1),
  learnt = learnt, parallel = 1
)

## the cumulative probabilities are indeed 0.055, 0.5, 0.945 within numerical error:
probs$values

## display the variability about the quantiles
quants$quantiles


## ## Example 2:
## Calculate the 5.5\%-, 50\%-, and 94.5\%-quantiles for the variate "bill lengt",
## for the subpopulation of species 'Adelie'

quants <- qPr(
  Yname = 'bill_len',
  X = data.frame(species = 'Adelie'),
  learnt = learnt, parallel = 1
)

## display the quantile values
quants$values

## verify these values using Pr():
probs <- Pr(
  Y = data.frame(bill_len = c(quants$values)),
  X = data.frame(species = 'Adelie'),
  tails = list(bill_len = -1),
  learnt = learnt, parallel = 1)

## the cumulative probabilities are indeed 0.055, 0.5, 0.945 within numerical error:
probs$values
}

}
\references{
\itemize{
\item Porta Mana (2025): \emph{What's special about 89\% credibility intervals?} \url{doi:10.5281/zenodo.17072199}.
}
}
\seealso{
\code{\link[=learn]{learn()}}, which generates the \code{learnt} objects required by \code{qPr()}.

\code{\link[=Pr]{Pr()}} to calculate joint and conditional probabilities.

\code{\link[=rPr]{rPr()}} to generate datapoints.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{util_qYXdiscr}
\alias{util_qYXdiscr}
\title{Calculate quantiles for discrete Y by bisection}
\usage{
util_qYXdiscr(
  iyx,
  params1,
  params2,
  auxmetadata,
  temporarydir,
  usememory = TRUE,
  doquantiles,
  quantiles,
  dosamples,
  nsamples,
  Qerror = NULL,
  tol = NULL
)
}
\description{
Used in 'qPr()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{rowcumsum}
\alias{rowcumsum}
\title{Cumulative sum along first dimension}
\usage{
rowcumsum(x)
}
\description{
Used in 'util_lprobsargsyx()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcmcstop.R
\name{funMCSELD}
\alias{funMCSELD}
\title{Calculate MC standard error using LaplacesDemon's batch means}
\usage{
funMCSELD(x)
}
\arguments{
\item{x}{matrix, each row being a "trace", that is a set of MC samples, whose MCSE is to be estimated.}
}
\value{
MCSE estimates, one for each trace. Division by sqrt(N) is already performed.
}
\description{
This function gives a good approximation of the "true" standard deviation in the case of independent samples. Multiply by \code{qnorm(x)} to obtain the \code{x}-quantile.
}
\details{
Modified from \url{https://github.com/LaplacesDemonR/LaplacesDemon/blob/master/R/ESS.R}.

Tested also on t-distributions with df=1.1 and Pareto with a=1.5 (mean exists, variance infinite).

\verb{sd() / sqrt(funESS3()} gives essentially identical results to \code{funMCSELD()}, but it's 20 times slower.

Used in 'util_combineYX()' in 'Pr()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_plotfunctions.R
\name{plotquantiles}
\alias{plotquantiles}
\title{Plot pairs of quantiles}
\usage{
plotquantiles(
  x,
  y,
  xdomain = NULL,
  alpha.f = 0.25,
  col = palette(),
  border = NA,
  type = "n",
  ...
)
}
\arguments{
\item{x}{Numeric or character: vector of x-coordinates. See \code{\link[=flexiplot]{flexiplot()}}.}

\item{y}{Numeric: a matrix having as many rows as \code{x} and an even number of columns, with one column per quantile. Typically these quantiles have been obtained with \code{\link[=Pr]{Pr()}}, as their \verb{$quantiles} value. This value is a three-dimensional array, and one of its columns (corresponding to the possible values of the \code{X} argument of \code{\link[=Pr]{Pr()}}) or one of its rows (corresponding to the possible values of the \code{Y} argument of \code{\link[=Pr]{Pr()}}) should be selected before being used as \code{y} input.}

\item{xdomain}{Character or numeric or \code{NULL} (default): vector of possible values of the variable represented in the x-axis, if the \code{x} argument is a character vector. The ordering of the values is respected. If \code{NULL}, then \code{unique(x)} is used.}

\item{alpha.f}{Numeric, default 0.25: opacity of the quantile bands, \code{0} being completely invisible and \code{1} completely opaque.}

\item{col}{Fill colour of the quantile bands. Can be specified in any of the usual ways, see for instance \code{\link[grDevices:col2rgb]{grDevices::col2rgb()}}. Default \verb{#4477AA}.}

\item{border}{Fill colour of the quantile bands. Can be specified in any of the usual ways, see for instance \code{\link[grDevices:col2rgb]{grDevices::col2rgb()}}. If \code{NA} (default), no border is drawn.}

\item{type}{see analogous argument in \code{\link[=flexiplot]{flexiplot()}}.}

\item{...}{Other parameters to be passed to \code{\link[=flexiplot]{flexiplot()}}.}
}
\value{
\code{NULL}, \link[base:invisible]{invisibly}; produces a plot, see \code{\link[graphics:matplot]{graphics::matplot()}}.
}
\description{
Utility function to plot pairs of quantiles obtained with \code{\link[=Pr]{Pr()}}.
}
\examples{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## create a grid of values for variate "bill length",
## based on the information in the dataset and metadata:
values <- vrtgrid(vrt = 'bill_len', learnt = learnt)

## calculate the probabilities and quantiles
probs <- Pr(Y = data.frame(bill_len = values), learnt = learnt, parallel = 1)

## plot the quantiles, setting lower plot range to zero
plotquantiles(x = values, y = probs$quantiles[, 1, ], ylim = c(0, NA),
  xlab = 'bill length', ylab = 'probability')

## add a plot of the probabilities in thick dashed red
flexiplot(x = values, y = probs$values, lwd = 5, lty = 2, col = 2, add = TRUE)

}
\seealso{
\code{\link[=Pr]{Pr()}} to calculate posterior probabilities and quantiles.

\code{\link[=plot.probability]{plot.probability()}} to directly plot posterior probabilities and quantiles contained in a probability object.

\code{\link[=flexiplot]{flexiplot()}} for more general plots.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_plotfunctions.R
\name{flexiplot}
\alias{flexiplot}
\title{Plot numeric or character values}
\usage{
flexiplot(
  x,
  y,
  type = NULL,
  lty = c(1, 2, 4, 3, 6, 5),
  lwd = 2,
  pch = c(1, 2, 0, 5, 6, 3),
  col = palette(),
  xlab = NULL,
  ylab = NULL,
  xlim = NULL,
  ylim = NULL,
  add = FALSE,
  xdomain = NULL,
  ydomain = NULL,
  alpha.f = 1,
  xjitter = NULL,
  yjitter = NULL,
  grid = TRUE,
  cex.main = 1,
  ...
)
}
\arguments{
\item{x}{Numeric or character: vector of x-coordinates. If missing, a numeric vector \code{1:...} is created having as many values as the rows of \code{y}.}

\item{y}{Numeric or character: vector of y coordinates. If missing, a numeric vector \code{1:...} is created having as many values as the rows of \code{x}.}

\item{type, lty, lwd, pch, col, xlab, ylab, add, cex.main}{see analogous arguments in \code{\link[graphics:matplot]{graphics::matplot()}}.}

\item{xlim, ylim}{\code{NULL} (default) or a vector of two values. In the latter case, if any of the two values is not finite (including \code{NA} or \code{NULL}), then the \code{min} or \code{max} \code{x}- or \code{y}-coordinates of the plotted points are used.}

\item{xdomain, ydomain}{Character or numeric or \code{NULL} (default): vector of possible values of the variables represented in the \code{x}- and \code{y}-axes, in case the \code{x} or \code{y} argument is a character vector. The ordering of the values is respected. If \code{NULL}, then \code{unique(x)} or \code{unique(y)} is used.}

\item{alpha.f}{Numeric, default 1: opacity of the colours, \code{0} being completely invisible and \code{1} completely opaque.}

\item{xjitter, yjitter}{Logical or \code{NULL} (default): add \code{\link[base:jitter]{base::jitter()}} to \code{x}- or \code{y}-values? Useful when plotting discrete variates. If \code{NULL}, jitter is added if the values are of character (or factor) class.}

\item{grid}{Logical: whether to plot a light grid. Default \code{TRUE}.}

\item{...}{Other parameters to be passed to \code{\link[graphics:matplot]{graphics::matplot()}}.}
}
\value{
\code{NULL}, \link[base:invisible]{invisibly}; produces a plot, see \code{\link[graphics:matplot]{graphics::matplot()}}.
}
\description{
Plot function that modifies and expands the \strong{graphics} package's \code{\link[graphics:matplot]{graphics::matplot()}} function in several ways.
}
\details{
This function is essentially a wrapper around \code{\link[graphics:matplot]{graphics::matplot()}}, augmenting the latter with some additional features useful for plotting data and results handled by \strong{Prova}. Some of the additional features provided by \code{flexiplot} are the following:
\itemize{
\item Either or both \code{x} and \code{y} arguments can be of class \code{\link[base:character]{base::character}}. In this case, axes labels corresponding to the unique values are used (see arguments \code{xdomain} and \code{ydomain}). This makes it easier to plot nominal and ordinal variates.
\item A jitter can also be added to the generated points, via the \code{xjitter} and \code{yjitter} switches. This makes it easier to generate scatter plots of nominal and ordinal variates.
\item It is possible to specify only a lower or upper limit in the \code{xlim} and \code{ylim} arguments, letting the other limit to be found automatically. This can be useful in plotting probabilities, in cases where we want to specify the lower, \code{0} limit, but want the upper limit to simply be the the maximum probability.
\item Transparency of lines or markers can be specified through argument \code{alpha.f}.
\item The plotting style is different, and default argument \code{type = 'l'} (line plot) rather than \code{type = 'p'} (point plot).
}

See the package's vignettes for more examples.
}
\examples{
## Scatter plot of the 'island' vs 'species' nominal variates of the penguins dataset;
## note how jitter is automatically added:
flexiplot(x = penguins[, 'species'], y = penguins[, 'island'])


## Scatter plot of the 'bill_len' vs 'species' variates of the penguins dataset:
flexiplot(x = penguins[, 'species'], y = penguins[, 'bill_len'])

## We can add jitter to separate the nominal values:
flexiplot(x = penguins[, 'species'], y = penguins[, 'bill_len'],
  xjitter = TRUE)


## Scatter plot of the 'bill_len' vs 'body_mass' variates;
## in this case we must specify the scatter-plot option `type = 'p'`:
flexiplot(x = penguins[, 'body_mass'], y = penguins[, 'bill_len'],
  type = 'p')

## Calculate the values of a normal distribution in a restricted range
x <- seq(from = -2, to = 2, length.out = 127)
y <- dnorm(x, mean = 0, sd = 1)

## plot the distribution, with 0 as the lower plot range:
flexiplot(x = x, y = y, ylim = c(0, NA))

}
\seealso{
\code{\link[=Pr]{Pr()}} to calculate posterior probabilities and quantiles.

\code{\link[=plot.probability]{plot.probability()}} to directly plot posterior probabilities and quantiles contained in a probability object.

\code{\link[=plotquantiles]{plotquantiles()}} to plot quantile ranges.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_buildauxmetadata.R
\name{buildauxmetadata}
\alias{buildauxmetadata}
\title{Build augmented metadata file}
\usage{
buildauxmetadata(data, metadata, Dthreshold = 1, tscalefactor = 4.266)
}
\arguments{
\item{data}{data.frame object}

\item{metadata}{data.frame object}

\item{Dthreshold}{Positive number: threshold of fraction
of unique datapoints to total datapoints, to decide
whether to treat a rounded variate as continuous}

\item{tscalefactor}{Positive number: scaling factor for variate conversion}
}
\value{
A \link[base:data.frame]{data frame} with auxmetadata.
}
\description{
Builds an augmented metadata \link[base:data.frame]{data frame} from the metadata and data given to 'learn()'. This augmented metadata object is saved in the 'learnt' object produced by 'learn()'.
}
\details{
In addition to the original metadata it contains info about transformed variates and their domains, estimated location- and scale-parameters, and similar metadata.

Used in 'learn()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{util_lprobsmi}
\alias{util_lprobsmi}
\title{Calculate pairs of log-probabilities for mutualinfo()}
\usage{
util_lprobsmi(xVs, params1, params2, lWnorm, lW)
}
\description{
Used in 'mutualinfo()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcmcstop.R
\name{funAC}
\alias{funAC}
\title{Compute autocovariance}
\usage{
funAC(y)
}
\arguments{
\item{y}{Time series}
}
\value{
Autocovariances at different lags
}
\description{
Modified from rstan \url{https://github.com/stan-dev/rstan/blob/develop/rstan/rstan/R/monitor.R}.
}
\details{
Used in 'funESS3()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{util_lprobsargsyx}
\alias{util_lprobsargsyx}
\title{Prepare arguments for util_lprobsyx from data}
\usage{
util_lprobsargsyx(x, auxmetadata, learnt, tails = NULL)
}
\description{
Used in 'Pr()', 'qPr()', 'rPr()', 'mutualinfo()'
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rPr.R
\name{rPr}
\alias{rPr}
\title{Generate datapoints}
\usage{
rPr(
  n,
  Ynames,
  X = NULL,
  learnt,
  tails = NULL,
  mcsamples = NULL,
  parallel = NULL
)
}
\arguments{
\item{n}{Positive integer: number of samples to draw.}

\item{Ynames}{Character vector: names of variates to draw jointly}

\item{X}{List or data.table or \code{NULL}: set of values of variates on which we want to condition the joint probability for \code{Y}. If \code{NULL} (default), no conditioning is made. Any rows beyond the first are discarded}

\item{learnt}{Either a character with the name of a directory or full path for a 'learnt.rds' object, produced by the \code{\link[=learn]{learn()}} function, or such an object itself.}

\item{tails}{Named vector or list, or \code{NULL} (default). The names must match some or all of the variates in arguments \code{X}. For variates in this list, the probability conditional is understood in an semi-open interval sense: \eqn{X \le x} or \eqn{X \ge x}, an so on. See analogous argument in \code{\link[=Pr]{Pr()}}.}

\item{mcsamples}{Vector of integers, or \code{'all'}, or \code{NULL} (default): which Monte Carlo samples calculated by the \code{\link[=learn]{learn()}} function should be used to draw the variate values. The default is to choose a random subset if \code{n} is smaller than their number, otherwise to recycle them as necessary.}

\item{parallel}{Not used: this function does not use parallelization.}
}
\value{
A \link[base:data.frame]{data frame} of joint draws of the variates \code{Ynames} from the posterior distribution, conditional on \code{X}. The row names of the data frame report the Monte Carlo sample (from \code{\link[=learn]{learn()}}) used for that draw, and the total number of draws from that sample so far.
}
\description{
This function generates datapoints of chosen joint variates, according to posterior probabilities and posterior conditional probabilities.
}
\details{
This function generates datapoints according to the posterior probability \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data})} or \eqn{\mathrm{Pr}(Y = y \vert X \le x, \text{data})} or combinations thereof, for the variates specified in the argument \code{Y}, and conditional on the variate values specified in the argument \code{X}. It is somewhat analogous to the \code{r}-variants of R distribution functions, such as \code{\link[stats:rnorm]{stats::rnorm()}}. If \code{X} is omitted or \code{NULL}, then the posterior probability \eqn{\mathrm{Pr}(Y | \text{data})} is used. Each variate in the argument \code{X} can be specified either as a point-value \eqn{X = x} or as a left-open interval \eqn{X \le x} or as a right-open interval \eqn{X \ge x}, through the argument \code{tails}.
}
\examples{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## ## Example 1:
## Generate 10 values of the 'species' variate,
## according to the frequency distribution estimated from the data

datapoints <- rPr(
  n = 10,
  Ynames = 'species',
  learnt = learnt
)

c(datapoints)


## ## Example 2:
## Generate 5 joint values of the 'species' and 'bill_len' variates.

datapoints <- rPr(
  n = 5,
  Ynames = c('species', 'bill_len'),
  learnt = learnt
)

print(datapoints, row.names = FALSE) ## row names give MCMC information


## ## Example 3:
## Generate 5 values of the 'species' variate,
## for the subpopulation of penguins having bill length shorter than 40 mm

datapoints <- rPr(
  n = 5,
  Ynames = 'species',
  X = data.frame(bill_len = 40),
  tails = list(bill_len = -1),
  learnt = learnt
)

c(datapoints)

}
\seealso{
\code{\link[=learn]{learn()}}, which generates the \code{learnt} objects required by \code{qPr()}.

\code{\link[=Pr]{Pr()}} to calculate joint and conditional probabilities.

\code{\link[=qPr]{qPr()}} to calculate quantiles.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/learn.R
\name{learn}
\alias{learn}
\title{Monte Carlo computation of posterior probability distribution}
\usage{
learn(
  data,
  metadata,
  auxdata = NULL,
  outputdir = NULL,
  nsamples = 3600,
  nchains = 8,
  nsamplesperchain = 450,
  parallel = TRUE,
  seed = NULL,
  cleanup = TRUE,
  appendinfo = TRUE,
  valueislearnt = TRUE,
  subsampledata = NULL,
  prior = missing(data) || is.null(data),
  startupMCiterations = 3600,
  minMCiterations = 0,
  maxMCiterations = +Inf,
  maxhours = +Inf,
  ncheckpoints = 12,
  maxrelMCSE = +Inf,
  minESS = 450,
  initES = 2,
  thinning = NULL,
  verbose = TRUE,
  plottraces = !cleanup,
  showKtraces = FALSE,
  showAlphatraces = FALSE,
  hyperparams = list(ncomponents = 64, minalpha = -4, maxalpha = 4, byalpha = 1, Rshapelo
    = 0.5, Rshapehi = 0.5, Rvarm1 = 3^2, Cshapelo = 0.5, Cshapehi = 0.5, Cvarm1 = 3^2,
    Dshapelo = 0.5, Dshapehi = 0.5, Dvarm1 = 3^2, Bshapelo = 1, Bshapehi = 1, Dthreshold
    = 1, tscalefactor = 4.266, Oprior = "Hadamard", Nprior = "Hadamard", avoidzeroW =
    NULL, initmethod = "datacentre", Qerror = pnorm(c(-1, 1)))
)
}
\arguments{
\item{data}{A dataset, given as a \code{\link[base:data.frame]{base::data.frame()}} or as a file path to a CSV file. If missing or \code{NULL}, then the prior probability distribution is calculated.}

\item{metadata}{\link{metadata} about the dataset's variates, given either as a \link[base:data.frame]{data frame} or as a file path to a CSV file.}

\item{auxdata}{A larger dataset, given as a data frame or as a file path to a CSV file. Such a dataset would be too large to use in the Monte Carlo sampling, but can still be used to help estimate some hyperparameters.}

\item{outputdir}{\code{NULL} (default) or \code{NA} or character: path to folder where output information and diagnostics should be saved. If \code{NULL}, a directory is created in the temporary-directory space given by \code{\link[base:tempdir]{base::tempdir()}}. If \code{NA}, a directory is created in the current working directory given by \code{\link[base:getwd]{base::getwd()}}. If character, this is taken to be the output directory; it should of course be writable by the user.}

\item{nsamples}{Integer, default 3600: number of desired, \emph{approximately independent} Monte Carlo samples. If this argument is changed, the user is also required to explicitly give either \code{nchains} or \code{nsamplesperchain}, but not both; the remaining third argument is determined from \eqn{\mathrm{nsamples} = \mathrm{nchains} \times \mathrm{nsamplesperchain}}.}

\item{nchains}{Integer, default 8: number of Monte Carlo chains. If this argument is changed, the user is also required to explicitly give either \code{nsamples} or \code{nsamplesperchain}, but not both; the remaining third argument is determined from \eqn{\mathrm{nsamples} = \mathrm{nchains} \times \mathrm{nsamplesperchain}}.}

\item{nsamplesperchain}{Integer, default 450: number of \emph{approximately independent} Monte Carlo samples per chain. If this argument is changed, the user is also required to explicitly give either \code{nsamples} or \code{nchains}, but not both; the remaining third argument is determined from \eqn{\mathrm{nsamples} = \mathrm{nchains} \times \mathrm{nsamplesperchain}}.}

\item{parallel}{Logical or positive integer or cluster object. \code{TRUE} (default): use roughly half of available cores; \code{FALSE} (default): use serial computation; integer: use this many cores. It can also be a cluster object previously created with \code{\link[parallel:makeCluster]{parallel::makeCluster()}}; in this case the parallel computation will use this object.}

\item{seed}{Integer or \code{NULL} (default): use this seed for the random number generator. If \code{NULL}, do not set the seed.}

\item{cleanup}{Logical, default \code{TRUE}: remove diagnostic files at the end of the computation?}

\item{appendinfo}{Logical, default \code{TRUE}: append information about number of variates ('V'), number of data points ('D'), number of Monte Carlo samples ('S'), and timestamp, to the name of the output directory \code{outputdir}? The appended string has the format 'Vn_Dn_Sn_YYMMDDTHHMMSS'.}

\item{valueislearnt}{Logical or \code{NULL}: should the \code{VALUE} returned be the \code{learnt} object containing the results from the Monte Carlo computation? Default \code{TRUE}. If \code{FALSE}, then \code{VALUE} is the output directory name. If \code{NULL}, then \code{VALUE} is \code{NULL}.}

\item{subsampledata}{Integer or \code{NULL} (default): if integer, use only that many datapoints from the original dataset in the \code{data} argument.}

\item{prior}{Logical: Calculate the prior distribution? Default is \code{FALSE} unless \code{data} argument is missing or \code{NULL}.}

\item{startupMCiterations}{Integer, default 3600: number of initial Monte Carlo iterations.}

\item{minMCiterations}{Integer, default 0: minimum number of Monte Carlo iterations to be doneby a chain.}

\item{maxMCiterations}{Integer, default \code{Inf}: Do at most this many Monte Carlo iterations per chain.}

\item{maxhours}{Numeric, default \code{Inf}: approximate time limit, in hours, for the Monte Carlo computation to last.}

\item{ncheckpoints}{Integer or \code{NULL}, default 12: number of datapoints (per chain) to use for checking when the Monte Carlo computation should end. If \code{NULL}, this is equal to number of variates + 2. If Inf, use all datapoints.}

\item{maxrelMCSE}{Numeric positive, default \code{+Inf}: desired maximal \emph{relative Monte Carlo Standard Error} of calculated probabilities with respect to their variability with new data. The default \code{+Inf} means that \code{minESS} is used instead. \code{maxrelMCSE} is related to \code{minESS} by \eqn{\mathrm{maxrelMCSE} = 1/\sqrt{\mathrm{minESS} + \mathrm{initES}}}.}

\item{minESS}{Numeric positive or \code{NULL}, default 450: desired minimal Monte Carlo \emph{Expected Sample Size}. If \code{NULL}, it is equal to the final \code{nsamplesperchain}. \code{minESS} is related to \code{maxrelMCSE} by \eqn{\mathrm{minESS} = 1/\mathrm{maxrelMCSE}^2 - \mathrm{initES}}.}

\item{initES}{Numeric positive, default 2: number of initial "burn-in" samples, separated by the Expected Sample Size, to be discarded. Note that the Monte Carlo chain typically starts in a high-probability region, so there is no reason to discard many initial samples.}

\item{thinning}{Integer or \code{NULL} (default): thin out the Monte Carlo samples by this value. If \code{NULL}: let the diagnostics decide the thinning value.}

\item{verbose}{Logical, default \code{TRUE}: output the progress to terminal? If \code{FALSE}, the progress is outputted to the file \code{'main.log'} in the \code{outputdir} directory.}

\item{plottraces}{Logical, default \code{TRUE}: save plots of the Monte Carlo traces of diagnostic values?}

\item{showKtraces}{Logical, default \code{FALSE}: save plots of the Monte Carlo traces of the K parameter?}

\item{showAlphatraces}{Logical, default \code{FALSE}: save plots of the Monte Carlo traces of the Alpha parameter?}

\item{hyperparams}{List: hyperparameters of the hyperprior; see values in "Usage".}
}
\value{
A "learnt" object, or name of directory containing such an object and other output files, or \code{NULL}, depending on argument \code{valueislearnt}.

\code{learn()} saves several files in a directory. By default this output directory is a temporary directory within the one used by \code{\link[base:tempdir]{base::tempdir()}}, but an alternative one can be chosen with the argument \verb{outputdir =}. The output directory contain several diagnostic files for the Monte Carlo computation; in particular:
\itemize{
\item \code{MCtraces.pdf}: shows several trace plots of the Monte Carlo sampling; the correspondin data are in the file \code{MCtraces.rds}.
\item \code{plotsamples_learnt.pdf}, \code{plotquantiles_learnt.pdf}: show the marginal posterior distributions of each individual variate, together with their variability (as samples or quantiles).
\item \verb{log-1.log}, \verb{log-2.log}, ... one for each parallel core; report the progress of each parallel Monte Carlo computation and notes about it.
\item \code{rng_seed.rds}: the state of the pseudorandom seed (see \link[base:Random]{base::Random}) when \code{learn()} was called.
\item \code{metadata.csv}: a copy of the metadata.
}

It is recommended that you give an explicit argument \verb{outputdir =} and save the directory with the files above for future reference. In particular, the \code{MCtraces.pdf} plot and \code{MCtraces.rds} data can be useful to report Monte Carlo convergence in any work of yours that used \strong{Prova}.
}
\description{
Compute the posterior joint probability distribution of the variates conditional on the given data, by means of Markov-chain Monte Carlo, using the package \strong{Nimble}.
}
\details{
This function takes as main inputs a set of data and metadata, and computes the full joint probability distribution for new data, including its variability. From this full joint distribution any other distributions of interest can subsequently be computed; see \code{\link[=Pr]{Pr()}} and related functions. This computation can also be interpreted as an estimation of the full joint frequency distribution of the variates in the \emph{whole population}, beyond the sample data, together with its uncertainty. The computation allows for the use of datapoints with partially missing variables: imputation is automatically made. This imputation is \emph{principled}, made according to the rules of probability theory.

The output is a "learnt" object, typically saved in a \code{learnt.rds} file, which is used in all subsequent probabilistic computations. Other information about the computation is provided in logs and plots, saved in a directory specified by the user.

See \code{vignette('intro')} for introductory examples.

The computation is "non-parametric": probability or frequency distributions are not assumed to be Gaussian or of any other specific shape; no "model" is assumed. The mathematical representation of the space of joint frequency distributions follows ideas of Dunson & Bhattacharya (2011); see \href{https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf}{technical manual} for details.

The computation is done via Markov-chain Monte Carlo, using the package \href{https://cran.r-project.org/package=nimble}{\strong{Nimble}}. "Convergence" of the Monte Carlo computation is automatically assessed with methods described in Vehtari & al. (2021) and Kwon & al. (2025); see \href{https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf}{technical manual} for details. The default values for convergence require that all of the following three conditions be fulilled:
\itemize{
\item The computation's numerical error (Monte-Carlo Standard Error) for the posterior probability must be smaller than 4.7\% of the standard deviation of the posterior's variability.
\item The computation's numerical error for the 0.055- and 0.945-percentiles of the posterior's variability should be smaller than 4.7\% of the distance between them.
}

Typically this requirement leads to final results obtained with the \code{\link[=Pr]{Pr()}} function having at least two significant digits.

The \code{learn()} function can take hours or even days to perform its computations, depending on the size of the dataset, number of variates, and the (initially unknown) "shape" of the underlying probability distribution. For this reason it is typically called within an R script, executed via \link[utils:Rscript]{utils::Rscript}. For example, a script 'myscript.R' could have the following structure:

\if{html}{\out{<div class="sourceCode">}}\preformatted{library('prova')

learn(
  data = 'filename_with_data.csv', # CSV file containing the dataset
  metadata = 'filename_with_metadata.csv', # CSV file containing the metadata
  outputdir = 'some_directory', # path to output directory
  parallel = 8 # machine has more than 8 cores, so we use 8
  ## possibly other arguments to learn()
)
}\if{html}{\out{</div>}}

and then be called on a bash terminal with

\if{html}{\out{<div class="sourceCode">}}\preformatted{$ Rscript myscript.R > learnoutput.log 2>&1 &
}\if{html}{\out{</div>}}

with such a call, the file 'learnoutput.log' will contain information about how the computation is proceeding and the estimated end time.
}
\examples{
\donttest{
### WARNING: the following example, if run, might even take a minute or more.

## Create dataset with 3 points of variate 'V' for demonstration:
dataset <- data.frame(V = rnorm(n = 3))

## Create metadata file:
metadata <- data.frame(name = 'V', type = 'continuous')

## Learn from the data:
learnt <- learn(
  data = dataset, metadata = metadata,
      ## the following parameters are unrealistic
      ## only used to reduce computation time for this example
  nsamples = 10, nchains = 1,
  startupMCiterations = 10, maxMCiterations = 10,
  minESS = 0, initES = 0
)

## Check structure of `learnt` object:
str(learnt)
}

}
\references{
For the mathematical representation of the frequency space:
\itemize{
\item Dunson, Bhattacharya (2011): \emph{Nonparametric Bayes regression and classification through mixtures of product kernels} \url{doi:10.1093/acprof:oso/9780199694587.003.0005}.
\item Ishwaran, Zarepour (2002): \emph{Exact and approximate sum representations for the Dirichlet process} \url{doi:10.2307/3315951}.
\item Porta Mana \url{https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf}.
}

About Bayesian inference under exchangeability ("population inference"):
\itemize{
\item Lindley, Novick (1981): \emph{The role of exchangeability in inference}, \url{doi:10.1214/aos/1176345331}.
\item Bernardo, Smith (2000): \emph{Bayesian Theory}. Wiley \url{doi:10.1002/9780470316870}.
\item Porta Mana \url{https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf}.
}

About nonparametrics:
\itemize{
\item Müller et al. (2015): \emph{Nonparametric Bayesian inference}. IMS \url{doi:10.1007/978-3-319-18968-0}.
\item Hjort et al. (2010): \emph{Bayesian Nonparametrics}. Cambridge University Press \url{doi:10.1017/CBO9780511802478}.
}

About Markov-chain Monte Carlo and "convergence":
\itemize{
\item de Valpine, Paciorek, Turek, & al. (2026): \emph{NIMBLE: MCMC, Particle Filtering, and Programmable Hierarchical Modeling} \url{doi:10.5281/zenodo.1211190}, \url{https://cran.r-project.org/package=nimble}.
\item Kwon & al. (2025): \emph{MCMC stopping rules in latent variable modelling} \url{doi:10.1111/bmsp.12357}.
\item Vehtari & al. (2021): \emph{Rank-normalization, folding, and localization: an improved R-hat for assessing convergence of MCMC} \url{doi:10.1214/20-BA1221}.
\item Roy (2020): \emph{Convergence diagnostics for Markov chain Monte Carlo} \url{doi:10.1146/annurev-statistics-031219-041300}.
\item Gilks & al. (1998): \emph{Markov Chain Monte Carlo in Practice}. Chapman & Hall/CRC \url{doi:10.1201/b14835}.
\item D. J. C. MacKay (2005): \emph{Information Theory, Inference, and Learning Algorithms}. Cambridge University Press \url{https://www.inference.org.uk/itila/book.html}.
\item Porta Mana \url{https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf}.
}
}
\seealso{
\code{\link[=metadatatemplate]{metadatatemplate()}} to help writing metadata files.

\code{\link[=Pr]{Pr()}} to calculate probabilities, and \code{\link[=qPr]{qPr()}} to calculate quantiles, given the data processed by \code{learn()}.

\code{\link[=rPr]{rPr()}} to generate datapoints similar to the data processed by \code{learn()}.

\code{\link[=mutualinfo]{mutualinfo()}} to calculate mutual information given the data processed by \code{learn()}.

\code{\link[=pread.csv]{pread.csv()}} and \code{\link[=pwrite.csv]{pwrite.csv()}} to read and write CSV files in the format used by \code{learn()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{metadataExample}
\alias{metadataExample}
\title{Example metadata file}
\format{
\subsection{\code{metadataExample}}{

A \link[base:data.frame]{data frame} with 2 rows and 10 columns.
}
}
\usage{
metadataExample
}
\value{
No return value.
}
\description{
A \link[base:data.frame]{data frame} containing the prior information about the variates \code{species} and \code{bill_len} of the \link[datasets:penguins]{datasets::penguins} dataset.
}
\seealso{
\code{\link[=metadatatemplate]{metadatatemplate()}} which helps producing this kind of metadata files from a given dataset.

\code{\link[=learn]{learn()}} which needs this kind of metadata files to "learn" from data.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vrtgrid.R
\name{vrtgrid}
\alias{vrtgrid}
\title{Create a grid of values for a variate}
\usage{
vrtgrid(vrt, learnt, length.out = 129)
}
\arguments{
\item{vrt}{Character: name of the variate, must match one of the names in the \code{metadata} file provided to the \code{\link[=learn]{learn()}} function.}

\item{learnt}{Either a character with the name of a directory or full path for a 'learnt.rds' object, produced by the \code{\link[=learn]{learn()}} function, or such an object itself.}

\item{length.out}{Numeric, positive (default 129): number of values to be created; used only for continuous, non-rounded variates (see \code{\link{metadata}}).}
}
\value{
A numeric or character vector of values.
}
\description{
This function creates a set of values for a variate, based on the information from data and metadata stored in a \code{learnt} object, created by the \code{\link[=learn]{learn()}} function. The set of values depends on the type of variate (nominal or continuous, rounded, and so on, see \link{metadata}). The range of values is chosen to include, and extend slightly beyond, the range observed in the data used in the \code{\link[=learn]{learn()}} function. Variate domains are always respected.
}
\examples{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## set of values for the variate "species";
## since this variate is of a nominal kind, all values are included
valuesSpecies <- vrtgrid(vrt = 'species', learnt = learnt)

print(valuesSpecies)

## create a set of values for the variate "bill length";
## this variate is continuous and rounded, only realistic values are included
valuesBill <- vrtgrid(vrt = 'bill_len', learnt = learnt)

range(valuesBill)

## let's take a subset of these values, to speed up computation
valuesBill <- valuesBill[seq(to = length(valuesBill), length.out = 65)]

## calculate the conditional probabilities for the 'bill_len' values above,
## given the values of 'species'
probs <- Pr(
  Y = data.frame(bill_len = valuesBill),
  X = data.frame(species = valuesSpecies),
  learnt = learnt, parallel = 1
)

## plot the conditional probability distributions, and their variability
plot(probs)

}
\seealso{
\code{\link[=learn]{learn()}}, which generates the \code{learnt} objects required by \code{vrtgrid()}.

\code{\link[=Pr]{Pr()}} to calculate probabilities and their variability.

\code{\link[=plot.probability]{plot.probability()}} to plot probabilities and quantiles calculated by \code{Pr()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_cleanup.R
\name{util_cleanup}
\alias{util_cleanup}
\title{Cleanup a learn()-output directory}
\usage{
util_cleanup(path)
}
\value{
No return value; called for side effects.
}
\description{
Cleanup a learn()-output directory
}
\details{
For deeper monitoring of the MCMC, the user can require the 'learn()' function not to clean intermediate MCMC-related files generated during the computation.

The present function can be used to remove these intermediate files from the output directory created by 'learn()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/learn.R
\name{workerfun}
\alias{workerfun}
\title{Worker function called by learn()}
\usage{
workerfun(
  acore,
  dirname,
  dashnameroot,
  avoidzeroW,
  initmethod,
  constants,
  datapoints,
  vn,
  showAlphatraces,
  Alphatoslice,
  Ktoslice,
  RWtoslice,
  changeSamplerOrder,
  minchainspercore,
  coreswithextrachain,
  nchains,
  maxhours,
  timestart0,
  showsamplertimes,
  startupMCiterations,
  maxMCiterations,
  showKtraces,
  ncomponents,
  plottraces,
  Qlo,
  Qhi,
  Qerror,
  minESS,
  initES,
  nsamplesperchain,
  minMCiterations,
  printtimediff,
  family,
  mainlog,
  verbose
)
}
\description{
This worker function is defined outside of learn() in order to avoid import of spurious objects into the parallel workers, and from the parallel workes into the main R session, with waste of memory.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_plotFsamples.R
\name{plotFsamples}
\alias{plotFsamples}
\title{Plot one-dimensional posterior probabilities}
\usage{
plotFsamples(
  filename,
  learnt,
  data,
  plotprobability = TRUE,
  plotvariability = "samples",
  nFsamples = NULL,
  datahistogram = !(missing(data) || is.null(data)),
  datascatter = !(missing(data) || is.null(data)),
  parallel = TRUE
)
}
\arguments{
\item{filename}{Character: name of plot output file}

\item{learnt}{Either a character with the name of a directory or full
path for an 'learnt.rds' object, or such an object itself}

\item{data}{data.table object or filepath: datapoints}

\item{plotprobability}{Logical: plot the resulting probability curve}

\item{plotvariability}{Character, either 'samples' or 'quantiles':
how to plot the variability of the probability distribution with new samples}

\item{nFsamples}{Positive number: if plotvariability='samples', then
number of samples of representative frequency distributions to display
as variability;
if plotvariability='quantiles', then
the quantiles (in range 0 to 0.5) to show}

\item{datahistogram}{Logical: plot the data as histogram?}

\item{datascatter}{Logical: plot the data as scatterplot along the x-axis?}

\item{parallel}{Logical or numeric: whether to use pre-existing parallel
workers, or how many to create and use}
}
\value{
No return value.
}
\description{
Used in 'learn()' to plot diagnostics.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_plotfunctions.R
\name{plot.probability}
\alias{plot.probability}
\title{Plot an object of class "probability"}
\usage{
\method{plot}{probability}(
  x,
  variability = NULL,
  subset = NULL,
  PvsY = NULL,
  legend = "top",
  lty = c(1, 2, 4, 3, 6, 5),
  lwd = 2,
  col = palette(),
  type = NULL,
  alpha.f = 1,
  var.alpha.f = NULL,
  xlab = NULL,
  ylab = NULL,
  main = NULL,
  ylim = c(0, NA),
  grid = TRUE,
  add = FALSE,
  ...
)
}
\arguments{
\item{x}{Object of class "probability", obtained with \code{\link[=Pr]{Pr()}}.}

\item{variability}{One of the values \code{'quantiles'}, \code{'samples'}, \code{'none'} (equivalent to \code{NA} or \code{FALSE}), or \code{NULL} (default), in which case the variability available in \code{p} is used. This argument chooses how to represent the variability of the probability; see \code{\link[=Pr]{Pr()}}. If the requested variability is not available in the object \code{p}, then a warning is issued and no variability is plotted.}

\item{subset}{Named list or named vector: which variate values to display. For the variates corresponding to the names in this list, only the vector of values corresponding to that variate is displayed.}

\item{PvsY}{Logical or \code{NULL}: should probabilities be plotted against their \code{Y} argument? If \code{NULL}, the argument between \code{Y} and \code{X} having larger number of values is chosen. As many probability curves will be plotted as the number of values of the other argument.}

\item{legend}{One of the values \code{'bottomright'}, \code{'bottom'}, \code{'bottomleft'}, \code{'left'}, \code{'topleft'}, \code{'top'}, \code{'topright'}, \code{'right'}, \code{'center'} (see \code{\link[graphics:legend]{graphics::legend()}}): plot a legend at that position. A value \code{FALSE} or any other does not plot any legend. Default \code{'top'}.}

\item{lty, lwd, col, type, xlab, ylab, main, ylim, grid, add}{see analogous arguments in \code{\link[graphics:matplot]{graphics::matplot()}}}

\item{alpha.f}{Numeric, default 0.25: opacity of the colours, \code{0} being completely invisible and \code{1} completely opaque.}

\item{var.alpha.f}{Numeric: opacity of the quantile bands or of the samples, \code{0} being completely invisible and \code{1} completely opaque.}

\item{...}{Other parameters to be passed to \code{\link[=flexiplot]{flexiplot()}}.}
}
\value{
\code{NULL}, \link[base:invisible]{invisibly}; produces a plot, see \code{\link[graphics:matplot]{graphics::matplot()}}.
}
\description{
This \code{\link[base:plot]{base::plot()}} method is a utility to plot probabilities obtained with \code{\link[=Pr]{Pr()}}, as well as their variabilities. The probabilities are plotted either against \code{Y}, with one curve for each value of \code{X}, or vice versa.
}
\examples{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## create a grid of values for variate "bill length",
## based on the information in the dataset and metadata:
values <- vrtgrid(vrt = 'bill_len', learnt = learnt)

## calculate the probabilities and quantiles
probs <- Pr(Y = data.frame(bill_len = values), learnt = learnt, parallel = 1)

## plot the probabilities and quantiles
plot(probs)

}
\seealso{
\code{\link[=Pr]{Pr()}} to calculate posterior probabilities and quantiles.

\code{\link[=hist.probability]{hist.probability()}} to plot the variability of the probabilities as a distribution.

\code{\link[=flexiplot]{flexiplot()}} (on which \code{plot.probability()} is based) for more general plots.

\code{\link[=plotquantiles]{plotquantiles()}} to plot quantile ranges.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{util_denorm}
\alias{util_denorm}
\title{Utility function to avoid finite-precision accuracys}
\usage{
util_denorm(lprob)
}
\description{
Used in 'rPr()', 'mutualinfo()', 'util_lprobsmi()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_plotfunctions.R
\name{hist.probability}
\alias{hist.probability}
\title{Plot the variability of an object of class "probability" as a histogram}
\usage{
\method{hist}{probability}(
  x,
  subset = NULL,
  breaks = NULL,
  legend = "top",
  lty = c(1, 2, 4, 3, 6, 5),
  lwd = 2,
  col = palette(),
  alpha.f = 1,
  fill.alpha.f = 0.125,
  showmean = TRUE,
  xlab = NULL,
  ylab = NULL,
  xlim = NULL,
  ylim = c(0, NA),
  main = NULL,
  grid = TRUE,
  add = FALSE,
  ...
)
}
\arguments{
\item{x}{Object of class "probability", obtained with \code{\link[=Pr]{Pr()}}.}

\item{subset}{Named list or named vector: which variate values to display. For the variates corresponding to the names in this list, only the vector of values corresponding to that variate is displayed.}

\item{breaks}{\code{NULL} or as in function \code{\link[graphics:hist]{graphics::hist()}}. If \code{NULL} (default), an optimal number of breaks for each probability distribution is computed.}

\item{legend}{One of the values \code{"bottomright"}, \code{"bottom"}, \code{"bottomleft"}, \code{"left"}, \code{"topleft"}, \code{"top"}, \code{"topright"}, \code{"right"}, \code{"center"} (see \code{\link[graphics:legend]{graphics::legend()}}): plot a legend at that position. A value \code{FALSE} or any other does not plot any legend. Default \code{"top"}.}

\item{lty, lwd, col, alpha.f, xlab, ylab, xlim, ylim, main, grid, add}{see analogous arguments in \code{\link[graphics:matplot]{graphics::matplot()}}}

\item{fill.alpha.f}{Numeric, default 0.125: opacity of the histogram filling. \code{0} means no filling.}

\item{showmean}{Logical, default \code{TRUE}: show the means of the probability distributions? The means correspond to the probabilities about the next observed unit.}

\item{...}{Other parameters to be passed to \code{\link[=flexiplot]{flexiplot()}}.}
}
\value{
\link[base:invisible]{Invisibly}, an object of class \link[graphics:hist]{"histogram"}.
}
\description{
The posterior probabilities calculated with the \code{\link[=Pr]{Pr()}} function, and outputted as a \code{probability} object, have an associated variability that comes from the finite size of the data sample. This variability can be interpreted in two ways:
\itemize{
\item How the probabilities would change, if we could collect a very large (infinite) amount of additional data, and how likely would such change be;
\item The relative frequency of a particular variate value in the full (sampled and unsampled) population is unknown; we can quantify our uncertainty about this relative frequency with a probability distribution.
}

The \code{hist()} method for a \code{probability} object is a utility to visualize this kind of variability, in the form of a distribution.
}
\examples{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## calculate the probability, and its variability,
## for the value 'Adelie' of the "species" variate
probs <- Pr(Y = data.frame(species = 'Adelie'), learnt = learnt, parallel = 1)
probs$values

## show the variability of this probability; equivalently show
## the probability distribution for the relative frequency of
## 'Adelie' penguins in the full population
hist(probs, legend = 'topright')

}
\seealso{
\code{\link[=Pr]{Pr()}} to calculate posterior probabilities and quantiles.

\code{\link[=plot.probability]{plot.probability()}} to plot the posterior probabilities.

\code{\link[=flexiplot]{flexiplot()}} (on which \code{hist.probability()} is based) for more general plots.

\code{\link[=plotquantiles]{plotquantiles()}} to plot quantile ranges.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcmcstop.R
\name{funESS3}
\alias{funESS3}
\title{Compute ESS}
\usage{
funESS3(x)
}
\arguments{
\item{x}{Vector of MC samples.}
}
\value{
Effective Sample Size.
}
\description{
Modified from 'rstan' \url{https://github.com/stan-dev/rstan/blob/develop/rstan/rstan/R/monitor.R}
}
\details{
Used in 'workerfun()' in 'learn()', and in 'funMCEQ()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_prsubset.R
\name{prsubset}
\alias{prsubset}
\title{Subset variates of an object of class "probability"}
\usage{
prsubset(x, subset)
}
\arguments{
\item{x}{Object of class "probability", obtained with \code{\link[=Pr]{Pr()}}.}

\item{subset}{Named list or named vector: variates to subset, given as list names, and corresponding values to subset.}
}
\value{
An object of class "probability", identical to the original object \code{x} except for a reduced range of values in some if its variates.
}
\description{
An object of class "probability", obtained with the \code{\link[=Pr]{Pr()}} function, holds the probabilities for all possible combinations of values of a set of joint variates \code{Y} conditional on a set of joint variates \code{X}, together with the variabilities of these probabilities and some other information. In some cases one may wish to exclude some of the values of the \code{Y} or \code{X} variates. For instance \code{Y} in the probability-class object could include the variate "age" with values from 18 to 100, and one may want to retain the values from 60 to 80.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcmcstop.R
\name{fftNGS}
\alias{fftNGS}
\title{Find optimal FFT size}
\usage{
fftNGS(N)
}
\arguments{
\item{N}{Integer.}
}
\value{
Optimal FFT size
}
\description{
Modified from rstan \url{https://github.com/stan-dev/rstan/blob/develop/rstan/rstan/R/monitor.R}.
}
\details{
Used in 'funAC()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_plotfunctions.R
\name{print.probability}
\alias{print.probability}
\title{Print an object of class "probability"}
\usage{
\method{print}{probability}(x, elements = NULL, subset = NULL, digits = TRUE, ...)
}
\arguments{
\item{x}{Object of class "probability", obtained with \code{\link[=Pr]{Pr()}}.}

\item{elements}{character or integer vector, or \code{NULL} (default): elements of the "probability" object to display. The syntax is the same as with \code{\link[base:Extract]{[}}. If \code{NULL}, the elements \verb{$values} and \verb{$quantiles} are displayed together in a special way.}

\item{subset}{Named list or named vector: which variate values to display. For the variates corresponding to the names in this list, only the vector of values corresponding to that variate is displayed.}

\item{digits}{positive number or \code{NULL} or \code{TRUE} (default): minimal number of significant digits, see \code{\link[base:print.default]{base::print.default()}}. If value is \code{TRUE}, then the significant digits for elements \verb{$values} and \verb{$quantiles} are determined from their respective \verb{$values.MCaccuracy} and \verb{$quantiles.MCaccuracy} elements of the \code{probability} object, see \code{\link[=Pr]{Pr()}}; whereas \verb{$samples} elements use 2 significant digits.}

\item{...}{Other parameters to be passed to \code{\link[base:print]{base::print()}}.}
}
\value{
Its \code{x} argument, \link[base:invisible]{invisibly}; see \code{\link[base:print]{base::print()}}.
}
\description{
This \code{\link[base:print]{base::print()}} method is a utility to display selected elements of a "probability" object obtained with \code{\link[=Pr]{Pr()}}; typically its posterior probabilies (element \verb{$values}) and their variabilities (element \verb{$quantiles}). If the \code{Y} or \code{X} variates are joint variates, this method also allow to display only selected values of them
}
\examples{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## Calculate the 3 x 2 probabilities for the 3 species
## given bill-lengths of 43 mm and 44 mm

Y <- data.frame(species = c('Adelie', 'Chinstrap', 'Gentoo'))
X <- data.frame(bill_len = c(43, 44))

probs <- Pr(Y = Y, X = X, learnt = learnt, parallel = 1)

## display the values and variabilities of these probabilities
print(probs)

## diplay 'values' only, and only for the species value 'Gentoo'
print(probs, elements = 'values', subset = list(species = 'Gentoo'))

}
\seealso{
\code{\link[=Pr]{Pr()}} to calculate posterior probabilities and quantiles.

\code{\link[=plot.probability]{plot.probability()}} to plot probabilities and quantiles calculated by `Pr()'.
\code{\link[=hist.probability]{hist.probability()}} to plot the variability of the probabilities as a distribution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{rowinvcumsum}
\alias{rowinvcumsum}
\title{Inverse cumulative sum along first dimension}
\usage{
rowinvcumsum(x)
}
\description{
Used in 'util_lprobsargsyx()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcsamples.R
\name{mcsubset}
\alias{mcsubset}
\title{Eliminate samples from mcsamples object}
\usage{
mcsubset(learnt, subsamples)
}
\description{
Used in 'learn()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{learntExample}
\alias{learntExample}
\title{Example \code{learnt} object produced by learn()}
\format{
\subsection{\code{learntExample}}{

A list containing results from Markov-chain Monte Carlo computation, including diagnostics and variate metadata.
}
}
\usage{
learntExample
}
\value{
No return value.
}
\description{
An example \code{learnt} object obtained by means of the \code{\link[=learn]{learn()}} function, using the \link[datasets:penguins]{datasets::penguins} dataset and the metadata in \link{metadataExample}, according to the call

\if{html}{\out{<div class="sourceCode">}}\preformatted{learn(data = penguins, metadata = metadataExample,
  nsamples = 225, nchains = 15)
}\if{html}{\out{</div>}}

It is a list that essentially contains posterior hyperparameters for drawing statistical inferences about the variates \code{species} and \code{bill_len}.

\strong{Note} that the \code{\link[=learn]{learn()}} function that produced \code{learntExample} was called with the option to create only a limited number (225) of Monte Carlo samples, in order to reduce its memory size. Thus the numerical error associated with the Monte Carlo approximation is relatively in inferences drawn from the posterior hyperparameters saved in \code{learntExample}. It is only meant to be used for illustration purposes of the package's capabilities.
}
\seealso{
\code{\link[=learn]{learn()}}, which produces this kind of object.

\code{\link[=Pr]{Pr()}}, \code{\link[=qPr]{qPr()}}, \code{\link[=rPr]{rPr()}}, \code{\link[=mutualinfo]{mutualinfo()}}: functions that require this kind of object in order to calculate probabilities and quantiles, generate data points, and calculate mutual information.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_vtransform.R
\name{vtransform}
\alias{vtransform}
\title{Transforms variates to different representations}
\usage{
vtransform(
  x,
  auxmetadata,
  Rout = NULL,
  Cout = NULL,
  Dout = NULL,
  Bout = NULL,
  Oout = NULL,
  Nout = NULL,
  variates = NULL,
  logjacobianOr = NULL
)
}
\arguments{
\item{x}{data.table object containing data to be transformed}

\item{auxmetadata}{auxmetadata object}

\item{Rout}{Character, output of R-type variate, with possible values:
'normalized': for internal MCMC use
'mi': for use in mutualinfo()
'original': original representation}

\item{Cout}{Character, output of C-type variate, with possible values:
'init': for internal MCMC use (init input)
'left', 'right': for internal MCMC use
'aux', 'lat': for internal MCMC use
'boundnormalized': for sampling functions
'boundisinf': for sampling functions
'mi': for use in mutualinfo()
'original': original representation}

\item{Dout}{Character, output of D-type variate, with possible values:
'init': for internal MCMC use (init input)
'left', 'right': for internal MCMC use
'aux': for internal MCMC use
'boundisinf': for sampling functions
'normalized': for sampling functions
'mi': for use in mutualinfo()
'original': original representation}

\item{Bout}{Character, output of B-type variate, with possible values:
'numeric': for internal MCMC use, values 0,1
'original': original representation}

\item{Oout}{Character, output of O-type variate, with possible values:
'numeric': for internal MCMC use, values 1,2,...
'original': original representation}

\item{Nout}{Character, output of N-type variate, with possible values:
'numeric': for internal MCMC use, values 1,2,...
'original': original representation}

\item{variates}{Character vector, names of variates
corresponding to columns of x (in case x misses column names)}

\item{logjacobianOr}{Logical or \code{NULL}: output is the log-Jacobian in orginal or transformed domain? \code{NULL} (default) means do not calculate the log-Jacobians}
}
\value{
A \link[base:data.frame]{data frame} of transformed variates, or their log-Jacobians.
}
\description{
See \url{https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Pr.R
\name{Pr}
\alias{Pr}
\title{Calculate posterior probabilities}
\usage{
Pr(
  Y,
  X = NULL,
  learnt,
  tails = NULL,
  priorY = NULL,
  nsamples = "all",
  quantiles = c(0.055, 0.25, 0.75, 0.945),
  parallel = TRUE,
  sep = ",",
  solidus = "|",
  verbose = FALSE,
  keepYX = TRUE
)
}
\arguments{
\item{Y}{Matrix or data.table: set of values of variates of which we want
the joint probability of. One variate per column, one set of values per row.}

\item{X}{Matrix or data.table or \code{NULL} (default): set of values of variates on which we want to condition the joint probability of \code{Y}. If \code{NULL}, no conditioning is made (except for conditioning on the learning dataset and prior assumptions). One variate per column, one set of values per row.}

\item{learnt}{Either a character with the name of a directory or full path for a 'learnt.rds' object, produced by the \code{\link[=learn]{learn()}} function, or such an object itself.}

\item{tails}{Named vector or list, or \code{NULL} (default). The names must match some or all of the variates in arguments \code{Y} and \code{X}. For variates in this list, the probability arguments are understood in an semi-open interval sense: \eqn{Y \le y} or \eqn{Y \ge y}, an so on. This is true for \code{Y} and \code{X} variates (on the left and on the right of the conditional sign \eqn{\,\vert\,}). A left-open interval \eqn{Y \le y} is indicated by \code{'<='} or \code{'left'} or \code{-1}; a right-open interval \eqn{Y \ge y} is indicated by \code{'>='} or \code{'right'} or \code{+1}. Values \code{NULL}, \code{'=='}, \code{0} indicate that a point value \code{Y = y} (not an interval) should be calculated. \strong{NB}: the semi-open intervals \emph{always} include the given value; this is important for ordinal or rounded variates. For instance, if \eqn{Y} is an integer variate, then to calculate  \eqn{\mathrm{Pr}(Y < 3)} you should require \eqn{\mathrm{Pr}(Y \le 2)}; for this reason we also have that \eqn{\mathrm{Pr}(Y \le 2)} and  \eqn{\mathrm{Pr}(Y \ge 2)} generally add up to \emph{more} than 1.}

\item{priorY}{Numeric vector with the same length as the rows of \code{Y}, or \code{TRUE}, or \code{NULL} (default): prior probabilities or base rates for the \code{Y} values. If \code{TRUE}, the prior probabilities are assumed to be all equal.}

\item{nsamples}{Integer or \code{NULL} or \code{'all'} (default): desired number of samples of the variability of the probability for \code{Y}. If \code{NULL}, no samples are reported. If \code{'all'} (or \code{Inf}), all samples obtained by the \code{\link[=learn]{learn()}} function are used.}

\item{quantiles}{Numeric vector, between 0 and 1, or \code{NULL}: desired quantiles of the variability of the probability for \code{Y}. Default \code{c(0.055, 0.25, 0.75, 0.945)}, that is, the 5.5\%, 25\%, 75\%, 94.5\% quantiles. These are typical quantile values in the Bayesian literature: they give 50\% and 89\% credibility intervals, which correspond to 1 shannons and 0.5 shannons of uncertainty (see \url{doi:10.5281/zenodo.17072199}). If \code{NULL}, no quantiles are calculated.}

\item{parallel}{Logical or positive integer or cluster object. \code{TRUE} (default): use roughly half of available cores; \code{FALSE}: use serial computation; integer: use this many cores. It can also be a cluster object previously created with \code{\link[parallel:makeCluster]{parallel::makeCluster()}}; in this case the parallel computation will use this object.}

\item{sep}{character, default \code{','}: character to separate variate names and values}

\item{solidus}{character, default \code{'|'}: character prepended to names of the variates in the conditional (typically the \code{X} variates).}

\item{verbose}{Logical, default \code{FALSE}: give messages about parallel processing?}

\item{keepYX}{Logical, default \code{TRUE}: keep a copy of the \code{Y} and \code{X} arguments in the output? This is used for the plot method.}
}
\value{
An object of class "probability", effectively a list consisting of the following elements:
\itemize{
\item \code{values}: a matrix with the probabilities \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data})}, for all joint values \eqn{y} of the \eqn{Y}-variates (rows) and  all joint values \eqn{x} of the \eqn{X}-variates (columns).
\item \code{quantiles} (possibly \code{NULL}): an array with the variability quantiles (3rd dimension of the array) for such probabilities.
\item \code{samples} (possibly \code{NULL}): an array with the variability samples (3rd dimension of the array) for such probabilities.
\item \code{values.MCaccuracy}, \code{quantiles.MCaccuracy}: arrays with the numerical accuracies (roughly speaking a standard deviation) of the Monte Carlo calculations for the \code{values} and \code{quantiles} elements.
\item \code{Y}, \code{X}: copies of the \code{Y} and \code{X} arguments.
}
}
\description{
This function calculates posterior probability densities, cumulative posterior probabilities, and mixtures thereof. It also outputs the variability of such probabilities if more training data were available, and the Monte Carlo Standard Error for the calculated posterior probabilities.
}
\details{
This function calculates the posterior probability \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data})}, where \eqn{Y = y} and \eqn{X = x} are two (non overlapping) sets of joint variate values, inputted as \link[base:data.frame]{data frame} arguments \code{Y} and \code{X}. It is somewhat analogous to the \code{d}-variants and \code{p}-variantes of R distribution functions, such as \code{\link[stats:dnorm]{stats::dnorm()}} and \code{\link[stats:pnorm]{stats::pnorm()}}. If \code{X} is omitted or \code{NULL}, then the posterior probability \eqn{\mathrm{Pr}(Y = y \vert \text{data})} is calculated.

For some variates in \code{Y} or \code{X}, tail values can also be prescribed, so that this function calculates mixed probabilities such as \deqn{\mathrm{Pr}(Y_1 = y_1, Y_2 \le y_2, \dotsc \vert X_1 = x_1, X_2 \ge x_2, \dotsc, \text{data})\ .} Tail values are inputted via the \code{'tails'} argument; see "Usage".

This function also outputs the variability of the posterior probabilities above, that is, probabilities such as \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{new\,data}, \text{data})} that we could have if more learning data were provided, as well as a number of samples of the possible values of such probability. This variability can be outputted in two ways; the user can choose either, or both, or none:
\itemize{
\item As samples (default 3600 samples, depending on the 'nsamples' argument given to the \code{\link[=learn]{learn()}} function) of the alternative values that the posterior probability could have.
\item As quantiles (default 5.5\%, 25\%, 75\%, 94.5\%) of the possible variability.
}

If several joint values are given for \code{Y} or \code{X}, the function will create a 2D grid of results for all possible combinations of the given \code{Y} and \code{X} values.

This function also allows for base-rate or other prior-probability corrections: If a prior (for instance, a base rate) for \code{Y} is given, the function will calculate the probability \eqn{\mathrm{Pr}(Y = y \vert X = x, \text{data}, \text{prior})} from \eqn{\mathrm{Pr}(X = x \vert Y = y, \text{data})} and the prior, by means of Bayes's theorem.

Each variate in each argument \code{Y}, \code{X} can be specified either as a point-value \eqn{Y = y} or as a left-open interval \eqn{Y \le y} or as a right-open interval \eqn{Y \ge y}, through the argument \code{tails}.

See \code{vignette('intro')} for example uses.
}
\examples{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## ## Example 1:
## Calculate the probability that an unknown penguin from this population
## is of species 'Adelie'

probs <- Pr(
  Y = data.frame(species = 'Adelie'),
  learnt = learnt, parallel = 1
)

## display the probability value
probs$values

## the full-population frequency of 'Adelie' penguins is unknown;
## display the 5.5\%- and 94.5\%-probability values
## for such frequency
probs$quantiles[, , c('5.5\%', '94.5\%')]

## we can also plot the probability distribution for this full-population frequency
hist(probs, legend = 'topright')


## ## Example 2:
## Calculate the 3 probabilities that an unknown penguin from this population
## is of species 'Adelie', 'Chinstrap', 'Gentoo'

probs <- Pr(
  Y = data.frame(species = c('Adelie', 'Chinstrap', 'Gentoo')),
  learnt = learnt, parallel = 1
)

## display the 3 probability values
probs$values

## the full-population frequencies of the three species are unknown;
## display the 5.5\%- and 94.5\%-probability values
## for such frequencies
probs$quantiles[, , c('5.5\%', '94.5\%')]

## plot the probabilities and quantiles
plot(probs)

## plot the probability distribution for the full-population frequency
## of each species
hist(probs)

## ## Example 3:
## Calculate the probability that an unknown penguin is of species 'Adelie'
## GIVEN that its bill length is 43 mm

probs <- Pr(
  Y = data.frame(species = 'Adelie'),
  X = data.frame(bill_len = 43),
  learnt = learnt, parallel = 1
)

## display the probability value
probs$values

## the full-subpopulation frequency of 'Adelie' penguins,
## among penguins having bill length of 43 mm, is unknown;
## display the 5.5\%- and 94.5\%-probability values
## for such conditional frequency
probs$quantiles[, , c('5.5\%', '94.5\%')]


## ## Example 4:
## Calculate the probability that
## an unknown penguin is of species 'Adelie' AND its bill length is 43 mm

probs <- Pr(
  Y = data.frame(species = 'Adelie', bill_len = 43),
  learnt = learnt, parallel = 1
)

## display the probability value
probs$values

## display the 5.5\%- and 94.5\%-probability values
## for the full-population frequency of 'Adelie' penguins with 43 mm bills
probs$quantiles[, , c('5.5\%', '94.5\%')]


## ## Example 5:
## Calculate the 3 x 2 probabilities for the 3 species
## GIVEN bill-lengths of 43 mm and 44 mm

Y <- data.frame(species = c('Adelie', 'Chinstrap', 'Gentoo'))

X <- data.frame(bill_len = c(43, 44))

probs <- Pr(Y = Y, X = X, learnt = learnt, parallel = 1)

## display the 3 x 2 probability values
probs$values

## display the 5.5\%- and 94.5\%-probability values
## for the full-population joint frequencies
probs$quantiles[, , c('5.5\%', '94.5\%')]

## plot the probabilities and quantiles
plot(probs)


## ## Example 6:
## Calculate the 3 x 2 joint probabilities for the 3 species
## AND bill-lengths of 43 mm and 44 mm

Y <- expand.grid(
  species = c('Adelie', 'Chinstrap', 'Gentoo'),
  bill_len = c(43, 44)
)

probs <- Pr(Y = Y, learnt = learnt, parallel = 1)

## display the 6 joint-probability values
probs$values

## display the 5.5\%- and 94.5\%-probability values
## for the full-population joint frequencies
probs$quantiles[, , c('5.5\%', '94.5\%')]


}
\references{
\itemize{
\item Lindley, Novick (1981): \emph{The role of exchangeability in inference}, \url{doi:10.1214/aos/1176345331}.
\item Bernardo, Smith (2000): \emph{Bayesian Theory}. Wiley \url{doi:10.1002/9780470316870}.
\item Jaynes (2003): \emph{Probability Theory: The Logic of Science}. Cambridge University Press \url{doi:10.1017/CBO9780511790423}.
\item MacKay (2005): \emph{Information Theory, Inference, and Learning Algorithms}. Cambridge University Press \url{https://www.inference.org.uk/itila/book.html}.
\item Porta Mana (2025): \emph{What's special about 89\% credibility intervals?} \url{doi:10.5281/zenodo.17072199}.
}
}
\seealso{
\code{\link[=learn]{learn()}}, which generates the \code{learnt} objects required by \code{Pr()}.

\code{\link[=plot.probability]{plot.probability()}} to plot probabilities and quantiles calculated by \code{Pr()}.

\code{\link[=hist.probability]{hist.probability()}} to plot histograms of the probability distributions calculated by \code{Pr()}.

\code{\link[=print.probability]{print.probability()}} to print the main elements of the probabilities calculated by \code{Pr()}.

\code{\link[=qPr]{qPr()}} to calculate quantiles for a specific variate, that is, the variate values having given probabilities.

\code{\link[=rPr]{rPr()}} to generate datapoints.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcsamples.R
\name{mcjoin}
\alias{mcjoin}
\title{Concatenate mcsample objects}
\usage{
mcjoin(x, y)
}
\description{
Used in 'learn()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutualinfo.R
\name{mutualinfo}
\alias{mutualinfo}
\title{Calculate mutual information between groups of joint variates}
\usage{
mutualinfo(
  Y1names,
  Y2names = NULL,
  X = NULL,
  learnt,
  tails = NULL,
  n = NULL,
  unit = "Sh",
  parallel = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{Y1names}{Character vector: first group of joint variates}

\item{Y2names}{Character vector or \code{NULL}: second group of joint variates}

\item{X}{Matrix or data.frame or \code{NULL}: values of some variates conditional on which we want the probabilities.}

\item{learnt}{Either a character with the name of a directory or full path
for an 'learnt.rds' object, or such an object itself.}

\item{tails}{Named vector or list, or \code{NULL} (default). The names must match some or all of the variates in arguments \code{X}. For variates in this list, the probability conditional is understood in an semi-open interval sense: \eqn{X \le x} or \eqn{X \ge x}, an so on. See analogous argument in \code{\link[=Pr]{Pr()}}.}

\item{n}{Integer or \code{NULL} (default): number of samples from which to approximately calculate the mutual information. Default as many as Monte Carlo samples in \code{learnt}.}

\item{unit}{Either one of 'Sh' for \emph{shannon} (default), 'Hart' for \emph{hartley}, 'nat' for \emph{natural unit}, or a positive real indicating the base of the logarithms to be used.}

\item{parallel}{Logical or positive integer or cluster object. \code{TRUE} (default): use roughly half of available cores; \code{FALSE}: use serial computation; integer: use this many cores. It can also be a cluster object previously created with \code{\link[parallel:makeCluster]{parallel::makeCluster()}}; in this case the parallel computation will use this object.}

\item{verbose}{Logical, default \code{FALSE}: give messages about parallel processing?}
}
\value{
A list consisting of the following elements:
\itemize{
\item \code{MI}, a vector of \code{value} and \code{accuracy}: the mutual information between (joint) variates \code{Y1names} and (joint) variates \code{Y2names}.
\item \code{CondEn12}, \code{CondEn21}, vectors of \code{value} and \code{accuracy}: the conditional entropy of the first variate given the second, and vice versa.
\item \code{En1}, \code{En2}, vectors of \code{value} and \code{accuracy}: the (differential) entropies of the first and second variates.
\item \code{MI.rGauss}, a vector of \code{value} and \code{accuracy}: the absolute value of the Pearson correlation coefficient \eqn{r} of a \emph{multivariate Gaussian distribution} having mutual information \code{MI}; the two are related by \eqn{\mathrm{MI} = -\ln(1 - r^2)/2}. It may provide a vague intuition for the \code{MI} value for people more familiar with Pearson's correlation, but should be taken with a grain of salt.
\item \code{unit}, \code{Y1names}, \code{Y1names}: same as the input arguments, included for the user's convenience.
}
}
\description{
This function calculates various entropic information measures between two grops of joint variates: the mutual information, the conditional entropies, and the entropies.
}
\details{
If \eqn{Y_1} and \eqn{Y_2} are two variates, each of which can be a joint variate such as \eqn{Y_1 = (Y_{1,1}, Y_{1,2}, \dotsc)}, and \eqn{X} a third, also possibly join, variate, then the mutual information \eqn{\mathit{MI}} between \eqn{Y_1} and \eqn{Y_2}, conditional on \eqn{X = x}, is given by
\deqn{\mathit{MI}(Y_1, Y_2 \vert X = x) \mathrel{:=}
\sum_{y_1, y_2}
\mathrm{Pr}(Y_1 = y_1, Y_2 = y_2 \vert X = x, \text{data})
\log_2\frac{
\mathrm{Pr}(Y_1 = y_1, Y_2 = y_2 \vert X = x, \text{data})
}{
\mathrm{Pr}(Y_1 = y_1 \vert X = x, \text{data})
\cdot
\mathrm{Pr}(Y_2 = y_2 \vert X = x, \text{data})
} \, \mathrm{Sh}
}
an expression which can also be written in several other equivalent ways. It is an information-theoretic measure of association that is model-free, that is, does not depend on assumptions such as linearity, gaussianity, and similar. See \code{vignette('mutualinfo')} for discussion and example uses, and also the "References" section.  If \eqn{Y_1, Y_2} are \emph{jointly gaussian variates}, then there is a mathematical correspondence between their mutual information and their Pearson correlation coefficient; see output \code{MI.rGauss} in the "Value" section.

The conditional entropy of \eqn{Y_1} with respect to \eqn{Y_2}, conditional on \eqn{X = x}, is given by
\deqn{\mathit{CondEn12}(Y_1, Y_2 \vert X = x) \mathrel{:=}
-\sum_{y_1, y_2}
\mathrm{Pr}(Y_1 = y_1 \vert Y_2 = y_2, X = x, \text{data})
\log_2
\mathrm{Pr}(Y_1 = y_1 \vert Y_2 = y_2, X = x, \text{data})
\cdot
\mathrm{Pr}(Y_2 = y_2 \vert X = x, \text{data})
\, \mathrm{Sh}
}

The (differential) entropy of \eqn{Y_1}, conditional on \eqn{X = x}, is given by
\deqn{\mathit{En1}(Y_1 \vert X = x) \mathrel{:=}
-\sum_{y_1}
\mathrm{Pr}(Y_1 = y_1 \vert X = x, \text{data})
\log_2
\mathrm{Pr}(Y_1 = y_1 \vert  X = x, \text{data})
\, \mathrm{Sh}
}

see "References" section for discussions about entropy and conditional entropy.

The function \code{mutualinfo()} calculates the quantities above for the joint variates specified in the arguments \code{Y1names} and \code{Y2names}, conditional on the values of the variates specified in the data frame \code{X}. If \code{X} is omitted or \code{NULL}, then the posterior probabilities \eqn{\mathrm{Pr}(Y_1 | \text{data})} etc. are used. Each variate in the argument \code{X} can be specified either as a point-value \eqn{X = x} or as a left-open interval \eqn{X \le x} or as a right-open interval \eqn{X \ge x}, through the argument \code{tails}.

The computation of these quantities is done via Monte Carlo integration, using the samples produced by the \code{\link[=learn]{learn()}} function. The present function also output the numerical error associated with this computation.
}
\examples{
## Load the example `learnt` object calculated from the "penguins" dataset;
## variates: 'species' and 'bill_len'
learnt <- learntExample

## mutual information between variates 'species' and 'bill_len'
MI <- mutualinfo(Y1names = 'species', Y2names = 'bill_len',
  learnt = learnt, parallel = 1)

paste0(MI$MI, ' ', MI$unit, collapse = ' +/- ')

## Shannon entropy of variate 'species'
paste0(MI$En1, ' ', MI$unit, collapse = ' +/- ')


\donttest{
## Shannon entropy of variate 'species',
## conditional on a bill length of 30 mm:
entr <- mutualinfo(
  Y1names = 'species',
  X = data.frame(bill_len = 30),
  learnt = learnt, parallel = 1
)

paste0(entr$En1, ' ', entr$unit, collapse = ' +/- ')

## the entropy is now lower; indeed a penguin with a short bill length
## is most probably of the 'Adelie' species:
probs <- Pr(
  Y = data.frame(species = c('Adelie', 'Gentoo', 'Chinstrap')),
  X = data.frame(bill_len = 30),
  learnt = learnt, parallel = 1
)

print(probs)
}

}
\seealso{
\code{\link[=Pr]{Pr()}} to calculate probabilities and their variability.

\code{\link[=learn]{learn()}}, which generates the \code{learnt} objects required by \code{mutualinfo()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadatatemplate.R
\name{metadatatemplate}
\alias{metadatatemplate}
\alias{metadata}
\title{Metadata and helper function for metadata}
\usage{
metadatatemplate(
  data,
  file = NULL,
  includevrt = NULL,
  excludevrt = NULL,
  addsummary2metadata = FALSE,
  backupfiles = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{data}{A dataset, given as a \link[base:data.frame]{data frame}
or as a file path to a csv file.}

\item{file}{Character or \code{NULL} (default): name of csv file where the metadata should be saved; if \code{NULL}: output metadata as \code{VALUE}.}

\item{includevrt}{Character or \code{NULL}: name of variates in dataset to be included.}

\item{excludevrt}{Character or \code{NULL}: name of variates in dataset to be excluded.}

\item{addsummary2metadata}{Logical: also output some diagnostic statistics
in the metadata? Default \code{FALSE}.}

\item{backupfiles}{Logical: rename previous metadata file if it exists?
Default \code{TRUE}.}

\item{verbose}{Logical: output heuristics for each variate? Default \code{TRUE}.}
}
\value{
A preliminary \link[base:data.frame]{data frame} containing the metadata, \link[base:invisible]{invisibly} if \code{file = NULL}. If argument \code{file} is a character, a preliminary metadata file is also created with that name or path.
}
\description{
Metadata and helper function to create a template metadata file or object.
}
\details{
The \code{\link[=learn]{learn()}} function needs metadata about the variates present in the data. Such metadata can be provided either as a \code{csv} file or as a \code{\link[base:data.frame]{base::data.frame()}}. The function \code{buildmetadata} creates a template metadata csv-file, or outputs a metadata data.frame, by trying to \emph{guess} metadata information from the dataset.The guesses may be very incorrect (as already said, metadata is information not contained in the data, so no algorithm can exist that extracts it from the data). \strong{The user \emph{must} modify and correct this template, using it as a starting point to prepare the correct metadata information.}
}
\section{Metadata information and format}{


In order to correctly learn from a dataset, the \code{\link[=learn]{learn()}} function needs information that is not contained in the data themeselves; that is, it needs \emph{meta}data. Metadata are provided either as a \code{csv} file or as a \code{\link[base:data.frame]{base::data.frame()}}.

A metadata file or data.frame must contain one row for each simple variate in the given inference problem, and the following fields (columns), even if some of them may be empty:

\code{name}, \code{type}, \code{domainmin}, \code{domainmax}, \code{datastep}, \code{minincluded}, \code{maxincluded}, \code{V1}, \code{V2}, (possibly additional \code{V}-fields, sequentially numbered)

The \code{type} field has three possible values: \code{nominal}, \code{ordinal}, \code{continuous}. The remaining fields that must be filled in depend on the \code{type} field. Here is a list of requirements:
\itemize{
\item \strong{\code{nominal}} and \strong{\code{ordinal}}: require \emph{either} \code{V1}, \code{V2}, ... fields \emph{or} \code{domainmin}, \code{domainmax}, \code{datastep} (all three) fields. No other fields are required.
\item \strong{\code{continuous}}: requires \code{domainmin}, \code{domainmax}, \code{datastep}, \code{minincluded}, \code{maxincluded}.
}

Here are the meanings and possible values of the fields:

\strong{\code{name}}: The name of the variate. This must be the same character string as it appears in the dataset (be careful about upper- and lower-case).

\strong{\code{type}}: The data type of variate \code{name}. Possible values are \code{nominal}, \code{ordinal}, \code{continuous}.
\itemize{
\item A \emph{nominal} (also called \emph{categorical}) variate has a discrete, finite number of possible values which have no intrinsic ordering. Examples could be a variate related to colour, with values "red", "green", "blue", and so on; or a variate related to cat breeds, with values "Siamese", "Abyssinian", "Persian", and so on. The possible values of the variate must be given in the fields \code{V1}, \code{V2}, and so on. It is important to include values that are possible but are \emph{not} present in the dataset. A variate having only two possible values (binary variate), for example "yes" and "no", can be specified as nominal.
\item An \emph{ordinal} variate has a discrete, finite number of possible values which do have an intrinsic ordering. Examples could be a Likert-scaled variate for the results of a survey, with values "very dissatisfied", "dissatisfied", "satisfied", "very satisfied"; or a variate related to the levels of some quantities, with values "low", "medium", "high"; or a variate having a numeric scale with values from 1 to 10. Whether a variate is nominal or ordinal often depends on the context. The possible values of the variate but be given in either one (but not both) or two ways: (1) in the fields \code{V1}, \code{V2}, ..., as for nominal variates; (2) as the fields \code{domainmin}, \code{domainmax}, \code{datastep}. Option (2) only works with numeric, equally spaced values: it assumes that the first value is \code{domainmin}, the second is \code{domainmin}+\code{datastep}, the third is \code{domainmin}+2*\code{datastep}, and so on up to the last value, \code{domainmax}.
\item A \emph{continuous} variate has a continuum of values with an intrinsic ordering. Examples could be a variate related to the width of an object; or to the age of a person; or one coordinate of an object in a particular reference system. A continuous variate requires specification of the fields \code{domainmin}, \code{domainmax}, \code{datastep}, \code{minincluded}, \code{maxincluded}. Some naturally continuous variates are often rounded to a given precision; for instance, the age of a person might be reported as rounded to the nearest year (25 years, 26 years, and so on); or the length of an object might be reported to the nearest centimetre (1 m, 1.01 m, 1.02 m, and so on). The minimum distance between such rounded values \strong{must} be reported in the \code{datastep} field; this would be \code{1} in the age example and \code{0.01} in the length example above. See below for further explanation of why reporting such rounding is important.
}

\strong{\code{domainmin}}: The minimum value that the variate (ordinal or continuous) can take on. Possible values are a real number or an empty value, which is then interpreted as \code{-Inf} (explicit values like \code{-Inf}, \code{-inf}, \code{-infinity} should also work). Some continuous variates, like age or distance or temperature, are naturally positive, and therefore have \code{domainmin} equal \code{0}. But in other contexts the minimum value could be different. For instance, if a given inference problem only involves people of age 18 or more, then \code{domainmin} would be set to \code{18}.

\strong{\code{domainmax}}: The maximum value that the variate (ordinal or continuous) can take on. Possible values are a real number, or an empty value, which is then interpreted as \code{+Inf} (explicit values like \code{Inf}, \code{inf}, \code{infinity} should also work). As with \code{domainmin}, the maximum value depends on the context. An age-related variate could theoretically have \code{domainmax} equal to infinity (empty value in the metadata file); but if a given study categorizes some people as "90 years old or older", then \code{domainmax} should be set to \code{90}.

\strong{\code{datastep}}: The minimum distance between the values of a variate (ordinal or continuous). Possible values are a positive real number or an empty value, which is then interpreted as 0 (the explicit value \code{0} is also accepted). For a numeric ordinal variate, \code{datastep} is the step between consecutive values. For a continuous \emph{rounded} variate, \code{datastep} is the minimum distance between different values that occurs because of rounding; see the examples given above. The function \code{buildmetadata} has some heuristics to determine whether the variate is rounded or not. See further details under the section Rounding below.

\strong{\code{minincluded}}, \strong{\code{maxincluded}}: Whether the minimum (\code{domainmin}) and maximum(\code{domainmax}) values of a \emph{continuous} variate can really appear in the data or not. Possible values are \code{true} (or \code{t} or \code{yes}) or \code{false} (or \code{f}, \code{no}, or an empty field); upper- or lower-case is irrelevant. Here are some examples about the meaning of these fields. (a) A continuous \emph{unrounded} variate such as temperature has 0 as a minimum possible value \code{domainmin}, but this value itself is physically impossible and can never appear in data; in this case \code{minincluded} is empty (or set to \code{false} or \code{no}). (b) A variate related to the \emph{unrounded} length, in metres, of some objects may take on any positive real value; but suppose that all objects of length 5 or less are grouped together under the value \code{5}. It is then possible for several datapoints to have value \code{5}: one such datapoint could originally have the value 3.782341...; another the value 4.929673..., and so on. In this case \code{domainmin} is set to \code{5}, and \code{minincluded} is set to \code{true} (or \code{yes}). Similarly for the maximum value of a variate and \code{maxincluded}. Note that if \code{domainmin} is minus-infinity (empty value in the metadata file), then \code{minincluded} is automatically empty (that is, \code{false}), and similarly for \code{maxincluded} if \code{domainmax} is infinity.
}

\examples{
## Create a preliminary data frame of metadata for the `penguins` dataset
metadata <- metadatatemplate(data = datasets::penguins, file = NULL)

## Note how the preliminary data frame includes additional spots
## for values of nominal and ordinal variates
## which could be missing from the data
print(metadata)


## Create a preliminary data frame of metadata for the `penguins` dataset,
## including only the 'species' and 'bill_len' variates:
metadata2 <- metadatatemplate(
  data = datasets::penguins, file = NULL,
  includevrt = c('species', 'bill_len')
)

print(metadata2)


## Create a preliminary data frame of metadata for the `penguins` dataset,
## excluding the 'year' variate:
metadata3 <- metadatatemplate(
  data = datasets::penguins, file = NULL,
  excludevrt = 'year'
)

print(metadata3)

## Generate 10 points for a continuous variate in (0, 1)
dataset <- runif(10)

## `metadatatemplate` correctly guesses the variate minimum,
## but not the maximum (`NA` is equivalent to `+Inf`)
metadata <- metadatatemplate(data = dataset, file = NULL)
print(metadata)

}
\seealso{
\code{\link[=learn]{learn()}}, which generates the information necessary to calculate posterior probabilities, based on data and metadata.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_cleanup.R
\name{util_joinPtraces}
\alias{util_joinPtraces}
\title{Join '____tempPtraces-' files}
\usage{
util_joinPtraces(path)
}
\value{
A \link[base:data.frame]{data frame} of MCMC traces.
}
\description{
Join '____tempPtraces-' files
}
\details{
For deeper monitoring of the MCMC, the user can require the 'learn()' function not to clean intermediate MCMC-related files generated during the computation. The files with prefix '____tempPtraces-' contain chunks of MCMC traces.

The present function can be used to join them into a single trace.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcmcstop.R
\name{funMCEQ}
\alias{funMCEQ}
\title{Calculate credibility quantiles on estimated quantile}
\usage{
funMCEQ(x, prob = c(0.055, 0.945), Qpair = pnorm(c(-1, 1)))
}
\arguments{
\item{x}{A vector of MC samples}

\item{prob}{numeric vector of probabilities: quantiles whose error interval is being estimated.}

\item{Qpair}{vector of length two (further elements are ignored): lower and higher credibility-quantiles requested. Default yields a credibility interval of 68\%, or one nominal normal standard deviation.}
}
\value{
A matrix with two rows and as many columns as elements in 'prob'. Forr each column, the first and second row determine the lower and upper bound of the credibility interval of width \code{Qpair[2] - Qpair[2]}.
}
\description{
Calculates the lower and upper bound of a credibility interval, for various quantiles of the empirical distribution of a vector of MC samples.
}
\details{
Tests show that it gives a very good approximation of the "true" credibility quantiles in the case of independent samples.

Tested also on t-distributions with df=1.1 and Pareto with a=1.5 (mean exists, variance infinite).

Used in 'workerfun()' in 'learn()'
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{util_lprobsbase}
\alias{util_lprobsbase}
\title{Calculate collection of log-probabilities for different components and samples}
\usage{
util_lprobsbase(xVs, params, logW, temporarydir = NULL, lab = "")
}
\value{
Matrix with as many rows as components and as many cols as samples
}
\description{
Used in 'Pr()', 'qPr()', 'rPr()', 'mutualinfo()', 'util_checkpoints()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lprobs.R
\name{util_combineYX}
\alias{util_combineYX}
\title{Calculate probabilities, quantiles, etc, for all Y and X combinations}
\usage{
util_combineYX(
  iyx,
  temporarydir,
  usememory = TRUE,
  doquantiles,
  quantiles,
  dosamples,
  nsamples,
  Qerror
)
}
\description{
Used in 'Pr()'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_mcsamples.R
\name{learnbind}
\alias{learnbind}
\title{Bind 3D arrays by first dimension}
\usage{
learnbind(x, y)
}
\description{
Used in 'util_checkpoints()' within 'learn()', and in various functions in 'util_lprobs.R'.
}
\details{
NB: the following variant is slower:

\if{html}{\out{<div class="sourceCode">}}\preformatted{function(x, y) \{
    out <- c(aperm(x), aperm(y))
    dim(out) <- c(rev(dim(x)[-1]), dim(x)[1] + dim(y)[1])
    aperm(out)
\}
}\if{html}{\out{</div>}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_createQfunction.R
\name{createQfunction}
\alias{createQfunction}
\title{Calculate and save transformation function for ordinal variates}
\usage{
createQfunction(
  nint = 3600,
  nsamples = 2^24L,
  mean = 0,
  sd = 3,
  shapelo = 0.5,
  shapehi = 0.5,
  rate = 1,
  file = paste0("__Qfunction", nint, "_", sd),
  plot = FALSE
)
}
\description{
It creates the interpolation functions 'util_Q', 'util_invQ', 'util_invDQ' and saves them into 'sysdata.rda'.
}
\details{
Those three functions are used to transform variates having bounded domains into variates with unbounded domains. See \url{https://github.com/pglpm/prova/raw/main/development/manual/pglpm2024-bayes_nonparam.pdf}.

NB: the functional form of this function does not depend on
the number of components, minalpha, and maxalpha parameters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_checkpoints.R
\name{util_Pcheckpoints}
\alias{util_Pcheckpoints}
\title{Calculate joint frequencies for MCMC-monitoring checkpoints}
\usage{
util_Pcheckpoints(testdata, learnt)
}
\arguments{
\item{testdata}{List of objects calculated with util_prepPcheckpoints}

\item{learnt}{mcsamples object}
}
\value{
The joint frequencies of Y corresponding to the Monte Carlo samples
}
\description{
Used in 'learn()'.
}
\keyword{internal}
